/*
** Copyright (c) 2003, Scott Lystig Fritchie.  All rights reserved.
** Some contributions Copyright (c) Chris Newcombe, 2006.  All rights reserved.
** See the file "../LICENSE" for license details.
*/

/* erl_driver_pipelib.c */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#ifdef	NEED_NETINET_IN_H
#include <netinet/in.h>		/* ntohl() */
#else	/* NEED_NETINET_IN_H */
#include <sys/param.h>		/* ntohl() */
#endif	/* NEED_NETINET_IN_H */

#include <erl_driver_pipelib.h>
#include <erl_driver_tk.h>

input_iovec_wrapper_t *
read_packet_make_input_iovec(int fd)
{
    input_iovec_wrapper_t * result_iovec_p;
    char	            lenbuf[4];
    int		            wanted, numread, got;
    
    wanted = 4;
    got = 0;
    while (got < wanted) {
	if ((numread = read(fd, lenbuf + got, wanted - got)) < 0) {
	    EDTK_LOG(EDTK_LOG_VERBOSE, "read 1 failed, errno = %d", errno);
	    return NULL;
	} else if (numread == 0) {
	    EDTK_LOG(EDTK_LOG_VERBOSE, "1 EOF");
	    return NULL;
	}
	got += numread;
    }
    wanted = ntohl(*(unsigned long *) lenbuf);
    EDTK_LOG(EDTK_LOG_VERBOSE, "wanted = %d", wanted);

    if ( NULL == (result_iovec_p = EDTK_MALLOC(sizeof(input_iovec_wrapper_t) + wanted)) ) {
        return NULL;
    }

    got = 0;
    while (got < wanted) {
	EDTK_LOG(EDTK_LOG_VERBOSE, "top got < wanted loop");
	if ((numread = read(fd, &result_iovec_p->input_data.orig_bytes[got], wanted - got)) < 0) {
	    EDTK_LOG(EDTK_LOG_VERBOSE, "read 2 failed, errno = %d", errno);
            EDTK_FREE(result_iovec_p);
	    return NULL;
	} else if (numread == 0) {
	    EDTK_LOG(EDTK_LOG_VERBOSE, "2 EOF");
            EDTK_FREE(result_iovec_p);
	    return NULL;
	}
	got += numread;
	EDTK_LOG(EDTK_LOG_VERBOSE, "got = %d", got);
    }

    result_iovec_p->ev.vsize = 2;
    result_iovec_p->ev.size  = wanted;
    result_iovec_p->ev.iov   = (SysIOVec *)&result_iovec_p->iov[0];
    result_iovec_p->ev.binv  = result_iovec_p->binv;

    result_iovec_p->iov[0].iov_base = NULL;
    result_iovec_p->iov[0].iov_len  = 0;
    result_iovec_p->iov[1].iov_base = result_iovec_p->input_data.orig_bytes;
    result_iovec_p->iov[1].iov_len  = wanted;

    result_iovec_p->binv[0] = NULL;
    result_iovec_p->binv[1] = Binary2ErlDrvBinary(&result_iovec_p->input_data);

    result_iovec_p->input_data.orig_size = wanted;
    result_iovec_p->input_data.refc      = 1;

    return result_iovec_p;
}

/*
 * Provide a work-alike versions for driver_*_binary().
 *
 * These must use ErlDrvBinary as provided erl_driver.h, as the
 * driver .so was compiled against that type.
 *
 * And the functions must meet the same guarantees
 * that the erl_driver functions have -- or at least the
 * subset of those guarantees that we actually need.
 *
 * In R10B (and before) ErlDrvBinary had a refc field (refcount).
 * In R11B (for SMP support) this was moved, and replaced with
 * functions to read/adjust the ref count.
 * We have to paste copies of some of those functions here,
 * and modify them slightly.  This will obviously break
 * badly (just pipe-mode) if the implementation changes
 * again in future releases of Erlang -- so we need to check
 * these things when uprading to new releases.
 *
 * Characteristics:
 *  In R10B, the refc member was not thread-safe (as either the driver or
 *  VM might be using a binary, but not both at once, as there was only a single
 *  VM thread and async workers were not allowed to call erl_driver functions).
 *  Also, driver_alloc_binary & co were not thread safe (for similar reasons).
 *  (The latter issue meant that the Berkeley DB driver had to malloc data
 *  inside BDB and then copy it into a new driver binary in ready_async (make_dbt),
 *  because private worker threads could not call driver_alloc_binary).
 *
 *  In R11B the refc field must be thread-safe if the emulator is built with SMP
 *  support, as other emulator threads might use the binary when the driver is using it.
 *  (this mostly applies to binaries that the VM sends us (i.e.arguments) as we don't
 *  currently hang on to any binaries after returning them to the VM, but I suppose we might in future).
 *
 *  However, in pipe mode we do *NOT* need to make the refcount thread safe as
 *
 *    - The VM never sees these driver binaries (they are purely internal to the driver instance)
 *
 *    - We never share a driver binary instance between multiple threads
 *      For input binaries (ariving from Erlang)
 *         1.  pipe-main.c  calls read_packet_make_input_iovec() to create the single driver binary holding the command
 *             and all arg data (sets it's refc to 1).
 *             Note: that driver binary is embedded in an input_iovec_wrapper_t struct which is all allocated as a single block
 *             -- i.e. that driver binary is NOT allocated with driver_alloc_binary, and must NOT be free'd by driver_free_binary.
 *             It CAN be processed by driver_free_binay so long as it's refcount never falls to 0 (which is true, unless there is a bug).
 *         2.  pipe-main.c calls outputv, which knows if this command uses an async worker, and if so, it
 *             increments the binary's refcount and appends the binary to callstate->refc_bin (to call
 *             driver_free_binary later, when the command is finished)
 *             (Actually it increments it once for each binary argument,
 *             and once for the tagbin because the same code processes binaries in linked-in mode, when they might be separate)
 *         3a. An async worker executes a command which accesses the binary (**but does NOT change it's refcount**)
 *         3b. pipe_main.c (after outputv returns) checks if refc == 1 and if it is, then it frees the input_iovec_wrapper_t (it must have been a non-async command).
 *             If refc != 1 then it registers the input_iovec_wrapper_t to be freed later.
 *         4.  pipe-main.c calls ready_input (ready_async) if there is some data ready (some async commands have finished)
 *         5.  ready_async calls driver_send_term (driver_output_term) which copies the data over stdout
 *         6   ready_async calls free_callstate, which calls driver_free_binary on all binaries in callstate->refc_bin
 *             (this reduces the refc back to 1)
 *         7.  pipe-main.c (after ready_async returnes) calls free_any_delayed_items_that_are_ready()
 *             which checks refc for all registered binaries, and if it finds a refc == 1 then it actually frees the input_iovec_wrapper_t.
 *             It is safe to examine all of the refcs as they can only be adjusted in ready_async, which can only
 *             be called by the pipe-main thread.
 *
 *      For output binaries
 *         1. outputv might pre-create them (we don't do that at the moment)
 *         2. An async worker might want to call driver_alloc_binary.
 *            In R10B and earlier this was not allowed, so the worker would have to malloc the data
 *            and wait to allocate a binary (and copy the data & free the malloced version) in ready_async.
 *            (That's how make_dbt worked in the BDB driver in R10B.)
 *
 *            But as an undocumented feature of R11B, driver_alloc_binary is now thread-safe.
 *            I found the following comments in  otp_src_R11B-1/erts/emulator/beam/io.c
 *           
 *               - driver_alloc_binary() is thread safe (efile driver depend on it). 
 *               - driver_realloc_binary(), and driver_free_binary() are *not* thread safe.
 *
 *            And an email note: from Ericsson:
 *               
 *               They are all thread safe in the emulator with smp support, but they will 
 *               probably never be thread safe in the emulator without smp support.
 *                
 *               The thread unsafe part is the reference counting of binaries. Making 
 *               them thread safe would cause overhead for all reference counted 
 *               binaries, not just driver binaries. We don't want to slow down all 
 *               binary management as it is now, but we might make them thread safe in 
 *               the future if threaded drivers should become very common (but don't 
 *               count on it).
 *               
 *               Rickard Green, Erlang/OTP
 *           
 *            This is good as it means the BDB driver can now use Scott's original scheme of telling BDB to use
 *            edtk_driver_alloc_wrapper() instead of malloc, to make BDB directly call driver_alloc_binary internally,
 *            so we won't have to allocate the data twice and copy it once.
 *            To do this, we DO therefore need driver_alloc_binary to be thread-safe even in pipe-mode, as multiple
 *            worker threads (in multiple driver instances) will be calling it simultaneously.
 *
 *         3. pipe-main.c calls read_input (ready_async)
 *         4. ready_async may call reply_xtra_XXX, which may call a user function to construct a driver term (e.g. make_dbt)
 *            and the user-function may call driver_alloc_binary (and appends a ptr to that binary to desc->tofree[]
 *         5. reply_xtra_XXX calls driver_send_term, which copies the data over stdout
 *         4. reply_xtra_XXX calls driver_free_binary on all binaries listed in desc->tofree
 *            (Note: if multiple VM threads could call ready_async in linked-in mode, this would not be thread-safe;
 *             a fix would be to append the binaries to callstate->refc_bin[], as they will be freed in free_callstate)
 */

/* As discussed above, these are NOT thread-safe */
long
driver_binary_get_refc(ErlDrvBinary *dbp)
{
    Binary* bp = ErlDrvBinary2Binary(dbp);
    return bp->refc;
}

long
driver_binary_inc_refc(ErlDrvBinary *dbp)
{
    Binary* bp = ErlDrvBinary2Binary(dbp);
    return ++(bp->refc);
}

long
driver_binary_dec_refc(ErlDrvBinary *dbp)
{
    Binary* bp = ErlDrvBinary2Binary(dbp);
    return --(bp->refc);
}

/* As discussed above, this MUST be thread-safe.
 * Also (from the documentation)
 *
 *    From erts version 5.5 (OTP release R11B), orig_bytes is guaranteed to be properly aligned
 *    for storage of an array of doubles (usually 8-byte aligned).
 */

ErlDrvBinary *
driver_alloc_binary(int size)
{
    Binary * bin;

    if ((bin = EDTK_MALLOC(size + sizeof(Binary))) != NULL) {
        bin->refc = 1;
        bin->orig_size = size;
    }
    return Binary2ErlDrvBinary(bin);
}

/* As discussed above this does not have to be thread-safe,
 * and in linked-in mode it is NOT thread-safe. */
ErlDrvBinary *
driver_realloc_binary(ErlDrvBinary *old_drvbin, int size)
{    
    ErlDrvBinary * new_drvbin = NULL;
    
    if (old_drvbin != NULL)
    {
        Binary * old_bin = ErlDrvBinary2Binary(old_drvbin);
        Binary * new_bin;
        if ( (new_bin = realloc(old_bin, size + sizeof(Binary))) != NULL)
        {
            new_drvbin = Binary2ErlDrvBinary(new_bin);
            new_drvbin->orig_size = size;
        }            
    }
    return new_drvbin;
}

/* As discussed above this does not have to be thread-safe */
void
driver_free_binary(ErlDrvBinary *drvbin)
{    
    if (drvbin != NULL)
    {
        Binary * bin = ErlDrvBinary2Binary(drvbin);
        if ( --(bin->refc) == 0) {
            EDTK_FREE(bin);
        }
    }
}


/* Rather than calling driver_output_term() directly, code generated by c_c_template.gsl often
 * calls driver_send_term(port, driver_caller(port), term, len).
 * In linked-driver mode, this send the reply directly to the Erlang process that send the
 * command to the port.
 * Of course, in pipe-mode, that is meaningless, and we have to send the reply to the
 * port's linked process.  However, we may arrange for that process to forward the reply
 * to the process that send the command (assuming that the stubs generated by erl_template.gsl
 * send the caller's pid, and we pass it back in term.
 * So in pipe-mode, driver_caller() does nothing and driver_send_term() just forwards to
 * driver_output_term() */
ErlDrvTermData driver_caller(ErlDrvPort port) {
    return 0;
}
int driver_send_term(ErlDrvPort port, ErlDrvTermData receiver, ErlDrvTermData* term, int n) {
    return driver_output_term(port, term, n);
}

int
driver_output_term(ErlDrvPort ix, ErlDrvTermData* data, int len)
{
    ErlDrvTermData	*ptr, *ptr_end, tag;
    size_t		bytes;
    char		bytesbuf[4];
    static char		*buf = NULL, *bufp;
    static int		bufsiz = 128 * 1024; /* initial/current buffer size, must be at least 4 */
    int			i, tmpi, res;
    ErlDrvBinary	*ebin;

    EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "start");

    if (buf == NULL) {
	if ((buf = (char *)EDTK_MALLOC(bufsiz)) == NULL) {
            return ENOMEM;
	}
    }

    bytes = 4;   /* space for length prefix */
    ptr = data;
    ptr_end = ptr + len;
    while (ptr < ptr_end) {
	tag = *ptr++;
	switch (tag) {
	case ERL_DRV_NIL:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: nil"); */
	    break;
	case ERL_DRV_ATOM:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: atom = %s", (char *) *ptr); */
	    bytes++;		/* name len, limited to 255! */
	    i = strlen((char *) *ptr);
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: i = %d", i); */
	    bytes += i;
	    ptr++;
	    break;
	case ERL_DRV_PORT:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: port = 0x%lx", *ptr); */
	    bytes += 4;
	    ptr++;
	    break;
	case ERL_DRV_INT:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: int = %ld", *ptr); */
	    bytes += 4;
	    ptr++;
	    break;
	case ERL_DRV_TUPLE:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: tuple size = %ld", *ptr); */
	    bytes++;		/* size of tuple, limited to 255! */
	    ptr++;
	    break;
	case ERL_DRV_LIST:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: list size = %ld", *ptr); */
	    bytes += 4;
	    ptr++;
	    break;
	case ERL_DRV_BINARY:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: binary/string 0x%lx + 0x%lx size = %ld", ptr[0], ptr[2], ptr[1]); */
	    bytes += 4;		/* Length indicator */
	    bytes += ptr[1];
	    ptr += 3;
	    break;
	case ERL_DRV_STRING:
	    bytes++;		/* Tag */
	    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: binary/string 0x%lx size = %ld", ptr[0], ptr[1]); */
	    bytes += 4;		/* Length indicator */
	    bytes += ptr[1];
	    ptr += 2;
	    break;
	default:
	    EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: UNKNOWN tag %lu", tag);
            return EINVAL;
	}
	/* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: bytes = %ld", bytes); */
    }
    /* EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "bc: total = %ld", bytes); */

    if (bytes > bufsiz) {
        int    newbufsiz = bufsiz + bytes;   /* upto 2x size factor*/
        char * newbuf    = realloc(buf, newbufsiz);
        if ( newbuf == NULL ) {
            EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "buf too small (%d) and realloc to size %d failed", bufsiz, newbufsiz);
            return ENOMEM;
        }
        buf    = newbuf;
	bufsiz = newbufsiz;
	return driver_output_term(ix, data, len); /* Lazy */
    }

    bufp = buf;

    /* length prefix */
    *(unsigned long *) bytesbuf = htonl(bytes - 4);   /* length does NOT include length of prefix */
    memcpy(bufp, (char *) bytesbuf, 4);
    bufp += 4;

    ptr = data;
    ptr_end = ptr + len;
    while (ptr < ptr_end) {
	tag = *ptr++;
	switch (tag) {
	case ERL_DRV_NIL:
	    *bufp++ = PIPE_DRIVER_TERM_NIL;
	    break;
	case ERL_DRV_ATOM:
	    *bufp++ = PIPE_DRIVER_TERM_ATOM;
	    i = strlen((char *) *ptr);
	    *bufp++ = i;
	    memcpy(bufp, (char *) *ptr, i);
	    bufp += i;
	    ptr++;
	    break;
	case ERL_DRV_PORT:
	    *bufp++ = PIPE_DRIVER_TERM_PORT;
	    *(unsigned long *) bytesbuf = htonl(*ptr);
	    memcpy(bufp, (char *) bytesbuf, 4);
	    bufp += 4;
	    ptr++;
	    break;
	case ERL_DRV_INT:
	    *bufp++ = PIPE_DRIVER_TERM_INT;
	    *(unsigned long *) bytesbuf = htonl(*ptr);
	    memcpy(bufp, (char *) bytesbuf, 4);
	    bufp += 4;
	    ptr++;
	    break;
	case ERL_DRV_TUPLE:
	    *bufp++ = PIPE_DRIVER_TERM_TUPLE;
	    *bufp++ = *ptr;
	    ptr++;
	    break;
	case ERL_DRV_LIST:
	    *bufp++ = PIPE_DRIVER_TERM_LIST;
	    tmpi = ntohl(*ptr);
	    memcpy(bufp, &tmpi, 4);
	    bufp += 4;
	    ptr++;
	    break;
	case ERL_DRV_BINARY:
	    *bufp++ = PIPE_DRIVER_TERM_BINARY;
	    tmpi = ntohl(ptr[1]);
	    memcpy(bufp, &tmpi, 4);
	    bufp += 4;
	    ebin = (ErlDrvBinary *) ptr[0]; 
	    memcpy(bufp, ebin->orig_bytes + ptr[2], ptr[1]);
	    bufp += ptr[1];
	    ptr += 3;
	    break;
	case ERL_DRV_STRING:
	    /* The pointers pushed on the stack differ from binary & string! */
	    *bufp++ = PIPE_DRIVER_TERM_STRING;
	    tmpi = ntohl(ptr[1]);
	    memcpy(bufp, &tmpi, 4);
	    bufp += 4;
	    memcpy(bufp, (char *) ptr[0], ptr[1]);
	    bufp += ptr[1];
	    ptr += 2;
	    break;
	default:
	    EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "UNKNOWN tag %lu.", tag);
            return EINVAL;
	}
    }
    EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "writing = %ul bytes", bytes);
    if ( (res = write(1, buf, bytes)) != bytes ) {
	EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "write(%uld) failed with %d.", bytes, res);
        return EINVAL;
    }

    return 0;
}

/* c_c_template.gsl generates code that refers to this -- see comments in that file.
 * This definition is just to satisfy the linker when in pipe mode.
 * (Actually, setting this to 0 _could_ be an alternative to setting pipe_mode_p
 * -- but this is a hack that depends on an internal Erlang VM symbol, so I don't
 * think that is a good idea */
int erts_async_max_threads = 0;

/* We have to emulate this erl_driver function (called from our driver_init()) */
void driver_system_info(ErlDrvSysInfo *sys_info, size_t size)
{
    if ( sizeof(ErlDrvSysInfo) != size ) {
        EDTK_LOG_AND_ABORT("size %ud != sizeof(ErlDrvSysInfo) %ud", size, sizeof(ErlDrvSysInfo));
    }
    sys_info->driver_major_version = ERL_DRV_EXTENDED_MAJOR_VERSION;
    sys_info->driver_minor_version = ERL_DRV_EXTENDED_MINOR_VERSION;
    sys_info->erts_version         = "pipe-main";           /* would be something like "5.5.3" */
    sys_info->otp_release          = "pipe-main";           /* would be something like "R11B"  */

    /* This says that we DON'T support Erlang-VM-style threadpool,
     * so the driver CANNOT use the driver_async() call.
     * Instead, EDTK-generated drivers may enabled \private_async_threadpool' in their XML files.
     * (Note: this replaces the old 'pipe_driver_p' global variable.) */
    sys_info->thread_support       = 0;

    /* We pretend that we ARE in SMP mode;
     * this means that driver_alloc/realloc/free_binary are thread-safe
     * (in pipe-mode we implement them ourself, below */
    sys_info->smp_support          = 1;
}

long driver_async(ErlDrvPort ix,
                  unsigned int* key,
                  void (*async_invoke)(void*), 
                  void* async_data,
                  void (*async_free)(void*))
{
    EDTK_LOG_AND_ABORT("This function should never be called in pipe-mode");
}

/*
** Since we always pass atoms by name, just return the pointer to the name.
** It's not our problem to worry about the longevity of the pointer....
*/

ErlDrvTermData
driver_mk_atom(char *name)
{
    EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "name = %s", name);
    return (ErlDrvTermData) name;
}

/*
** Just return the port number as-is.
*/

ErlDrvTermData
driver_mk_port(ErlDrvPort port)
{
    EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "port = 0x%p", (void*)port);
    return (ErlDrvTermData) port;
}


/*
** This is necessary for the private_async_threadpool feature.
** For that purpose it only supports ONE fd in DO_READ mode.
*/

unsigned int  num_fds_to_poll = 1;  /* 1 for stdin, but may be set to 2 by driver_select */

struct pollfd fds_to_poll[2] =
{
   {
       0,				       /* 0 is stdin */
       POLLIN | POLLERR | POLLHUP | POLLNVAL,  /* requested events */
       0				       /* returned events */
   },
   {
       -1,				       /* optionally installed by driver_select() */
       POLLIN | POLLERR | POLLHUP | POLLNVAL,  /* requested events */
       0				       /* returned events */
   },
};

int driver_select(ErlDrvPort port, ErlDrvEvent event, int mode, int on)
{
    if ( on )
    {
        if ( fds_to_poll[1].fd != -1 ) {
            EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "failed: an fd has already been installed (%d)", fds_to_poll[1].fd);
            return -1;
        }
        if ( mode != DO_READ ) {
            EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "failed: mode must be DO_READ");
            return -1;
        }        
        fds_to_poll[1].fd = (int)event;
        num_fds_to_poll = 2;
    }
    else
    {
        if ( fds_to_poll[1].fd != (int)event ) {
            EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "failed: attempt to uninstall wrong fd (%d but should be %d)", (int)event, fds_to_poll[1].fd);
            return -1;
        }
        if ( mode != DO_READ ) {
            EDTK_LOG(EDTK_LOG_PIPE_VERBOSE, "failed: mode must be DO_READ");
            return -1;
        }        
        fds_to_poll[1].fd = -1;
        num_fds_to_poll = 1;
    }
    return 0; /*success*/
}

    
