/*
** Copyright (c) 2003, Scott Lystig Fritchie.  All rights reserved.
** Some contributions Copyright (c) Chris Newcombe, 2006.  All rights reserved.
** See the file "../LICENSE" for license details.
*/

#include <stdarg.h>
#include <erl_driver_tk.h>


#ifdef DRIVER_USING_PTHREADS
pthread_mutex_t logging_output_global_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

FILE * edtk_log_file;

unsigned int edtk_log_mask = 0;



void *
edtk_driver_binary_malloc(size_t size)
{
    ErlDrvBinary *eb;

    if ((eb = driver_alloc_binary(size)) != NULL) {
	EDTK_LOG(EDTK_LOG_ALLOC_FREE, "size %zu eb = 0x%p orig_bytes = 0x%p", size, (void*)eb, (void*)(eb->orig_bytes));
	return eb->orig_bytes;
    } else {
	EDTK_LOG(EDTK_LOG_ALLOC_FREE, "failed");
	return NULL;
    }
}

void *
edtk_driver_binary_realloc(void *p, size_t new_size)
{
    ErlDrvBinary *eb, *neweb;

    if ( (eb = edtk_alloced_ptr2ErlDrvBinary(p)) != NULL )
    {
        int prev_size = eb->orig_size;
	if ( (neweb = driver_realloc_binary(eb, new_size)) != NULL )
        {
            EDTK_LOG(EDTK_LOG_ALLOC_FREE, "eb 0x%p, memory 0x%p, size %zu moved to eb 0x%p, memory 0x%p, new size %zu", (void*)eb, p, prev_size, (void*)neweb, (void *)(neweb->orig_bytes), new_size);
	    return neweb->orig_bytes;
	}
        else
        {
            if ( new_size )
            {
                EDTK_LOG(EDTK_LOG_ALLOC_FREE, "failed to resize eb 0x%p, memory 0x%p, size %zu to new size %zu", (void*)eb, p, prev_size, new_size);
            }
            else
            {
                EDTK_LOG(EDTK_LOG_ALLOC_FREE, "freed eb 0x%p, memory 0x%p, size %zu", (void*)eb, p, prev_size);
            }
	    return NULL;
	}
    }
    else
    {
        /* If p/eb is NULL, realloc acts like malloc */
	if ( (neweb = driver_alloc_binary(new_size)) != NULL )
        {
            EDTK_LOG(EDTK_LOG_ALLOC_FREE, "allocated new eb 0x%p, memory 0x%p, size %zu", (void*)neweb, (void*)(neweb->orig_bytes), new_size);
	    return neweb->orig_bytes;
	}
        else
        {
            EDTK_LOG(EDTK_LOG_ALLOC_FREE, "failed to allocate new binary of size %zu", new_size);
	    return NULL;
	}
    }
}

void
edtk_driver_binary_free(void *p)
{
    ErlDrvBinary * eb;
    if ( (eb = edtk_alloced_ptr2ErlDrvBinary(p)) != NULL )
    {
	EDTK_LOG(EDTK_LOG_ALLOC_FREE, "eb = 0x%p p = 0x%p", (void*)eb, (void*)p);
	driver_free_binary(eb);
    }
    else
    {
	EDTK_LOG(EDTK_LOG_ALLOC_FREE, "asked to free NULL");
    }
}

ErlDrvBinary *
edtk_alloced_ptr2ErlDrvBinary(void *p)
{
    ErlDrvBinary	stackeb;
    int			offset = stackeb.orig_bytes - (char *) &stackeb;

    if ( p != NULL )
    {
	EDTK_LOG(EDTK_LOG_VERBOSE, "p = 0x%p eb = 0x%p", (void*)p, (void *)((char *)p - offset));
	return (ErlDrvBinary *) ((char *) p - offset);
    }
    else
    {
	return NULL;
    }
}


/* malloc/realloc/free that do optional driver-level logging */

void * EDTK_MALLOC(size_t size)
{
    if ( edtk_log_mask & EDTK_LOG_ALLOC_FREE )
    {
        void * res;
        if ( (res = malloc(size)) != NULL )
        {
            EDTK_LOG_FORCE("malloc'ed size %zu at 0x%p", size, res);
            return res;
        }
        else
        {
            EDTK_LOG_FORCE("failed to malloc size %zu", size);
            return NULL;
        }
    }
    else
    {
        return malloc(size);
    }
}

void * EDTK_REALLOC(void * mem, size_t new_size)
{
   if ( edtk_log_mask & EDTK_LOG_ALLOC_FREE )
   {
       void * res;
       if ( (res = realloc(mem, new_size)) != NULL )
       {
           if ( mem != NULL )
           {
               EDTK_LOG_FORCE("moved block from 0x%p to 0x%p, with new size %zu", mem, res, new_size);
           }
           else
           {
               EDTK_LOG_FORCE("allocated new block of size %zu at 0x%p", new_size, res);
           }
           return res;
       }
       else
       {
           if ( new_size )
           {
               if ( mem != NULL )
               {
                   EDTK_LOG_FORCE("failed to resize block at 0x%p to new size %zu", mem, new_size);
               }
               else
               {
                   EDTK_LOG_FORCE("failed to alloc new block of size %zu", new_size);
               }
           }
           else
           {
               EDTK_LOG_FORCE("freed 0x%p", mem);
           }
           return NULL;
       }
   }
   else
   {
       return realloc(mem, new_size);
   }
}


/*
** Arg want_contiguous: if true, return error if the 'n' we're forwarding
** past are in a single contiguous buffer.
** 
** Return value:
**      -1 = error
**       0 = Success, there is no more data to be read
**       1 = Success
*/

int
edtk_ev_forward_N(ErlIOVec *ev, int n, int *pp, int *qp, int want_contiguous)
{
    int	forward_pos = *pp + n;

    if ((*qp) >= ev->vsize) {
	return -1;
    }
    if (forward_pos < ev->iov[*qp].iov_len) {
	*pp += n;
	return 1;
    } else if (forward_pos == ev->iov[*qp].iov_len) {
	(*qp)++;
	*pp = 0;
	if ((*qp) < ev->vsize) {
	    return 1;
	} else {
	    return 0;
	}
    } else {
	n -= (ev->iov[*qp].iov_len - *pp);
	(*qp)++;
	(*pp) = 0;
	if ((*qp) < ev->vsize && ! want_contiguous) {
	    return edtk_ev_forward_N(ev, n, pp, qp, want_contiguous);
	} else {
	    return -1;
	}
    }
}

