/*
** Copyright (c) 2003, Scott Lystig Fritchie.  All rights reserved.
** Some contributions Copyright (c) Chris Newcombe, 2006.  All rights reserved.
** See the file "../LICENSE" for license details.
*/

/* pipe_main.c */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <erl_driver.h>
#include <erl_driver_tk.h>
#include <erl_driver_pipelib.h>
#include <errno.h>
#include <memory.h>
#include <limits.h>
#include <unistd.h>
#include <assert.h>


extern unsigned int edtk_driver_log_level;  /* in erl_driver_tk.h/.c */


static input_iovec_wrapper_t ** delayed_free_input_iovecs = NULL;
static int                      num_delayed_free_slots = 0;
static int                      num_used_delayed_free_slots = 0;

static const int grow_by_slots = 100;

void register_for_delayed_free(input_iovec_wrapper_t * input_iovec_p)
{
    input_iovec_wrapper_t ** p;

    /* Do we need to make some room? (This will happen at least on the very first call) */
    if ( num_used_delayed_free_slots == num_delayed_free_slots )
    {
        if ( (delayed_free_input_iovecs = EDTK_REALLOC(delayed_free_input_iovecs,
                                                       sizeof(delayed_free_input_iovecs[0])
                                                       * (grow_by_slots + num_delayed_free_slots)))
             == NULL )
        {
            /* Note that we overwrote the old value of delayed_free_input_iovecs, 
             * so that would be a memory-leak if we weren't aborting immediately. */
            EDTK_LOG_AND_ABORT("realloc failed");
        }
        memset(&delayed_free_input_iovecs[num_delayed_free_slots],
               0,
               grow_by_slots * sizeof(delayed_free_input_iovecs[0]));
        
        num_delayed_free_slots += grow_by_slots;
        EDTK_LOG(EDTK_LOG_VERBOSE, "grown to %d slots", num_delayed_free_slots);
    }

    /* Find a free slot (there must be one).
     * (This array should stay fairly empty, so there is probably not much value
     * in keeping a marker on the last-free slot and searching from there.) */
    p = delayed_free_input_iovecs;
    while ( *p != NULL ) {
        ++p;
    }
    assert( p < &delayed_free_input_iovecs[num_delayed_free_slots] );
    
    *p = input_iovec_p;
    ++num_used_delayed_free_slots;
    EDTK_LOG(EDTK_LOG_VERBOSE, "marked 0x%p for delayed free (now using %d slots out of %d)", (void*)input_iovec_p, num_used_delayed_free_slots, num_delayed_free_slots);
}

void free_any_delayed_items_that_are_ready()
{
    input_iovec_wrapper_t ** p    = delayed_free_input_iovecs;
    int num_used_slots_to_examine = num_used_delayed_free_slots;

    while ( num_used_slots_to_examine != 0 )
    {
        /* Find the next used slot (we know there is one) */
        while ( *p == NULL ) {
            /* EDTK_LOG(EDTK_LOG_VERBOSE, "slot %d is NULL (using %d slots out of %d)", p - delayed_free_input_iovecs, num_used_delayed_free_slots-1, num_delayed_free_slots); */
            ++p;
        }
        assert( p < &delayed_free_input_iovecs[num_delayed_free_slots] );
        if ( (*p)->input_data.refc == 1 )
        {
            EDTK_LOG(EDTK_LOG_VERBOSE, "freeing slot #%d (0x%p) (now using %d slots out of %d)", p - delayed_free_input_iovecs, (void*)*p, num_used_delayed_free_slots-1, num_delayed_free_slots);
            EDTK_FREE(*p);
            *p = NULL;
            assert(num_used_delayed_free_slots != 0);
            --num_used_delayed_free_slots;
        }
        else {
            EDTK_LOG(EDTK_LOG_VERBOSE, "NOT freeing slot #%d (0x%p) as refc=%ld (now using %d slots out of %d)", p - delayed_free_input_iovecs, (void*)*p, (*p)->input_data.refc, num_used_delayed_free_slots - 1, num_delayed_free_slots);
        }
        ++p;
        --num_used_slots_to_examine;
        /* we have not stepped past the end of the array (OR we have finished the loop and the while statement is about to exit) */
        assert( (p < &delayed_free_input_iovecs[num_delayed_free_slots]) || (num_used_slots_to_examine == 0) );
    }
}

                    
int
main(int argc, char *argv[])
{
    void *                  libhandle;
    ErlDrvEntry *           (*driver_init)(void *);
    ErlDrvEntry *           driver_entry;
    ErlDrvData              driver_data;
    input_iovec_wrapper_t * input_iovec_p;
    int                     res;
    int                     i;
    int                     num_fds_with_returned_events;

    edtk_log_mask = 0;
    
    if (argc != 2) {
	fprintf(stderr, "usage: %s path-to-driver-shared-lib (pipe version)\r\n", argv[0]);
	exit(1);
    }
    EDTK_LOG(EDTK_LOG_INIT, "Attempting to dlopen");
    if ((libhandle = dlopen(argv[1], RTLD_NOW)) == NULL) {
	fprintf(stderr, "dlopen failed on '%s'", argv[1]);
	perror(dlerror());
	exit(2);
    }

    if ((driver_init = (ErlDrvEntry* (*)(void *)) dlsym(libhandle, "driver_init")) == NULL) {
	fprintf(stderr, "Cannot find 'driver_init', exiting\r\n");
	exit(3);
    }
    EDTK_LOG(EDTK_LOG_INIT, "calling 'driver_init'");
    if ((driver_entry = (*driver_init)(NULL)) == NULL) {
	fprintf(stderr, "'driver_init' returned NULL\r\n");
	exit(4);
    }
    EDTK_LOG(EDTK_LOG_INIT, "calling 'init'");
    if ( (res = (*driver_entry->init)()) != 0) {
	fprintf(stderr, "Driver's 'init' function did not return 0 (%d)\r\n", res);
	exit(5);
    }
    EDTK_LOG(EDTK_LOG_INIT, "calling 'start'");
    driver_data = (*driver_entry->start)((ErlDrvPort)-1, "Call from pipe-main");
    if ( driver_data == ERL_DRV_ERROR_GENERAL ) {
	fprintf(stderr, "Driver's 'start' function returned ERL_DRV_ERROR_GENERAL\r\n");
        exit(6);
    } else if ( driver_data == ERL_DRV_ERROR_ERRNO ) {
	fprintf(stderr, "Driver's 'start' function returned ERL_DRV_ERROR_ERRNO (%d)\r\n", errno);
        exit(7);
    } else if ( driver_data == ERL_DRV_ERROR_BADARG ) {
	fprintf(stderr, "Driver's 'start' function returned ERL_DRV_ERROR_BADARG\r\n");
        exit(8);
    }

    EDTK_LOG(EDTK_LOG_INIT, "entering polling-for-input loop");

    for (;;)
    {
        for (i=0; i<num_fds_to_poll; ++i) {
            fds_to_poll[i].revents = 0;
        }
        /* (timeout is very large) */
        if ( 0 == (num_fds_with_returned_events = poll(fds_to_poll, num_fds_to_poll, INT_MAX)) ) {
            /* timeout occured -- just loop */
            continue;
        }

        /* stdin */
        if ( fds_to_poll[0].revents & (POLLERR | POLLHUP | POLLNVAL) )
        {
            if ( fds_to_poll[0].revents & POLLHUP ) {
                /* too noisy: EDTK_LOG(EDTK_LOG_VERBOSE, "stdin closed from Erlang side (graceful shutdown)"); */
            }
            else {
                EDTK_LOG(EDTK_LOG_SHUTDOWN, "Error event on stdin (%d), exiting", (int)fds_to_poll[0].revents);
            }
            break;
        }
        if ( fds_to_poll[0].revents & POLLIN )
        {
            if ( (input_iovec_p = read_packet_make_input_iovec(0)) == NULL ) {
                /* error in input, or exit */
                break;
            }

            /* Process the command */
            (*(driver_entry->outputv))(driver_data, &input_iovec_p->ev);

            if ( input_iovec_p->input_data.refc == 1 ) {
                EDTK_FREE(input_iovec_p);
            }
            else {
                /* outputv has incremented the refcount (so that ready_async can access the input data).
                 * Therefore we have to delay freeing this, until it's refcount drops to 1 */
                register_for_delayed_free(input_iovec_p);
            }
        }

        /* fd installed by driver_select (for private threadpool) */
        if ( num_fds_to_poll > 1 )
        {
            if ( fds_to_poll[1].revents & (POLLERR | POLLHUP | POLLNVAL) ) {
                EDTK_LOG(EDTK_LOG_SHUTDOWN, "Error event on driver_select fd (%d), exiting", (int)fds_to_poll[1].revents);
                break;
            }
            if ( fds_to_poll[1].revents & POLLIN )
            {
                if ( driver_entry->ready_input )
                {
                    EDTK_LOG(EDTK_LOG_VERBOSE, "fd[1] is readable so calling entry->ready_input");
                    (*(driver_entry->ready_input))(driver_data, (ErlDrvEvent)(fds_to_poll[1].fd));

                    free_any_delayed_items_that_are_ready();
                }
                else
                {
                    char buff[1024];
                    EDTK_LOG(EDTK_LOG_VERBOSE, "fd[1] is readable but driver_entry->ready_input is NULL!");

                    /* Discard (most of) the readable data, so we don't busy-spin.
                     * It is NOT correct to loop on read until all of the data is gone, because
                     * if there happened to be one exact buffer full, the next read will block.
                     * We would need to put the fd into non-blocking mode and loop until we get
                     * EWOULDBLOCK, and that's too fiddly for the benefit (I don't know how portable it is) */
                    read(fds_to_poll[1].fd, buff, sizeof(buff));
                }
            }
        }
    }

    EDTK_LOG(EDTK_LOG_VERBOSE, "calling driver's 'stop' function");
    (*driver_entry->stop)(driver_data);

    EDTK_LOG(EDTK_LOG_SHUTDOWN, "freeing final pipe-main resources");
    /* Need to do this one last time after calling stop */
    free_any_delayed_items_that_are_ready();
    assert(num_used_delayed_free_slots == 0);
    
    EDTK_LOG(EDTK_LOG_SHUTDOWN, "closing dynamic library");
    dlclose(libhandle);

    EDTK_LOG(EDTK_LOG_SHUTDOWN, "exiting");
    exit(0);
}
