%%%-------------------------------------------------------------------
%%% File    : berkeley_db_global_transaction_manager.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%    In our model of distributed transactions the GTM 
%%%    is a passive store of the authoratative status of eacn 
%%%    distributed transaction.
%%%
%%%    ?BDB_RG_H:do_distributed_txn calls this to get 
%%%    a GUID for the transaction, and commit the txn
%%%    (the GTM uses the 'presumed abort' optimization, so 
%%%    the caller never needs to explicitly abort).
%%%
%%%    berkeley_db_port_coordinator calls this to determine how to 
%%%    resolve distributed transactions 
%%%
%%%    - at startup after recovery, if txn_recover finds any 
%%%      prepared distributed transactions
%%%
%%%    - if a client process crashes while owning a prepared txn
%%%      (this is detected by group_valmap_cleanup, and the C code 
%%%      sends an event (containing the txn's GUID) to the port-coordinator).
%%%
%%%    - after a site becomes master (after implicit recovery), 
%%%      if txn_recover finds any prepared distributed transactions
%%%
%%%
%%%    Notes on Weakly Replicated Applications
%%%
%%%    It is essential that the GTM (and it's persistent data) be
%%%    highly available, as if a participant in a distributed
%%%    transaction crashes when it has a prepared transaction, it will
%%%    block forever until the GTM tells it how to resolve the
%%%    transction.  Therefore we need to replicate the GTM's data. 
%%%
%%%    We currently achieve that entirely in **caller** code.
%%%
%%%    The GTM is an example of a 'weakly replicated application'.  It
%%%    has no notion of replication or replication groups itself
%%%    (other than the single helper function
%%%    get_proc_name_from_replication_group_site_name() which is not
%%%    used by the GTM, and only located in this module as a
%%%    convenience).
%%%
%%%    Instead, the GTM is written as a non-replicated applicated, and
%%%    replication is achieved by a few conventions in the *caller*.
%%%
%%%    i.e. 'Weakly replicated applications' themselves *only* use
%%%    do_txn for their internal operations.  But such applications
%%%    are *called* via do_txn_on_master.  The *caller* uses a TxnFun
%%%    (passed to do_txn_on_master) that knows it is hitting a
%%%    replication master, and uses that knowledge to find/construct
%%%    the name of the local server for the 'weakly replicated
%%%    application' (there is a convention for registering such names
%%%    -- see below). The TxnFun then calls the necessary API on that
%%%    local server, which performs normal BDB transactions, which are
%%%    transparently replicated.
%%%
%%%    See ?BDB_RG_H:do_distributed_txn for examples.
%%%
%%%    Details of the conventions:
%%%
%%%    1. The top-level application choose a replication group for the
%%%       GTM to use. 
%%%
%%%       *CRITICAL* The GTM MUST NOT be hosted on a replication group
%%%       that will also be a participant in distributed transactions
%%%       managed by that GTM.  Otherwise a catastrophic deadlock
%%%       (chicken-and-egg) situation can arise in recovery, if the
%%%       entire group fails with unresolved transactions:
%%%        
%%%       a. the group restarts and a site is elected master 
%%%       b. the master port-coordiantor runs txn_recovery and finds an
%%%          unresolved txn 
%%%       c. the master port-coordinator calls the GTM
%%%          to resolve the txn, using do_txn_on_master (which of
%%%          course will be the current site) 
%%%       d. do_txn_on_master calls BDB_PC:acquire_bdb_port_handle, 
%%%          which will timeout as the port-coordinator is tied up doing 
%%%          recovery. Even if recovery was done in a different process, BDB 
%%%          would reject the write that is likely to be done when resolving a
%%%          presumed-abort transaction)
%%%        
%%%       Note that the GTM can use a very lightweight replication group
%%%       (i.e. configured to only a few threads), alongside a 'normal'
%%%       replication group on some chosen cluster of hosts.  
%%%        
%%%       e.g. If the application is a partitioned data-store, then they
%%%       application could represent each 'weakly replicated
%%%       application' as constant key -- e.g. "GTM" -- and run that key
%%%       through the partitioning function (Key -> ReplicationGroupName)
%%%       to locate the application.
%%%
%%%    2. At application init (immediately after replication_group
%%%    servers have been started) the application must start (and
%%%    supervise) an instance of the GTM server for each site in the
%%%    chosen replication group.  
%%%
%%%       - Each GTM instance is told to locally register a process
%%%       name based on the replication group name and site name.
%%%       The function get_proc_name_from_replication_group_site_name()
%%%       is provided as a convenience for this.  (Alternatively, 
%%%       start_link could take a replication group name and site-name,
%%%       but that then binds the GTM to the concept of a replication 
%%%       group, and that should only be done if the binding is 
%%%       fully implemented -- e.g. the replication_group starts 
%%%       and supervises the GTM).
%%%        
%%%       (Note: if several of the group's sites run on the same node
%%%       -- e.g. during development/testing -- then there will be
%%%       several named GTM instances on that node).
%%%        
%%%       - Each GTM instance is told to use the named port-coordinator
%%%       for that group+site to access its persistent state.
%%%
%%%    3. Each node in the application maintains (probably in an ETS
%%%       table) an instance of
%%%       #berkeley_db_replication_group_caller_state{} for the
%%%       replication group that is associated with the GTM.
%%%       (Partitioned applications will have a caller_state for each
%%%       partition replication group, and don't need a special copy
%%%       for the GTM).
%%%
%%%    4. To perform a distributed transaction, the application calls
%%%       ?BDB_RG_H:do_distrbuted_txn, passing the caller_state for
%%%       the replication group that is associated with the GTM.
%%%
%%%    5. ?BDB_RG_H:do_distrbuted_txn always calls the GTM from
%%%       TxnFuns that it executes via do_txn_on_master, using the
%%%       caller_state passed by the application to find that
%%%       replication group/master.
%%%
%%%       These are the TxnFuns that actually 'call' the replicated
%%%       application.  Their first job is to find the GTM server for
%%%       the master (replication group + site) that is handling the
%%%       request. This is done by the following example:
%%%
%%%       txn_fun_get_new_distributed_txn_guid(BdbPort, _) ->
%%%        
%%%           {ReplicationGroupName, SiteName} = get(txn_fun_replication_context),
%%%           GTM_ServerName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(ReplicationGroupName, 
%%%                                                                                   SiteName),
%%%           ?BDB_GTM:get_new_distributed_txn_guid(GTM_ServerName).
%%%
%%% The upside of this 'weakly replicated' approach is extreme
%%% simplicitly -- the 'target' application need have no notion of
%%% replication at all, so long as it follows a few basic rules (using
%%% a named port coordinator, each instance a stand-alone server).
%%%
%%% The downside is that the calling application must know about (and
%%% indeed track changes in) the location of the GTM.  Ideally the
%%% application wouldn't even know that the GTM existed.
%%%
%%% It is arguable that there should be a stronger encapsulation of a
%%% 'replicated application'.  e.g. berkeley_db_replication_group
%%% might be configured witn a list of the replicated applications
%%% that are running on its group, and be responsible for starting &
%%% supervising them.  However, that still doesn't solve the problem
%%% of how all possible *application* (caller) sites have an initial
%%% correct copy of RgCallerState.  Without a global registry (and all
%%% of the problems that implies), that is difficult to do.
%%% At the moment I don't consider the extra code to be worth the 
%%% relatively benefit. The current burdens on the application are 
%%% not too onerous.  It would have to choose a replication group for 
%%% the GTM anyway, so actually adding it to the supervisor tree is 
%%% only a few extra lines of code.  The main burden is tracking 
%%% the GTM's caller_state per node -- and that's only a few simple 
%%% lines of code.
%%%
%%%
%%% APIs    
%%%
%%% get_new_distributed_txn_guid()
%%%    Returns a 128-byte binary, suitable for use with BDB's txn_prepare call.
%%%    This uses a berkeley_db_sequence_server to generate 64-bit 
%%%    unique ids. Currently the rest of the 128-byte guid is all 0's.
%%%    If we ever support multiple GTMs then each GTM will need a unique 
%%%    ID of its own, and that ID will also need to be placed in the guid.
%%%
%%% commit(DistTxnGUID)
%%%    If and only if the transaction has not been recorded as aborted 
%%%    this writes 'committed' to stable storage, and if that succeeds,
%%%    it returns ok, and the caller should tell the participants to do 
%%%    local commit.
%%%
%%%    If this returns {error, {GUID, already_aborted}} then the
%%%    caller should tell all participants to do local abort.
%%%
%%%    If this returns any other error or times-out then the
%%%    transaction MAY still have been committed.  The caller must NOT
%%%    not tell partiticpants to abort (or commit), and should call
%%%    resolve().
%%%
%%% abort(DistTxnGUID)
%%%    If and only if the transaction has not been recorded as committed 
%%%    this writes 'aborted' to stable storage, and if that succeeds,
%%%    it returns ok, and the caller should tell the participants to do 
%%%    local aborted.
%%%
%%%    If this returns {error, {GUID, already_committed}} then the
%%%    caller should tell all participants to do local commit, and
%%%    should probably ask a human for help, as this should never
%%%    happen.
%%%
%%%    If this returns any other error or times-out then the
%%%    transaction MAY still have been aborted.  The caller must NOT
%%%    not tell partiticpants to abort (or commit), and should call
%%%    resolve().
%%%
%%% resolve(DistTxnGUID or ListDistTxnGUIDs)
%%%    If successful, returns {ok, {GUID, committed|aborted}}.
%%%
%%%    IMPORTANT: If no record exists in the database for this
%%%    transaction , we use the 'presumed abort' optimization.  So
%%%    calling 'resolve' must implicitly abort such a transaction.
%%%    This is necessary because of a race condition exists: It is
%%%    possible that a participant failed and recovered, and is asking
%%%    for a decision without the transaction's coordinator being
%%%    aware of the fact.  So the transaction coordinator may later
%%%    calling commit() on the same transaction.  To guard against
%%%    this, the the first call to resolve() on a txn for which we
%%%    have no record on disk will write 'aborted' to disk before it
%%%    returns to the caller.  This causes any subsequent call to
%%%    commit() to fail with {error, already_aborted}.
%%%
%%%    If resolve() returns an error or times-out then the result is
%%%    implicitly UNKNOWN and the caller must NOT locally abort (or
%%%    commit).  They should retry the call to resolve until it
%%%    succeeds (and ask a human for help).
%%%
%%%    This is why it is ESSENTIAL that the GTM is highly-available.
%%%    Another alternative is for all participant port_coordinators to
%%%    locally record the decision for the most recent N distributed
%%%    transactions, and the {Node,Name} of each participant.  Then if
%%%    the GTM is not available, any participant may ask any other for
%%%    the decision (and we are only blocked if no-one knows what that
%%%    decision is).  However, is much easier to just use BDB-HA for
%%%    the GTM, and an appropriately high number of replicas.
%%%
%%% TODO: 
%%% - It is ESSENTIAL that a GTM is highly available. 
%%%   Therefore this needs to use berkeley_db_replication_group.
%%% 
%%% - Adding (extremely conservative) purging of old state records based 
%%%   on timestamps.
%%%
%%% - Add retry-limits for deadlock handling (not done now as this will 
%%%   change completely when replication is used -- deadlock retries 
%%%   will be automatic then)
%%%-------------------------------------------------------------------
-module(berkeley_db_global_transaction_manager).

-behaviour(gen_server).

-include("berkeley_db_drv.hrl").  %% required for DB_XIDDATASIZE
-define(BDB,     berkeley_db_drv).
-define(BDB_H,   berkeley_db_helpers).
-define(BDB_U,   berkeley_db_utilities).
-define(BDB_PC,  berkeley_db_port_coordinator).
-define(BDB_SEQ, berkeley_db_sequence_server).

%% API
-export([start_link/2, 
         start_link/3,
         stop/1,
         get_new_distributed_txn_guid/1,
         commit/2,
         abort/2,
         resolve/2,
         get_proc_name_from_replication_group_site_name/2  % convenience function
        ]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% Internal APIs exported for use by do_txn
-export([txn_fun_commit_or_abort/2, txn_fun_resolve_one_txn/2, txn_fun_resolve_multiple_txns/2]).


%% This is read-only state (I would normally keep it in the process dictionary, but as we have so little mutable state,
%% I keep these for speed).
-record(state, {port_coordinator_name, sequence_server_pid, num_deadlock_retries, txn_states_db_id}).

%% These are the terms (plus an erlang:now timestamp) that are stored on disk.
%% Obviously we only need a single byte.
-define(COMMITTED_IN_BDB_RECORD, c).
-define(ABORTED_IN_BDB_RECORD,   a).

-define(GTM_DB_NAME,             "global_transaction_manager.db").

%% These config options have no defaults:
%%
%%    {port_coordinator_name, Name}   cannot default to 'berkeley_db_port_coordinator' as GTM should be used with a replication group.
%%

-define(DEFAULT_CONFIG_PROPLIST, [{num_deadlock_retries,            5},
                                  {persist_every_N_consumed_values, 100},
                                  {txn_states_db_id,                {?GTM_DB_NAME, "txn_states"}},
                                  {meta_data_db_id,                 {?GTM_DB_NAME, "meta_data"}}]).


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).


%%====================================================================
%% API
%%====================================================================

-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}                    -> Result;
               {error, {txn_fun_error, Error}} -> throw({bdb_gtm_error, Error});
               {error, Error}                  -> throw({bdb_gtm_error, Error})
           end).

%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link(ServerName, PortCoordinatorName) ->
    start_link(ServerName, PortCoordinatorName, []).

start_link(ServerName, PortCoordinatorName, ConfigProplist) ->
    gen_server:start_link({local, ServerName}, 
                          ?MODULE, 
                          [{port_coordinator_name, PortCoordinatorName} | ConfigProplist],
                          []).

stop(Server) ->
    gen_server:cast(Server, stop).

get_new_distributed_txn_guid(Server) ->
    ?THROW_ERRORS( gen_server:call(Server, get_new_distributed_txn_guid) ).
    
commit(Server, DistributedTxnGUID) ->
    ?THROW_ERRORS( gen_server:call(Server, {commit, DistributedTxnGUID}) ).

abort(Server, DistributedTxnGUID) ->
    ?THROW_ERRORS( gen_server:call(Server, {abort, DistributedTxnGUID}) ).

resolve(Server, DistributedTxnGUID) ->
    ?THROW_ERRORS( gen_server:call(Server, {resolve, DistributedTxnGUID}) ).


get_proc_name_from_replication_group_site_name(ReplicationGroupName, SiteName) ->
    %% This intentionally does not follow the scheme for replication_group servers and 
    %% port_coordinators owned by replication group servers. (They have the 
    %% replication group name and site name before the type of the server.
    %% The GTM has the type of the server before the replication group name and 
    %% site name, to show that the GTM just happens to be hosted/using this 
    %% particular replication group -- it's not a component of all replication groups).
    list_to_atom("bdb_GTM_" ++ atom_to_list(ReplicationGroupName) ++ "_" ++ atom_to_list(SiteName)).


%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%
%%--------------------------------------------------------------------
init(SpecifiedConfigProplist) ->

    ?INFO_LOGGER("starting (~w)~n", [self()]),

    %% Append the default config to the list specified by the caller, and extract all of the first occurences (the final config).
    ConfigProplist = ?BDB_U:get_dominant_proplist_values([SpecifiedConfigProplist,
                                                          ?DEFAULT_CONFIG_PROPLIST]),

    %% Store the final config in the process dictionary so that it will be logged if the server crashes.
    put(config_proplist, ConfigProplist),

    %% Display the final config.
    lists:foreach(fun({Key, Value}) -> ?INFO_LOGGER_NO_TIMESTAMP("init: ~32w = ~200p~n", [Key, Value]) end, lists:keysort(1, ConfigProplist)),

    %% Validate the config options
    lists:foreach(fun({port_coordinator_name,           Name})               when is_atom(Name)                      -> ok;
                     ({num_deadlock_retries,            N})                  when is_integer(N), N >= 0              -> ok;
                     ({persist_every_N_consumed_values, N})                  when is_integer(N), N >= 0              -> ok;
                     ({txn_states_db_id,                Filename})           when is_list(Filename)                  -> ok;
                     ({txn_states_db_id,                {Filename, DBName}}) when is_list(Filename), is_list(DBName) -> ok;
                     ({meta_data_db_id,                 Filename})           when is_list(Filename)                  -> ok;
                     ({meta_data_db_id,                 {Filename, DBName}}) when is_list(Filename), is_list(DBName) -> ok;
                     (Other) -> throw({unrecognized_config_option, Other})
                  end,
                  ConfigProplist),

    %% Start the sequence_server, with a name based on our name.
    %% Note: we DON't trap exits.  If sequence_server crashes then so do we (so we don't have to supervise it).
    {registered_name, GTMProcName} = process_info(self(), registered_name),

    %% sequence_server takes these config items in the same form as GTM
    {value, PersistEvery}   = lists:keysearch(persist_every_N_consumed_values, 1, ConfigProplist),
    {value, DeadlocRetries} = lists:keysearch(num_deadlock_retries, 1, ConfigProplist),

    {ok, SeqServerPid} = ?BDB_SEQ:start_link(list_to_atom(atom_to_list(GTMProcName) ++ "_sequence_server"),
                                             proplists:get_value(port_coordinator_name, ConfigProplist), 
                                             proplists:get_value(meta_data_db_id, ConfigProplist),
                                             <<"next_distributed_txn_counter">>,
                                             [PersistEvery, DeadlocRetries]),

    %% I'd normally put all of this immutable state in the process dictionary,
    %% to avoid creating superfluous garbage whenever State is modified.
    %% But the GTM never modifies State, so I put these in State for speed.

    {ok, #state{port_coordinator_name = proplists:get_value(port_coordinator_name, ConfigProplist),
                sequence_server_pid   = SeqServerPid,
                num_deadlock_retries  = proplists:get_value(num_deadlock_retries, ConfigProplist),
                txn_states_db_id      = proplists:get_value(txn_states_db_id,     ConfigProplist)}}.


%%--------------------------------------------------------------------
%% Function: handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(get_new_distributed_txn_guid, _From, State) ->

    TxnCounter = ?BDB_SEQ:consume_one_value(State#state.sequence_server_pid),

    DistributedTxnGUID = <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>>,
    %%?INFO_LOGGER("get_new_distributed_txn_guid: returning ~w~n", [TxnCounter]),
    {reply, {ok, DistributedTxnGUID}, State};


handle_call({commit, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID}, _From, State) ->
    {reply, 
     ?BDB_H:do_txn_no_throw(State#state.port_coordinator_name,
                            {?MODULE, txn_fun_commit_or_abort, {?COMMITTED_IN_BDB_RECORD, TxnCounter, DistributedTxnGUID, State}},
                            State#state.num_deadlock_retries),
     State};


handle_call({abort, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID}, _From, State) ->
    %% TODO: we don't really need to write an abort record on explicit
    %% abort -- we could just use the presumed-abort feature.
    %% Currently we only write it because it is less code and we
    %% rarely (perhaps never) explicitly abort, (we have to read the
    %% record anyway, to see if it has already been explicitly
    %% committed).
    %% (Note: as described elsewhere, we DO need to write an abort
    %% record if resolve is called on a txn with no current record, to
    %% prevent a later attempt to commit that record.)
    {reply, 
     ?BDB_H:do_txn_no_throw(State#state.port_coordinator_name,
                            {?MODULE, txn_fun_commit_or_abort, {?ABORTED_IN_BDB_RECORD, TxnCounter, DistributedTxnGUID, State}},
                            State#state.num_deadlock_retries),
     State};


handle_call({resolve, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID}, _From, State) when is_binary(DistributedTxnGUID) ->    
    Reply = ?BDB_H:do_txn_no_throw(State#state.port_coordinator_name,
                                   {?MODULE, txn_fun_resolve_one_txn, {TxnCounter, DistributedTxnGUID, State}},
                                   State#state.num_deadlock_retries),

    %%?INFO_LOGGER("resolve: for ~w returning ~p~n", [TxnCounter, Reply]),
    {reply, Reply, State};


handle_call({resolve, ListDistributedTxnGUIDs}, _From, State) when is_list(ListDistributedTxnGUIDs) ->    
    Reply = ?BDB_H:do_txn_no_throw(State#state.port_coordinator_name,
                                   {?MODULE, txn_fun_resolve_multiple_txns, {ListDistributedTxnGUIDs, State}},
                                   State#state.num_deadlock_retries),

    %%?INFO_LOGGER("resolve (multiple): returning ~p~n", [Reply]),
    {reply, Reply, State}.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_info(Unrecognized, _State) ->
    exit({unrecognized_message, Unrecognized}).

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, #state{sequence_server_pid = SeqServerPid}) ->

    ?INFO_LOGGER("terminate: (reason: ~p)~n", [Reason]),

    %% Shut down support processes cleanly, and wait for them to exit (or 
    %% they will receive exit signals when this process exits after terminate returns).
    %% Important: the support processes might *already* have died
    %% and we might already have eaten the EXIT message.
    %% Fortunately erlang:monitor does what we want -- we will get a DOWN message
    %% even if the process is not running when we try to monitor it.

    SeqServerMonitorRef = erlang:monitor(process, SeqServerPid),
    ?BDB_SEQ:stop(SeqServerPid),

    ?INFO_LOGGER("terminate: waiting for sequence_server ~p to exit~n", [SeqServerPid]),
    receive {'DOWN', SeqServerMonitorRef, process, _, _} -> ok end,
    ?INFO_LOGGER("terminate: got exit for global_transaction_manager_sequence_server~n"),

    ?INFO_LOGGER("terminate: success~n"),
    ok.


%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%% TxnFuns
%%
%% NewTxnStateInBdbRecord must be one of 
%%    ?COMMITTED_IN_BDB_RECORD
%%    ?ABORTED_IN_BDB_RECORD
%%
%% This function will ONLY write NewTxnState if there is no existing record for 
%% that transaction.  It will never overwrite any existing state
%% (will return {error, {already_committed|already_aborted, DistTxnGUID}}
%%
%% The non-overwrite guarantee so allows 'resolve' to do a separate read first, 
%% (outside of this function), without having to worry about an explicit commit happening 
%% between the read and calling this function to write an abort record 
%% (it does the latter if no record is found).
%%
txn_fun_commit_or_abort(BdbPort, {NewTxnStateForBdbRecord, TxnCounter, DistributedTxnGUID, State}) ->

    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, State#state.txn_states_db_id),
    do_commit_or_abort(BdbPort, DB, NewTxnStateForBdbRecord, TxnCounter, DistributedTxnGUID).


do_commit_or_abort(BdbPort, DB, NewTxnStateForBdbRecord, TxnCounter, DistributedTxnGUID) ->

    %%?INFO_LOGGER("txn_fun_commit_or_abort: called with '~w' for txn ~w~n", [NewTxnStateForBdbRecord, TxnCounter]),
    Txn    = ?BDB:txn_begin(BdbPort, void, []),
    Cursor = ?BDB:db_cursor(BdbPort, DB, Txn, []),

    Key = <<TxnCounter:64>>,

    %% Important: we take a read-modify-write lock to prevent race conditions while we 
    %% examine any existing record for this transaction.
    Reply = 
        try ?BDB:c_get(BdbPort, Cursor, Key, <<>>, [bdb_DB_SET, bdb_DB_RMW]) of 

            {Key, ValueBinary} -> 
                case binary_to_term(ValueBinary) of

                    {?ABORTED_IN_BDB_RECORD, _} ->
                        case NewTxnStateForBdbRecord of
                            ?COMMITTED_IN_BDB_RECORD ->
                                %% already aborted (and we want to commit), so we return an error
                                throw({DistributedTxnGUID, already_aborted});
                            ?ABORTED_IN_BDB_RECORD ->
                                %% already aborted (and we want to abort), so we pretend success
                                {DistributedTxnGUID, aborted} 
                        end;

                    {?COMMITTED_IN_BDB_RECORD, _} -> 
                        case NewTxnStateForBdbRecord of
                            ?COMMITTED_IN_BDB_RECORD ->
                                %% already committed (and we want to commit), so we pretend success
                                {DistributedTxnGUID, committed};
                            ?ABORTED_IN_BDB_RECORD ->
                                %% already committed (and we want to abort), so we return an error
                                throw({DistributedTxnGUID, already_committed})
                        end
                end
        catch
            {bdb_error, bdb_DB_NOTFOUND} -> 
                %%?INFO_LOGGER("do_commit_or_abort: writing '~w' for txn ~w~n", [NewTxnStateForBdbRecord, TxnCounter]),
                ?BDB:c_put(BdbPort, 
                           Cursor, 
                           Key, 
                           term_to_binary({NewTxnStateForBdbRecord, erlang:now()}), 
                           bdb_DB_KEYFIRST),

                case NewTxnStateForBdbRecord of
                    ?COMMITTED_IN_BDB_RECORD ->
                        {DistributedTxnGUID, committed};
                    ?ABORTED_IN_BDB_RECORD ->
                        {DistributedTxnGUID, aborted}
                end
        end,    

    ?BDB:c_close(BdbPort, Cursor),
    ?BDB:txn_commit(BdbPort, Txn, []),
    Reply.


txn_fun_resolve_one_txn(BdbPort, {TxnCounter, DistributedTxnGUID, State}) ->

    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, State#state.txn_states_db_id),
    resolve_one_txn(BdbPort, DB, TxnCounter, DistributedTxnGUID).


resolve_one_txn(BdbPort, DB, TxnCounter, DistributedTxnGUID) ->

    %% If 'resolve' finds no record for a transaction it 
    %% MUST write an explicit abort record before returning.
    %%
    %% If that write is necessary we DON'T need to do it in the same
    %% transaction as the original read.  If a commit happens between 
    %% the two then the write will fail, as do_commit_or_abort will 
    %% ONLY write something if there is no existing record (and if a 
    %% record does exist which matches the 'new' state -- in this case 
    %% an 'aborted' record, then do_commit_or_abort just returns as 
    %% though it succeeded).
    Key = <<TxnCounter:64>>,
    try ?BDB:db_get(BdbPort, DB, void, Key, <<>>, []) of 

        {Key, ValueBinary} -> 
            case binary_to_term(ValueBinary) of
                {?ABORTED_IN_BDB_RECORD, _} ->
                    {DistributedTxnGUID, aborted};

                {?COMMITTED_IN_BDB_RECORD, _} -> 
                    {DistributedTxnGUID, committed}
            end
    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            ?INFO_LOGGER("resolve_one_txn: no record found for txn ~w so writing explicit 'aborted' record~n", [TxnCounter]),
            do_commit_or_abort(BdbPort, DB, ?ABORTED_IN_BDB_RECORD, TxnCounter, DistributedTxnGUID)
    end.


txn_fun_resolve_multiple_txns(BdbPort, {ListDistributedTxnGUIDs, State}) ->

    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, State#state.txn_states_db_id),

    lists:map(fun(<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID) ->
                      try resolve_one_txn(BdbPort, DB, TxnCounter, DistributedTxnGUID) of
                          Result -> {ok, Result}
                      catch
                          Error  -> {error, Error}
                      end
              end,
              ListDistributedTxnGUIDs).

