
%%
%% berkeley_db_msg_queue.erl
%%
%% Example of a transactional, persistent queue for sending arbitrary Erlang messages.
%%
%% This uses berkeley_db_port_coordinator and berkeley_db_sequence_server.
%%
%% Design/API:
%%
%% Messages are arbitrary Erlang terms.
%%
%% The QName is the identity of the message channel.
%% - several processes are registered locally and with global:register() with names based on QName.
%% - several BDB databases are created with names based on QName
%% Therefore there are restrictions on the characters that can appear in QName, 
%% and it should be appropriately unique.
%%
%% The application simply enqueues messages as part of normal work transactions.
%% i.e. The application does not send the messages anywhere, just commits them.
%%
%% All messages enqueued in a channel are sent to a receivepump process (presumably on another node)
%% that is found using 'global' (and the QName).
%%
%% The receive pump stores the messages locally.  An application on the same host as the receivepump 
%% can dequeue the messages (with blocking or polling consume).
%%
%% The advantage of storing messages locally at both ends is that the applications at either end 
%% are not blocked if the other end (or link) goes down.
%%
%% Duplicate messages may be received by receivepump if the connection is dropped after receivepump 
%% commits the message locally but before sendpump gets the 'can_commit' ack and successfully commits it
%% it's consume.
%% We detect and prevent this using 64-bit transactionally increasing message ids.
%% receivepump stores all recent message ids in a database (in the same transaction that enqueues the 
%% message to the receiver\s local queue), and discards messages whose ids are in that database.
%% Obviously we need to purge old message ids after a while.
%% This period defines the maximum time that two nodes might be disconnected from each other 
%% before there is a (slight) risk that a duplicate message might be accepted.
%% For testing purposes this is set very low.  It might typically be 1day - 1week.
%%
%% TODO
%%
%% - Test with different QRecordSizes
%%
%% - Add replication (use berkeley_db_replication_group_helpers:do_txn_on_master())
%%
%% - Remove use of global:whereis and add per-message routing, so we only need one 
%%   incoming and one outgoing message queue per replication group (message queues use 
%%   OS threads for db_get(CONSUME_WAIT), so we cannot have lots of message queues)
%%   e.g. Destination would be replication group name.
%%
%% - Package nicely, e.g. spawn sendpump (and sequence_server) locally, receivepump remotely, 
%%   and run supervisors for allof them
%%
%% - Handle deadlocks (done automatically when we use ?BDB_RG_H:do_txn_on_master()).
%%   Force deadlocks in tests.
%%
%% - Optionally allow for multiple send-pump and multiple receive-pump processes, to increase throughput 
%%   (**this would break in-order delivery**).
%%
%% - blocking_consume should use db_get_non_block.  Then pump processes (and sinks) can also wait 
%%   for a msg to tell then to exit gracefully.
%%
%% - Possibly significant performance improvement -- sendpump should send the 
%%   one/two database records separately, rather than reconstructing the full msg binary.
%%   receivepump should not construct it either - should just call append_prepared_msg 
%%   with the 2 parts. Usually QRecLen will match and the parts can just be inserted 
%%   into the receive queue.  If QRecLen does not match then we will need to split/recombine 
%%   it efficiently so we can enqueue it.

-module(berkeley_db_msg_queue).

-include("berkeley_db_msg_queue.hrl").

-include("berkeley_db_drv.hrl").   %% required for bdb_port_handle record, so receive_reply_timeout can be changed for DB_CONSUME_WAIT


%% Public functions
-export([ %% Sender node calls these
          spawn_link_sendpump/3,

          start_link_sequence_server/3,
          stop_sequence_server/1,

          %% Receiver (sink) node calls this
          spawn_link_receivepump/4,

          %% Workers on both sides need to call this.
          create_bdb_msg_q/3,

          %% Sender workers call these
          enqueue_term_msg/3,              % for Erlang terms
          enqueue_external_term_msg/3,     % for Erlang terms that are already encoded via term_to_binary
          enqueue_raw_binary_msg/3,        % for raw binaries that are not encoded terms

          %% Receiver (sink) workers call this
          blocking_consume_one_msg/2,

          %% Workers on both sides call this.
          close_bdb_msg_q/1
        ]).


%% Messages are arbitrary Erlang terms that are serialized with term_to_binary().
%%
%% The QName is the identity of the message channel.
%%
%% All messages enqueued in a channel are sent to a receivepump process that 
%% globally registers itself under the queue name (using global:register_name/2).
%%
%% The receive pump stores the messages locally.  Any process (or number of processes) 
%% on that node may then consume the received messages.
%%
%% Databases are created based on the queue name.
%%
%%  Sender:
%%   "msg_q_" ++ QName ++ "_q.db",
%%   "msg_q_" ++ QName ++ "_bt.db", 
%%
%%  Receiver:
%%   "msg_q_" ++ QName ++ "_q.db",
%%   "msg_q_" ++ QName ++ "_bt.db", 
%%   "recent_msg_ids_" ++ QName ++ ".db"
%%
%% QRecordSize is used to size DB_QUEUE records.
%% It must be >= ?SIZEOF_MSG_HEADER (currently 14 bytes) to allow for the header
%%
%%    HeaderVersion   (1 byte  :  8-bit unsigned integer)
%% version #0:
%%    Flags           (1 byte  :  8-bit unsigned integer)   : Not used yet (padding)
%%    MsgId           (8 bytes : 64-bit unsigned integer big-endian)
%%    MsgBinLen       (4 bytes : 32-bit unsigned integer big-endian)
%%
%% Any remaining space in the queue record is used for the msg body.
%% If the message won't all fit in the remaining space then the excess will be 
%% stored in the BTREE.
%%
%% QRecordSize should be chosen so that most messages will fit in 
%% the queue record, without wasting too much space.
%% It can only be set when the queue is first created.  
%% Once the database exists, the value is ignored.

-define(SIZEOF_MSG_HEADER, 14).
-define(MSG_HEADER_VERSION, 0).

-define(get_DB_QUEUE_filename(QName), "msg_q_" ++ QName ++ "_q.db").
-define(get_DB_BTREE_filename(QName), "msg_q_" ++ QName ++ "_bt.db").
-define(SEQUENCE_KEY_NAME,  "startOfNextMsgIdRange").


-define(BDB,      berkeley_db_drv).
-define(BDB_U,    berkeley_db_utilities).
-define(BDB_PC,   berkeley_db_port_coordinator).
-define(BDB_SEQ,  berkeley_db_sequence_server).


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).


-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_msg_queue_error, Error})
           end).


%% This needs to run on the enqueue-side before any new messages can be enqueued.
start_link_sequence_server(BDBPortCoordName, QName, PersistEveryNConsumedValues) ->
    ?BDB_SEQ:start_link(get_sequence_server_name(QName),
                        BDBPortCoordName, 
                        ?get_DB_BTREE_filename(QName),
                        ?SEQUENCE_KEY_NAME,
                        [{persist_every_N_consumed_values, PersistEveryNConsumedValues}]).

stop_sequence_server(QName) ->
    ?BDB_SEQ:stop(get_sequence_server_name(QName)).


%% This needs to run on the enqueue-side before any new messages can be sent 
%% (but messages can be enqueued without it)
spawn_link_sendpump(BDBPortCoordName, QName, QRecordSize) ->
    proc_lib:spawn_link(fun() -> sendpump_init(BDBPortCoordName, QName, QRecordSize) end).


%% This needs to run on the dequeue side before any new messages can be sent or 
%% consumed.  
%% (Messages that have been previously received may be consumed if this is not running.)
spawn_link_receivepump(BDBPortCoordName, QName, QRecordSize, PurgeMsgIdsOlderThanSeconds) ->
    proc_lib:spawn_link(fun() -> receivepump_init(BDBPortCoordName, QName, QRecordSize, PurgeMsgIdsOlderThanSeconds) end).


%% Returns a bdb_msg_q record (contains handles to the local databases).
%%
%% Both enqueue and dequeue workers create one of these.
create_bdb_msg_q(BdbPortHandle, QName, QRecordSize) when QRecordSize >= ?SIZEOF_MSG_HEADER ->

    DB_Q = ?BDB_PC:get_db_handle_with_db_config(BdbPortHandle,
                                                bdb_DB_QUEUE, 
                                                ?get_DB_QUEUE_filename(QName),
                                                [{set_fixed_record_size_bytes,     QRecordSize},
                                                 {set_fixed_record_pad_byte,       0},
                                                 {set_queue_extent_size_db_pages,  100}]),

    DB_BT = ?BDB_PC:get_db_handle(BdbPortHandle,
                                  bdb_DB_BTREE, 
                                  ?get_DB_BTREE_filename(QName)),

    %% sequence_server_name is only used (i.e. server has to be running) if the application calls 
    %% one of the public enqueue functions.  But we don't know if the caller is a sender or 
    %% a sink, so we always create the name.  (Note that the name needs to be unique across BDB 
    %% environments and QName).
    #bdb_msg_q{qname                = QName, 
               bdb_port_handle      = BdbPortHandle, 
               db_q                 = DB_Q, 
               q_reclen             = QRecordSize, 
               db_btree             = DB_BT,
               sequence_server_name = get_sequence_server_name(QName)}.


%% Workers on the send-side call this.
%%
%% Returns
%%      {CreatedMsgId, CreatedSendSideQKey}    [the latter is mostly useful for logging in order to diagnose problems]
%%
%% Or throws/exits if there is a problem.
%%
%% This creates a child transaction, so the optional parent transaction is not polluted if this fails part way through.
enqueue_term_msg(MSG_Q, OptionalParentTxn, MsgAsTerm) ->
    enqueue_binary_msg(MSG_Q, OptionalParentTxn, term_to_binary(MsgAsTerm), ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM).

%% This variant is available for users that already have the message in external-term format
%% (i.e. MsgTermAsBinary is the result of an earlier call to term_to_binary).
%% Note that the receiver WILL call binary_to_term() on the message, so this is not 
%% suitable for sending raw binaries.
%% Raw binaries (that are not encoded terms) should be sent using enqueue_raw_binary_msg
enqueue_external_term_msg(MSG_Q, OptionalParentTxn, MsgTermAsBinary) when is_binary(MsgTermAsBinary) ->
    enqueue_binary_msg(MSG_Q, OptionalParentTxn, MsgTermAsBinary, ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM).

%% This variant is available for users that want to send raw binary data.
%% A flag is passed that suppresses the call to binary_to_term on the receiver side.
enqueue_raw_binary_msg(MSG_Q, OptionalParentTxn, MsgRawBinary) when is_binary(MsgRawBinary) ->
    enqueue_binary_msg(MSG_Q, OptionalParentTxn, MsgRawBinary, 0).  % MsgFlags does *not* have ?MSG_FLAGS__IS_EXTERNAL_TERM set

%% This is a private helper which implements the above two public functions.
enqueue_binary_msg(MSG_Q, OptionalParentTxn, MsgBinary, MsgFlags) ->

    CreatedMsgId = ?BDB_SEQ:consume_one_value(MSG_Q#bdb_msg_q.sequence_server_name),

    {CreatedMsgId, 
     %% this returns CreatedQKey
     append_binary_msg_with_id(MSG_Q,
                               OptionalParentTxn, 
                               CreatedMsgId, 
                               MsgFlags, 
                               MsgBinary)}.


close_bdb_msg_q(_MSG_Q) ->
    %% Nothing to do here as the port_coordinator owns the DB handles (we would have to refcount them 
    %% to close them safely, as other instances of bdb_msg_q might still have them open), 
    %% and other things may also be using the sequence server.
    ok.


%% This is a public function that can be used by 'sink' (worker) processes.
%% It is (almost) the opposite of enqueue_persistent_msg()
%%
%% Returns  {MsgId, MsgFlags, MsgBodyBinary}
%%
%% Or throws/exits if there is a problem.
%%
%% See comments in the .hrl file for values of MsgFlags and what they mean.
%%
%% This creates a child transaction, so the optional parent transaction is not polluted if this fails part way through.
blocking_consume_one_msg(MSG_Q, OptionalParentTxn) ->

    Txn = ?BDB:txn_begin(MSG_Q#bdb_msg_q.bdb_port_handle, OptionalParentTxn, []),

    %% This call uses DB_CONSUME_WAIT, so it blocks this process until a message is enqueued, 
    %% or until someone closes the port (or sends a command that forces a lock timeout).
    %% (Really the process is in a receive clause in EDTK-generated code, waiting for a message from the driver.
    %% And the driver has an async IO worker thread that is blocking in the Berkeley DB API.)
    %% Note that db_get(CONSUME_WAIT) *must* have a receive_reply_timeout of infinity -- see comments 
    %% in berkeley_db_drv.erl for db_get/6.  

    {QKey, <<?MSG_HEADER_VERSION:8,  % we only support a single header version at the moment
            MsgFlags:8,
            MsgId:64,
            MsgBodyBinTotalLen:32,
            RestFixedRecBin/binary>> }

        = ?BDB:db_get((MSG_Q#bdb_msg_q.bdb_port_handle)#bdb_port_handle{receive_reply_timeout = infinity}, 
                      MSG_Q#bdb_msg_q.db_q, 
                      Txn, 
                      <<>>, 
                      <<>>, 
                      [bdb_DB_CONSUME_WAIT]),

    %% Reconstruct full MsgBody if it is fragmented
    MsgBodyBin = 
        if MsgBodyBinTotalLen =< size(RestFixedRecBin) ->

                %% entire message is in the remainder of the fixed-length record 
                %% (not necessarily the entirety of the remainder of the record though)
                <<RestFixedRecBin:MsgBodyBinTotalLen/binary>>;

           true ->
                %% fetch rest of message from btree
                {QKey, VarBin} = ?BDB:db_get(MSG_Q#bdb_msg_q.bdb_port_handle, MSG_Q#bdb_msg_q.db_btree, Txn, QKey, <<>>, []),

                %% For some reason we cannot construct a binary on its own here, we have to bind it to a variable.
                TMP = <<RestFixedRecBin/binary, VarBin/binary>>,
                TMP
        end,

    MsgBodyBinTotalLen = size(MsgBodyBin), % just an assert

    ?BDB:txn_commit(MSG_Q#bdb_msg_q.bdb_port_handle, Txn, []),
    {MsgId, MsgFlags, MsgBodyBin}.


%%
%% Internal functions.
%%

%% Helper for start_link_sequence_server and create_bdb_msg_q.
%%
%% Note that the sequence_server_name only needs with respect to QName.
%% Different BDB environments cannot have different queues that have the same QName
%% as receivepump registers globally with a process name based only on QName.
get_sequence_server_name(QName) ->
    list_to_atom("berkeley_db_msg_queue_sequence_" ++ QName).


%% This appends to a bdb_msg_queue and returns the CreatedQKey
%% The operation is not called 'enqueue' to distinguish it from operations on enqueablePersistMsgQ,
%% -- the latter generats a new MsgId, but this one doesn'tl
append_binary_msg_with_id(MSG_Q, OptionalParentTxn, MsgId, MsgFlags, MsgBodyBin) ->

    {FixedRecBin, VarRecBin} = format_msg_for_databases(MSG_Q#bdb_msg_q.q_reclen, MsgId, MsgFlags, MsgBodyBin),

    Txn = ?BDB:txn_begin(MSG_Q#bdb_msg_q.bdb_port_handle, OptionalParentTxn, []),
    
    CreatedQKey = ?BDB:db_put_append(MSG_Q#bdb_msg_q.bdb_port_handle, MSG_Q#bdb_msg_q.db_q, Txn, FixedRecBin, []),

    if size(VarRecBin) > 0 -> 
            ?BDB:db_put(MSG_Q#bdb_msg_q.bdb_port_handle, MSG_Q#bdb_msg_q.db_btree, Txn, CreatedQKey, VarRecBin, []);
       true -> 
            ok
    end,
    ?BDB:txn_commit(MSG_Q#bdb_msg_q.bdb_port_handle, Txn, []),
    CreatedQKey.


%% Private helper
%% Returns 
%%   {FixedRecBin, VariableRecBin}
%% (VariableRecBin may be <<>>, in which case we can choose not to insert it of course)
%%
%% Useful shell command:
%%   berkeley_db_test:format_msg_for_databases(4000, 0, term_to_binary(hello1)).
format_msg_for_databases(QRecordSize, MsgId, MsgFlags, MsgBodyBin) ->

    SizeOfOverflowFromQRecord = ?SIZEOF_MSG_HEADER + size(MsgBodyBin) - QRecordSize,

    if SizeOfOverflowFromQRecord =< 0 ->
            %% All fits in db_q
            {<<?MSG_HEADER_VERSION:8, MsgFlags:8, MsgId:64, (size(MsgBodyBin)):32, MsgBodyBin/binary>>, 
             <<>>};
       true ->
            {MsgBinPt1, MsgBinPt2} = split_binary(MsgBodyBin, size(MsgBodyBin) - SizeOfOverflowFromQRecord),
            {<<?MSG_HEADER_VERSION:8, MsgFlags:8, MsgId:64, (size(MsgBodyBin)):32, MsgBinPt1/binary>>,
             MsgBinPt2}
    end.
    


%%
%% The sendpump process.
%% 

sendpump_init(BDBPortCoordName, QName, QRecordSize) ->

    %% If the receivepump is started after us then it may not be registered yet.
    %% global:send/2 will *exit* with {badarg, {Name,Msg}} if the process is not registered.
    %% As a convenience, we wait until it is registeredl
    %% Sadly, 'global' does not publish events when it registers names, so we have to poll.
    case global:whereis_name(QName) of

        undefined ->
            ?INFO_LOGGER("sendpump_init: no receivepump process globally registered for ~p. Sleeping for 1 second.~n", [QName]),
            timer:sleep(1000),
            sendpump_init(BDBPortCoordName, QName, QRecordSize);

       Pid when is_pid(Pid) ->
            ?INFO_LOGGER("sendpump_init: entereing loop (initial receiver process globally registered as ~p is ~p)~n", [QName, Pid]),
            sendpump_loop( create_bdb_msg_q(?BDB_PC:acquire_bdb_port_handle(BDBPortCoordName), 
                                                QName, 
                                                QRecordSize) )
        end.

sendpump_loop(MSG_Q) ->

    Txn = ?BDB:txn_begin(MSG_Q#bdb_msg_q.bdb_port_handle, void, []),

    %% We can only commit this consume transaction after receiving an ack that confirms the 
    %% receiver has commited it.  But the receiving node/bdb environment may have it's own 
    %% send queue back to another receive on this node/bdb environment.
    %% Therefore we have to worry about distributed deadlocks 
    %% (e.g. if the receiving site has an enqueue that is sending to a recevier on this site
    %% and the receiver on this site blocks on a lock held by this transaction.
    %% By far the simplest (the only practical) way is to just use lock/txn timeouts.
    %% IMPORTANT: this requires that the deadlock detector is being called at regular 
    %% intervals, as BDB does not generate timing events.
    %% We spawn a process to do regular deadlock detection in sendpump_process.
    %% Timeout is u_int32 microseconds (not millisecons)
    ?BDB:txn_set_timeout(MSG_Q#bdb_msg_q.bdb_port_handle, 
                         Txn, 
                         ?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS * 1000, 
                         bdb_DB_SET_TXN_TIMEOUT),

    %%?INFO_LOGGER("sendpump_loop blocking~n"),
    {MsgId, MsgFlags, MsgBodyBin} = blocking_consume_one_msg(MSG_Q, Txn),

    ?INFO_LOGGER("sendpump_loop got MsgId=~p Msg=~p~n", 
              [MsgId, 
               if ((MsgFlags band ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) =:= ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) -> 
                       binary_to_term(MsgBodyBin);
                  true -> MsgBodyBin
               end]),

    %% We assume that channel name has been registered with 'global'

    %% We can't use MsgId as a tag as we may send any message more than once 
    %% (until the receive commits its enqueue and we commit our consume)
    Tag = make_ref(),

    ?INFO_LOGGER("sendpump sending to QName=~p which 'global' resolves to pid=~p~n", [MSG_Q#bdb_msg_q.qname, global:whereis_name(MSG_Q#bdb_msg_q.qname)]),
    global:send(MSG_Q#bdb_msg_q.qname, {msg, self(), Tag, MsgId, MsgFlags, MsgBodyBin} ),
    
    receive
        {Tag, can_commit} -> ?BDB:txn_commit(MSG_Q#bdb_msg_q.bdb_port_handle, Txn, []),
                             sendpump_loop(MSG_Q);

        Other             -> ?BDB:txn_abort(MSG_Q#bdb_msg_q.bdb_port_handle, Txn),
                             throw( {unrecognized_ack_msg, Other} )
    after
        5000              -> ?BDB:txn_abort(MSG_Q#bdb_msg_q.bdb_port_handle, Txn),
                             throw( timeout_waiting_for_ack )
    end.



%%
%% The receiver
%%

receivepump_init(BDBPortCoordName, QName, QRecordSize, PurgeMsgIdsOlderThanSeconds) ->
    %%?INFO_LOGGER("receivepump_process spawning~n"),

    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(BDBPortCoordName),

    %% This process must ack the message before the sender can commit it's consume transaction.
    %% Therefore we want to do the least work possible -- i.e. no business logic.
    %% So we simply enqueue the message in a local bdb_msg_queue.
    %% (Note that we DON'T need an enqueablePersistMsgQueue here -- despite the name -- as 
    %% we already have MsgIds, and enqueablePersistMsgQueue's only extra feature is that it 
    %% creates new message-ids -- which we definitely do not want).
    %% Some other process can consume messages from that queue and do the work (and commit the 
    %% consume as part of the work transaction).

    %% QRecordSize can be different to the sender's value.
    MSG_Q = create_bdb_msg_q(BdbPortHandle, QName, QRecordSize),
    
    %% We also need a database to store some number of recent message-ids, to detect duplicate 
    %% messages. The key is MsgId, the value is a timestamp with minute resolution.
    %% There are several things to consider on how we store these:
    %% - Msg-ids are monotonically increasing numbers
    %% - DB_BTREE has an optimization that minimizes page-splits (database size/cache misses) 
    %%   if inserts are sequentially higher keys in the rightmost leaf node).
    %% - To use that feature We would have to ensure that our sort function preseved the desired 
    %%   ordering (e.g. little-endian binary integers don't sort in numerical increasing order with the 
    %%   default sort function).
    %%   Changing the sort function is not yet supported by the bdb driver (but could be quite 
    %%   easily -- especially for a few well-known functins like a big-endian transformation).
    %%   Or we could make the keys ASCII msg-ids (maybe padded with leading zeroes),
    %%   but that's not very space-efficient.
    %%   Fortunately Erlang makes it trivial to construct a big-endian binary containing the 64-bit number.
    %% - If we ever have more than one receive_pump then the rightmost page would become 
    %%   a lock hotspot (although BDB v4.6 is supposed to have record-level locking).
    %% - We also need to purge expired msg-ids.  In the past I did this as part of inserting 
    %%   a new msg-id; looking before & after the inserted item for expired ones (and assuming 
    %%   that this was almost free as they would be on the same page).  Obviously this relies 
    %%   on the keys NOT being sorted, as if they are sorted, the older keys will be at the 
    %%   exact opposite side of the btree.  
    %%   Also, this takes more time while the sender is blocking for it's ack, and I'd like 
    %%   to avoid that.
    %%   If we do use a btree with increasing keys, then the oldest keys will be the first 
    %%   ones in the database, which means it won't be contending for locks with the receivepump
    %%   process unless the database is almost empty.
    %% So, assuming that we only have 1 receive pump (no lock contention), and arrange for the 
    %% msg-ids to sort in numerically increas8ng order, and we spawn a separate houskeeping process 
    %% that examines the lowest  msg-ids for timestamp-based purging, then DB_BTREE looks like the right choice.

    MsgIdDB = ?BDB_PC:get_db_handle(BdbPortHandle, 
                                    bdb_DB_BTREE, 
                                    "recent_msg_ids_" ++ QName ++ ".db"),

    %% Start the housekeeping process
    spawn_link_purge_expired_msg_ids(BDBPortCoordName, MsgIdDB, PurgeMsgIdsOlderThanSeconds),

    %% Everthing is ready, so we can register now.
    ?INFO_LOGGER("receivepump registering globally as ~p~n", [MSG_Q#bdb_msg_q.qname]),
    yes = global:register_name(MSG_Q#bdb_msg_q.qname, self()),

    receivepump_loop(MSG_Q, MsgIdDB).

receivepump_loop(MSG_Q, MsgIdDB) ->

    %%?INFO_LOGGER("receivepump_loop blocking~n"),
    receive
        {msg, SenderPid, Tag, MsgId, MsgFlags, MsgBodyBin} -> 
            ?INFO_LOGGER("receivepump_loop received MsgId=~p Msg=~p~n", 
                      [MsgId, if ((MsgFlags band ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) =:= ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) -> 
                                      binary_to_term(MsgBodyBin);
                                 true -> MsgBodyBin
                              end]),

            Txn = ?BDB:txn_begin(MSG_Q#bdb_msg_q.bdb_port_handle, void, []),

            %% detect duplicate messages
            try ?BDB:db_put(MSG_Q#bdb_msg_q.bdb_port_handle, 
                            MsgIdDB, 
                            Txn, 
                            <<MsgId:64>>, 
                            term_to_binary(erlang:now()), 
                            [bdb_DB_NOOVERWRITE]) of 

                ok ->  
                    %% this is a new msg, not a duplicate
                    %%?INFO_LOGGER("receivepump_loop not a duplicate MsgId=~p Msg=~p~n", [MsgId, binary_to_term(MsgBodyBin)]),

                    %% Enqueue this message locally
                    %% We really don't want to transform this message back to Erlang terms
                    %% - it's potentially expensive, and we don't need the term form at all here.
                    %% So we will enqueue it locally in it's binary form.
                    append_binary_msg_with_id(MSG_Q, Txn, MsgId, MsgFlags, MsgBodyBin),

                    ?BDB:txn_commit(MSG_Q#bdb_msg_q.bdb_port_handle, Txn, []),
                    SenderPid ! {Tag, can_commit},
                    receivepump_loop(MSG_Q, MsgIdDB)

            catch
                {bdb_error, bdb_DB_KEYEXIST} ->  
                    %% we have seen this msg-id before, so just discard it
                    ?INFO_LOGGER("receivepump_loop received duplicate MsgId=~p Msg=~p~n", 
                              [MsgId, if ((MsgFlags band ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) =:= ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) -> 
                                              binary_to_term(MsgBodyBin);
                                         true -> MsgBodyBin
                                      end]),

                    %% We can tell sender to commit immediately -- they don't have to wait for us to abort
                    SenderPid ! {Tag, can_commit},

                    ?BDB:txn_abort(MSG_Q#bdb_msg_q.bdb_port_handle, Txn),
                    receivepump_loop(MSG_Q, MsgIdDB)
            end;

        Other -> throw( {unrecognized_msg, Other} )
    end.


%%
%% Housekeeping process to purge expired message-id, based on their timestamps being older than erlang:now().
%% (This is spawned, owned, and killed by receivepump).
%%

spawn_link_purge_expired_msg_ids(BDBPortCoordName, MsgIdDB, PurgeMsgIdsOlderThanSeconds) ->
    spawn_link(fun() -> purge_expired_msg_ids_init(BDBPortCoordName, MsgIdDB, PurgeMsgIdsOlderThanSeconds) end).

purge_expired_msg_ids_init(BDBPortCoordName, MsgIdDB, PurgeMsgIdsOlderThanSeconds) ->
    %%?INFO_LOGGER("purge_expired_msg_ids_process spawning~n"),

    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(BDBPortCoordName),
    purge_expired_msg_ids_loop(BdbPortHandle, MsgIdDB, PurgeMsgIdsOlderThanSeconds).

purge_expired_msg_ids_loop(BdbPortHandle, MsgIdDB, PurgeMsgIdsOlderThanSeconds) ->

    %% We may need to modify the database, so we *HAVE* to use a transactional cursor 
    %% (modifications by non-transactional cursors in a transactional environment will 
    %% break recovery -- DB_AUTO_COMMIT does NOT work with cursor operations).

    Txn    = ?BDB:txn_begin(BdbPortHandle, void, []),
    Cursor = ?BDB:db_cursor(BdbPortHandle, MsgIdDB, Txn, []),
    
    try ?BDB:c_get(BdbPortHandle, Cursor, <<>>, <<>>, [bdb_DB_FIRST, bdb_DB_RMW]) of 

        {MsgId, Timestamp} ->  
            MsgAgeSeconds = timer:now_diff(erlang:now(), binary_to_term(Timestamp)) / 1000000,

            if MsgAgeSeconds > PurgeMsgIdsOlderThanSeconds ->

                    ?INFO_LOGGER("purge_expired_msg_ids_loop: Purging MsgId ~p (age ~p seconds) and sleeping for 1 second~n", [MsgId, MsgAgeSeconds]),
                    ?BDB:c_del(BdbPortHandle, Cursor, []),
                    ?BDB:c_close(BdbPortHandle, Cursor),
                    ?BDB:txn_commit(BdbPortHandle, Txn, []),

                    timer:sleep(1000),  % 1 second
                    purge_expired_msg_ids_loop(BdbPortHandle, MsgIdDB, PurgeMsgIdsOlderThanSeconds);

               true ->
                    ?BDB:c_close(BdbPortHandle, Cursor),
                    ?BDB:txn_abort(BdbPortHandle, Txn),  % fast, as nothing was modified

                    %% Rather than polling at some fixed interval, we should 
                    %% sleep until the oldest msg becomes stale.
                    ?INFO_LOGGER("purge_expired_msg_ids_loop: First msg (age ~p seconds) is not old enough (~p): Sleeping for ~p seconds~n", [MsgAgeSeconds, PurgeMsgIdsOlderThanSeconds, MsgAgeSeconds + 1]),
                    timer:sleep(round((MsgAgeSeconds + 1) * 1000)),
                    purge_expired_msg_ids_loop(BdbPortHandle, MsgIdDB, PurgeMsgIdsOlderThanSeconds)
            end

    catch
        {bdb_error, bdb_DB_NOTFOUND} ->
            ?BDB:c_close(BdbPortHandle, Cursor),
            ?BDB:txn_abort(BdbPortHandle, Txn),  % fast, as nothing was modified

            ?INFO_LOGGER("purge_expired_msg_ids_loop: No messages yet: Sleeping for ~p + 1 seconds~n", [PurgeMsgIdsOlderThanSeconds]),
            timer:sleep((PurgeMsgIdsOlderThanSeconds + 1) * 1000),
            purge_expired_msg_ids_loop(BdbPortHandle, MsgIdDB, PurgeMsgIdsOlderThanSeconds)
    end.

