%%%-------------------------------------------------------------------
%%% File    : berkeley_db_replication_group_helpers.erl
%%% Author  : Chris Newcombe <>
%%% Description : 
%%%
%%% Helper functions to make it reasonably convient for applications to 
%%% access replication groups without requiring a global directory 
%%% service to find sites/masters etc.
%%%
%%% Problem statement:
%%%
%%% berkeley_db_replication_group:do_txn_on_master() will find the 
%%% master if it moves, but the caller must first find _some_ site in 
%%% the replication group, in order to call do_txn_on_master at all.
%%% Having a MasterLocationHint is no use if it is the master that fails.
%%%
%%% Also, clients might want to run a txn on a replication_group that 
%%% does not currently have a site on the current node (either the 
%%% site has failed, or it really is a different cluster).
%%% 
%%% Also (in the latter case), clients might want to run a read-only
%%% txn on _any_ site.  They would presumeably  prefer to use a local
%%% site if one exists, but would like to select a random site 
%%% otherwise.
%%%
%%% If a site fails (cannot be contacted), the caller would like to avoid
%%% trying it again for some period.  But if all sites appear to fail,
%%% we probably _do_ want to try them again (rather than waiting for 
%%% some timeout period).
%%%
%%% Solution:
%%%
%%% The caller can create a 'caller_state' object (one per replication group)
%%% and pass it to do_txn_on_master() and do_txn_on_any_site().
%%%
%%% refresh_master_location_hint() will wait until a master is found and
%%% return an RgCallerState containing an updated master_location_hint,
%%% and will write the updated RgCallerState to persistent storage if it has 
%%% changed (and is the latest -- see comments in code).
%%% (This function is the parallel of ?BDB_RG:get_master_location_hint.
%%%  The caller may use get_MLH_from_RgCallerState() if they want to 
%%%  extract the #master_location_hint.)
%%% It picks a site to ask for the location of the master (prefering
%%% sites on the current node if there are any).  If that site is down
%%% or fails, it picks another site to ask, until there are no sites
%%% left or the timeout is reached.
%%%
%%% do_txn_on_master() will use the #master_location_hint that is 
%%% currently cached in the caller-state.  If the master is unknown 
%%% or is lost before the txn completes, it uses refresh_master_location_hint 
%%% to try to find another master (with the appropriate timeout).
%%%
%%% do_txn_on_any_site{} picks any site that is believed to be 'up', 
%%% but prefers local sites first, if one exists.  It also prefers 
%%% in_sync sites over not_in_sync sites.
%%% TODO: by default this should be restricted to in_sync sites only, 
%%% and should take a list of options so that the caller can explicitly 
%%% allow not_in_sync sites.
%%%
%%% The caller_state contains 2 funs, to read and write itself from
%%% persistent storage (e.g. an ETS table).  If the caller_state
%%% changes (e.g. a site is discovered to be down, or the master
%%% moves), then do_txn_on_... will use these funs to store the
%%% update.  (The 'read' fun is required as multiple concurrent
%%% processes may be using copies of the same caller_state object, and
%%% therefore reading and writing the same persistent copy. To avoid
%%% overwriting a newly updated record in persistent storage with
%%% changes based on an older version, the caller_state contains a
%%% last_persisted_timestamp. To updated the copy in persistent
%%% storage, the code first reads the latest version from persistent
%%% storage (without requiring a lock, so there is still a race
%%% condition) and only writes the new version if it is based on a
%%% version that is not older than the current persistent version.  We
%%% might also try merging caller_state objects, but it's unlikely the
%%% potential (small) improvement would be worth the extra complexity.
%%%
%%% Created : 26 Oct 2006 by Chris Newcombe <>
%%%
%%% TODO: 
%%%
%%% - do_txn_on_any_site wants to do it on **in_sync** sites.
%%%   So instead of using rpc:call, there should be a method
%%%   ?BDB_RG:do_txn_on_any_in_sync_site().
%%%   We pick a random site (prefering local sites), 
%%%   and call the method on that site.
%%%   If that site happens to not be in-sync, in picks a site 
%%%   that is and forwards the request.
%%%   It should return an indication of whether the _original_ 
%%%   site was in/not-in sync.  caller_state should keep that
%%%   so up_site_locators should be in_sync_site_locators.
%%%   and not_in_sync flags should be moved to down_sites and polled.
%%%
%%% - In pick_random_up_site, if all sites are down and we try
%%%   to contact one (random) site, then if we _do_ contact it we want
%%%   to get the current group state (or all traffic will be sent to
%%%   that single site until the others emerge from the timeout
%%%   period).  So we need the same mechanism as the next point
%%%   (?BDB_RG:get_current_caller_state()).  And that means we don't
%%%   use rpc:call here either -- we call the above method instead.
%%%
%%% - When replication_group supports removing sites, caller_state
%%%   will drift out of sync with the real site membership.  So
%%%   whenever we fail to connect to a site, and later successfully
%%%   connect to a different site, we should ask that site for the new
%%%   membership.  i.e. There should be a
%%%   ?BDB_RG:get_current_caller_state() method, which would be a
%%%   smarter version of make_initial_caller_state().  It could be
%%%   invoked via do_txn_on_any_site() using the current (flawed) ?CS,
%%%   to get a better one.  Either than, or clients need to hook into
%%%   replication_group events -- and even exposing those to the
%%%   clients is a problem (which sites send them?)  or every call to
%%%   every ?BDB_RG method all returns the current group status -
%%%   obviously far too heavyweight.
%%%
%%% - Is a timeout the correct mechanism to limit these fumctions?
%%%   e.g. should caller_state contain a max_num_sites_to_try as well?
%%%
%%% - In pick_random_up_site: As RgCallerState is presumably
%%%   rarely/never transmitted across nodes, we could cache
%%%   LocalUpSiteLocators and avoid the list comprehension here.
%%%
%%% - Ideally we would have cached pids for each UpSiteLocator (like
%%%   master_location_hint) does.  That would save a call to whereis
%%%   (the first clause of do_txn_with_retry_on_deadlock_no_throw),
%%%   but that of course risks flagging a site as down if it has
%%%   crashed since we got the cached pid -- even if it is back up
%%%   now.  (Using process names fixes that.)
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_replication_group_helpers).

%% API
-export([initialize_caller_state/5,
         get_replication_group_site_pid/1,
         refresh_master_location_hint/2,          % timeout is FindMasterTimeout which can be 0  (the gen_server:call timeout is still the default of 5000 ms)
         refresh_master_location_hint/3,          % takes both FindMasterTimeout and ClientServerTimeout
         get_MLH_from_RgCallerState/1,
         do_txn_on_master/2,                  % will call refresh_master_location_hint if master is not known
         do_txn_on_master/3,                  % ditto
         do_txn_on_master/4,                  % ditto
         do_txn_on_any_site/2,
         do_txn_on_any_site/3,
         do_txn_on_any_site/4,
         do_distributed_txn_unordered/4,      % executes against masters of several different replication groups
         prepare_and_commit_distributed_txn/4 % TxnFuns participating in distributed transactions must tail-call this.
        ]).

%% Internal API
%% Exported for use with rpc:call, do_txn etc
-export([get_local_replication_group_site_pid/2,
         txn_fun_get_new_distributed_txn_guid/2, 
         txn_fun_commit_distributed_txn/2]).


-define(BDB,     berkeley_db_drv).
-define(BDB_U,   berkeley_db_utilities).
-define(BDB_H,   berkeley_db_helpers).
-define(BDB_PC,  berkeley_db_port_coordinator).
-define(BDB_RG,  berkeley_db_replication_group).
-define(BDB_GTM, berkeley_db_global_transaction_manager).

-include("berkeley_db_replication_group_helpers.hrl").
-define(CS,      berkeley_db_replication_group_caller_state).

-include("berkeley_db_replication_group_internal.hrl").


-define(random_list_element(List),
    lists:nth( random:uniform(length(List)), List)).


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).


% Debug helpers (poor-man's tracing)
-define(AT, ?INFO_LOGGER("At line ~w~n", [?LINE])).   %% remove with regex    ^ *\?AT, *^J


%%====================================================================
%% API
%%====================================================================

%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%
%%   This also calls WriteFun to make the new caller_state persistent.
%%
%%   AllSites is the same list required by ?BDB_RG:start_sites()
%%   ReadFun  =   fun() -> caller_state
%%   WriteFun =   fun(caller_state) -> void
%%--------------------------------------------------------------------

initialize_caller_state(ReplicationGroupName, AllSites, IgnoreDownSitesForMilliseconds, ReadFun, WriteFun) ->

    %% Note: we ONLY sort AllSiteLocators to make assert_invariants_RgCallerState a little faster 
    AllSiteLocators = lists:sort(
                        lists:map(
                          fun({{SiteName, SiteNode}, _BdbDataDir, _BDBHAHostNamePort, _SiteSpecificConfig}) ->
                                  {ReplicationGroupName, SiteName, SiteNode}
                          end,
                          AllSites)),

    %% We set master_location_hint to 'unknown' to indicate that at the present time 
    %% we don't know who is master. 
    %% We *don't* want to call get_master_location hint to try to find the master, 
    %% as this initializer is often used BEFORE any of the sites have been started, 
    %% and we want it to be fast.
    %% When do_txn_on_master is called with master_location_hint = 'unknown' then 
    %% it calls refresh_master_location_hint (blocking based on the client-supplied txn
    %% timeout) and therefore discovers the master as quickly as possible.
    %%
    %% TODO: as an optimization, if any of AllSites are on the *local* node, we 
    %% could try a call here to refresh_master_location_hint against such a node
    %% using a timeout of 0, which explicitly means 'poll' and return immediately.
    %%
    %% Another alternative might be to pick a random SiteLocator and 
    %% set #master_location_hint to that site-locator, but with pid = unknown.
    %% But that would require more code in the functions that accept master location hints.
    %% Yet another alternative (previously tried) would be to allow a plain SiteLocator 
    %% master_location_hint, as all public ?BDB_RG APIs that accept a #master_location_hint 
    %% also accept a SiteLocator.
    %% The big problem with both of those alternatives is that they violate the invariant 
    %% that master_location_hint refers to a site that is in UpSiteLocators -- not initially, 
    %% (as all sites are in UpSiteLocators initially), but after a site is shown to be down.
    %%
    %% To avoid that we have to allow 'unknown' master as well, and then it is simpler 
    %% just to require master_location_hint to be a real #master_location_hint that 
    %% (as an invariant) refers to a site in UpSiteLocators, or is 'unknown' (with 
    %% last_poll_for_new_master_time set appropriately).

    RgCallerState = #?CS{replication_group_name             = ReplicationGroupName,
                         all_site_locators                  = AllSiteLocators,         
                         up_site_locators                   = AllSiteLocators,   % we assume that all sites are running
                         down_sites                         = [],
                         ignore_down_sites_for_milliseconds = IgnoreDownSitesForMilliseconds,
                         master_location_hint               = unknown,
                         last_persisted_time                = now(),
                         persist_read_fun                   = ReadFun,
                         persist_write_fun                  = WriteFun},

    assert_invariants_RgCallerState(RgCallerState),

    %% Write it to the persistent store.
    %% We DON"T call persist_caller_state_if_latest() as this function is intended to be called at startup time
    %% and ReadFun() is not guaranteed to work until after the first call to WriteFun.
    %% This is that first call.
    WriteFun(RgCallerState),

    RgCallerState.   


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Parameters:  SiteLocator
%% Returns:     site_not_running | Pid
%%--------------------------------------------------------------------

get_replication_group_site_pid({ReplicationGroupName, SiteName, _SiteNode}) when _SiteNode == node() ->
    %% Optimization for when the site is on the current node
    get_local_replication_group_site_pid(ReplicationGroupName, SiteName);

get_replication_group_site_pid({ReplicationGroupName, SiteName, SiteNode}) ->    
    try 
        rpc:call(SiteNode, ?MODULE, get_local_replication_group_site_pid, [ReplicationGroupName, SiteName])
    catch
        {badrpc, nodedown} ->  site_not_running
    end.

%% This is passed to rpc:call for execution on the remote-node.
get_local_replication_group_site_pid(ReplicationGroupName, SiteName) ->
    case whereis(?BDB_RG:get_proc_name(ReplicationGroupName, SiteName)) of
        undefined -> site_not_running;
        Pid       -> case erlang:is_process_alive(Pid) of
                         true  ->  Pid;
                         false ->  site_not_running
                     end
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%  Returns NewRgCallerState with a possibly updated master_location_hint
%%  (and perhaps other changes)
%% FindMasterTimeout may be 0 (for an immediate poll/return)
%%--------------------------------------------------------------------

refresh_master_location_hint(RgCallerState, FindMasterTimeout) ->
    refresh_master_location_hint(RgCallerState, FindMasterTimeout, 5000).

refresh_master_location_hint(#?CS{} = RgCallerState, FindMasterTimeout, ClientServerTimeout) ->

    StartTime = now(),

    %% This prefers sites on the current node, if there are any.
    {SiteLocator, RgCallerState2} = pick_random_up_site(RgCallerState),

    try 
        ?BDB_RG:get_master_location_hint(SiteLocator, FindMasterTimeout, ClientServerTimeout)
    of
        #master_location_hint{} = MLH ->

            %% We need to insert MLH into an RgCallerState to return.
            %%
            %% We should use RgCallerState2 (returned from pick_random_up_site), as it is the latest 
            %% one we have (might be different from the input RgCallerState - e.g. the site referred to by 
            %% MLH might have been in DownSites in RgCallerState, but have been resurrected in RgCallerState2).
            %%
            %% However, even then there is _still_ a chance that MLH refers to a site that is in DownSites
            %% in RgCallerState2.  If that is the case then we need to resurrect that site 
            %% (note: do_txn_on_master also does this if a new MLH is returned).
            %% This helper function takes care of all that.

            update_MLH_in_RgCallerState(MLH, RgCallerState2)
    catch
        %% See comments in catch clause of do_txn_on_master
        {bdb_replication_group_error, {site_is_down, SiteLocator, _Reason}} = Error ->
            ?INFO_LOGGER("refresh_master_location_hint: Got ~w. Considering retry~n", [Error]),
            refresh_master_location_hint(adjust_for_site_is_now_down(RgCallerState2, SiteLocator),
                                         reduced_timeout(FindMasterTimeout, StartTime),
                                         ClientServerTimeout)

        %% All other exceptions are allowed to propagate
    end.


%% private helper, used by refresh_master_location_hint and do_txn_on_master.
update_MLH_in_RgCallerState(#master_location_hint{site_locator = NewMasterSiteLocator} = NewMLH, 
                            #?CS{up_site_locators = UpSiteLocators,
                                 down_sites       = DownSites} = RgCallerState) ->

    NewRgCallerState =
        case lists:member(NewMasterSiteLocator, UpSiteLocators) of
            true ->
                RgCallerState#?CS{master_location_hint = NewMLH};
            false ->
                ?INFO_LOGGER("NewMLH ~w refers to a site that we think is down, so resurrecting that site.~n", [NewMLH]),
                RgCallerState#?CS{master_location_hint = NewMLH,
                                  up_site_locators     = UpSiteLocators ++ [NewMasterSiteLocator],
                                  down_sites           = [DS || {DSL, _Time} = DS <- DownSites, DSL =/= NewMasterSiteLocator]}
        end,

    assert_invariants_RgCallerState(NewRgCallerState),

    %% Important: persist_caller_state_if_latest might find a more recent (and entirely different)
    %% caller state in persistent storage, and return that.
    %% But that 'newer' state might not contain MLH.
    %% We could just force-write NewRgCallerState to persistent storage, but 
    %% the 'newer' state may have other changes that we don't want to loose 
    %% (e.g. it might have been refreshed from real group status).
    %% For now we just attempt to write NewRgCallerState, and ignore 
    %% any 'newer' state -- i.e. we return the MLH that we just obtained, 
    %% and NewRgCallerState as we know it contains that MLH (is self-consistent).
    persist_caller_state_if_latest(NewRgCallerState),

    NewRgCallerState.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Returns:  #master_location_hint (e.g. from result of refresh_master_location_hint)
%%--------------------------------------------------------------------
get_MLH_from_RgCallerState(#?CS{master_location_hint = MLH}) ->
     MLH.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Returns:
%%       {TxnResult, PossiblyUpdatedRgCallerState}
%%
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_master(RgCallerState, TxnFun) ->
    do_txn_on_master(RgCallerState, TxnFun, 0, 5000). % default NumDeadlockRetries & gen_server timeout

do_txn_on_master(RgCallerState, TxnFun, NumDeadlockRetries) ->
    do_txn_on_master(RgCallerState, TxnFun, NumDeadlockRetries, 5000). % default gen_server timeout


do_txn_on_master(#?CS{master_location_hint = unknown} = RgCallerState, 
                 TxnFun, 
                 NumDeadlockRetries, 
                 Timeout) ->

    %% The master is believed to be down (or just not yet discovered -- i.e. the system might be starting up).
    %%
    %% As elections are usually fast, we can minimize the intrusiveness of a failover by calling 
    %% refresh_master_location_hint, blocking for the client's txn Timeout .
    %%
    %% TODO: perhaps we should also call refresh_master_location_hint if ?BDB_RG:do_txn_on_master (below) throws 
    %% no_master, or throws site_is_down for the site that is believed to be master.
    StartTime = now(),

    try 
        refresh_master_location_hint(RgCallerState, Timeout, Timeout)
    of
        NewRgCallerState ->

            ?INFO_LOGGER("?BDB_RG_H:do_txn_on_master: refresh_master_location_hint returned ~p~n", [NewRgCallerState]),
            do_txn_on_master(NewRgCallerState, 
                             TxnFun, 
                             NumDeadlockRetries, 
                             reduced_timeout(Timeout, StartTime))
    catch
        {bdb_replication_group_error, {no_sites_available, _}} ->
            throw({bdb_replication_group_error, no_master})
    end;

do_txn_on_master(#?CS{master_location_hint = CurrMLH} = RgCallerState, 
                 TxnFun, 
                 NumDeadlockRetries, 
                 Timeout) when Timeout > 0 ->

    StartTime = now(),

    try ?BDB_RG:do_txn_on_master(CurrMLH, TxnFun, NumDeadlockRetries, Timeout) of

        {TxnFunResult, CurrMLH} -> 
            {TxnFunResult, RgCallerState};

        {TxnFunResult, #master_location_hint{} = NewMLH} ->

            %% The master has changed somehow, so update the master_location_hint 
            %% in RgCallerState.
            %%
            %% Note that the site referred to by NewMLH might have
            %% been in DownSites in RgCallerState, and have been
            %% resurrected since then (without our knowledge).  If
            %% that is the case then we need to resurrect that site.
            %% This helper function takes care of all that.

            NewRgCallerState = update_MLH_in_RgCallerState(NewMLH, RgCallerState),

            {TxnFunResult, NewRgCallerState}
    catch

        %% TODO: if we get a txn_fun_error that suggests that the site we reached 
        %% was not the master (e.g. db_put/c_put failed), then we might 
        %% want to retry another site.  Or maybe that should be handled in 
        %% ?BDB_RG:do_txn_on_master (i.e. a lower level than here).

        %% ?BDB_RG:do_txn_on_XXX() converts a variety of errors into 
        %% this summary exception:
        {bdb_replication_group_error, {site_is_down, DownSiteLocator, _}} = Error ->
            %% A target replication_group server is down/not reachable.
            %% This may refer to the site in CurrMLH, or (if the master has 
            %% moved) a different site -- DownSiteLocator tells which one.
            %%
            %% Try another site, if there is one.
            %% (adjust_for_site_is_now_down() will change master_location_hint to
            %% a different (random) site, or will throw if there are no more sites.)
            ?INFO_LOGGER("do_txn_on_master: Got ~w. Considering retry~n", [Error]),
            do_txn_on_master(adjust_for_site_is_now_down(RgCallerState, DownSiteLocator),
                             TxnFun,
                             NumDeadlockRetries,
                             reduced_timeout(Timeout, StartTime))

        %% All other exceptions are allowed to propagate
    end;

%% Here Timeout must be =< 0
do_txn_on_master(#?CS{replication_group_name = ReplicationGroupName}, _, _, _) ->
    throw({bdb_replication_group_error, {do_txn_on_master, timeout, ReplicationGroupName}}).


%% Private helper
%% Returns: NewRgCallerState
%% Throws:  {bdb_replication_group_error, {no_sites_available, NewRgCallerState}}
%%
%% Assumes that DownSiteLocator is currently a member of UpSiteLocators.
%%
%% Does not assert invariants on input RgCallerState as it is done by the caller.
adjust_for_site_is_now_down(#?CS{master_location_hint = MLH,
                                 up_site_locators     = UpSiteLocators,
                                 down_sites           = DownSites} = RgCallerState,
                            DownSiteLocator) ->

    %% Remove DownSite from UpSiteLocators, 
    %% and simultaneously assert that it is present in assert that DownSiteLocator is valid
    NewUpSiteLocators =
        case ?BDB_U:lists_find_and_remove_first(DownSiteLocator, UpSiteLocators) of
            {found, ListAfterRemoval} -> ListAfterRemoval;
            not_found                 -> throw({bdb_replication_group_error, {down_site_not_in_up_sites, DownSiteLocator, RgCallerState}})
        end,

    %% Does the master_location_hint need to change?
    NewMLH  = case MLH of
                  #master_location_hint{site_locator = DownSiteLocator} ->
                      unknown;
                  _ ->
                      %% Current master location hint is not affected
                      MLH
              end,

    NewRgCallerState = 
        RgCallerState#?CS{master_location_hint = NewMLH,
                          up_site_locators     = NewUpSiteLocators,
                          down_sites           = [{DownSiteLocator, now()} | DownSites]},

    assert_invariants_RgCallerState(NewRgCallerState),

    persist_caller_state_if_latest(NewRgCallerState),

    case NewUpSiteLocators of
        [] ->
            throw({bdb_replication_group_error, {no_sites_available, NewRgCallerState}});
        _ ->
            %% At least 1 other site is still up.
            NewRgCallerState
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Returns      {TxnResult, PossiblyUpdatedRgCallerState}
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_any_site(RgCallerState, TxnFun) ->
    do_txn_on_any_site(RgCallerState, TxnFun, 0, 5000). % default NumDeadlockRetries & gen_server timeout

do_txn_on_any_site(RgCallerState, TxnFun, NumDeadlockRetries) ->
    do_txn_on_any_site(RgCallerState, TxnFun, NumDeadlockRetries, 5000). % default gen_server timeout


do_txn_on_any_site(#?CS{} = RgCallerState, TxnFun, NumDeadlockRetries, Timeout) ->

    StartTime = now(),

    %% This prefers sites on the current node, if there are any.
    {SiteLocator, RgCallerState2} = pick_random_up_site(RgCallerState),

    %%?INFO_LOGGER("do_txn_on_any_site: Chosen site ~500p~n", [SiteLocator]),

    try 
        ?BDB_RG:do_txn_on_any_site(SiteLocator, TxnFun, NumDeadlockRetries, Timeout)
    of
        Result -> {Result, RgCallerState2}
    catch
        %% See comments in catch clause of do_txn_on_master
        {bdb_replication_group_error, {site_is_down, SiteLocator, _Reason}} = Error ->
            ?INFO_LOGGER("do_txn_on_any_site: Got ~w. Considering retry~n", [Error]),
            do_txn_on_any_site(adjust_for_site_is_now_down(RgCallerState2, SiteLocator),
                               TxnFun,
                               NumDeadlockRetries,
                               reduced_timeout(Timeout, StartTime))

        %% All other exceptions are allowed to propagate
    end.



%%
%% This executes a single distributed transaction on the masters of
%% multiple replication groups.
%%
%%
%% IMPORTANT: this spawns helper processes ('local clients' in the
%% comments below) to talk to each replication group. Those helpers
%% send messages to the process that called this function.  If this
%% function returns normally (success) then it guarantees that none of
%% these messages will arrive in the caller's mailbox.
%%
%% ** But if the function throws (e.g. if we reach Timeout) then **
%% ** some of the helper processes may still be running, and may **
%% ** send messages to the caller process.                       **
%%
%% All such messages are tagged with a new TxnGUID, and all receives
%% done inside this function are selective based on that TxnGUID.  But
%% if the caller doesn't want to risk polultion of it's message queue
%% then it should spawn its own intermediate worker to call this
%% function, and act as a bit-bucket for stray messages after it
%% returns.  Or, the caller could ignore such messages, and perhaps
%% flush its message queue periodically if it is a long-lived process.
%%
%% [Note that all of the helper processes are spawned with plain
%% spawn, not spawn-link, so there is no danger of an EXIT signal
%% affecting the caller's process after this function has returned.]
%%
%%
%% There are two variants of this function:
%%
%% unordered: 
%%   - This executes all TxnFuns in parallel
%%   - To prevent distributed deadlock, all port-coordinators are 
%%     ASSUMED to be using BDB lock or transaction timeouts and an 
%%     appropriate periodic deadlock detector.
%%
%% ordered:
%%   - This executes TxnFun's serially -- it waits for each TxnFun to 
%%     prepare its local transaction before calling the next TxnFun.
%%     This is obviously potentially a lot slower, but it has the 
%%     benefit that if the ordering is constant (and no unordered 
%%     distributed txns are performed) then there is no possibility
%%     of distributed deadlock.
%%
%% e.g. The ordered variant may be a win if there are only a small
%% number of groups participating in the transaction, or if 
%% each TxnFun is very fast.
%%    
%% Both variants have identical interfaces, so applications can 
%% easily experiment with both.
%%
%% It is possible to use a mix of both variants so long as all 
%% replication groups meet the requirements of the 'unordered'
%% variants (use BDB lock/txn timeouts etc).  The more that the 
%% ordered variant is used, the less often distributed deadlocks 
%% will occur.
%%
%% Parameters:
%%
%%    RgGroupTxnFuns = [{RgParticipantTag, RgCallerState, TxnFun}, 
%%                             ...]
%%
%% RgParticipantTag = term() -- any opaque application-specific tag (passed
%%                        through as a convenience for associating 
%%                        results with inputs)
%%
%% RgCallerState = #berkeley_db_replication_group_caller_state{}
%%
%% TxnFun = {Module, Function, ExtraArgs}
%%
%% Each TxnFun must conform to to the following signature & protocol
%%     
%%    txn_fun(BdbPort, {DistTxnInfo, CallerSuppliedExtraArgs}) ->
%%        
%%        %% Normal BDB application code, for example:
%%        DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "my.db"),
%%        Txn = ?BDB:txn_begin(BdbPort, void, []),
%%        ?BDB:db_put(BdbPort, DB, Txn, Key, term_to_binary(Data), []),
%%     
%%        %% CRITICAL: the TxnFun must NOT unilaterally commit, abort, or prepare Txn.
%%        %% Instead it must tail-call the following function:
%%
%%        ?BDB_RG_H:prepare_and_commit_distributed_txn(Txn, TxnFunResult, DistTxnInfo),
%%     
%%
%% Returns a 2-tuple:
%%
%%  {TxnGUID, ParticipantResults}
%%
%%  TxnGUID = binary of size DB_XIDDATASIZE
%%
%%  ParticipantResults = {RgParticipantTag, TxnFunResult}, 
%%                      ...]
%%
%% Important: ParticipantResults is NOT necessarily in the same order as 
%% the input RgGroupTxnFuns (hence the inclusion of RgParticipantTag).
%%
%%
%% Or throws  a 2-tuple
%%
%%  {berkeley_db_distributed_txn_error, Reason}
%%
%%
%% Implementation details.
%%
%% Obviously we need to call do_txn_on_master for eacn of the
%% replication groups.  But do_txn_on_master is synchronous (blocking)
%% and we also need to communicate with the TxnFun while
%% do_txn_on_master is running it.  Therefore we need to spawn a local
%% process to call do_txn_on_master.
%%
%% So there are 3 tiers of processes involved
%%
%%  - One 'DistTxnCoordPid' -- the process that calls do_distributed_txn
%%  - N 'local client' processes whicn call do_txn_on_master
%%  - N 'do_txn_on_master worker' processes which execute the TxnFun(s)
%%
%% The 'unordered' variant starts all of the local clients in
%% parallel, and the 'ordered' variant starts them one at a time,
%% after the previous one has said that it has successfully prepared.
%% Both variants end up with the same number of processes in
%% existance.
%%
%% DistTxnCoordPid first spawns-links the 'local client' processes,
%% and then enters a wait clause to receive 'successfully prepared'
%% messages from participants, or errors from 'local client' processes
%% or partipants. (Note: DistTxnCoordPid is also linked to all 'local
%% client' processes and DistTxnCoordPid does not trap exits, so if
%% any of them crash then they are all torn down and any
%% 'do_txn_on_master workers' (whicn are NOT linked to
%% DistTxnCoordPid, or even to their own 'local client' processes)
%% will eventually reach their timeouts and exit, or return results
%% that are sent to nowhere).
%% 
%% Each freshly-spawned 'local client' process calls
%% ?BDB_RG_H:do_txn_on_master to forward WrappedTxnFun to the master
%% site, spawn a worker process on that site, and start executing the
%% TxnFun.  (Currently a new worker process is always spawned to
%% execute it, even if the master site is on this node so the client
%% could theorectically have called ?BDB_H:do_txn. In future
%% do_txn_on_master may execute WrappedTxnFun in *the current process*
%% (via do_txn) if the master is on this node. But that change will be
%% transparent to this code.)
%%
%% In the newly spawned worker process on the master site,
%% WrappedTxnFun begins a transaction, does some work, and then when a
%% non-distributed transaction would call txn_commit, it instead calls
%% the 'other half' of this code:
%% ?BDB_RG_H:prepare_and_commit_distributed_txn.
%%
%% That function will call txn_prepare. If txn_prepare succeeds, then
%% it (prepare_and_commit_...) then sends a 'successfully prepared'
%% message directly to DistTxnCoordPid, and then enters a receive
%% clause (with timeout) waiting for instructions from
%% DistTxnCoordPid.  (Note that prepare_and_commit_... does not use
%% the TxnFun's associated 'local client' process as a message route
%% -- that process will be blocked in a gen_server:call, waiting for
%% do_txn_on_master to return it's result.)
%%
%% If txn_prepare fails then prepare_and_commit_distributed_txn throws
%% a 'prepare_failed' exception (which as a side-effect causes the the
%% local transaction to be aborted -- that is safe as we know that
%% DistTxnCoordPid has not seen 'successful prepare' from all
%% participants (as this participant at least did not send any such
%% message), so the GTM cannot have been told to record this
%% transaction as committed) The prepare_failed error is propagated
%% back (via normal do_txn_on_master handling of exceptions) to the
%% 'local client' process, which catches the exception as it is
%% (re)thrown from the client side of do_txn_on_master.  The 'local
%% client' process sends the error to DistTxnCoordPid and then exits
%% normally.  It is better to sending of an explicit error message to
%% DistTxnCoordPid, rather than simply crash, as we want
%% DistTxnCoordPid to tell all of the other participants to abort too.
%% (If the local client process crashed then DistTxnCoordPid would 
%% exit, and all of the other workers would wait for their full timeout
%% -- while holding locks via their prepared transactions).
%% Note that txn_prepare should rarely fail (it does not attempt to
%% acquire locks -- although it does write to disk and do synchronous
%% replication) and the error path achieves the same end).
%%
%% If a 'local client' process hits a problem when calling
%% do_txn_on_master (e.g. site_down exception) then it sends an
%% appropriate error to DistTxnCoordPid, and exits normally.  If
%% DistTxnCoordPid receives such an error, or if DistTxnCoordPid gets
%% a timeout or an explicit error from the 'do_txn_on_master worker'
%% process while waiting for 'successfully prepared' message from any
%% participant then it will send messages to all other participants,
%% telling them to abort their local transactions.
%%
%% After any of these 'abort' scenarios, DistTxnCoordPid waits (with a
%% timeout) for responses from the remaining local client processes
%% (i.e. for do_txn_on_master to return).  TODO: In theory,
%% DistTxnCoordPid does not need to wait for the responses -- it could
%% just throw.  But waiting may help avoid piling up of retry
%% attempts.
%%
%% If DistTxnCoordPid crashes (this should never happen) for some
%% other reason when waiting for 'successfully prepared' message from
%% a participant then all of the local client processes will be killed
%% via link signales, and the spawned do_txn_on_master processes will
%% wait until their timeouts for a message from DistTxnCoordPid and
%% then exit, or will return a result to nowhere.
%%
%% Once DistTxnCoordPid has gathered 'successfully prepared' messages
%% from all participants, it will tell the GTM to record that the
%% transaction is commited.
%%
%% If the GTM returns success then DistTxnCoordPid will tell all of
%% the participants to commit their local transactions. They do so,
%% and then return from ?BDB_RG_H:prepare_and_commit_distributed_txn,
%% which allows the TxnFun to return its result. The results arrive at
%% local spawned client processes (i.e. do_txn_on_master finally
%% returns), and the local client processes forward the results to
%% DistTxnCoordPid (which is waiting for them).  When DistTxnCoordPid
%% has received all of the results, it returns to the caller.
%%
%% If the GTM returns *explicit failure* (e.g. 'already_aborted') then
%% DistTxnCoordPid will tell all of the participants TxnFuns to to
%% abort their local transactions and throw a 'told_to_abort'
%% exception.  Therefore ?BDB_RG_H:prepare_and_commit_distributed_txn
%% will NOT return control to WrappedTxnFun.  In the local client
%% process, do_txn_on_master will throw the same error.  The local
%% client process then simply exits normally. (It does not need to
%% crash as DistTxnCoordPid already knows that all this will happen).
%%
%%
%% TODO:
%%
%% - If a participant's TxnFun fails with {bdb_error, bdb_DB_LOCK_NOTGRANTED} 
%%   then it might be the result of a distributed deadlock -- so 
%%   we should retry NumDeadlockRetries times.
%%
%% See comments at the top of berkeley_db_global_transaction_manager.erl
%%
%% - Currently the caller must know about (and indeed track changes
%%   in) the location of the GTM.  Ideally the caller wouldn't even
%%   know that the GTM existed.
%%
%%   This stems from the fact that The GTM can be considered a
%%   'replicated application', and that is currently a very weak
%%   abstraction.  The GTM server is assumed to be running on every
%%   site in GTM_RgCallerState, and this code calls it via a call to
%%   do_txn_on_master.  A slightly better (but still weak) abstraction
%%   would be for the GTM's own client functions to contain the call
%%   to do_txn_on_master.  However, the caller would still need to
%%   pass the GTM_RgCallerState and accept it as part of the result.
%%   Really we would like for the GTM's client functions to retrieve
%%   RgCallerState from some agreed-upon location (e.g. an ETS table)
%%   and store it there afterwards. The open question being, how does
%%   that location be initially populated on every possible caller
%%   node.  Also, the top-level application is responsible for
%%   starting the GTM on every node mentioned in GTM_RgCallerState.
%%   Agaim that is very weak -- and inconvenient if we ever allow new
%%   sites to be dynamically added to replication groups (we have to
%%   have a mechanism to start all of the 'replicated applications' on
%%   each such site.  It's arguable that berkeley_db_replication_group
%%   should have some notion of the replicated applications that are
%%   running on this group, and be responsible for starting &
%%   supervising them.  However, that still doesn't solve the problem
%%   of how all possible caller sites have an initial correct copy of
%%   RgCallerState.  So the current thinking is, as the abstraction is
%%   weak, let's not take any pains to hide the fact.  The caller has
%%   to track the location of the GTM and start the GTM where they
%%   want it to live.  If the caller adds more sites to the GTM then
%%   they need to adjust GTM_RgCallerState accordingly. That's not too
%%   onerous.

do_distributed_txn_unordered(#?CS{} = GTM_RgCallerState, 
                             RgCallerStatesTxnFuns, 
                             NumDeadlockRetries, 
                             Timeout) ->

    {<<_TxnCounter:64, 0:((128*8)-64)>> = TxnGUID, GTM_RgCallerState2}
        = do_txn_on_master(GTM_RgCallerState, 
                           {?MODULE, txn_fun_get_new_distributed_txn_guid, []}, 
                           NumDeadlockRetries, 
                           Timeout),

    DistTxnCoordPid = self(),

    %% spawn-link the calls to do_txn_on_master.
    ParticipantLocalPids = 
        lists:foldl(
          fun({ParticipantTag, #?CS{} = RgCallerState, {Module, Function, CallerArgs}}, Acc) ->

                  %% We use spawn, not spawn_link, as we return before this process 
                  %% ends and we don't want the exit signal to kill our process after 
                  %% we return.  The process catches all exceptions anyway, 
                  %% so linking to it is not required.

                  LocalClientPid = 
                      spawn(
                        fun() ->
                                %% Package of info for ?BDB_RG_H:prepare_and_commit_distributed_txn
                                DistTxnInfo = {ParticipantTag, TxnGUID, DistTxnCoordPid, Timeout},

                                %%?INFO_LOGGER("dist_txn_local_client_proc: calling do_txn_on_master for ~w with ~w~n", [ParticipantTag, {Module, Function, {DistTxnInfo, CallerArgs}}]),
                                try do_txn_on_master(RgCallerState,
                                                     {Module, Function, {DistTxnInfo, CallerArgs}}, 
                                                     NumDeadlockRetries, 
                                                     Timeout) 
                                of
                                    %% By the time do_txn_on_master returns the RemoteWorkerProcess will 
                                    %% already have sent the TxnFunResult back to DistTxnCoordPid as part 
                                    %% of the 'successfully prepared'message.
                                    %% The value of do_txn_on_master so this is actually the result of the 
                                    %% TxnFun's tail-call to prepare_and_commit_distributed_txn.
                                    %% We assert that here to catch malformed TxnFuns (i.e. 
                                    %% those that don't tail-call prepare_and_commit_distributed_txn.
                                    {prepare_and_commit_finished, _RgCallerState2} -> 
                                        ?INFO_LOGGER("dist_txn_local_client_proc: group ~w returned prepare_and_commit_finished for ~w~n", [ParticipantTag, _TxnCounter])
                                catch
                                    %% We have to tell DistTxnCoordPid that this participant 
                                    %% **appears** to have failed to prepare.
                                    %% NOTE: if the reason is a timeout, then the remote-worker process might 
                                    %% be running, so it might still actually prepare.

                                    throw:Reason ->
                                        DistTxnCoordPid ! {ParticipantTag, do_txn_on_master_failed, TxnGUID, Reason};

                                    Class:Reason ->
                                        DistTxnCoordPid ! {ParticipantTag, do_txn_on_master_failed, TxnGUID, {{Class, Reason}, erlang:get_stacktrace()}}
                                end
                        end),

                  [{ParticipantTag, LocalClientPid} | Acc]
          end,
          [],
          RgCallerStatesTxnFuns),

    ?INFO_LOGGER("do_distributed_txn: spawned participants ~w for ~w~n", [ParticipantLocalPids, _TxnCounter]),

    %% Wait for responses from ALL participant (even if it means waiting the full timeout)
    %% We must get one of these as do_txn_on_master has a timeout
    %%
    %%  - do_txn_on_master error messages from 'local client' process
    %%  - 'sucessfully prepared' messages from do_txn_on_master server-side worker
    %%  - 'failed to prepare'    messages from do_txn_on_master server-side worker
    %%
    %% This returns {PreparedParticipants, FailedParticipants}
    %%
    %% A 'prepared' participant is represented by {ParticipantTag, RemoteWorkerPid, TxnFunResult}
    %% A 'failed'   participant is represented by {ParticipantTag, Error}

    case dist_txn_receive_responses(TxnGUID, ParticipantLocalPids, [], [], Timeout) of

        {PreparedParticipants, []} ->
            ?INFO_LOGGER("do_distributed_txn: all participants successfully prepared ~w~n", [_TxnCounter]),

            %% We have 'successful prepare' messages from all participants.
            %% So we can now commit the transaction

            try do_txn_on_master(GTM_RgCallerState2,
                                 {?MODULE, txn_fun_commit_distributed_txn, TxnGUID}, 
                                 NumDeadlockRetries, 
                                 Timeout)
            of
                {{TxnGUID, committed}, _MLH3} ->

                    %% Success.  Asychronously tell all participants
                    %% to commit their local transactions.  Then we
                    %% can return the result

                    ?INFO_LOGGER("do_distributed_txn: GTM successfully commited ~w~n", [_TxnCounter]),
                    ParticipantResults =
                        lists:foldl(fun({ParticipantTag, RemoteWorkerPid, TxnFunResult}, Acc) ->
                                            ?INFO_LOGGER("do_distributed_txn: telling ~w ~w to commit ~w~n", [ParticipantTag, RemoteWorkerPid, _TxnCounter]),
                                            RemoteWorkerPid ! {ParticipantTag, commit, TxnGUID},
                                            [{ParticipantTag, TxnFunResult} | Acc]
                                    end,
                                    [],
                                    PreparedParticipants),

                    {TxnGUID, ParticipantResults}

            catch
                {bdb_gtm_error, {TxnGUID, already_aborted}} = Error ->

                    %% Failure.  This can happen if some prepared
                    %% participant crashed (after sending us their
                    %% message -- so probably a timeout) and it's
                    %% port-coordinator asked the GTM to resolve the
                    %% txn.  The GTM doesn't have any record of
                    %% TxnGUID, so returns 'aborted' but *also*
                    %% (first) writes an explicit 'aborted' record to
                    %% disk, so that any attempt to commit that txn in
                    %% future will fail.  (This exception means
                    %% precisely the above.)
                    %%
                    %% Asynchronously tell all participants to abort, 
                    %% and then rethrow.

                    ?INFO_LOGGER("do_distributed_txn: GTM explicitly failed to commit ~w~n", [_TxnCounter]),
                    lists:foreach(fun({ParticipantTag, RemoteWorkerPid, _TxnFunResult}) ->
                                          ?INFO_LOGGER("do_distributed_txn: telling ~w ~w to abort ~w~n", [ParticipantTag, RemoteWorkerPid, _TxnCounter]),
                                          RemoteWorkerPid ! {ParticipantTag, abort, TxnGUID}
                                  end,
                                  PreparedParticipants),
                    throw(Error);

                Reason ->
                    %% Note: we CANNOT tell the participants to abort
                    %% on *any* other kind of failure, because it
                    %% might just be a lost message, broken connection
                    %% etc.  and the GTM might actually *have*
                    %% committed the txn.  We could just allow the
                    %% exception to propagate (and we do so if it is
                    %% an 'exit' or 'error).  But to avoid causing the
                    %% participant workers to wait for their full
                    %% timeout for our decision, we send them an
                    %% explicit message to throw an exception (cause
                    %% the do_txn_on_master workers to exit).  That
                    %% causes group_valmap_cleanup to find the
                    %% prepared transactions, and send events to the
                    %% relevant port-coordinators, to tell them to ask
                    %% the GTM to resolve the transactions.

                    lists:foreach(fun({ParticipantTag, RemoteWorkerPid, _TxnFunResult}) ->
                                          ?INFO_LOGGER("do_distributed_txn: telling ~w ~w to that txn is indeterminate ~w~n", [ParticipantTag, RemoteWorkerPid, _TxnCounter]),
                                          RemoteWorkerPid ! {ParticipantTag, indeterminate, TxnGUID}
                                  end,
                                  PreparedParticipants),
                    throw(Reason)
            end;

        {PreparedParticipants, FailedParticipants} ->
            %% Some participants failed.  Tell all prepared
            %% participants to abort, and then throw.  Asynchronously
            %% tell all participants to abort, and then rethrow.
            ?INFO_LOGGER("do_distributed_txn: some partitpants failed ~w for ~w~n", [FailedParticipants, _TxnCounter]),
            lists:foreach(fun({ParticipantTag, RemoteWorkerPid, _TxnFunResult}) ->
                                  RemoteWorkerPid ! {ParticipantTag, abort, TxnGUID}
                          end,
                          PreparedParticipants),
            throw({berkeley_db_distributed_txn_error,
                   {some_participants_failed, TxnGUID, FailedParticipants}})
    end.


%% TxnFuns to call the Global Transaction Manager in the style of a
%% 'weakly replicated application' (see comments at top of
%% berkeley_db_global_transaction_manager.erl)
%%
%% Note: each TxnFun is executed on the GTM master node.  But it still
%% needs to find the name of GTM server for that replication group +
%% site.  That means we need to know which site we were sent to, in
%% order to call
%% ?BDB_GTM:get_proc_name_from_replication_group_site_name().
%% (Unfortunately the GTM instances cannot simply be named
%% 'berkeley_db_global_transaction_manager' as we might need more than
%% one instance per node.)
%%
%% The caller of do_txn_on_master does not know the site, as the
%% previous master may be down and do_txn_on_master might just
%% have discovered this new master.
%%
%% The current TxnFun interface does not have a mechanism for
%% that.
%%
%% Options:
%%
%% 1. The TxnFun is passed a BdbPort handle which contains a
%%    port_coordinator_name member. Replication group sites have
%%    port-coordinator names which follow a specific scheme, so we
%%    could parse the replication-group name and site name from
%%    the port-coordinator name.  Aside from being a hack, and
%%    being slow, this is weak; the current port-coordinator
%%    naming scheme uses underscores as the delimiter between
%%    parts of the name, but the replication group name and site
%%    names might contain underscores too.  Rejected.
%%
%% 2. We could add a SiteName argument to TxnFun, but that would
%%    mean that a TxnFun's signature would need to vary depending
%%    on whether it was used with a replicated call
%%    (do_txn_on_master) or a non-replicated call.  That's too
%%    heavy a price to pay for a parameter that most TxnFuns won't
%%    use.
%%
%% 3. The 'replication aware' forms of do_txn
%%    (i.e. do_txn_on_master, do_txn_on_any_site, could store 'out
%%    of band' parameters in the process dictionary of the worker
%%    process that executes TxnFun.  So TxnFun can just use get().
%%    Of course, this feels like a hack too.  e.g. We have a
%%    potential (slight) optimization for do_txn_on_master to
%%    execute the TxnFun in the *callers* process if the master
%%    happens to be on this node.  If we do that then
%%    do_txn_on_master would be writing to the caller processes'
%%    dictionary, which in my view is unacceptable.  So this
%%    approach would preclude that optimization.  (It's not much
%%    of an optimization anyway -- only saves a spawn, and as
%%    already noted elsewhere it does change the behaviour a
%%    little as the Timeout arg to do_txn_on_master would nolonger
%%    be enforced, as the caller would be executing the TxnFun not
%%    blocked in a gen_server:call. So maybe it was a bad idea to
%%    begin with.)
%%
%%    The benefits of this approach (if we formalize it) are that
%%    the TxnFun could have access (if it wished) to a bunch of
%%    information about the replication group in which it was
%%    executing -- the list of all the sites, the current 'down
%%    sites' and 'up sites' lists, the name of the local
%%    replication_group server.  Of course if the TxnFun uses any
%%    of that stuff then it *is* replication-aware, so the
%%    objection in #2 (using a different signature (arity) for
%%    TxnFun's that are replicated) does not really add anything.
%%    And it doesn't affect how we write 'weakly replicated
%%    applications', only how we *call* them.  ('Weakly replicated
%%    applicationes' are (by definition) written in a style that
%%    are not replication-aware and only use do_txn internally --
%%    they are *called* via do_txn_on_master, but don't known
%%    that. The *caller* uses a TxnFun that knows it is hitting a
%%    replication master (e.g. this one) but the application that
%%    is being called is ignorant of the fact. The caller just
%%    chooses the site that is currently master.)  But the benefit
%%    of not doing #2 is that *most* TxnFun's don't need to know
%%    whether they are being used in replicated or non-replicated
%%    mode.  
%%
%%    Another alternative would be for all TxnFuns to take 3
%%    args. In replicated mode the second arg would be some
%%    replication-group info. In non-replicated mode it would be
%%    something else -- the name of the port-coordinator (although
%%    that is already available in the BdbPort handle).
%%
%%
%% The current code uses approach #3, as it is a minimal code change.
%% do_txn_on_master puts the site_id {ReplicationGroupName, SiteName} into 
%% the process dictionary of the spawned worker, before it calls do_txn.

txn_fun_get_new_distributed_txn_guid(_BdbPort, _) ->

    {ReplicationGroupName, SiteName} = get(txn_fun_replication_context),
    GTM_ServerName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(ReplicationGroupName, 
                                                                             SiteName),
    ?BDB_GTM:get_new_distributed_txn_guid(GTM_ServerName).


txn_fun_commit_distributed_txn(_BdbPort, TxnGUID) ->

    {ReplicationGroupName, SiteName} = get(txn_fun_replication_context),
    GTM_ServerName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(ReplicationGroupName, 
                                                                             SiteName),
    ?BDB_GTM:commit(GTM_ServerName, TxnGUID).


%% Helpers for do_distributed_txn_unordered

%% Returns 
%%    {PreparedParticipants, FailedParticipants}
%%
%% PreparedParticipants = {ParticipantTag, RemoteWorkerPid, TxnFunResult}
%% FailedParticipants   = {ParticipantTag, Reason}
%%
dist_txn_receive_responses(_, [], PreparedParticipants, FailedParticipants, _) ->
    %% We have received 'successful prepare' messages from all participants.
    {PreparedParticipants, FailedParticipants};

dist_txn_receive_responses(<<_TxnCounter:64, 0:((128*8)-64)>> = TxnGUID, PartipantLocalPids, PreparedParticipants, FailedParticipants, Timeout) ->
    StartTime = now(),
    receive
        {ParticipantTag, prepared, TxnGUID, RemoteWorkerPid, TxnFunResult} ->

            ?INFO_LOGGER("dist_txn_receive_responses: participant ~w has successfully prepared ~w~n", [ParticipantTag, _TxnCounter]),
            dist_txn_receive_responses(TxnGUID, 
                                       lists:keydelete(ParticipantTag, 1, PartipantLocalPids), 
                                       [{ParticipantTag, RemoteWorkerPid, TxnFunResult} | PreparedParticipants],
                                       FailedParticipants,
                                       reduced_timeout(Timeout, StartTime));
            
        {ParticipantTag, do_txn_on_master_failed, TxnGUID, Reason} ->

            %% Either we failed to contact this participant, or it
            %% failed to successfully prepare the transaction.  We
            %% only need one such error to force us to abort the
            %% distrbuted transaction.  
            %%
            %% TODO: we *could* just return here (i.e. treat this like
            %% the timeout case below, and convert all remaining
            %% members of ParticipantLocalPids to FailedParticipants).
            %% Currently we don't return immediately.  We try to get a
            %% response (or error) from all remaining partipants, so
            %% we know who to tell to locally abort their prepared
            %% transactions. That's potentially better than 
            %% abandoning everything now, because the participants 
            %% that we _don't_ explicitly tell to abort will 
            %% hold locks for the full duration of their timeout 
            %% (waiting for a message from us) and then have to 
            %% ask their local port-coordiantor to resolve their 
            %% part of the transaction with the GTM.

            ?INFO_LOGGER("dist_txn_receive_responses: participant ~w returned an error ~w for ~w~n", [ParticipantTag, Reason, _TxnCounter]),
            dist_txn_receive_responses(TxnGUID, 
                                       lists:keydelete(ParticipantTag, 1, PartipantLocalPids), 
                                       PreparedParticipants,
                                       [{ParticipantTag, Reason} | FailedParticipants],
                                       reduced_timeout(Timeout, StartTime))
    after
        Timeout ->
            %% We consider all remaining 'open' participants to have failed.
            %% Note that they might have actually successfully prepared, 
            %% and their messages simply been lost or delayed.  That's 
            %% unfortunate (as we can't send them 'abort' instructions 
            %% so they must now wait for their full timeout and then 
            %% ask the GTM for a decision), but we cannot wait forever.
            ?INFO_LOGGER("dist_txn_receive_responses: reached timeout without responses from ~w for ~w~n", [PartipantLocalPids, _TxnCounter]),
            {PreparedParticipants, [lists:foldl(fun({ParticipantTag, _}, Acc) -> 
                                                        [{ParticipantTag, timeout_waiting_for_prepared} | Acc] 
                                                end, 
                                                [], 
                                                PartipantLocalPids)
                                    | FailedParticipants]}
    end.


%% This is tail-called called by an application's TxnFun.
%% It always returns 'ok' (to the do_txn_on_master worker process)
%% or throws.
%% The TxnFuns' result is actually sent back to DistTxnCoordPid
%% as part of the 'successfully prepared' message.  Doing so allows 
%% do_distributed_txn to return as soon as the transaction has been 
%% successfully recorded as committed by the GTM and asynchronous 
%% 'commit' messages have been sent back to the processes waiting 
%% in this function.  (The caller doesn't care if one of those messages 
%% fails to arrive, or fails to be executed. The fact that the GTM
%% has recorded commitment is enough.)
%% So do_txn_on_master eventually just returns 'ok' in each 'local 
%% client' process -- or throws an error, Either way that result is 
%% ignored.
prepare_and_commit_distributed_txn(BdbPort, Txn, TxnFunResult, {ParticipantTag, TxnGUID, DistTxnCoordPid, Timeout}) ->
    
    try ?BDB:txn_prepare(BdbPort, Txn, TxnGUID) of

        ok ->
            %% CRITICAL: we CANNOT now unilaterally abort or commit.
            %% We have to get an explicit instruction.

            %% Tell the coordinator that we are waiting for instructions.
            DistTxnCoordPid ! {ParticipantTag, prepared, <<_TxnCounter:64, 0:((128*8)-64)>> = TxnGUID, self(), TxnFunResult},

            receive
                %% These are asynchronous messages -- there's no 
                %% need to reply (although when we return the 
                %% result will be propagated back as the TxnFunResult
                {ParticipantTag, commit, TxnGUID} -> 
                    ?INFO_LOGGER("prepare_and_commit_distributed_txn: told to commit ~w~n", [_TxnCounter]),
                    ?BDB:txn_commit(BdbPort, Txn, []),

                    %% The process that called do_txn_on_master
                    %% asserts that it returns the following, to catch
                    %% malformed TxnFuns (that don't tail-call
                    %% prepare_and_commit_distributed_txn)
                    prepare_and_commit_finished;

                {ParticipantTag, indeterminate, TxnGUID} ->
                    %% This means that DistTxnCoordPid failed to get a response from the GTM.
                    %% The GTM may or may not have commited the transaction.
                    %% To handle this we only need to throw an exception.
                    %% That causes group_valmap_cleanup to process the transaction.  
                    %% As it is a prepared (unresolved) txn, group_valmap_cleanup 
                    %% will NOT auto-abort it. Instead it sends a 'need_to_resolve' event
                    %% to the port-coordinator (including TxnGUID and the Txn valmap handle).
                    %% That event causes the port-coordinator to ask the GTM whether to
                    %% commit or abort (which the port-coordinator then does).
                    ?INFO_LOGGER("prepare_and_commit_distributed_txn: told to ask GTM to resolve~w~n", [_TxnCounter]),
                    throw({indeterminate_distributed_txn, TxnGUID});

                {ParticipantTag, abort, TxnGUID} ->
                    %% This is rare. It means that the GTM told DistTxnCoordPid that 
                    %% the transaction was already aborted, because another participant 
                    %% asked the GTM to resolve the transaction before DistTxnCoordPid 
                    %% tried to commit it.
                    ?INFO_LOGGER("prepare_and_commit_distributed_txn: told to abort ~w~n", [_TxnCounter]),
                    ?BDB:txn_abort(BdbPort, Txn),

                    %% We return rather than throwing because DistTxnCoordPid 
                    %% told us to do this.
                    prepare_and_commit_finished

            after
                Timeout ->
                    %% If we don't hear from DistTxnCoordPid then we have the same situation 
                    %% as 'indeterminate' above. 
                    ?INFO_LOGGER("prepare_and_commit_distributed_txn for ~w reached timeout~n", [_TxnCounter]),
                    throw({berkeley_db_distributed_txn_error,
                           {timeout_waiting_for_decision, TxnGUID, DistTxnCoordPid, Timeout}})
            end
    catch
        Exception ->
            ?INFO_LOGGER("Prepare failed ~w on ~w~n", [Exception, TxnGUID]),
            %% Throwing causes the local transaction to be aborted, 
            %% which is safe as we know that DistTxnCoordPid was 
            %% never told that all partipants had successfully prepared.
            %% This error propagates back to DistTxnCoordPid via 
            %% the local client proc for this call to do_txn_on_master,
            %% and causes DistTxnCoordPid to tell all other participants 
            %% to abort.
            throw({ParticipantTag, failed_to_prepare, TxnGUID, Exception})
    end.


%%====================================================================
%% Internal functions
%%====================================================================

update_caller_state_from_group_status(#?CS{replication_group_name = ReplicationGroupName,
                                           persist_write_fun      = PersistWriteFun} = RgCallerState, 
                                      GroupStatus) ->

    UpSiteLocators = [{ReplicationGroupName, SiteName, SiteNode} || {{SiteName, SiteNode}, SiteStatus} <- GroupStatus, SiteStatus =/= unknown],
    
    %% Note: this resets the ignore_down_sites_for_milliseconds time of any site that is already known to be down.
    %% (i.e. such a site might be about to recover).
    %% TODO: If that is a problem we could extract the times from RgCallerState#?CS.down_sites
    Now = now(),
    DownSites = [{{ReplicationGroupName, SiteName, SiteNode}, Now} || {{SiteName, SiteNode}, SiteStatus} <- GroupStatus, SiteStatus =:= unknown],

    MasterSiteLocator
        = case [SiteId || {SiteId, {master, available}} <- GroupStatus] of
              [] ->
                  %% No site is currently an advertizing master, at least according to the site that provided GroupStatus.
                  %% It is still best to specify _some_ up site.
                  %% We prefer a site on *this* node if there is one.
                  %% Otherwise we prefer any other site that is running.
                  case [SiteLocator || {_, _, SiteNode} = SiteLocator <- UpSiteLocators, SiteNode == node()] of
                      [] -> 
                          %% No local site, so consider them all
                          ?random_list_element(UpSiteLocators);

                      LocalUpSiteLocators ->
                          ?random_list_element(LocalUpSiteLocators)
                  end;

              [{SiteName, SiteNode}] ->
                  %% TODO: with the new replication_group state-machine code there _might_ be more than 
                  %% one master listed for short periods.  The way to find the 'right' one 
                  %% is to call ?BDB_RG:get_master_location_hint on the site that we called get_group_status,
                  %% with FindMasterTimeout of 0
                  {ReplicationGroupName, SiteName, SiteNode}
          end,

    NewRgCallerState = 
        RgCallerState#?CS{master_location_hint = #master_location_hint{site_locator = MasterSiteLocator,
                                                                       pid          = get_replication_group_site_pid(MasterSiteLocator)},
                          up_site_locators     = UpSiteLocators,
                          down_sites           = DownSites,
                          last_persisted_time  = Now},

    assert_invariants_RgCallerState(NewRgCallerState),

    %% NOTE: as we've asked for 'authoratative' state from the replication_group server, 
    %% we DON'T call persist_caller_state_if_latest() here.
    %% We simply force the new state into persistent storage.
    ?INFO_LOGGER("update_caller_state_from_group_status: force-writing NewRgCallerState~n"),
    PersistWriteFun(NewRgCallerState), % assumes that NewRgCallerState has same PersistWriteFun

    NewRgCallerState.


%% Note: this function assumes that RgCallerState#?CS.last_persisted_time has not been 
%% modified since RgCallerState was read from persistent storage.
%%
%% Returns the latest version we know about (possibly input version
%% with an updated last_persisted_time, possibly the latest version
%% read from persistent storage, which of course ***may be arbitrarily
%% different from the input version**)
persist_caller_state_if_latest(#?CS{persist_read_fun    = PersistReadFun,
                                    persist_write_fun   = PersistWriteFun,
                                    last_persisted_time = LastPersistedTime} = RgCallerState) ->

    %% We only update the persistent copy if it has not changed since
    %% we read our original version, otherwise our changes may have
    %% been based on out-of-date information, so we risk losing
    %% information. 
    %%
    %% However, note that we don't have a lock on the persistent entry
    %% between the read and the write below, and we don't have an
    %% atomic compare-exchange operation, so there is still a
    %% race-condition -- we may overwrite the latest change.  This is
    %% deemed to be acceptable; each written caller_state is still
    %% valid, and if we do overwrite a 'good' change with a 'less
    %% good' change then the system should adapt and fix that quickly.
    %%
    %% We could try comparing/merging NewRgCallerState with the one we
    %% just got from the table, but I don't think there is any need
    %% for that extra complexity. (Also the notes on the race
    %% condition still apply -- we'd still need some kind of 'test and
    %% set' operation to ensure that we didn't overwrite with a still
    %% newer version).

    LatestRgCallerState 
        = #?CS{last_persisted_time = LatestLastPersistedTime}
        = PersistReadFun(),

    %% We don't call assert_invariants_RgCallerState on LatestRgCallerState 
    %% as it is a read and we don't do that elsewhere. 
    %% i.e. We assume that nothing is trampling on the ETS table (any gross 
    %% trampling will break the type-check above anyway).

    if LatestLastPersistedTime =< LastPersistedTime ->

            %% update last_persisted_time, and write 
            NewRgCallerState = RgCallerState#?CS{last_persisted_time = now()},
            ?INFO_LOGGER("persist_caller_state_if_latest: writing new persist version~n"),
            PersistWriteFun(NewRgCallerState),  % assumes that NewRgCallerState has same PersistWriteFun
            NewRgCallerState;

       true -> 
            %% Return the latest one we know about
            ?INFO_LOGGER("persist_caller_state_if_latest: NOT writing persist version, but returning version read from persist store~n"),
            LatestRgCallerState
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%           This prefers sites on the current node, if there are any.
%% Returns
%%           SiteLocator
%% Throws:
%%           {bdb_replication_group_error, {no_sites_available, NewRgCallerState}}
%%--------------------------------------------------------------------

pick_random_up_site(RgCallerState) ->

    try 
        pick_random_up_site_fast(RgCallerState)
    catch
        {bdb_replication_group_error, {no_sites_available,
                                       #?CS{down_sites = DownSites} = NewRgCallerState}} = Error ->

            %% Here we believe that *all* sites are down, as they have all failed to respond 
            %% to attempts within the last ignore_down_sites_for_milliseconds period.
            %% But that timeout is probably conservative. 
            %% To avoid waiting for the full timeout before being able to access the group again,
            %% we attempt to check a one of the sites to see if they are already running.
            %%
            %% If we do manage to contact a site that is up, we want to get the current group state 
            %% from the replication_group server (and convert it to a caller_state).
            %% Otherwise _all_ 'any site' traffic will be sent to that single site, until the others emerge from 
            %% the timeout period.
            %%
            %% Currently we only check one site (even if that site really is still down) to 
            %% minimize overhead per call.  As the site is randomly chosen, if this function is called 
            %% frequently it will find a site in reasonable time.
            %% The alternative is to check more sites -- but if they really are all down we may 
            %% incur significant overhead (connection attempts) until one of them comes back up.
            %% If the caller really does want to know the 'live' status of the whole group then 
            %% then can map is_site_running() over #CS.site_locators.
            %%
            %% We choose a random site rather than, say, the site that has been down for longest
            %% as this call will be made frequently and repeatedly. We obviously don't want to make the 
            %% same choice each time (or add yet more state to record past choices).

            {ProbeSiteLocator, _} = ?random_list_element(DownSites),

            %%?INFO_LOGGER("*********************************Probing randomly chosen site ~w~n", [ProbeSiteLocator]),
            case get_replication_group_site_pid(ProbeSiteLocator) of

                site_not_running ->
                    %%?INFO_LOGGER("Randomly chosen site is down, rethrowing original error~n"),
                    throw(Error);   % rethrow original exception

                _Pid ->
                    %%?INFO_LOGGER("Asking randomly chosen site for group status~n"),
                    %% We found a site that is up
                    %% Ask it for the current group state, and convert that to caller_state.
                    try ?BDB_RG:get_group_status(ProbeSiteLocator) of
                        GroupStatus ->  
                            {ProbeSiteLocator, update_caller_state_from_group_status(NewRgCallerState, GroupStatus)}
                    catch
                        %% It appears that ProbeSiteLocator site really is not up after all (e.g. may have just crashed).
                        %% Rethrow the original error.
                        _ -> throw(Error)
                    end
            end
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%           This variant does not attempt to connect to any down site (to see if it is up again yet).
%% Returns
%%           {SiteLocator, NewRgCallerState}}
%% Throws:
%%           {bdb_replication_group_error, {no_sites_available, NewRgCallerState}}
%%--------------------------------------------------------------------

pick_random_up_site_fast(#?CS{up_site_locators                   = UpSiteLocators,
                              down_sites                         = DownSites,
                              ignore_down_sites_for_milliseconds = IgnoreDownSitesForMs} = RgCallerState) ->

    %% First resurrect any down_sites that have been down for more than IgnoreDownSitesForMs.
    Now = now(),

    {ResurrectedUpSiteLocators, StillDownSites, CallerStateStatus} 
        = lists:foldl(
            fun({SiteLocator, TimeLastTried} = DownSite, {NewUpSiteLocators, StillDownSites, CallerStateStatus}) ->

                    case ?BDB_U:now_diff_milliseconds(Now, TimeLastTried) of

                        T when T > IgnoreDownSitesForMs ->
                            %% resurrect this site
                            ?INFO_LOGGER("pick_random_up_site_fast: resurrecting site ~w~n", [SiteLocator]),
                            {[SiteLocator|NewUpSiteLocators], StillDownSites, caller_state_has_changed};

                        _ ->
                            %% we guess that this site is probably still down
                            %%?INFO_LOGGER("pick_random_up_site_fast: we still guess that site ~w is down~n", [SiteLocator]),
                            {NewUpSiteLocators, [DownSite|StillDownSites], CallerStateStatus}
                    end
            end,
            {UpSiteLocators, [], caller_state_not_changed},
            DownSites),

    %%?INFO_LOGGER("ResurrectedUpSiteLocators = ~w~n", [Resurrected]),
    %%?INFO_LOGGER("           StillDownSites = ~w~n", [StillDownSites]),

    NewRgCallerState
        = #?CS{up_site_locators = NewUpSiteLocators}

        = case CallerStateStatus of 
              caller_state_has_changed -> 
                  RgCallerState2 = RgCallerState#?CS{up_site_locators = ResurrectedUpSiteLocators,
                                                     down_sites       = StillDownSites},
                  assert_invariants_RgCallerState(RgCallerState2),

                  %% Important: this MAY find a newer version in the persistent store and return that.
                  %% so the return value might be quite diff3rent from the input.
                  persist_caller_state_if_latest(RgCallerState2);

              caller_state_not_changed ->
                  RgCallerState
          end,

    %% Important: NewRgCallerState_might be a quite different version (the latest version read from persistent storage).
    %% So we can't use ResurrectedUpSiteLocators etv, we have to get them from NewRgCallerState.

    case NewUpSiteLocators of 
        [] ->
            %% We don't change master_location_hint, even though it is probably invalid.
            throw({bdb_replication_group_error, {no_sites_available, NewRgCallerState}});

        LatestUpSiteLocators ->
            %% Pick a random site from NewSiteLocators

            %% Optimization::prefer a site local to this node, if there are any
            %% TODO: as RgCallerState is presumably rarely/never transmitted across nodes, 
            %% we could cache LocalUpSiteLocators and avoid the list comprehension here.
            
           SiteLocator =
                case [SiteLocator || {_, _, SiteNode} = SiteLocator <- LatestUpSiteLocators, SiteNode == node()] of
                    [] -> 
                        %% No local site, so consider them all
                        Result = ?random_list_element(LatestUpSiteLocators),
                        %%?INFO_LOGGER("Chose ~w from LatestUpSiteLocators ~w~n", [Result, LatestUpSiteLocators]),
                        Result;

                    LocalUpSiteLocators ->
                        Result = ?random_list_element(LocalUpSiteLocators),
                        %%?INFO_LOGGER("Chose ~w from LocalUpSiteLocators ~w~n", [Result, LocalUpSiteLocators]),
                        Result
                end,

            %% We don't change master_location_hint as it might be fine.
            {SiteLocator, NewRgCallerState}
    end.



reduced_timeout(infinity, _StartTime) ->
    infinity;
reduced_timeout(CurrTimeout, StartTime) ->
    case CurrTimeout - ?BDB_U:now_diff_milliseconds(now(), StartTime) of
        T when T >= 0 -> T;
        _             -> 0
    end.


assert_invariants_RgCallerState(#?CS{replication_group_name             = ReplicationGroupName,
                                     all_site_locators                  = SortedAllSiteLocators,         
                                     master_location_hint               = MasterLocationHint,
                                     up_site_locators                   = UpSiteLocators,
                                     down_sites                         = DownSites,
                                     ignore_down_sites_for_milliseconds = IgnoreDownSitesForMilliseconds,
                                     last_persisted_time                = {LPT1, LPT2, LPT3},
                                     persist_read_fun                   = PersistReadFun,
                                     persist_write_fun                  = PersistWriteFun} = RgCallerState)
  when is_atom(ReplicationGroupName),
       is_list(SortedAllSiteLocators),
       is_list(UpSiteLocators),
       is_list(DownSites),
       is_integer(IgnoreDownSitesForMilliseconds),
       is_integer(LPT1), is_integer(LPT2), is_integer(LPT3), 
       is_function(PersistReadFun),
       is_function(PersistWriteFun) ->

    %% UpSiteLocators ++ all sites in DownSites should == AllSites
    %% This relies on all_site_locators already being in sorted order (see make_initial_caller_state)
    %% This indirectly verifies that all SiteLocators are well-formed, **assuming**
    %% that all_site_locators was properly initialized and is immutable (both of which are true).
    SortedAllSiteLocators = lists:sort(UpSiteLocators ++ [SL || {SL, {_T1, _T2, _T3}} <- DownSites]),

    %% If MasterLocationHint is not unknown then it must be a site from UpSiteLocators
    case MasterLocationHint of

        #master_location_hint{site_locator = MasterSiteLocator,
                              pid          = MasterPid} when is_pid(MasterPid) -> 
            {true, _} = {lists:member(MasterSiteLocator, UpSiteLocators), RgCallerState};   % include RgCallerState in badmatch result for debugging

        unknown -> 
            ok
    end.

%% assert_invariants_SiteLocator({ReplicationGroupName, SiteName, Node}) 
%%   when is_atom(ReplicationGroupName),
%%        is_atom(SiteName),
%%        is_atom(Node) ->   % no is_node() guard predicate
%%     ok.
