%%%-------------------------------------------------------------------
%%% File    : berkeley_db_sequence_server.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%% A server that issues unique 64-bit integers on demand. It
%%% guarantees that (if BDB replication is used) the same integer will
%%% never be issued more than once.
%%%
%%% This manipulates a persistent counter in Berkeley DB
%%% at a local specified by BDBPortCoordinatorName,Id,Key.
%%% Multiple instances of the server can safely share the same 
%%% counter -- although there is no need for that, and it is 
%%% discouraged.
%%%
%%% This is fast. It caches ranges of this value of a user-specified
%%% size, and performs a BDB transaction when the cached range is
%%% exhausted.
%%% 
%%% This can be used as (or by) a 'weakly replicated application' see
%%% comments at the top of berkeley_db_global_transaction_manager.erl
%%% for details.).
%%%
%%% ***IMPORTANT***: The ONLY guarantee offered is that of uniqueness.
%%% In particular, there is **NO** guarantee that the sequence of
%%% integers produced will be contiguous, or even increasing.  For
%%% example:
%%%
%%% - Restarting the server for any reason will cause gaps in the
%%% sequence, as the first value issued is always the start of the
%%% *nex* complete 'cache-range'.  
%%%
%%% - If the sequence_server run as a 'weakly replicated application'
%%% -- i.e. on all sites of a replication, and the caller finds the
%%% master site before calling the server instance on that site (so
%%% sequence_server itself remains ignorant of replication), then it
%%% is possible that an application might see lower values returned 
%%% after higher values.
%%% e.g. 
%%%   1. Site A is master, and application consumes a value, which 
%%%      consumes an entire cached range on that site.  
%%%
%%%   2. The port-coordinator or replication group on site A fails for 
%%%      some reason, and site B is elected master.  Note that this 
%%%      does NOT automatically kill the sequence_server on site A 
%%%      (the sequence_server is not linked to the port-coordiantor,
%%%      and might go for long periods without calling the port-coordiantor).
%%%
%%%   3. Application consumes a new value -- from site B, so the 
%%%      value is the start of the next cached range (a gap).
%%%
%%%   4. The port-coordinator or replication group on site B now fails  
%%%      for some reason, and site A is elected master again.
%%%
%%%   5. The application consumes another value. This is returned 
%%%      from the original OLD cached range on site A.
%%%      i.e. The value returned will be lower than the value returned 
%%%      in step 3.  If more values are consumed they will be
%%%      sequential until the value in step 3 is about to be 
%%%      reached.  Then a new cached range will be started 
%%%      and the next value will be a full cached-range higher 
%%%      than the value in step 3 (leapfrog).
%%%
%%% TODO: 
%%%
%%% - The dependency on the port_coordinator is perhaps a 
%%%   case for having the port_coordinator own these
%%%   (creating them via a port_coordinator API).
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_sequence_server).

-behaviour(gen_server).

%% API
-export([start_link/4, start_link/5, consume_one_value/1, stop/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% Internal functions exported for use with do_txn
-export([txn_fun_consume_next_range/2]).


-define(BDB,    berkeley_db_drv).
-define(BDB_U,  berkeley_db_utilities).
-define(BDB_PC, berkeley_db_port_coordinator).
-define(BDB_H,  berkeley_db_helpers).


%% These config options have no defaults:
%%
%%    {db_id,  Filename}   or {Filename, DBName}
%%    {db_key, DBKey}
%%

-define(DEFAULT_CONFIG_PROPLIST, [{port_coordinator_name,           berkeley_db_port_coordinator},
                                  {num_deadlock_retries,            5},
                                  {persist_every_N_consumed_values, 100}]).


%% When next_value = start_next_batch_value (or if next_value = 'undefined'), we need to consume another batch
-record(state, {next_value, start_next_batch_value}).


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).



%%====================================================================
%% API
%%====================================================================

-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_sequence_server_error, Error})
           end).

%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%
%% Notes
%%
%% PersistEveryNConsumedValues
%% - If this is small we do frequent transactions to consume message-ids.
%%   (If it is 1 then we do a transaction every time).
%%   If it is large we do less transactions but more counter values are wasted 
%%   if the server exits (even normally), because on restart we have to start 
%%   consuming the next batch.
%%--------------------------------------------------------------------
start_link(SequenceServerName, PortCoordinatorName, DBId, DBKey) ->
    start_link(SequenceServerName, PortCoordinatorName, DBId, DBKey, []).


start_link(SequenceServerName, PortCoordinatorName, DBId, DBKey, ConfigProplist) ->
    gen_server:start_link({local, SequenceServerName}, 
                          ?MODULE, 
                          [{port_coordinator_name, PortCoordinatorName},
                           {db_id,                 DBId},
                           {db_key,                DBKey}
                           | ConfigProplist], 
                          []).

consume_one_value(SequenceServerName) ->
    ?THROW_ERRORS( gen_server:call(SequenceServerName, consume_one_value) ).

stop(SequenceServerName) ->
    gen_server:cast(SequenceServerName, stop).


%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%--------------------------------------------------------------------
init(SpecifiedConfigProplist) ->

    ?INFO_LOGGER("starting (~w)~n", [self()]),

    %% Append the default config to the list specified by the caller, and extract all of the first occurences (the final config).
    ConfigProplist = ?BDB_U:get_dominant_proplist_values([SpecifiedConfigProplist,
                                                          ?DEFAULT_CONFIG_PROPLIST]),

    %% Store the final config in the process dictionary so that it will be logged if the server crashes.
    put(config_proplist, ConfigProplist),

    %% Display the final config.
    lists:foreach(fun({Key, Value}) -> ?INFO_LOGGER_NO_TIMESTAMP("init: ~32w = ~200p~n", [Key, Value]) end, lists:keysort(1, ConfigProplist)),

    %% Validate the config options
    lists:foreach(fun({port_coordinator_name,           Name})             when is_atom(Name)                          -> ok;
                     ({db_key,                          DBKey})            when is_list(DBKey) orelse is_binary(DBKey) -> ok;
                     ({db_id,                           Filename})         when is_list(Filename)                      -> ok;
                     ({db_id,                           {Filename, DBId}}) when is_list(Filename), is_list(DBId)       -> ok;
                     ({num_deadlock_retries,            N})                when is_integer(N), N >= 0                  -> ok;
                     ({persist_every_N_consumed_values, N})                when is_integer(N), N > 0                   -> ok;
                     (Other) -> throw({unrecognized_config_option, Other})
                  end,
                  ConfigProplist),

    %% Store immutable config in process dictionarly not State, so
    %% that State is as small as possible (to minimize garbage
    %% collection -- this server may be called very frequently).

    put(port_coordinator_name,           proplists:get_value(port_coordinator_name, ConfigProplist)),
    put(db_id,                           proplists:get_value(db_id, ConfigProplist)),
    put(db_key,                          case proplists:get_value(db_key, ConfigProplist) of
                                             DBKey when is_binary(DBKey) -> DBKey; 
                                             DBKey                       -> list_to_binary(DBKey) 
                                         end),
    put(persist_every_N_consumed_values, proplists:get_value(persist_every_N_consumed_values, ConfigProplist)),
    put(num_deadlock_retries,            proplists:get_value(num_deadlock_retries, ConfigProplist)),
    

    %% We DON"T consume a range to begin with as 
    %%   a. in replicated applications we are unlikely to have a master at startup
    %%   b. this server might never be called in its lifetime, so consume a range would simply be wasteful

    {ok, #state{next_value = undefined}}.


%%--------------------------------------------------------------------
%% Function: %% handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(consume_one_value, _FromPid, #state{next_value             = NextValue,
                                                start_next_batch_value = StartNextBatchValue} = State) ->

    if is_integer(NextValue) andalso (NextValue < StartNextBatchValue) ->

            %%?INFO_LOGGER("handle_call(consume_one_value): Returning ~w to ~w (~w values left in range)~n", [NextValue, _FromPid, StartNextBatchValue - (NextValue + 1)]),
            {reply, {ok, NextValue}, State#state{next_value = NextValue + 1}};

       NextValue =:= undefined orelse NextValue == StartNextBatchValue ->

            %% Current batch is exhausted (or we havenpt even start one yet).
            %% So we need to start a batch now.

            NewState = ?BDB_H:do_txn(get(port_coordinator_name),
                                     {?MODULE, txn_fun_consume_next_range, State},
                                     get(num_deadlock_retries)),

            NewNextValue = NewState#state.next_value,

            %%?INFO_LOGGER("handle_call(consume_one_value): NEW RANGE: Returning ~w to ~w (~w values left in range)~n", [NewNextValue, _FromPid, NewState#state.start_next_batch_value - (NewNextValue + 1)]),
            {reply, {ok, NewNextValue}, NewState#state{next_value = NewNextValue + 1}}
    end.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_info(Unrecognized, _State) ->
    exit({unrecognized_message, Unrecognized}).

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, State) ->

    ?INFO_LOGGER("terminate: (reason: ~p) State = ~p~n", [Reason, State]),
    
    %% Even though we may be shutting down cleanly, it is not safe to attempt to 
    %% avoid wasting values by writing the (unconsumed) State#state.next_value 
    %% back to the database.  If this were guaranteed to be the only 
    %% process using this database entry, then it would be OK.  
    %% But we cannot assume that.
    ok.

%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%% Returns a new State.
%% This IGNORES (overwrites) the current State#state.next_value and State#state.start_next_batch_value 
%% (it uses the database to generate new such values).
%% This works even if the counter did not previously exist (so long as the port and database are writeable).

txn_fun_consume_next_range(BdbPort, #state{next_value = NextValue} = State) ->

    DBId   = get(db_id),
    DB     = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DBId),
    DBKey  = get(db_key),

    Txn    = ?BDB:txn_begin(BdbPort, void, []),    
    Cursor = ?BDB:db_cursor(BdbPort, DB, Txn, []),

    NewState = 
        %% TODO: use c_get_data
        try ?BDB:c_get(BdbPort, Cursor, DBKey, <<>>, [bdb_DB_SET, bdb_DB_RMW]) of 

            {_, <<ActualStartOfNextBatchValue:64>>}
            %% ... do some sanity checks, but remember that we might still be initializing our 
            %% ... in-memory value (i.e. next_value might be undefined)
            when NextValue =:= undefined 
                 orelse (is_integer(NextValue) andalso NextValue =< ActualStartOfNextBatchValue) ->

                %% There is a value in the database.  

                StartNextBatchValue = ActualStartOfNextBatchValue + get(persist_every_N_consumed_values),

                %%?INFO_LOGGER("consume_next_range: START NEW RANGE: Storing StartNextBatchValue = ~w~n", [StartNextBatchValue]),
                ?BDB:c_put(BdbPort, 
                           Cursor, 
                           <<>>, 
                           <<StartNextBatchValue:64>>, 
                           bdb_DB_CURRENT),

                State#state{next_value             = ActualStartOfNextBatchValue,
                            start_next_batch_value = StartNextBatchValue};

            {_, Value} -> 
                %%?ERROR_LOGGER("consume_next_range: Invalid sequence value in database: ~p~n", [Value]),
                exit({invalid_sequence_value_in_database, DBId, DBKey, Value})

        catch
            {bdb_error, bdb_DB_NOTFOUND} 
            %% ... sanity check: we better not think we already have a valid next_value...
            when NextValue =:= undefined ->

                %% No persistent entry at all (yet)
                %%?INFO_LOGGER("consume_next_range: START NEW SEQUENCE: Key '~p'~n", [get(db_key)]),
                PersistEveryNConsumedValues = get(persist_every_N_consumed_values),
                ?BDB:c_put(BdbPort, 
                           Cursor, 
                           DBKey, 
                           <<PersistEveryNConsumedValues:64>>, 
                           bdb_DB_KEYFIRST),

                State#state{next_value             = 0,
                            start_next_batch_value = PersistEveryNConsumedValues}
        end,    

    ?BDB:c_close(BdbPort, Cursor),
    ?BDB:txn_commit(BdbPort, Txn, []),
    NewState.


