
-module(examples_no_replication).
-export([example_1/0]).

%% It is possible to use the berkeley_db driver in several different
%% ways (the regression tests give examples of each).
%%
%% This file shows the recommended way.
%%
%% Again, please pay attention to the serious warning at the top of 
%% README-cnewcom : read the BDB documentation first.  
%% Misuse of BDB APIs can cause system crashes and/or loss or corruption 
%% of data.

%% Note that we don't need to include any .hrl files
-define(BDB_PC,berkeley_db_port_coordinator).
-define(BDB_H, berkeley_db_helpers).
-define(BDB,   berkeley_db_drv).


%% Internal functions exported for use with ?BDB_H:do_txn
-export([txn_fun_simple_put/2, txn_fun_simple_get/2, 
         txn_fun_put_account/2, txn_fun_get_account_by_email/2]). 


example_1() ->

    %% The port-coordiantor can be owned by a normal OTP supervisor
    %% (not shown here).
    %%
    %% This call creates a process that is locally registered as
    %% 'berkeley_db_port_coordinator'.  (You can also run multiple
    %% named port-coordinators, but most applications don't need to do
    %% that.)
    %%
    %% The only required argument is a BDB 'environment' directory,
    %% which must already exist.  (There are a lot of optional
    %% configuration arguments, not shown here.)

    {ok, _ServerPid} = ?BDB_PC:start_link("example_bdb_data_dir"),


    %% Now some example transactions
    %%
    %% Applications should call ?BDB_H:do_txn/2, which executes a
    %% {Module, Function, ExtraArgs} and will retry that operationq
    %% some number of times if BDB throws a deadlock exception. (See
    %% BDB documentation for an explanation of deadlocks and how to
    %% minimize them).

    %% Do a simple write (to a single key in a single database table)
    %% Note that key and data arguments may be strings or binaries.
    %% Strings will be converted to binaries internally, 
    %% so it is better to use binaries where possible.

    ?BDB_H:do_txn({?MODULE, txn_fun_simple_put, 
                   {<<"my-key">>, <<"my-data">>}}, 3),

    %% Read it back (note that retrieved data is always returned as
    %% binaries)

    <<"my-data">> = ?BDB_H:do_txn({?MODULE, txn_fun_simple_get, 
                                   "my-key"}, 3),


    %% A slightly more complex example: a multi-part transaction
    %% involving several databases

    ExampleAccountObj = [{email_addr, "my-email-address"}, 
                         {balance, 100}],

    ?BDB_H:do_txn({?MODULE, txn_fun_put_account, 
                   {"my-account-name", ExampleAccountObj}}, 3),

    {<<"my-account-name">>, ExampleAccountObj}
        = ?BDB_H:do_txn({?MODULE, txn_fun_get_account_by_email, 
                         "my-email-address"}, 3),

    %% Clean shutdown
    %%
    %% If owned by a supervisor, the port-coordinator should have a
    %% 'Shutdown policy' that is an integer (maybe 5000) - to allow it
    %% to cleanly close BDB.  Note that as BDB uses transactions and
    %% write-ahead logging, any committed data is safe even if the
    %% port-coordinator crashes or is killed.

    ok = ?BDB_PC:stop(),
    
    io:format("~nExample finished successfully.~n~n"),

    %% As this is just an example, we don't bother waiting for the EXIT 
    %% signal for _ServerPid.  
    %% As ?BDB_PC:stop is asynchronous, this means that BDB doesn't 
    %% get a chance to shutdown cleanly.  But that's OK as it is transactional
    %% and if/when we re-open that BDB environment, recovery will fix things.

    erlang:halt().


%% Important: TxnFuns may be run multiple times due to BDB deadlocks.
%% So these should be idempotent and side-effect free.

txn_fun_simple_put(BdbPort, {Key, Data}) ->
    %% ... this creates the database on first use
    %% ... (database names must be strings, not binaries)
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),

    %% Passing 'void' as the transaction handle means 'auto-commit')
    ?BDB:db_put(BdbPort, DB, void, Key, Data, []).


txn_fun_simple_get(BdbPort, Key) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),
    ?BDB:db_get_data(BdbPort, DB, void, Key, []).
    

txn_fun_put_account(BdbPort, {AccountName, AccountObj}) ->
    {value, {email_addr, EmailAddr}} = lists:keysearch(email_addr, 1, AccountObj),

    AccountDB    = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "account.db"),
    IndexEmailDB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "index_email_to_account.db"),

    %% Minor optimization; transactions should be as short as possible to minimize 
    %% the time that locks/resources are held.  So it's probably worth doing any work 
    %% possible before starting the transaction.
    AccountObjBin = term_to_binary(AccountObj),

    Txn = ?BDB:txn_begin(BdbPort, void, []),
    %% Optimization: we access IndexEmailDB first, as get_account_by_email must do so, 
    %% and consistent ordering of lock acquisition greatly helps to reduce deadlocks.
    ?BDB:db_put(BdbPort, IndexEmailDB, Txn, EmailAddr, AccountName, []),
    ?BDB:db_put(BdbPort, AccountDB, Txn, AccountName, AccountObjBin, []),
    ?BDB:txn_commit(BdbPort, Txn, []).


txn_fun_get_account_by_email(BdbPort, EmailAddr) ->
    AccountDB    = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "account.db"),
    IndexEmailDB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "index_email_to_account.db"),

    Txn = ?BDB:txn_begin(BdbPort, void, []),
    AccountName   = ?BDB:db_get_data(BdbPort, IndexEmailDB, Txn, EmailAddr, []),
    AccountObjBin = ?BDB:db_get_data(BdbPort, AccountDB, Txn, AccountName, []),
    %% commit as soon as possible to release locks as soon as possible 
    %% (e.g. defer calling binary_to_term until afterwards)
    ?BDB:txn_commit(BdbPort, Txn, []),
    {AccountName, binary_to_term(AccountObjBin)}.
