
-module(examples_with_replication).

-export([example_API_set_1/0,
         example_API_set_2/0]).


%% When using the berkeley_db driver with replication enabled, 
%% there are two sets of APIs that may be used.
%% This file gives examples of both.
%%
%% (For simplicity of example, this site runs one replication-group of 3 sites 
%%  all on the same Erlang node.  This may make startup a little sluggish, 
%%  depending on your hardware, as 3 BDB environments are opened in parallel.
%%  It should be easy to modify this file to run on 3 separate nodes 
%%     - change the ?BDB_RG:start_sites to ?BDB_RG:start_link
%%       (the arguments are only slightly different -- see docs.)
%%     - change the node names in RgAllSites.)
%%
%% The two sets of replication APIs are documented at the top of the relevant source files.
%% Read them in the following order, after reading BDB's 
%% own documentation on replication
%%
%%  http://www.oracle.com/technology/documentation/berkeley-db/db/gsg_db_rep/C/index.html
%%  http://www.oracle.com/technology/documentation/berkeley-db/db/ref/rep/intro.html
%%
%% 1. The first set of APIs can/should be used by code that 
%%    is running on the same Erlang node as a site in the replication-group 
%%    that it wants to access.  The APIs are called on the 
%%    'local' replication-group server, and the request is transparently 
%%    forwarded to another site if necessary (e.g. if do_txn_on_master is 
%%    called on a site/node that is not master).
%%  
%%    berkeley_db_replication_group.erl
%%
%%        do_txn_on_site
%%        do_txn_on_any_site
%%        get_master_location_hint
%%        do_txn_on_master
%%        get_node_from_master_location_hint
%%        get_site_locator_from_master_location_hint
%%        get_group_status
%%    
%% 2. The second set of APIs has similar functionality to the first set,
%%    but offers better support for accessing a 'remote' replication-group
%%    -- i.e. from an Erlang node that is not running a site that belongs 
%%    to that group. The main difference is that these APIs 
%%    attempt to work around sites/nodes that are unreachable, whereas the 
%%    first set of APIs will usually fail if the 'local' site is down. 
%%    To achieve this, these APIs keep a small 'caller state' record 
%%    on the caller node (usually storing it in an ETS table).
%%    This record tracks the current location of the master, which
%%    sites are down (and any back-off/retry times), etc.
%%
%%    Note; this second set of APIs may of course be used like the first set;
%%    on nodes that are running a site of the target replication-group.  
%%    The benefit is that if the local site is down (e.g. restarting/recovering)
%%    but the node/user-application is still running, then the APIs will usually 
%%    be able to find another site to use until the local site is running again.
%%
%%    berkeley_db_replication_group_helpers.erl
%%
%%        do_txn_on_any_site
%%        refresh_master_location_hint
%%        do_txn_on_master
%%        get_MLH_from_RgCallerState
%%
%%    This second set of APIs also offers a convenient way to do fully 
%%    distributed transactions across the masters of several independent 
%%    replication-groups:
%%
%%        do_distributed_txn_unordered
%%        prepare_and_commit_distributed_txn
%%
%% Again, please pay attention to the serious warning at the top of 
%% README-cnewcom : read the BDB documentation first.  
%% Misuse of BDB APIs can cause system crashes and/or loss or corruption 
%% of data.


%% Note that we don't need to include any .hrl files
-define(BDB_RG_H,  berkeley_db_replication_group_helpers).
-define(BDB_RG,    berkeley_db_replication_group).
-define(BDB_PC,    berkeley_db_port_coordinator).
-define(BDB,       berkeley_db_drv).


%% Internal functions exported for use with ?BDB_H:do_txn
-export([txn_fun_simple_put/2, txn_fun_simple_get/2, 
         txn_fun_put_account/2, txn_fun_get_account_by_email/2]). 


example_API_set_1() ->

    %% Init code common to both exaples -- please read the comments in this function.

    start_example_replication_group_sites(),

    %% Once started, sites are addressed with SiteLocators which are simply 3-tuples:
    %%
    %%   {ReplicationGroupName, SiteName, SiteNode}.
    %%
    %% In this case, all of the sites are 'local'.  For the same of example we choose 
    %% to pretend that the local site is one of the (probably) replicas (as we 
    %% specified that site1 is our preferred master).

    LocalSiteLocator = {example_group, site3, node()},

    %% Wait for a master to be elected and find it's location 
    %% (the location can be passed to do_txn_on_master as an optimization 
    %% to avoid hops).
    %% Typically this only needs to be done on startup.

    MasterLocationHint = ?BDB_RG:get_master_location_hint(LocalSiteLocator, 20000),

    %% Now some example transactions.
    %%
    %% To do write transactions, applications should call ?BDB_RG:do_txn_on_master,
    %% which is the equivalent of ?BDB_H:do_txn shown in examples_no_replication.erl.
    %%
    %% To do read transactions, the application nees to choose the level of 
    %% consistency it desires:
    %%
    %%  - for stronger consistency, use do_txn_on_master (and replicas 
    %%    are only used as stand-by sites incase the master fails)
    %%
    %%  - for weaker consistency but higher performance, use do_txn_on_any_site.
    %%    
    %% Even if a blocking ack_policy (e.g. bdb_DB_REPMGR_ACKS_QUORUM) is used, 
    %% any individual replica may lag behind the others.  (The replication-manager 
    %% has features to detect persistently lagging replicas and restart them.)
    %% The bdb_DB_REPMGR_ACKS_ALL_PEERS policy could be used to require 
    %% all replicas to ack before txn_commit returns, but at the *major* cost 
    %% of being unable to do any writes if even a single replica crashes or is 
    %% unreachable.
    %%
    %% The "Not Durably Replicated" condition:
    %%
    %% Note that the 'best' consistency above is only a "stronger", not "strong" (absolute).
    %% Even when doing all reads on the master, a consisntecy issue can arise:
    %% With replication enabled, txn_commit on a write transaction may throw
    %%
    %%     {bdb_error, bdb_DB_REP_NOTPERM} 
    %%
    %% This means that the specified acks_policy was not met for that transaction 
    %% -- i.e. the desired number of replicas did not apply and acknowledge the 
    %% transaction within the specified acks timeout.
    %%
    %% This means that consistency is currently unstable:
    %%
    %%    - HAS been applied on the master (and has already been exposed to reads)
    %%
    %%    - MAY have been applied on one or more replicas (but not all of them),
    %%      and has already been exposed to reads on those replicas.
    %%
    %%    - If the master fails, then the next master might NOT have the transaction
    %%      (i.e. the transaction might be lost).
    %%      Note that the election algorithm prefers to elect the site with 
    %%      the 'most' transactions, so unless zero replicas received the transaction
    %%      then it _should_ be safe.  But it is not guaranteed.
    %%
    %% Note that BDB-HA automatically attempts to repair any such inconsistency 
    %% issues -- the application does not need to do anything (indeed it cannot).
    %% The problem may simply be that the acks timeout is too small for the 
    %% current write-load or network-congestion, and the transaction will 
    %% become durable shortly after the error.
    %%
    %% However, a more serious problem may have occured; e.g. 
    %%
    %%   - The master's network connection may have failed.  This is serious because 
    %%     if a majority of sites believe that the master has failed or is unreachable,
    %%     that majority can and should elect another master.  
    %%     Writing to two masters in the same group is extremely bad, as some 
    %%     transactions will *certainly* be lost when the masters finally discover 
    %%     each other, and one syncs-up with the other.
    %%
    %%   - Too many replicas may have failed to fulfull the acks_policy.
    %%
    %% So the main issue is, should the application tell users/other applications 
    %% that the transaction is 'committed' or not?  This issue cannot be avoided 
    %% as the data modified by the transaction has ALREADY been exposed to reads
    %% on the master and possibly other replicas.
    %% 
    %% Solutions:
    %%
    %% 1. For strong (perfect) consistency, use ?BDB_RG_H:do_distributed_txn_unordered
    %%    (see berkeley_db_test.erl for examples).  However, it is well known that 
    %%    distributed transactions (two phase commit) does not scale very well, due to 
    %%    holding locks on all sites, the latency of the extra round-trips involved, 
    %%    and (ultimately) the cost of distributed deadlocks (these are only detected 
    %%    with timeouts).
    %%
    %% 2. Application-level recognition of BDB-HA semantics; e.g. add explicit version 
    %%    numbers to data when it is modified (this at least allows clients to know for sure
    %%    which version of the data they have read.)
    %%
    %% 3. Wait for the master-lease feature in BDB v4.6.
    %%    If all reads are done on the master and this forthcoming feature is enabled 
    %%    then the master won't expose not-durably-replicated data to other transactions.
    %%    If the acks_policy is violated, instead of txn_commit/txn_prepare throwing 
    %%    bdb_DB_REP_NOTPERM then an election will be called and the ex-master will 
    %%    be quarantined until it has synced-up with the new master.
    %%    This gives strong consistency, at the cost of an election/sync-up 
    %%    whenever bdb_DB_REP_NOTPERM might have happened.


    %% Do a simple write (to a single key in a single database table).
    %%
    %% Note that this is almost identical to the call in examples_no_replication.erl
    %% Main differences are:
    %%
    %%  - do_txn_on_master requires a new (first) argument; 
    %%    either a SiteLocator (e.g. our LocalSiteLocator),
    %%    or (better) a MasterLocationHint.
    %%
    %%  - do_txn_on_master returns a MasterLocationHint, 
    %%    which may be different from the input hint if the master 
    %%    has moved (or if a SiteLocator was passed).

    {_TxnResult, _PossiblyNewMasterLocationHint} 
        = ?BDB_RG:do_txn_on_master(MasterLocationHint, 
                                   {?MODULE, txn_fun_simple_put, 
                                    {"my-key", "my-data"}}, 
                                   3),

    %% Do a read txn.
    %%
    %% Here we choose 'weak' consistency (read on a replica), so there is a 
    %% chance that the specified replica has not yet applied the txn 
    %% (our acks_policy only requires that a majority of sites have applied it.)
    %%
    %% do_txn_on_any_site prefers the local site if it is 'in sync' 
    %% with the master.

    <<"my-data">> = ?BDB_RG:do_txn_on_any_site(LocalSiteLocator,
                                               {?MODULE, txn_fun_simple_get, 
                                                "my-key"}, 
                                               3),

    %% A slightly more complex example: a multi-part transaction
    %% involving several databases

    ExampleAccountObj = [{email_addr, "my-email-address"}, 
                         {balance, 100}],

    ?BDB_RG:do_txn_on_master(MasterLocationHint,
                             {?MODULE, txn_fun_put_account, 
                              {"my-account-name", ExampleAccountObj}}, 
                             3),

    {<<"my-account-name">>, ExampleAccountObj}
        = ?BDB_RG:do_txn_on_any_site(LocalSiteLocator,
                                     {?MODULE, txn_fun_get_account_by_email, 
                                      "my-email-address"}, 
                                     3),

    %% Clean shutdown
    %%
    %% If owned by a supervisor, the replication-group server should have a
    %% 'Shutdown policy' that is an integer (maybe 5000) - to allow it
    %% to cleanly close BDB.
    %%
    %% For this example, as we used the 'start_sites' convenience 
    %% function, we use the corresponding 'stop_entire_group' convenience 
    %% function.   This function currently has fairly weak guarantees 
    %% -- in rare circumstances (e.g. if it is restarting) a site might 
    %% miss the order to shutdown.  It is mainly used for tests.
 
    ok = ?BDB_RG:stop_entire_group(LocalSiteLocator),
    
    io:format("~nExample (API set #1) finished successfully.~n~n"),

    %% As this is just an example, we don't bother waiting for the sites to shut down.
    %% As ?BDB_RG:stop_entire_group is asynchronous, this means that the sites don't
    %% get a chance to shutdown cleanly.  But that's OK as it is transactional
    %% and if/when we re-open those BDB environments, recovery will fix things.

    erlang:halt().


example_API_set_2() ->

    %% Init code common to both exaples -- please read the comments in tyis function.

    RgAllSites = start_example_replication_group_sites(),

    %% Now initialize the 'caller state' record that will allow us to 
    %% use the 'remote' replication-group.
    %%
    %% As multiple processes on this node might want to access the remote
    %% replication group, we'd like to share the caller state.
    %% The easiest way is via an ETS table containing a single record.
    %% Note: if an application needs to talk to multiple replication groups 
    %% then one ETS table can be used to hold multiple caller states.
    %%
    %% Add these lines of code to the top of the module:
    %%
    %%   -include("berkeley_db_replication_group_helpers.hrl").
    %%   -define(CS,  berkeley_db_replication_group_caller_state).
    %%
    %% and add this option to the ets:new call:
    %%
    %%   {keypos, #?CS.replication_group_name}
    %%
    %% And then use ets:lookup to find the caller-state for a given 
    %% replication group. 
    %% We don't need that for this example.

    RgCallerStateTableId = ets:new(rg_caller_state, [public]),

    ReadRgCallerState    = fun() -> 
                                   [RgCallerState] = ets:tab2list(RgCallerStateTableId),
                                   RgCallerState
                           end,

    WriteRgCallerState   = fun(RgCallerState) -> 
                                   io:format("RgCallerState being written to persistent store: ~p~n", [RgCallerState]),
                                   ets:insert(RgCallerStateTableId, RgCallerState) 
                           end,

    _RgCallerState1 = ?BDB_RG_H:initialize_caller_state(
                         example_group, 
                         RgAllSites,  % as passed to ?BDB_RG:start_sites
                         5000,        % ignore_down_sites_for_milliseconds
                         ReadRgCallerState,
                         WriteRgCallerState),

    %%
    %% Now we can do our example transactions:
    %%

    %% In general, if the application doesn't have a 'recent' RgCallerState, it should read 
    %% the latest one from ETS table.  (e.g. A server request-handler would do this 
    %% for each request.)

    RgCallerState = ReadRgCallerState(),

    %% Wait for a master to be elected (normally only required on startup).
    %% (When the master is elected this will write am updated RgCallerState to the ETS table,
    %% as well as returning it.)
    
    RgCallerState2 = ?BDB_RG_H:refresh_master_location_hint(RgCallerState, 20000),

    %% Note that the ?BDB_RG_H APIs don't use MasterLocationHints directly,
    %% the hint is stored in the caller state.  If you need to extract the 
    %% hint, use 
    %%
    %%    MasterLocationHint = ?BDB_RG_H:get_MLH_from_RgCallerState(RgCallerState2),
    %% then
    %%    MasterSiteLocator  = ?BDB_RG_H:get_site_locator_from_master_location_hint(MasterLocationHint),
    %% or 
    %%    MasterNode         = ?BDB_RG:get_node_from_master_location_hint(MasterLocationHint),
    %%
    %% Hint: this can be used by a feature that 'moves the application' to the master node, 
    %% to avoid making an inter-node hop per transaction.

    %% Most of the BDB_RG_H APIs return a PossiblyUpdatedRgCallerState in addition to their result.
    %% When making consecutive calls as we do below, it is best to use 
    %% the latest caller state that you have received. (Usually the caller state has not changed, 
    %% and these are references to the same record.)
    %% But even if you don't use the latest state, the APIs should still work
    %% (perhaps incuring an unecessary hop or attempt to contact a site that is down.)

    {_TxnResult, RgCallerState3} 
        = ?BDB_RG_H:do_txn_on_master(RgCallerState2, 
                                     {?MODULE, txn_fun_simple_put, 
                                      {"my-key", "my-data"}}, 
                                     3),

    %% Do a read txn.
    %%
    %% Here we choose 'weak' consistency (read on a replica), so there is a 
    %% chance that the specified replica has not yet applied the txn 
    %% (our acks_policy only requires that a majority of sites have applied it.)
    %%
    %% do_txn_on_any_site prefers the local site if it is 'in sync' 
    %% with the master.

    {<<"my-data">>, RgCallerState4}
        = ?BDB_RG_H:do_txn_on_any_site(RgCallerState3,
                                       {?MODULE, txn_fun_simple_get, 
                                        "my-key"}, 
                                       3),

    %% A slightly more complex example: a multi-part transaction
    %% involving several databases

    ExampleAccountObj = [{email_addr, "my-email-address"}, 
                         {balance, 100}],

    %% ... here's an example of where we don't bother to 
    %% ... bind the PossiblyUpdatedRgCallerState
    ?BDB_RG_H:do_txn_on_master(RgCallerState4,
                               {?MODULE, txn_fun_put_account, 
                                {"my-account-name", ExampleAccountObj}}, 
                               3),
    
    {{<<"my-account-name">>, ExampleAccountObj}, RgCallerState5}
        = ?BDB_RG_H:do_txn_on_any_site(RgCallerState4,
                                       {?MODULE, txn_fun_get_account_by_email, 
                                        "my-email-address"}, 
                                       3),

    %% Shutdown
    %%
    %% Currently there is no 'stop_entire_group' API that uses RgCallerState.
    %% So we fake it...
    
    MasterLocationHint = ?BDB_RG_H:get_MLH_from_RgCallerState(RgCallerState5),
    MasterSiteLocator  = ?BDB_RG:get_site_locator_from_master_location_hint(MasterLocationHint),

    ok = ?BDB_RG:stop_entire_group(MasterSiteLocator),

    io:format("~nExample (API set #2) finished successfully.~n~n"),

    %% As this is just an example, we don't bother waiting for the sites to shut down.
    %% As ?BDB_RG:stop_entire_group is asynchronous, this means that the sites don't
    %% get a chance to shutdown cleanly.  But that's OK as it is transactional
    %% and if/when we re-open those BDB environments, recovery will fix things.

    erlang:halt().


start_example_replication_group_sites() ->

    %% A site is an instance of a replication-group server, usually 
    %% started with ?BDB_RG:start_link.   Each site an be owned 
    %% by a normal OTP supervisor, not shown here.
    %%
    %% (For this example we use a convenience function that allows us 
    %%  to start several sites at once. This is usually only used 
    %%  for tests/examples.  The arguments are very similar to 
    %%  those for ?BDB_RG:start_link.)
    %%
    %% Each replication-group server starts its own private port-coordinator,
    %% and exits if the port-coordinator exits.  Applications usually 
    %% don't interact directly with port-coordinators for replication sites, 
    %% they use the APIs provided by the replication-group server.
    %%
    %% A replication-group has a name (an atom).  This name is 
    %% included in the locally-registered names for the replication-group
    %% servers and port-coordinators that are part of the site 
    %% (the process-naming convention is fixed, and used to find 
    %% sites on remote-nodes).
    %%
    %% When a replication-group site is started, it must be told 
    %% about all other sites in that group -- i.e. with static configuration.
    %% (Currently there is no support for dynamically adding or removing sites,
    %% as BDB-HA's repmgr does not have such support.  I hope to add 
    %% that in future.)

    RgAllSites =
        [{{site1, node()}, "example_bdb_data_dir_site1", {"localhost", 4000}, [{replication, [{this_site_priority, 2}]}]},
         {{site2, node()}, "example_bdb_data_dir_site2", {"localhost", 4010}, [{replication, [{this_site_priority, 1}]}]},
         {{site3, node()}, "example_bdb_data_dir_site3", {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
        ], 

    %%
    %% In the example configuration below, the per-site "replication" options 
    %% are entirely optional. Here we use it to set a preference for which site is master.
    %%
    %% The common_port_coordinator_config options are also optional 
    %% -- these are just some exaples.

    {_GoodSitesPids, [] = _BadSitesErrors} 
        = ?BDB_RG:start_sites(
             example_group, 
             RgAllSites,
             [{log_removal_period_seconds,     disabled},  % just to demonstrate a replication-group config parameter
              %% Common port_coord config should not include any 
              %%      'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
              %% (as those are all set internally)
              %% Note that site-specific port-coordinator config (above) obviously overrides any setting here.
              {common_port_coordinator_config, 
               [{port_start_function,          start},
                {replication,                  [{initial_role_and_threadcount, {bdb_DB_REP_ELECTION, 2}}, 
                                                {ack_policy,		     bdb_DB_REPMGR_ACKS_QUORUM}
                                               ]}]}
             ]),
    
    %% We return this for use by example_API_set_2()
    RgAllSites.


%% The following example transactions are identical to those in examples_no_replication.erl.
%% This shows that replication is almost indepedent of other BDB APIs, 
%% and also independent of the choice of 'set 1' or 'set 2' replication APIs.

%% Important: TxnFuns may be run multiple times due to BDB deadlocks.
%% So these should be idempotent and side-effect free.

txn_fun_simple_put(BdbPort, {Key, Data}) ->
    %% ... this creates the database on first use
    %% ... (database names must be strings, not binaries)
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),

    %% Passing 'void' as the transaction handle means 'auto-commit')
    ?BDB:db_put(BdbPort, DB, void, Key, Data, []).


txn_fun_simple_get(BdbPort, Key) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),
    ?BDB:db_get_data(BdbPort, DB, void, Key, []).
    

txn_fun_put_account(BdbPort, {AccountName, AccountObj}) ->
    {value, {email_addr, EmailAddr}} = lists:keysearch(email_addr, 1, AccountObj),

    AccountDB    = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "account.db"),
    IndexEmailDB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "index_email_to_account.db"),

    %% Minor optimization; transactions should be as short as possible to minimize 
    %% the time that locks/resources are held.  So it's probably worth doing any work 
    %% possible before starting the transaction.
    AccountObjBin = term_to_binary(AccountObj),

    Txn = ?BDB:txn_begin(BdbPort, void, []),
    %% Optimization: we access IndexEmailDB first, as get_account_by_email must do so, 
    %% and consistent ordering of lock acquisition greatly helps to reduce deadlocks.
    ?BDB:db_put(BdbPort, IndexEmailDB, Txn, EmailAddr, AccountName, []),
    ?BDB:db_put(BdbPort, AccountDB, Txn, AccountName, AccountObjBin, []),
    ?BDB:txn_commit(BdbPort, Txn, []).


txn_fun_get_account_by_email(BdbPort, EmailAddr) ->
    AccountDB    = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "account.db"),
    IndexEmailDB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "index_email_to_account.db"),

    Txn = ?BDB:txn_begin(BdbPort, void, []),
    AccountName   = ?BDB:db_get_data(BdbPort, IndexEmailDB, Txn, EmailAddr, []),
    AccountObjBin = ?BDB:db_get_data(BdbPort, AccountDB, Txn, AccountName, []),
    %% commit as soon as possible to release locks as soon as possible 
    %% (e.g. defer calling binary_to_term until afterwards)
    ?BDB:txn_commit(BdbPort, Txn, []),
    {AccountName, binary_to_term(AccountObjBin)}.
