%%%-------------------------------------------------------------------
%%% File    : berkeley_db_replication_group.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%%
%%%   This module implements a gen_server that manages one site in a
%%%   BDB-HA replication group and provides a convenient API for using 
%%%   BDB-HA from Erlang.
%%%
%%%   [See also berkeley_db_replication_group_helpers.erl, which provides 
%%%    a similar set of APIs to that provided in this module but which 
%%%    support redirects and retries if one or more sites might be down,
%%%    and also supports basic load-balancing across replicas.]
%%%
%%% Primary Responsibilities:
%%%
%%% 1. Each replication_group server spawns and monitors a local
%%%    port_coordinator server, which is the actual BDB-HA site.
%%%    Currently if that spawned port-coordinator exits for any
%%%    reason, the parent replication_group server also exits.
%%%    i.e. The two processes act as a coupled unit, and leave restart
%%%    strategy to a higher-level supervisor.  (There is some
%%%    rationale for this at the end of this top-of-file comment.)
%%%
%%% 2. Provide a naming-scheme for sites in a replication
%%%    group.  These are SiteLocators, defined as
%%%        {ReplicationGroupName, SiteName, SiteNode}.
%%%    An API (do_txn_on_site) is provided to run transactions to be
%%%    run on a given named site whether that site is on the local
%%%    node or remote-node.
%%%   
%%% 3. Track the location of the master and transparently forward
%%%    transactions to that site on demand (see do_txn_on_master).
%%%    A mechanism is provided (#master_location_hint) to minimize 
%%%    inter-node hops when reaching the master (basically it should 
%%%    almost never be more than 1).
%%%
%%% 4. Provide a way of waiting until a master is elected without
%%%    polling (see get_master_location_hint).
%%%
%%% 5. Track the status of replicas, in particular whether they are
%%%    'in sync' with the current master.  See do_txn_on_any_site,
%%%    which is intended to run read-requests on replicas.  Even
%%%    though reading from replicas implies weak consistency (they
%%%    lag), there is a *huge* difference in reading from in 'in-sync'
%%%    replica (which is lagging behind the master by a few
%%%    transactions) and an 'not in sync' replica which might be
%%%    arbitrarily out of date, or even still in the middle of
%%%    bootstrapping.
%%%
%%% 6. Monitor the rate of PERM_FAILED events on the master and
%%%    shutdown the master (to force an election) if too many such
%%%    events are generated in a given time-window.
%%%
%%%    PERM_FAILED events mean that a write transaction failed with 
%%%    {bdb_error, bdb_DB_NOT_PERM}.  This means that the current 
%%%    policy (threshold) for replication acks was not reached, 
%%%    for that transaction, which means that the transaction was 
%%%    not durably replicated.  If the master crashes then a new 
%%%    master _might_ be elected that has not seen that transaction,
%%%    so when the ex-master recovers and syncs with that new master,
%%%    a commited transaction will be rolled back.
%%%    See the comments in the handler for DB_REP_EVENT_PERM_FAILED 
%%%    for more detail on this, and a discussion of master leases 
%%%    (a feature coming in BDB v4.6 that provides strong consistency 
%%%    by quaratining a master if it gets PERM_FAILED).
%%%
%%% 7. Monitoring whether any replicas are lagging signficantly behind
%%%    the master by polling their DB_REP_STAT.st_next_lsn fields and
%%%    logging a warning if the difference with the master becomes too
%%%    large.
%%%
%%%    Also, if an in_sync replica persistently lags (greater than 
%%%    threshold for N consecutive tests) then we can stop that 
%%%    replica (with the intention that it superviosr will restart it).
%%%    This restart feature is only useful to work around possible 
%%%    bugs in BDB-HA in which replicas become stuck.  This has 
%%%    happened in testing on logfile changeover and checkpoints.
%%%    I don't know of any remaining such bugs, but this is a defensive 
%%%    measure just in case.
%%%
%%% 8. Perform regular dummy write transactions whenever necessary:
%%%
%%%    - After a site is elected master, then **before** advertizing
%%%      it as available for writes (via get_master_location_hint or
%%%      do_txn_on_master) we want to wait for enough replicas to fall
%%%      into sync that application write transactions will be durable
%%%      (i.e. will not fail with {bdb_error, bdb_DB_REP_NOTPERM}.  We
%%%      don't want to second-guess the acks_policy by counting
%%%      in_sync sites.  Instead we look for our dummy writes to
%%%      transition from getting {bdb_error, bdb_DB_REP_NOTPERM} to
%%%      succeeding (obviously the very first dummy write might
%%%      succeed too -- i.e. we might never see bdb_DB_REP_NOTPERM).
%%%    
%%%    - If there are any connected but not_in_sync replicas then do
%%%      writes to stimulate those replicas into generating STARTUPDONE
%%%      when they are in-sync with the master (this is necessary in
%%%      BDB v4.5 but might be fixed in later versions)
%%%
%%% 9. Coordiante *safe* removal of txn logs on all sites in the
%%%    group.  The BDB 'AUTOREMOVE' facility is NOT SAFE for use with
%%%    replication as it makes local decisions and aggressive.  The
%%%    master will remove its own logs as soon as it has finished with
%%%    them, but if a replica is lagging behind then it may request a
%%%    log record from a log file that the master has already remove.
%%%    That will either cause the site to fail (panic if
%%%    repmgr_set_config NOAUTOINIT is used) or delete it's local
%%%    environment and begin bootstrapping the entire environment from
%%%    the master/peers.  Obviously neither of these scenarios is
%%%    acceptable so we must do conservative logging; we gather the
%%%    list of removable logs from all sites, and only remove logs up
%%%    to the latest one that can be removed from ALL sites.  This
%%%    also ensures that if a master fails then the new master will
%%%    have the same set of logs as the old master, so other (lagging)
%%%    replicas won't immediately fail/auto-init when they attempt to
%%%    sync-up with the new master.  Note that if a site is down for
%%%    some period we cannot safely remove logs.  But as a site might
%%%    fail permanently, we have to have a threshold for how many logs
%%%    we allow to build up before we _do_ start deleting logs again.
%%%    (If the down site restarts after that threshold has passed then
%%%    it may indeed require auto-init; but this is unavoidable -- we
%%%    just need to control it and prevent it happening to in-progress
%%%    sites).
%%%
%%%
%%% Design Overview 
%%% (described before the API as it explains the consistency model,
%%%  which obviously affects the semantics of the API)
%%%
%%% Tracking status of BDB-HA sites:
%%%
%%%   This is actually a fairly tricky job.  The set of
%%%   replication_group servers form a distributed system which is
%%%   attempting to track/shadown the internal state of another
%%%   distributed system (BDB-HA), when that state is only exposed via
%%%   local asychronous messages (events) from BDB-HA, and only some
%%%   of those events reveal partial state about selected other sites.
%%%
%%%   e.g. BDB-HA tells all sites about changes in master (the NEWMASTER
%%%   event), and that event does also say which site is the new
%%%   master. 
%%%
%%%   BDB-HA also tells individual sites when they are 'in-sync' (the
%%%   STARTUPDONE event), which means they are returning IS_PERM live
%%%   transactions from the current master.  However, STARTUPDONE does
%%%   not identify the site that master -- the application is assumed
%%%   to have received and processed an earlier NEWMASTER event, and
%%%   remembered which site is master for that replica.  (In bdb v4.5
%%%   it is possible for thread- scheduling to cause STARTUPDONE to be
%%%   delivered before NEWMASTER which is obviously confusing. We
%%%   don't support this as it is very rare -- I've only seen it if
%%%   the sites are running on the same node -- and it is fixed in
%%%   v4.6).
%%%
%%%   BDB-HA tells a site if it wins an election (MASTER event)
%%%   but that event doesn[t say anything about the state of any of the
%%%   replicas, other than an election quorum was present (note that 
%%%   any or even all of the sites in the election quorum might be 
%%%   significantly out of sync with the master that they elected).
%%%
%%%   We need to make sense of all this, and construct local copy of
%%%   the entire group's status on each site, and ensure that the
%%%   local copies are eventually-consistent with each other (assuming
%%%   that the group's state stabilizes -- i.e. site's aren't
%%%   persistently and frequently restarting/reconnecting).  
%%%
%%%   This is just the type of stateful distributed-system problem
%%%   that we are using BDB-HA to solve, but we can't write the
%%%   authoratative state in a transaction on the master because of the 
%%%   obvious catch-22 that we are trying to decide where the master is.
%%%   And consensus protocols such as Paxos don't seem useful as 
%%%   we aren't trying to achieve a new consensus, we are trying 
%%%   to track the result of BDB-HA's own consensus protocol (the 
%%%   election algorithm is very similar to paxos).
%%%
%%%   Our solution is :
%%%
%%%   - All sites run a local state machine per site in the group, 
%%%     including the local site
%%%
%%%   - All sites attempt to connect to (or even spawn!) any sites
%%%     that are known (via configuration) to be in the group but
%%%     which currently have 'unknown' state (i.e. are not connected).
%%%     This constructs a fully-connected network, which should mirror
%%%     that established by the repmgr BDB-HA layer. However, nothing
%%%     prevents the two mirror networks from driftint out of sync
%%%     (e.g. connections at the Erlang layer rely on Distributed
%%%     Erlang, connections at the BDB-HA layer are raw TCP/IP and
%%%     rely on different ports being open.).  We are not attempting
%%%     to assist BDB-HA (or provide failure-detection for it) - we
%%%     merely have a requirement to exchange information between all
%%%     sites in the group and do not have access to the BDB-HA's
%%%     private connections for that purpose, so we have to make our
%%%     own.
%%%
%%%   - When a site discovers the existance of another running site it 
%%%     'subscribes' to the other site by sending a 'subscribe' request 
%%%     (a non-blocking cast; subscribe is retried until it succeeds).
%%%     
%%%     This achieves the following:
%%%
%%%     1. The subsriber starts an Erlang process monitor on the subscribee, 
%%%        so the subscriber gets a 'DOWN' message if the subscribee 
%%%        crashes or shuts-down, or if Distributed Erlang loses the
%%%        network connection to the subscribee's node.
%%%
%%%     2. The subscribee sends an initial snapshot of its local 
%%%        state to the subscriber, and promises to send ordered 
%%%        status updates to that subscriber in future until 
%%%        it gets a 'DOWN' message for that subscriber.
%%%
%%%        Distributed Erlang guarantees that messages between two
%%%        processes are delivered in order so long as there is a
%%%        connection.  And an Erlang process monitor is guaranteed to
%%%        deliver a 'DOWN' message if the monitored process exits or
%%%        the node (VM) that the monitored process is running on
%%%        crashes, or if the connection between the nodes running the
%%%        two processes failes (e.g. network partition, detected via
%%%        Distributed Erlang node-to-node heartbeats; see
%%%        net_kernel:set_net_ticktime/1 and note that the default is
%%%        60 seconds which is probably too low).  This gives the only
%%%        ordering and failure-detection guarantee that we need.  The
%%%        only thing that monitor does not detect is deadlock or
%%%        livelock -- i.e. it (obviously) does not poll the monitored
%%%        process to see if it is responsive.  So we must ensure that
%%%        all of our failure modes cause crashes. OTP's gen_server
%%%        does something to guarante this as handle_call uses a
%%%        timeout.  But eventually we might want to run a periodic
%%%        healthcheck in which sites poll all of the sites to which
%%%        they are subscribed.  (TODO: implement this health-check;
%%%        although I have never seen an indident of this in
%%%        practice).
%%%
%%%     Note that subscribe is uni-directional.  Each site will
%%%     independently discover and subscribe to the other.  This makes
%%%     the code significantly less complicated, but does mean that
%%%     each pair of sites will be using 2 monitors (one running in
%%%     each direction), rather than a link.  However, I don't think
%%%     that is remotely an issue. Also the semantics of link/1 are
%%%     significantly worse than the semantics of monitor (as link is
%%%     much older).  e.g. Look at code in
%%%     berkeley_db_port_coordinator:acquire_bdb_port_handle for the
%%%     mess required to detect if link fails due to the linked
%%%     process having just crashed.  Monitor just looks like it
%%%     succeeds, and then sends the normal 'DOWN' message -- much
%%%     cleaner).
%%%
%%%   - If a site receives a local BDB-HA event that affects the 
%%%     state of that (local) site then it processes the event, changes 
%%%     it's state, and sends the new (absolute) state to all sites which 
%%%     are subscribed to it.  
%%%     NOTE that the site whose state has change only sends the
%%%     change in its OWN state. In particular it does NOT send its
%%%     view of any other site's state. We rely completely on BDB-HA
%%%     correctly delivering events to all sites; e.g. NEWMASTER
%%%     should be triggered on all sites.  Therefore we do not
%%%     propagate/gossip about general BDB-HA events.  We only send
%%%     status changes for individual sites (to all other sites).
%%%     e.g. If site A believes that site B is master and then site A
%%%     receives a message from site C that site C now thinks that
%%%     site D (or even site A!)  is master then site A _only_ updates
%%%     it's local copy of site C's state.  Site A does NOT change its
%%%     own view of the current master until BDB-HA tells it directly
%%%     via a local event -- even if other sites are claiming that it
%%%     (site A) is now master.
%%%
%%%   - All status-update messages generate by a site carry a
%%%     microsecond-resolution timestamp from the generating site.
%%%     This lets us reject duplicates.  However, there is NO attempt
%%%     to construct a group-wide total-ordering or even causal
%%%     ordering of events/state-changes (e.g. running Lamport Logical
%%%     Clocks, or Vector Clocks).  We assume that BDB-HA's consensus
%%%     (election) protocol has taken care of all that, and all we
%%%     care about is propagating the results so that all sites have
%%%     an eventually-consistent local copy.  All status updates from 
%%%     remote-sites are accepted ('delivered') at once; they are not 
%%%     buffered to ensure an ordering.
%%%    
%%%     This means that each per-site local copy of the group state is
%%%     often not even self-consistent.  e.g. Replicas might be (and
%%%     indeed are frequently in practice) told which site is master
%%%     before the new master site has itself been told).  Therefore
%%%     it is possible that sites appear to temporarily disagree on
%%%     which site is the master, even if only a single site is asked
%%%     for status (of the whole group).  e.g. Site A might observe
%%%     the master fail and a new master be elected before it receives
%%%     a status update from site B reflecting the same change, so for
%%%     a while site A snapshot of site B's status is out of sync with
%%%     site A's knowledge.  This can also happen in reverse; site A
%%%     might receive an update from site B claiming that site B has
%%%     discovered (and even synced with) a new master, when site A as
%%%     yet knows nothing about that change.  Site B might even claim
%%%     that site A is now the master before BDB-HA tells site A that
%%%     it (site A) won the election.)
%%%
%%%     Note: it might be possible to establish a consistent ordering
%%%     simply by exposing a bit more information from the underlying
%%%     BDB-HA layer. e.g. If BDB-HA expose it's internal 'egen'
%%%     generation-id as part of MASTER, NEWMASTER, and STARTUPDONE
%%%     event payloads then we would have both a total-ordering and a
%%%     way to pair events with status-updates.  (egen is a durable
%%%     generation id, so this would fix the problem even if sites
%%%     crash and restart rapidly.)
%%%    
%%%     In the absence of that, we observe that masters don't change
%%%     very rapidly, and the current last-message-arrival works in
%%%     practice.
%%%
%%%     This whole thing is therefore a very weak form of a classical
%%%     group communication system.  There are many intentional
%%%     departures from the classical design:
%%%
%%%     - Incomplete support for changes to group-membership.
%%%       This is one of the hard problems that classical
%%%       group-communication mechanisms exist to solve.  e.g. To add
%%%       a new site to the group. e.g. Many such systems have a
%%%       concept of epochs (periods of frozen group membership in
%%%       which we pretend there are no failures and no additional
%%%       members).  
%%%
%%%     - When a site learns of another site it does performs
%%%       unidirectional (polling) 'state transfer' (asks for the
%%%       other site's full state) as well simultaneously subscribing
%%%       to all future state changes.  Sites are expected to discover
%%%       each other by polling.  Sites do not currently tell all
%%%       other existing sites about the new site that they just
%%%       located.  (This will probably need to change to allow sites
%%%       to be added dynamically).
%%%
%%%     - No causal order tracking of messages (vector clocks) or
%%%       causal-order delivery of messages (i.e. no buffering of
%%%       received messages until all those that causally preceed it
%%%       have been received and acted upon)
%%%      
%%%       Currently all status-update messages are timestamped with
%%%       erlang:now() from their originating host, to allow ordering
%%%       messages from one site (but not across all sites).
%%%      
%%%       Also, all status-updates updates are *currently* absolute
%%%       (idempotent).  This simplifies things. e.g. There is a race
%%%       condition in the subscribe_and_get_status operation we have
%%%       to subscribe before returning the status or we might miss
%%%       events.  But then some events might be sent that preceed the
%%%       status value that is returned by the function.  Because we
%%%       have idempotent events we can simply discard those that
%%%       preceed the one obtained by the function.  (note: those
%%%       timestamps cannot be meaningfully compared with our own
%%%       local time, but we don't need to do that, we just need to
%%%       order the messages from one site).  If the messages were not
%%%       idempotent (deltas, not absolute) then we would need to
%%%       discard 0..N events (those that are timestamped before the
%%%       OtherSiteStatus received from get_site_status) before
%%%       returning the absolute value we got from
%%%       subscribe_and_get_status, and then apply any remaining
%%%       events.
%%%      
%%%     - No message retransmit, and no blocking for acks from each
%%%       destination.  Instead, we rely on Erlang's guarantee for
%%%       sending to monitored/linked processes: that either the
%%%       message is delivered or eventually the sender will receive a
%%%       DOWN/EXIT message for the destination process.  See 'medium
%%%       safe' :
%%%       http://www.erlang.org/ml-archive/erlang-questions/199910/msg00043.html
%%%
%%%
%%%   State Machine
%%%
%%%   current_master is 'the advertized master', and may be
%%%
%%%     unknown
%%%     this_site
%%%     {this_site, waiting_for_durability}
%%%     OtherSiteId
%%%
%%%   this_site_status can have the following states:   (MSL is MasterSiteLocator}
%%%
%%%     {master, waiting_for_durability}          - has received MASTER but is still verifying that it can write durable transactions, so is NOT useable by do_txn_on_master
%%%     master                                    - has received both MASTER and achieved_durablity : is now usable by do_txn_on_master
%%%
%%%     {replica, not_in_sync, unknown}                       - has not yet found a master
%%%     {replica, not_in_sync, {MSL, transitioning_to}}       - has received NEWMASTER but not yet received other_site_status_update with status 'master'
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}} - received an other_site_status_update with status '{master, waiting_for_durability}' and has not yet received STARTUPDONE
%%%     {replica, not_in_sync, MSL}                           - received an other_site_status_update with status 'master' but not yet received STARTUPDONE (clearly some other replicas have received STARTUPDONE as the master can write durable transactions)
%%%
%%%   If NEWMASTER and/or STARTUPDONE happens to arrive before other_site_status_update:
%%%     {replica, in_sync, {MSL, transitioning_to}}           - we got NEWMASTER and STARTUPDONE before the status update arrived from the new master see comments in handle_sleepycat_event(NEWMASTER) -- this means 
%%%     {replica, in_sync, {MSL, waiting_for_durability}}     - we got NEWMASTER and STARTUPDONE before the status update arrived from the new master see comments in handle_sleepycat_event(NEWMASTER) -- this means 
%%%     {replica, in_sync, MSL}                               - has received STARTUPDONE (live txns) from Master
%%%
%%%   Note that these states are illegal as STARTUPDONE should not arrive before NEWMASTER:
%%%     {replica, in_sync, unknown}
%%%
%%%   other_site_status can have the same states as this_site_status, plus:
%%%
%%%     unknown                                   - not connected
%%%
%%%   Note: BDB-HA does NOT currently provide an event if an in-sync
%%%   replica starts to fall behind the Master (i.e. the opposite of
%%%   STARTUPDONE).  Receipt of PERM_FAILED event would mean that
%%%   several (too many) replicas have fallen behind, but some
%%%   replicas may fall behind without triggering that event. Even in
%%%   the 'base' replication API (below the repmgr layer we are using)
%%%   receipt of NOT_PERM from the replica does not necessarily mean
%%%   it is falling behind as out-of-order messages cause that
%%%   (temporarily) -- and in any case, we use repmgr which hides
%%%   NOT_PERM events from us.  Therefore, the master runs a process which 
%%%   periodically compares st_next_lsn on all connected replicas to 
%%%   that of the master.
%%%
%%%
%%%   State Machine Management:
%%%
%%%   All other_site_status_update messages are accepted and stored.
%%%
%%%   We don't ever alter other sites' status based on local information
%%%   (e.g. if we become master or learn of a master, we don't reset the 
%%%   the status of any other site that has previously claimed to be master,
%%%   or has previously claimed to be synced to another master).
%%%   This limits the invariants that we can assert, but that reflects the 
%%%   nature of the disjoint timeframes of our local information.
%%%
%%%   Only other_site_status_updates refering to master or {master, waiting_for_durability}
%%%   affect our local state (this_site_status).
%%%
%%%   Legal state transition sequences for this_site_status when discovering a new master:
%%%
%%%   - We assume that STARTUPDONE can *only* be delivered after NEWMASTER (this reduces the combinatorics considerably)
%%%     IMPORTANT: in v4.5, BDB-HA might deliver STARTUPDONE before NEWMASTER.  
%%%     It is very rare, and I have only seen it happen when the new master is on the 
%%%     same host as the replica (i.e. in berkeley_db_test.erl).
%%%     Sleepycat have fixed this in v4.6 ([#15265]); they will guarantee that 
%%%     the application's event-handler for NEWMASTER returns before they trigger 
%%%     the STARTUPDONE event, and/or provide the master envid with STARTUPDONE 
%%%     as well as NEWMASTER. 
%%%
%%%   - We assume that {other_site_status_update, {master, waiting_for_durability}} 
%%%     MUST be delivered before {other_site_status_update, {master, available}}
%%%
%%%   - If at any point we get a DOWN from the site that we have an MSL for, 
%%%      then we reset this_site_status to {replica, not_in_sync, unknown}
%%%
%%%   1. Perhaps the 'ideal':                                  Event:
%%%
%%%     {replica, not_in_sync, unknown}
%%%             |                                              DB_EVENT_REP_NEWMASTER
%%%     {replica, not_in_sync, {MSL, transitioning_to}}
%%%             |                                              {other_site_status_update, {master, waiting_for_durability}}
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              DB_EVENT_REP_STARTUPDONE
%%%     {replica, in_sync, {MSL, waiting_for_durability}}
%%%             |                                              {other_site_status_update, master}
%%%     {replica, not_in_sync, MSL}
%%%
%%%   2. STARTUPDONE one step earlier than #1; before {other_site_status_update, {master, waiting_for_durability}
%%%
%%%     {replica, not_in_sync, unknown}
%%%             |                                              DB_EVENT_REP_NEWMASTER
%%%     {replica, not_in_sync, {MSL, transitioning_to}}
%%%             |                                              DB_EVENT_REP_STARTUPDONE
%%%     {replica, in_sync, {MSL, transitioning_to}}
%%%             |                                              {other_site_status_update, {master, waiting_for_durability}}
%%%     {replica, in_sync, {MSL, waiting_for_durability}}
%%%             |                                              {other_site_status_update, master}
%%%     {replica, in_sync, MSL}
%%%
%%%   3. STARTUPDONE one step later than #1 before {other_site_status_update, master} (i.e. later than most other sites)
%%%
%%%     {replica, not_in_sync, unknown}
%%%             |                                              DB_EVENT_REP_NEWMASTER
%%%     {replica, not_in_sync, {MSL, transitioning_to}}
%%%             |                                              {other_site_status_update, {master, waiting_for_durability}}
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              {other_site_status_update, master}
%%%     {replica, not_in_sync, MSL}
%%%             |                                              DB_EVENT_REP_STARTUPDONE
%%%     {replica, in_sync, MSL}
%%%
%%%   4. NEWMASTER one step later than #1, after {other_site_status_update, {master, waiting_for_durability}}
%%%
%%%     {replica, not_in_sync, unknown}
%%%             |                                              {other_site_status_update, {master, waiting_for_durability}}
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              DB_EVENT_REP_NEWMASTER  (does not change status but must have same MSL)
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              {other_site_status_update, master}
%%%     {replica, not_in_sync, MSL}
%%%             |                                              DB_EVENT_REP_STARTUPDONE
%%%     {replica, in_sync, MSL}
%%%
%%%   5. NEWMASTER one step later than #1 (after {other_site_status_update, {master, waiting_for_durability}}) and STARTUPDONE one step earlier than #1
%%%
%%%     {replica, not_in_sync, unknown}
%%%             |                                              {other_site_status_update, {master, waiting_for_durability}}
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              DB_EVENT_REP_NEWMASTER  (does not change status but must have same MSL)
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              DB_EVENT_REP_STARTUPDONE
%%%     {replica, in_sync, {MSL, waiting_for_durability}}
%%%             |                                              {other_site_status_update, master}
%%%     {replica, in_sync, MSL}
%%%
%%%   6. NEWMASTER two steps later than #1: after {other_site_status_update, master}
%%%
%%%     {replica, not_in_sync, unknown}
%%%             |                                              {other_site_status_update, {master, waiting_for_durability}
%%%     {replica, not_in_sync, {MSL, waiting_for_durability}}
%%%             |                                              {other_site_status_update, master}
%%%     {replica, not_in_sync, MSL}
%%%             |                                              DB_EVENT_REP_NEWMASTER  (does not change status but must have same MSL)
%%%     {replica, not_in_sync, MSL}
%%%             |                                              DB_EVENT_REP_STARTUPDONE
%%%     {replica, in_sync, MSL}
%%%
%%%
%%%   Legal state transition sequences for this_site_status when becoming master:
%%%
%%%   1.
%%%     {replica, _, _}
%%%             |                                              DB_EVENT_REP_MASTER
%%%     {master, waiting_for_durability} 
%%%     (and current_master = {this_site, waiting_for_durability}
%%%             |                                              achieved_durablity
%%%     master                           
%%%     (and current_master = {this_site, available})
%%%
%%%   - If DB_EVENT_REP_CLIENT occurs while we are in either master state 
%%%     then currently this site EXITs (live downgrades are not supported).
%%%     Some of the asserts (e.g. in handle_other_site_status_update) 
%%%     assume that the steps to becoming master are a ratchet.
%%%
%%%
%%% Public API
%%%
%%% Function: do_txn_on_master()
%%%
%%% Description: 
%%%
%%%   Forwards a transaction fun to be executed on the master site.
%%%
%%%   This is similar in intent to berkeley_db_helpers:do_txn/3 but with 
%%%   some important differences:
%%%
%%%      - if the current host happens to be the master then the TxnFun will be executed by a spawned process 
%%%        (TODO: we could just execute it in the client process).
%%%
%%%      - if the current host is NOT the master then the unless the current host is the master, the TxnFun will be sent to another 
%%%
%%%      - the return value is different (contains more information) -- see below.
%%%
%%%   TxnFun must be  
%%%
%%%      {Module, Function, ClientDataTerm}
%%%
%%%   where the following exported function must exist on the master node:
%%%
%%%      Module:Function(BdbPortHandle, ClientDataTerm) -> ResultTerm
%%%
%%%   i.e. TxnFun is (unfortunately) NOT a normal fun (closure) or even a new-style 'fun M:F/A'.
%%%   This is to allow the code on the server to be upgraded independently of the code on the caller.
%%%   Normal Erlang hot-code loading fails here (and normal funs break) because TxnFun will almost 
%%%   certainly contain BDB API constants (converted from #defines in db.h to -defines in berkeley_db.xml
%%%   and berkeley_db_drv.hrl), and those constants can (and usually) do change between BDB releases 
%%%   (even patch releases).  So using the wrong constants could do anything -- including corrupting data.
%%%   Therefore the version of the fun's code used to actually call BDB (the version on the server) must
%%%   be compiled with the new berkeley_db_drv.hrl.  And the fun invocation sent by the client must match 
%%%   that new version.  fun invocations *can* invoke the *old* code (because funs contain an internal 
%%%   module-id, they don't find the module by name) -- but in this case we absolutely don't want that.
%%%   Fortunately upgrading BDB cannot be done via hot-code loading -- the VM must be stopped and restarted 
%%%   which clears out the old code completely.
%%%
%%%   Using the Module:Function form to specify a TxnFun means that the client code can be a different version
%%%   than the version on the server.
%%%
%%%   IMPORTANT: TxnFun must **NOT** call berkeley_db_port_coordinator:release_bdb_port_handle() on BdbPortHandle
%%%   as the server may choose to re-use that handle.   This should be obvious as the TxnFun is passed the handle 
%%%   and does not obtain it itself.
%%%
%%% Returns    
%%%
%%%   {ResultOfTxnFun, MasterLocationHint}
%%%
%%%   Note: MasterLocationHint is an opaque term which may be passed as the first argument of do_txn_on_master/2
%%%   to target future requests directly to the master, so avoiding extra process/node hops.
%%%   A function is provided to obtain the Node from a MasterLocationHint, to allow applications to move 
%%%   their work to the current master node and so avoid multiple inter-node hops if they intend to call 
%%%   do_txn_on_master multiple times.
%%%
%%% or throws one of...
%%%
%%%   {txn_fun_error, Reason}                            (e.g. if TxnFun called throw)
%%%   {txn_fun_error, {bdb_error, bdb_DB_LOCK_DEADLOCK}} (if TxnFun hit BDB deadlocks and exhausted NumDeadlockRetries)
%%%   {txn_fun_error, {return_with_open_resources, 
%%%                    NumResources, 
%%%                    TxnFunResult}}                    (if TxnFun returned with an open cursor/txn)
%%%
%%% or a system error...
%%%
%%%    {bdb_replication_group_error, Reason}             (*** e.g. Reason might be 'no_master' ***)
%%%
%%% or an internal error (these are examples, *not* an exhaustive list)...
%%%
%%%   {bdb_helper_error, {acquire_bdb_port_handle, Reason}}
%%%   {bdb_helper_error, {release_bdb_port_handle, Reason, TxnFunResult}}
%%%   {bdb_helper_error, {group_valmap_cleanup, Reason}}
%%%   {bdb_helper_error, Reason}
%%% 
%%%
%%% NOTE: a TxnFun will be executed in it's own process and can live as long as it wishes.
%%% It can send messages to other processes (via pids bound to closure variables), and wait to receive 
%%% messages from them.  In this way a TxnFun may open a cursor and then send batches of key/values to 
%%% other processes on demand.  Note: a TxnFun should guarantee to terminate.  If it uses receive then 
%%% there should be a non-infinite timeout, or the TxnFun should monitor/link to the process it is expecting 
%%% to receive messages from.  If a TxnFun crashes then any cursors that it opened will be closed and any 
%%% txns it opened will be aborted.
%%%
%%% Example of a simple TxnFun
%%%
%%%         my_txn(BdbPortHandle, {MyDbName, MyKey, MyData}) -> 
%%%
%%%                 %% we almost always need at least one DB handle (which we must NOT close)
%%%                 {ok, DB} = berkeley_db_port_coordinator:get_db_handle(BdbPort, bdb_DB_BTREE, MyDbName),
%%%          
%%%                 %% now do some work and auto-commit it...
%%%                 {ok, 0}  = ?BDB:db_put(BdbPort, DB, void, MyKey, MyData, []),
%%%          
%%%                 %% and/or do a multi-part transaction and explicitly commit it...
%%%                 {ok, Txn} = ?BDB:txn_begin(BdbPort, void, []),
%%%                 {ok, 0}   = ?BDB:db_put(BdbPort, DB, Txn, MyKey, MyData, []),
%%%                 {ok, 0}   = ?BDB:db_put(BdbPort, DB, Txn, <<"some other key">>, MyData, []),
%%%                 {ok, 0}   = ?BDB:txn_commit(BdbPort, Txn, []),
%%%         end
%%% 
%%%
%%% Function: get_master_location_hint()
%%%
%%%   The return value is described under do_txn_on_master.
%%%
%%%   Obviously do_txn_on_master with a no-op TxnFun can be used to find the location of the master.
%%%   However, do_txn_on_master will fail immediately if no master currently exists.
%%%
%%%   get_master_location hint accepts a FindMasterTimeout which can be used to 
%%%   block until a master is elected (and then provide a shortcut to accessing it).
%%%   FindMasterTimeout may be 0 (for an immediate poll and return), greater than zero, or 'infinity'.
%%%
%%%   This API also avoids the need to provide a useless TxnFun (and the overhead on the master of 
%%%   spawning a process, acquiring a port-handle, running the TxnFun, releasing the port handle,
%%%   replying etc).   
%%%
%%%   Note: it appears inconvenient (for applications) that do_txn_on_master and get_master_location_hint 
%%%   are separate.  Why doesn't do_master_on_txn block if there isn't a master?
%%%   The answer is that this would be slower for normal operations in a stable system (i.e. 
%%%   the master is assumed to be running for long periods).
%%%   See berkeley_db_replication_group_helpers.erl for an additional layer that indirectly provides 
%%%   a version of do_txn_on_master that will attempt to find a new master if necessary.
%%%   (The functions in the helper module are aware of the status of the entire group, and 
%%%   will try to direct requests via any site that is up, and retry as necessary.)
%%%
%%%
%%% Function: do_txn_on_site()
%%%
%%%   Executes a READ-ONLY transaction (as site might not be master)
%%%
%%%   This is equivalent to ?BDB_H:do_txn, but is usable from a remote node 
%%%   (i.e. it spawns a worker process on the target node, even if that node is the caller's node).
%%%
%%% 
%%% Function: do_txn_on_any_site()
%%%
%%% Description: 
%%%
%%%   Similar to do_txn_on site, but chooses a site from the following preference order:
%%%
%%%       - if the local site is in_sync (or the master) then choose the local site
%%%       - else if there are any in_sync sites (*excluding* the master) choose a random such site
%%%       - the master (as it is in-sync)
%%%       - the local site (which is not in-sync)
%%%       - [TODO] if the local site is down, we could pick any other site
%%%
%%%  TODO: the last entries above means that this function does NOT guarantee to exectute 
%%%  on an in-sync site.  I intend to add a parameter to control this, and the default will be changed 
%%%  to only allow reads on in_sync sites unless the caller explicitly allows not_in_sync sites.
%%%
%%% Returns    
%%%
%%%   ResultOfTxnFun
%%%
%%% or throws one of the errors described earlier.
%%%
%%% 
%%% Minor implementation details:
%%%
%%% Rationale for exiting if the spawned port-coordinator dies:
%%% 
%%%   Currently, if a replication_group server dies then its port_coordinator is killed 
%%%   (i.e. they are linked).
%%%   It isn't completely obvious that this is the right thing to do.  Here is the rationale for the choice:
%%%
%%%   1. If there is a bug in replication_group server (e.g. inconsistent tracking of master) that kills the 
%%%      replication_group server then we DON'T want the port_coordinator to die if the replication_group server dies, 
%%%      because we want BDB-HA replication to continue while replication_group server restarts.
%%%
%%%   2. If the replication_group server detects a problem with the port_coordinator it can 
%%%      explicitly terminate it to restart it.
%%%       
%%%   3. On the other hand, if there is a bug in the port coordinator because if the the replication_group server 
%%%      cannot handle (and therefore dies) then the replication_server will thrash -- die/spawn/die/spawn
%%%      unless something also kills the port_coordinator.
%%%
%%%   On balance the danger of #3 outweighs the minor availability issue of #1, because the port_coordinator will be 
%%%   restarted as soon as the replication_group server is restarted by a peer, which should happen almost immediately 
%%%   as every other site in the group will attempt to restart it.
%%%       
%%%
%%%   Optionally, all replication_group servers (that are part of the same named group) form a 
%%%   'peer supervision' ring.  Each site attempts to find all of the others, and will 
%%%   attempt to remote-spawn any that appear to be down -- with infinite retry attempts
%%%   unless the target site is explicitly removed from the group via an API.
%%%   So starting one site (with appropriate config) is sufficient to start them all.
%%%
%%%   There is a convenience API called start_sites that will start (in parallel) an entire 
%%%   group of replication_group servers at once, but it does NOT link them all to the caller,
%%%   This is because gen_server:start_link is hardwired to termiante if an EXIT message 
%%%   is received from the process that called start_link.
%%%   Typically we do NOT want to kill the entire replication group if the single process 
%%%   that called caller crashes.  Instead the caller should use erlang:monitor on the 
%%%   pids returned by start_sites.  The caller MUST also remember that if a replication_group 
%%%   server dies then the other sites will attempt to re-spawn it.  So if the caller gets 
%%%   a 'DOWN' message it will need to start looking for a new registered process being 
%%%   created, and then monitor that new process.)
%%%   If some application does want the death of a single admin process to cause a 
%%%   replication_group to be torn down entirely, then a second admin process should monitor 
%%%   the first and issue 'group_shutdown' commands to all sites in the replication group.
%%%   That will be the ONLY way to shutdown all of the sites, other than all of the Erlang 
%%%   nodes failing simultaneously. 
%%%
%%%   Note that start_sites is only a convenience.  It is perfectly fine for an application's init
%%%   function to to start an individual replication_group server, with appropriate config to spawn 
%%%   the others. N.B. Again, if the local application instance dies then the replication_group server 
%%%   will continue to run (and be restarted) unless the entire group is shutdown or that 
%%%   particular site is explicitly removed from the group via an API call.   
%%%   So the model is significantly different from a pure OTP hierarchy.
%%%
%%%
%%% TODO
%%%
%%% - If do_txn_on_master crashes in a way that suggests that the site might not be master (i.e. db_put returns 'cannot write to database')
%%%   then reset entire replication_group's state (don't kill all port_coords, but poll local bdb rep_state to get status).
%%%
%%% - Fix shutdown_group API to make it use persistent consensus (so if a site is restarted during the shutdown, it doesn't ressurect the 
%%%   entire group).
%%%
%%% - Add add_site/remove_site APIs (we need corresponding repmgr functionality from Sleepycat)
%%%
%%% - Why can't 2 sites elect a master (with full or normal elections)?  Alan says that there is special-case code for 2 sites
%%%   - ask about that.
%%%
%%% - consistency 
%%%
%%%  1. The underlying BDB call to the set_event_notify callback for
%%%     NEWMASTER and STARTUPDONE is implicitly a race condition
%%%     because it is subject to thread scheduling.  This can lead to
%%%     inverted order of delivery for these events. (This is fixed in
%%%     v4.6)
%%%
%%%  2. That callback enqueues an operation to be executed by the Erlang VM in ready_async.
%%%     The queue then defines an order on the events, but the order of enqueing is again subject to 
%%%     thread scheduling.  Using a static counter as a logical clock (to send with the message) 
%%%     would fix this but not affect #1, so would be useless.
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_replication_group).

-behaviour(gen_server).

%% API
-export([start_sites/3,              % returns [GoodPidsSites, BadSites], does NOT link
         start/4,                    % returns Pid or throws, does NOT link. Used by start_sites.
         start_link/4,               % returns {ok, Pid} or {error, Error} --- for use by conventional supervisors;
         get_proc_name/2,
         get_port_coordinator_proc_name/2,
         get_master_location_hint/2, % timeout is FindMasterTimeout which can be 0  (the gen_server:call timeout is still the default of 5000 ms)
         get_master_location_hint/3, % takes both FindMasterTimeout and ClientServerTimeout
         do_txn_on_master/2, 
         do_txn_on_master/3,         % also takes TxnFunTimeout   (that timeout does NOT imply waiting for a master to be discovered)
         do_txn_on_master/4,         % also takes NumDeadlockRetries
         do_txn_on_site/2,           % always runs txn on specified site, no matter what its state
         do_txn_on_site/3,
         do_txn_on_site/4,
         do_txn_on_any_site/2,       % will use an in-sync site if there is one. Prefers local site, then any other non-master site, then master.
         do_txn_on_any_site/3,
         do_txn_on_any_site/4,
         remove_site/1,              % TODO: not fully implemented; for correctness this requires dynamic group-membership (consensus)
         get_group_status/1,
         stop_entire_group/1,        % TODO: not fully implemented
         get_site_locator_from_master_location_hint/1,
         get_node_from_master_location_hint/1
        ]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% Private API
-export([remote_find_or_spawn_site/3,      % exported for use by rpc:call
         %% exported for use by ?BDB_H:do_txn
         txn_fun_probe_to_sync_replicas/2,
         txn_fun_rev_list_removable_txn_logs/2,
         txn_fun_remove_specified_txn_logs/2,
         txn_fun_rep_stat_next_lsn/2,
         txn_fun_compare_rep_stat_next_lsn_to_master/2,
         txn_fun_start_checkpointer_if_enabled/2,
         txn_fun_stop_checkpointer_if_running/2,
         txn_fun_set_master_cache_throttle_io/2,
         txn_fun_stop_site/2
        ]).


%% A 'site id' is:
%%
%%    {PortCoordName, RemoteNodeName}
%%
%% Note this does NOT include doesn't include the ReplicationGroupName 
%% (but that is always known by context as it is the same for all sites)
%% And it does NOT include replication_group server's process name 
%% (but that can be calculated from the site's PortCoordName name
%% and ReplicationGroupName).

-include("berkeley_db_drv.hrl").
-include("berkeley_db_replication_group_internal.hrl").

%% Support processes are find_or_spawn_peers_process and master_dummy_write_txn_process.
-record(support_proc, {name, pid, stop_fun}).

-record(state, {current_master,     % unknown | {this_site, waiting_for_durability} | {this_site, available} | {OtherSiteId, waiting_for_durability} | {OtherSiteId, available}
                this_site_status,   % #site_status (when status cannot be 'unknown' and both replication_group_pid and monitor_ref are 'unknown')
                other_sites_status, % [{OtherSiteId, #site_status}, ...] 
                other_sites,        % stored in State because we can add and remove sites (eventually this will be in a BDB on disk)'
                notify_clients,     % list of {GenServerCallFrom, TimerRef|infinity} of clients to notify when we discover a master
                perm_failed_times,  % list of now() timestamps in reverse chronological order of all PERM_FAILED events with the allowed period (cannot be more than the configured threshold)
                support_procs       % list of support_proc records
               }).
%% TODO: Eventually 'OtherSites' will be stored in a group-membership record in a replicated BDB.


%% Note: common_port_coordinator_config is not present in this list,
%% as it has a separate default (see below), which is always used 
%% as an inheritance-fallback, even if the call supplies a custom list

-define(DEFAULT_CONFIG_PROPLIST, 
        [{find_other_sites_period_milliseconds,        5000},     % TODO: add random fuzz to de-phase sites?
         {rpc_call_timeout_milliseconds,               5000},         
         {spawn_other_sites,                           disabled},

         %% If necessary the master runs a process probe_to_sync_replicas_process.
         %% It is used for 2 things:
         %%
         %% - To do write transactions until it sees a successful transaction (i.e. not a NOTPERM txn)
         %%   at which point it sends us a achieved_durability message, which
         %%   which allows us to advertize ourself as master.
         %%   This is critical for write-availability
         %%
         %% - In BDB v4.5 the STARTUPDONE (replica-is-in-sync) event is only sent on receipt of new 
         %%   write transactions by a replica.  Until that is fixed, whenever there are any connected 
         %%   replicas that are not in sync we need to do writes to stimulate STARTUPDONE just incase 
         %%   the application isn't doing any.
         %%   This is important for read-availability if the application is not generating writes.
         %%
         {probe_to_sync_replicas_periods_milliseconds, {{not_durable,  500},  % nested tuple, *not* a tagged list (order is significant)
                                                        {durable,     2000}}},

         %% If this site receives more than Count PERM_FAILED events in WindowMillseconds 
         %% then we gracefully shutdown.
         %% We differentiate between when the master is waiting_for_durability and when it is available.
         %% This lets us bound the time that we are willing to sit in waiting_for_durability,
         %% and also lets a site be more sensitive to PERM_FAILED when in 'normal running'
         %% than when trying to establish durability.
         %% Valid values are:
         %%   'disabled'
         %% or a nested tuple *not* a tagged list (order is significant):
         %%   {{waiting_for_durability, {MaxCount1, WindowMilliseconds1}},
         %%    {available,              {MaxCount2, WindowMilliseconds2}}
         {perm_failed_events_limits,                   disabled},

         %% If cache_throttle_io_write_ops_and_sleep_microseconds is used then 
         %% it may be advantageous to have higher rate of IO on the replicas than 
         %% the master (i.e. master is throttled back more).
         %%
         %% Rationale:
         %% If the master sends a checkpoint record to a replica but the 
         %% replica is still performing the dirty-page IO required for that 
         %% checkpoint then that replica starts returning NOTPERM acks until 
         %% the IO is finished (the checkpoint record cannot be written to 
         %% the txn log until all of the earlier LSNs have been written to 
         %% the database files, and as the log is sequential and fixed-order,
         %% buffering the checkpoint record means that no subsequent records can 
         %% be written to the log until that checkpoint record is written 
         %% (the log records are buffered in __db.rep.db until the blockage is 
         %% cleared).
         %% If enough replicas hit this condition then the master will see 
         %% all new txns hit PERM_FAILED errors until the problem clears .
         %%
         %% A recent patch makes the replicas do the necessary IO in parallel 
         %% with the master, but if they fall behind for some reason 
         %% (e.g. disk/cpu contention) then the problem can still happen.
         %% Another recent patch temporarily suppresses any throttling if 
         %% a replica receives a chkp record before it finishes the IO for 
         %% that checkpoint -- i.e. the problem should be cleared as quickly 
         %% as possible, but at the cost of an spike in IO.
         %% Despite those patches (and identical hardware for replicas and 
         %% master) this condition does seem to happen a lot when cache-IO-throttling 
         %% is being used.
         %%
         %% Ideally the master would not send the checkpoint record until a
         %% majority of replicas have confirmed that they have finished the IO
         %% but that feature is not yet implemented.
         %% Until then, Sleepycat/Oracle have provided a way to make 
         %% the master (in txn_checkpoint) sleep for a configurable amount of 
         %% time (see he port-coordinator's {internal_sleep, master_checkpoint_seconds})
         %% after finishing the IO on the master's side but before sending the 
         %% checkpoint record to the replicas.
         %% i.e. the master will 
         %%   - tell replicas to start doing the IO for the checkpoint
         %%   - do its own (supposedly identical) IO for the checkpoint
         %%   - sleep for {internal_sleep, master_checkpoint_seconds} 
         %%   - send the checkpoint record to the replicas.
         %% However, during that sleep period, the only dirty-page IO being done 
         %% is coming from the cache_trickle calls (which happen in parallel to 
         %% the txn_checkpoint calls).  So only half the rate of IO is being done, 
         %% and txns might start to find no clean pages in the cache.
         %%
         %% So we want to minimize the duration of {internal_sleep, master_checkpoint_seconds}.
         %% One way to do that is to configure the replicas to do slightly faster 
         %% rate of dirty-page IO than the master. Hence this config item.
         %%
         %% For convenience, we invert that:
         %t
         %% - the port-coordinator's {cache_throttle_io_write_ops_and_sleep_microseconds, _} 
         %%   is used for all sites (replicas)
         %%
         %% - when a site receives DB_EVENT_REP_MASTER, this config value will be applied 
         %%   **unless it is set to 'same_as_replicas'** (see run_txn_funs_on_BDB_event below).         
         {master_cache_throttle_io,                     same_as_replicas},

         %% The master runs a process to check if any replicas are significantly lagging behind
         {check_for_lagging_replicas_period_seconds,    60},
         {lagging_replica_threshold_bytes,              100000},  % don't want to be _too_ sensitive
         {lagging_replica_restart_after_N_consecutive,  3},       % if a replica falls persistently behind, stop it (intended to restart it)

         %% The master runs a coordinated log-removal process at this frequency
         {log_removal_period_seconds,                   300},      % or 'disabled'

         %% On some file systems, deleting lots of large files can cause major IO spikes,
         %% so we need to throttle the rate at which we call file:delete.
         {log_removal_delete_file_period_milliseconds, 1000},

         %% We cannot safely remove logs on any site if even a single site is down 
         %% (because that site might come back up later and need the removed logs 
         %% to catch up -- if they are not available the site must re-init from 
         %% scratch or simply panic).
         %% But a site might fail permanently, so we must limit our exposure.
         %% (i.e. risk making the down site fail with 'unable to join environment'
         %% or do auto-init, if it restarts).         
         {log_removal_max_num_trailing_log_files,       100}      % assuming each log is 10 MB, this is 1 GB of acculumulated logs
        ]).


-define(DEFAULT_COMMON_PORT_COORDINATOR_CONFIG,
        %% Note: init() constructs some custom config that is inserted before this, 
        %% to specify replication sites, register as a handler for replication events etc.
        %% Also, the caller may of course specify site-specific config,
        %% and their own common_port_coordinator_config.  
        %% The defaults given here are always used, with lower priority than the 
        %% caller-specified values, but higher priority than the defaults defined 
        %% in the ?BDB_PC module.

        %% We MUST disable DB_LOG_AUTOREMOVE as it is not safe with replication -- see comment at top of file.
        %% Also, init() will *reject* any caller-supplied site-specific or common_port_coordinator_config
        %% that attempts to enable LOG_AUTOREMOVE (i.e. which attempts to override this setting).
        [{env_set_flags,                         [bdb_DB_AUTO_COMMIT, 
                                                  bdb_DB_REGION_INIT, 
                                                  bdb_DB_TIME_NOTGRANTED]},

         %% We only want to run checkpoints on the master
         %% Note: if the port-coordinator is configured with {do_parallel_trickle_and_checkpoint, enabled}
         %% then this currently means that cache-trickler will only be run on the
         %% master too (i.e. the replicas won't do any dirty-page IO).
         %% TODO: in that case txn_fun_stop_checkpointer could be more intelligent, and 
         %% stop cache_trickler_checkpointer process and *start* the cache_tricker process.
         {run_txn_funs_on_BDB_event,

          [{bdb_DB_EVENT_REP_MASTER, [{?MODULE, txn_fun_start_checkpointer_if_enabled, [], 0},
                                      %% this code is inserted into init(), which binds this variable:
                                      {?MODULE, txn_fun_set_master_cache_throttle_io, MasterCacheIO, 0}
                                      ]},

           {bdb_DB_EVENT_REP_CLIENT, [{?MODULE, txn_fun_stop_checkpointer_if_running, [], 0}]}
          ]}
        ]).


%% Macros to enforce assertion of state invariants.
%% Because Erlang has such clean management of modifications to state we can optimize use of 
%% asserts.  
%%   - We assert invariants *immediately* after every change to State, even if it is about to be modified again.
%%   - We *don't* assert when only reading State (the most common case)
%%   - We *don't* assert when returning unchanged State from a gen_server callback.
%%
%% This policy has several benefits
%%   - almost optimally efficient (only assert on change)
%%   - safe (if enforced rigorously)
%%   - makes it clear who should assert invariants if a helper function returns a changed state 
%%     (i.e. the helper function should do it internally before returning; the caller of the function should not do it)
%%   - amenable to automation via a macro that removes all explicit calls to assert_state_invariants and eliminates lots of temporary state variables.
%%
%% Implementing this policy is simple: The following macro should be used when generating a new value for State.
%% However, see the discussion immediately below.

%% This introduces a variable, NewState, and doesn't have any cleverness to avoid multiple occurences per scope 
%% (e.g. using a fun), because the nature of this macro is that it should only be used once per scope anyway.
-define(NEW_STATE(Expr), begin
                             NewState = Expr, % only evaluate it once (may have side-effects)
                             assert_state_invariants(NewState),
                             NewState
                         end).

%% The only signficant downside of this simple scheme appears to be that 
%% it does not allow fine-grained multi-step modifications to State 
%% -- i.e. chains of composed modifications, each stage which might produce a State
%% that would not pass assert_state_invariants, but which (as a chain) finally generate 
%% a valid state.  
%% This practice might not rest on very solid formal ground, but is very useful for 
%% separation-of-concerns (i.e. functions which modify a partial state and then 
%% return another partial state).
%% We want to support this and make it very visible/explicit when it is occuring.
%% So we use the following convention:
%%
%%  - Anything that creates a *partial* State (that might not pass invariants)
%%    must use the macro ?NEW_PARTIAL_STATE().
%%    Currently this is only an identity macro, used to make the intent visible.
%%    (Obviously it does not call assert_state_invariants.)
%%    In future we might consider setting a State member to record the fact that it 
%%    inconsistent. (However, using that flag to suppress asserts would be 
%%    very dangerous -- if it ever 'escapes' it would suppress all future asserts).
%%
%%  - All functions that return a *partial* State (and therefore do not internally 
%%    assert invariants on the value that they return) must have names that 
%%    end with the suffix "_UNSAFE".
%%
%%  - All variables that are bound to temporary *partial* copies of State 
%%    also have names that end with the suffix "_UNSAFE".
%%
%%  - The only way to remove the 'unsafe' property from a State value 
%%    is to pass it through the macro ?NEW_CONSISTENT_STATE(),
%%    which obviously does call assert_state_invariants.
%%    (If we do ever set a State member to record the 'partial' condition 
%%    then this macro would clear that value before calling asserts).

-define(NEW_PARTIAL_STATE(NewState),    NewState).             % no assert check
-define(NEW_CONSISTENT_STATE(NewState), ?NEW_STATE(NewState)).


-record(site_status, {status,                 % see comment at top of file
                      last_updated_time,      % local erlang:now() of last change to status | unknown
                      replication_group_pid,  % ReplicationGroupPid | unknown
                      monitor_ref             % Ref | unknown
                     }).
%% Note that in term comparisons, atom (e.g. 'unknown') < any erlang:now(), as any atom < any tuple

-define(BDB,       berkeley_db_drv).
-define(BDB_H,     berkeley_db_helpers).
-define(BDB_U,     berkeley_db_utilities).
-define(BDB_PC,    berkeley_db_port_coordinator).

-define(STRINGIZE(Arg), ??Arg).

-define(seed_rng(), 
        {RNGSeed1, RNGSeed2, RNGSeed3} = now(),
        random:seed(RNGSeed1, RNGSeed2, RNGSeed3)).


% Debug helpers (poor-man's tracing)
-define(AT, io:format("~nAt line ~w~n", [?LINE])).   %% remove with regex    ^ *\?AT, *C-qC-j


%%====================================================================
%% API
%%====================================================================

-define(THROW_SIMPLE_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_replication_group_error, Error})
           end).

%% One of these macros must be used for any API that takes a SiteLocator or MasterLocationHint.
%%
%% Note that it throws most {error, Reason} as just Reason i.e. it does *not* wrap (most) of them in {bdb_replication_group_error, _}
%% That's because some errors (specifically txn_fun_error) are not replication_group errors - they are caller errors, 
%% that we are simply passing through to the client.

-define(THROW_REMOTE_SITE_ERRORS(Expr, SiteLocator),
           try Expr of
               {ok, Result}    -> Result;
               {error, Reason} -> throw(Reason)
           catch
               exit:Reason ->
                   %% gen:server:call will exit() if the destination process is down,
                   %% such that we get:
                   %%
                   %%   exit:{noproc,   {gen_server, call, [Server, Args]}}
                   %%   exit:{killed,   {gen_server, call, [Server, Args]}}
                   %%   exit:{nodedown, {gen_server, call, [Server, Args]}}
                   %%   exit:{timeout,  {gen_server, call, [Server, Args]}}
                   %%
                   %% That is safe here (it won't kill this site) because we are in a spawned process.
                   %% But if we don't catch it we won't reply to the client, so the client will 
                   %% block for the full timeout.
                   %% We convert it to an error return that is thrown as an exception on the 
                   %% client, so the client can retry with another site 
                   throw({bdb_replication_group_error, {site_is_down, SiteLocator, Reason}})
           end).

%% This variant must be used for any API that calls ?BDB_H:do_txn_no_throw() on the server.
%% See also forward_command_to_remote_site(), which contains a non-throwing a variant of this.
-define(THROW_TXN_REMOTE_SITE_ERRORS(Expr, SiteLocator),
           try Expr of
               {ok, Result}    -> Result;

               %% Even if the replication_group server is up, the port-coordinator on that 
               %% site might have just crashed, or be hung.
               %% Such errors are returned from the local call to do_txn_no_throw() on the target site.
               %% We re-throw these as a summary 'site_is_down' exception, for convenience of the caller.
               {error, {bdb_helper_error, {acquire_bdb_port_handle, {{exit, {noproc,{gen_server,call,_}}}, _}}} = Reason} ->
                   throw({bdb_replication_group_error, {site_is_down, SiteLocator, Reason}});

               {error, {bdb_helper_error, {release_bdb_port_handle, {{exit, {bdb_receive_reply_timeout, _}}}, _}} = Reason} ->
                   throw({bdb_replication_group_error, {site_is_down, SiteLocator, Reason}});

               {error, {bdb_helper_error, {release_bdb_port_handle, {port_is_unreachable, _}}} = Reason} ->
                   throw({bdb_replication_group_error, {site_is_down, SiteLocator, Reason}});

               %% All other errors are re-thrown unchanged.
               {error, Reason} -> throw(Reason)
           catch
               exit:Reason ->
                   %% see comments in THROW_REMOTE_SITE_ERRORS
                   throw({bdb_replication_group_error, {site_is_down, SiteLocator, Reason}})
           end).


%%--------------------------------------------------------------------
%% Function: start_sites() -> {ok,[ListOfPids]} | {error,Error}
%%
%% Description: Starts MULTIPLE sites on any combination of local and remote-nodes.
%%
%% AllSites must be 
%%
%%           [ {{SiteName, NodeName}, BdbEnvDir, {HostName, Port}, SiteSpecificPortCoordConfigProplist}, 
%%             ... ]
%%
%% See also definition of DEFAULT_CONFIG_PROPLIST.
%%
%% Returns   {GoodSitesPids, ErrorSiteaReasons}
%%
%%              where a 'good site/pid' is     {Site, Pid}
%%              and an 'error' site/reason' is {Site, Reason}
%%
%% The order of sites in the result is not guaranteed to match the order of the parameter list.
%%
%% This does NOT link or monitor the resulting pids - the caller should do that if desired.
%% The caller can easily call erlang:monitor on the returned GoodPids.
%%
%% Rationale:
%%
%% This is a convenience function that is somewhat redundant as 
%% starting a single site somewhere will attempt to spawn all of the others, so this function  
%% is somewhat redundant.
%%
%% It exists because: 
%%
%%  - the caller might not want a site on it's local node
%%  - the caller can learn which sites were successfully spawned and which ones failed
%%
%% This calls start() on the required nodes (inc local node if necessary).
%%--------------------------------------------------------------------
start_sites(ReplicationGroupName, AllSites, ConfigProplist) ->    

    %% gen_server does not have a remote-spawn-link facility, so we must use an rpc to 
    %% spawn on remote nodes.
    %%
    %% There may be many sites (almost certainly more than 1), so we start them in parallel.
    %%
    %% For now we use this (which spawns a local process per list element, each of which calls rpc:call and blocks until it returns):
    %%   ?BDB_U:pmap(Fun/1, List) -> ResultsList
    %%   Evaluates Fun for every element Elem in List, in parallel. Returns the list of return values, in the same order as the input list.
    %%
    %% An alternative would be to use
    %%   rpc:async_call(Node, Module, Function, Args) -> Key
    %%   Implements call streams with promises, a type of RPC which does not suspend the caller until the result is finished. Instead, a key is returned which can be used at a later stage to collect the value. The key can be viewed as a promise to deliver the answer. 
    %%   In this case, the key Key is returned, which can be used in a subsequent call to yield/1 or nb_yield/1,2 to retrieve the value of evaluating apply(Module, Function, Args) on the node Node. 
    %%
    %% Important: pmap spawns short-lived processes to do call our Fun, so obviously the Fun cannot use start_link as the spawned 
    %% server would be linked to the shor-lived pmap helper process.  But were not using links anyway.
   
    lists:foldl(fun({Site, {ok, Pid}},       {Good, Bad})  -> {[{Site, Pid} | Good], Bad};
                   ({Site, {error, Reason}}, {Good, Bad})  -> {Good, [{Site, Reason} | Bad]};
                   ({Site, Other},           {Good, Bad})  -> {Good, [{Site, Other} | Bad]}
                end,
                {[], []},
                ?BDB_U:pmap(
                   fun({{_, NodeName}, _, _, _} = Site) ->
                           {Site, 
                            try rpc:call(NodeName, ?MODULE, start, [ReplicationGroupName, 
                                                                    Site,
                                                                    AllSites -- [Site], 
                                                                    ConfigProplist]) 
                            of
                                Pid when is_pid(Pid)  -> {ok, Pid};
                                Other                 -> {error, Other}
                            catch
                                Class:Reason          -> {error, {Class, Reason}}
                            end}
                   end,
                   AllSites)).


%%--------------------------------------------------------------------
%% Function: start() -> Pid or throws
%%
%% Description: Starts ONE site on the LOCAL node.
%%
%% This does NOT link the caller to the spawned process.
%% If that is desired then the caller should use erlang:monitor() on the returned pid.
%%
%% ThisSite      must be   {{SiteName, node()}, BdbEnvDir, {HostName, Port}, SiteSpecificPortCoordConfig}
%% OtherSites    must be [ {{SiteName, NodeName}, BdbEnvDir, {HostName, Port}, SiteSpecificPortCoordConfig}, ... ]
%%
%%--------------------------------------------------------------------
start(ReplicationGroupName, 
      {{SiteName, NodeName}, _, {_, _}, _} = ThisSite,
      OtherSites, 
      ConfigProplist) ->

    NodeName = node(), % assert
    ?THROW_SIMPLE_ERRORS( gen_server:start({local, get_proc_name(ReplicationGroupName, SiteName)}, 
                                           ?MODULE, 
                                           {ReplicationGroupName, ThisSite, OtherSites, ConfigProplist},
                                           []) ).


%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%%
%% Description: Starts ONE site on the LOCAL node.
%%
%% This is for use with conventional supervisors.
%% Returns 
%%         {ok, Pid}
%%         {error, Error}
%%
%%--------------------------------------------------------------------
start_link(ReplicationGroupName, 
           {{SiteName, NodeName}, _, {_, _}, _} = ThisSite,
           OtherSites, 
           ConfigProplist) ->

    NodeName = node(), % assert

    gen_server:start_link({local, get_proc_name(ReplicationGroupName, SiteName)}, 
                          ?MODULE, 
                          {ReplicationGroupName, ThisSite, OtherSites, ConfigProplist},
                          []).


%%--------------------------------------------------------------------
%% Function: get_proc_name(ReplicationGroupName, SiteName)
%%
%% Description: Finds the locally-registered name of the replication_group process 
%%              for a given group/site
%%
%%--------------------------------------------------------------------
get_proc_name(ReplicationGroupName, SiteName) ->
    list_to_atom(get_proc_name_as_list(ReplicationGroupName, SiteName)).

get_proc_name_as_list(ReplicationGroupName, SiteName) ->
    "bdb_" ++ atom_to_list(ReplicationGroupName) ++ "_" ++ atom_to_list(SiteName) ++ "_replication_group".    


%%--------------------------------------------------------------------
%% Function: get_port_coordinator_proc_name(ReplicationGroupName, SiteName)
%%
%% Description: Finds the locally-registered name of the port_coordinator process 
%%              for a given group/site
%%
%%--------------------------------------------------------------------
get_port_coordinator_proc_name(ReplicationGroupName, SiteName) ->
    list_to_atom("bdb_" ++ atom_to_list(ReplicationGroupName) ++ "_" ++ atom_to_list(SiteName) ++ "_port_coordinator").


%%--------------------------------------------------------------------
%% Function: get_master_location_hint(SiteLocator, FindMasterTimeout)
%%
%% Description: See API documentation at top of file.
%%
%% This supports FindMasterTimeout 0 .. infinity, but ClientServerTimeout should be 1..N (i.e. not infinity).
%% If FindMasterTimeout is infinity then ClientServerTimeout is ignored (i.e. set to infinity too).
%%
%% SiteLocator is {ReplicationGroupName, SiteName, SiteNode}
%%
%% Returns an opaque MasterLocationHint object.
%%
%%--------------------------------------------------------------------

%% This variant uses the default gen_server 5000ms client-server timeout
get_master_location_hint({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, FindMasterTimeout) ->
    get_master_location_hint({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, FindMasterTimeout, 5000).

get_master_location_hint({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, FindMasterTimeout, ClientServerTimeout) ->

    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               {get_master_location_hint, FindMasterTimeout},
                                               case FindMasterTimeout of 
                                                   infinity -> infinity;
                                                   _        -> FindMasterTimeout + ClientServerTimeout
                                               end),
                               SiteLocator).

%%--------------------------------------------------------------------
%% Function: do_txn_on_master
%%
%% Description: See API documentation at top of file.
%%
%% There are 2 versions:
%%
%%  - first clause takes a MasterLocationHint which contains a Distributed Erlang pid.
%%
%%  - second clause takes {ReplicationGroupName, SiteName, Node} and is used by 
%%    clients when they don't have a MasterLocationHint.
%%
%% There are also convenience functions provide a default timeout and NumDeadlockRetries.
%%
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_master(MasterLocationHint, TxnFun) ->
    do_txn_on_master(MasterLocationHint, TxnFun, 0, 5000). % default NumDeadlockRetries & gen_server timeout

do_txn_on_master(MasterLocationHint, TxnFun, NumDeadlockRetries) ->
    do_txn_on_master(MasterLocationHint, TxnFun, NumDeadlockRetries, 5000). % default gen_server timeout

%% The main function
do_txn_on_master(#master_location_hint{pid          = MasterPid,
                                       site_locator = MasterSiteLocator}, 
                 TxnFun, 
                 NumDeadlockRetries, 
                 TxnFunTimeout) ->

    ?THROW_TXN_REMOTE_SITE_ERRORS( gen_server:call(MasterPid,
                                                   {do_txn_on_master, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                                   TxnFunTimeout),
                                   MasterSiteLocator );

do_txn_on_master({ReplicationGroupName, SiteName, SiteNode} = SiteLocator,
                 TxnFun, 
                 NumDeadlockRetries,
                 TxnFunTimeout) -> 

    ?THROW_TXN_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                                   {do_txn_on_master, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                                   TxnFunTimeout),
                                   SiteLocator ).


%%--------------------------------------------------------------------
%% Function: do_txn_on_site
%%
%% Description: See API documentation at top of file.
%%
%% There are convenience functions provide a default timeout and NumDeadlockRetries.
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_site(SiteLocator, TxnFun) ->
    do_txn_on_site(SiteLocator, TxnFun, 0, 5000).  % default NumDeadlockRetries & gen_server timeout

do_txn_on_site(SiteLocator, TxnFun, NumDeadlockRetries) ->
    do_txn_on_site(SiteLocator, TxnFun, NumDeadlockRetries, 5000). % default gen_server timeout

do_txn_on_site({ReplicationGroupName, SiteName, SiteNode} = SiteLocator,
                   TxnFun, 
                   NumDeadlockRetries,
                   TxnFunTimeout) -> 

    ?THROW_TXN_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                                   {do_txn_on_site, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                                   TxnFunTimeout),
                                   SiteLocator ).


%%--------------------------------------------------------------------
%% Function: do_txn_on_any_site
%%
%% Description: See API documentation at top of file.
%%
%% There are convenience functions provide a default timeout and NumDeadlockRetries.
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_any_site(SiteLocator, TxnFun) ->
    do_txn_on_any_site(SiteLocator, TxnFun, 0, 5000).  % default NumDeadlockRetries & gen_server timeout

do_txn_on_any_site(SiteLocator, TxnFun, NumDeadlockRetries) ->
    do_txn_on_any_site(SiteLocator, TxnFun, NumDeadlockRetries, 5000). % default gen_server timeout

do_txn_on_any_site({ReplicationGroupName, SiteName, SiteNode} = SiteLocator,
                   TxnFun, 
                   NumDeadlockRetries,
                   TxnFunTimeout) -> 

    ?THROW_TXN_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                                   {do_txn_on_any_site, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                                   TxnFunTimeout),
                                   SiteLocator ).


%%--------------------------------------------------------------------
%% Function: remove_site(ReplicationGroupName, {SiteName, SiteNode})
%%
%% Description: 
%%
%%  Removes a given site from the group and shuts down that site.
%%  The site can be added back into the group with a call to add_site().
%%
%% TODO:
%%
%%  Currently this just makes all sites forget about the target site, and 
%%  then shuts it down.
%%  We we have persistent group membership, this will change that membership.
%%
%%  If the target site happens to be the master then we must do a clean 
%%  safe fail-over
%%    - suspend writes on that site (not all sites, so not same as stop_entire_group)
%%    - shutdown that site gracefully (i.e. so that an election is called immediately 
%%      and the ex-master does not participate)
%%      e.g. set ex-master's site-priority to   and then call rep_start(CLIENT) to downgrade it,
%%      (check with Sleepycat that repmgr  calls an election immediately if the master does it). 
%%
%%--------------------------------------------------------------------
remove_site({ReplicationGroupName, SiteName, SiteNode} = SiteLocator) ->
    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               remove_self),
                               SiteLocator ).


%%--------------------------------------------------------------------
%% Function: get_group_status(SiteLocator)
%%
%% Description: 
%%
%%  Returns current status of all the sites.
%%
%%           [{SiteId, Status}, ... ]
%%
%%  where SiteId = {SiteName, SiteNode}
%%        Status = see comment at top of file
%%
%%  Of course, this information may be out of date even before the  caller receives it.
%%
%%--------------------------------------------------------------------
get_group_status({ReplicationGroupName, SiteName, SiteNode} = SiteLocator) ->
    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               get_group_status),
                               SiteLocator ).


%%--------------------------------------------------------------------
%% Function: stop_entire_group(SiteLocator)
%%
%% Description: 
%%
%%   Performs a *temporary* shutdown of ALL sites in the group.
%%   'Temporary' means that the group membership is NOT changed 
%%   (when we eventually store group membership in a replicated database).
%%
%% This take the same kind of 'site locator' as do_txn_on_master
%% -- i.e. a MasterLocationHint or any {ReplicationGroupName, SiteName, Node}
%%
%% TODO: 
%%
%% Currently this command just shuts down all sites and RETURNS WITHOUT WAITING.
%% (it does not wait as it is hard to avoid deadlocks).
%%
%% It needs to be completely rewritten to do the following:
%%
%%    - suspend writes (on all sites not just current master, so even if master fails we won't start taking writes again)
%%    - wait for all sites to sync-up (with a timeout)
%%    - shutdown all sites.
%%    - return when complete (or an error occurs)
%%
%% This will use the persistent group-membership database as a group consensus persistent state machine.
%%
%%--------------------------------------------------------------------
stop_entire_group(#master_location_hint{pid = MasterPid}) ->
    gen_server:cast(MasterPid, {stop_entire_group, self()});

stop_entire_group({ReplicationGroupName, SiteName, SiteNode}) ->
    gen_server:cast({get_proc_name(ReplicationGroupName, SiteName), SiteNode}, 
                    {stop_entire_group, self()}).


%%--------------------------------------------------------------------
%% Function: get_site_locator_from_master_location_hint(MasterLocationHint)
%%
%% Description: Exposes master site ID, for logging etc.
%%--------------------------------------------------------------------
get_site_locator_from_master_location_hint(#master_location_hint{site_locator = SL}) ->
    SL.

%%--------------------------------------------------------------------
%% Function: get_node_from_master_location_hint(MasterLocationHint)
%%
%% Description: Allows application to move work to the master node,
%%              to reduce number of hops when invoking do_txn_on_master 
%%              multiple times.
%%--------------------------------------------------------------------
get_node_from_master_location_hint(#master_location_hint{site_locator = {_, _, Node}}) ->
    Node.


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).


%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%
%%--------------------------------------------------------------------
init({ReplicationGroupName, 
      {{ThisSiteName, ThisSiteNode} = ThisSiteId, ThisSiteEnvDir, {ThisSiteHostName, ThisSitePort}, ThisSitePortCoordConfig} = ThisSite, 
      OtherSites, 
      SpecifiedConfigProplist}) ->

    ThisSiteNode = node(),  % assert

    ?INFO_LOGGER("starting (~w)~n", [self()]),
    process_flag(trap_exit, true),

    ThisSitePortCoordName = get_port_coordinator_proc_name(ReplicationGroupName, ThisSiteName),

    %% Reconcile _other_ specified config with our defaults
    %% -- and then tack-on the generated PortCoordConfig above, just for future reference     
    ConfigProplist = ?BDB_U:get_dominant_proplist_values([SpecifiedConfigProplist,
                                                         ?DEFAULT_CONFIG_PROPLIST]),

    %% Validate config
    lists:foreach(
      fun({common_port_coordinator_config, L}) when is_list(L)                         -> ok;
         ({find_other_sites_period_milliseconds, N}) when is_integer(N), N > 0         -> ok;
         ({rpc_call_timeout_milliseconds, N}) when is_integer(N), N > 0                -> ok;
         ({spawn_other_sites, enabled})                                                -> ok;
         ({spawn_other_sites, disabled})                                               -> ok;
         ({perm_failed_events_limits = K, disabled = V})                               -> put(K, V); 
         ({perm_failed_events_limits = K, {{waiting_for_durability, {C1, W1} = V1},
                                           {available,              {C2, W2} = V2}}})
          when is_integer(C1), C1 >= 0, is_integer(W1), W1 > 0,
               is_integer(C2), C2 >= 0, is_integer(W2), W2 > 0                         -> put(K, {V1, V2});
         ({master_cache_throttle_io, same_as_replicas})                                -> ok;
         ({master_cache_throttle_io, {W, S}})
         when is_integer(W), W >= 0, is_integer(S), S >= 0                             -> ok;
         ({probe_to_sync_replicas_periods_milliseconds = K, {{not_durable, ND}, 
                                                             {durable,      D}}})
         when is_integer(ND), ND >= 0, is_integer(D), D >= 0                           -> put(K, {ND, D});
         ({check_for_lagging_replicas_period_seconds, N}) when is_integer(N), N > 0    -> ok;
         ({check_for_lagging_replicas_period_seconds, disabled})                       -> ok;
         ({lagging_replica_threshold_bytes, N}) when is_integer(N), N >= 0             -> ok;
         ({lagging_replica_restart_after_N_consecutive, N}) when is_integer(N), N > 0  -> ok; 
         ({log_removal_period_seconds, N}) when is_integer(N), N > 0                   -> ok;
         ({log_removal_period_seconds, disabled})                                      -> ok;
         ({log_removal_delete_file_period_milliseconds, N}) when is_integer(N), N >= 0 -> ok;
         ({log_removal_max_num_trailing_log_files, N}) when is_integer(N), N > 0       -> ok;
         (Other) -> 
              throw({unrecognized_config_option, Other})
      end,
      ConfigProplist),

    %% The macro ?DEFAULT_COMMON_PORT_COORDINATOR_CONFIG (used below) has a free-variable reference
    %% to this:
    MasterCacheIO = get_value_for_key(master_cache_throttle_io, ConfigProplist),

    %% Prepare port_coordinator configuration for this site.
    %%  - tell BDB-HA which host IP/names and ports to use.
    %%  - tell the local port_coordinator that we we want to receive all replication_gen_event messages 

    PortCoordConfig = 
        [{replication, [{this_hostname_port, {ThisSiteHostName, ThisSitePort}},
                        {peer_hostnames_ports, lists:map(fun({_, _, HostNamePort, _}) -> 
                                                                 HostNamePort
                                                         end,
                                                         OtherSites)},
                        {event_handlers, [{berkeley_db_replication_group_gen_event_handler,
                                           %% It is OK to use 'self()' here, rather than a process name.  
                                           %% If we die and restart and find that port_coordinators are 
                                           %% already running then we will need to get an update of absolute state anyway.
                                           {forward_events_as_messages_to, local_sleepycat_event, self()}}]}
                       ]}]
        ++ ThisSitePortCoordConfig
        ++ get_value_for_key(common_port_coordinator_config, SpecifiedConfigProplist)
        ++ ?DEFAULT_COMMON_PORT_COORDINATOR_CONFIG,

    %% Our ?DEFAULT_COMMON_PORT_COORDINATOR_CONFIG disables LOG_AUTOREMOVE as it is unsafe.
    %% Reject any caller-supplied config that attempts to enable LOG_AUTOREMOVE.
    case lists:member(bdb_DB_LOG_AUTOREMOVE, 
                      proplists:get_value(env_set_flags, PortCoordConfig, []))
    of
        true -> throw({unsafe_port_coordinator_config, {env_set_flags, bdb_DB_LOG_AUTOREMOVE}});
        _    -> ok
    end,

    %% Display the final config
    lists:foreach(fun({Key, Value}) -> ?INFO_LOGGER_NO_TIMESTAMP("init: ~55w = ~200p~n", [Key, Value]) end, lists:keysort(1, ConfigProplist)),

    %% Display site-info
    display_site(this_site, ThisSite),
    lists:foreach(fun(OtherSite) -> display_site(other_site, OtherSite) end, OtherSites),

    %% The port-coordinator sends NEWMASTER events which include the BDB-HA HostPort of the site that is 
    %% now master. We need to translate them to our SiteIds {SiteName, SiteNode}
    put(peer_hostnameports_to_site_ids,
        lists:map(fun({SiteId, _, HostNamePort, _}) -> {HostNamePort, SiteId} end,
                  OtherSites)),

    %% Spawn our local port_coordinator first, as this will validate the config.
    %%
    %% If the port_coordinator claims to already be running then 
    %% tell it to shutdowm (it is almost certainly already shutting down), 
    %% wait for it to be unregistered, and then try again.
    %% This can happen if someone does exit(RgPid, kill) and then 
    %% a supervisor restarts the replication-group server while the 
    %% old port-coordinator is still shuting down.  We send the port-coordinator 
    %% a non-killing exit signal (as a private async version of 'stop' API)
    %% just incase there is some other circumstance that the port-coord 
    %% might be running -- e.g. someone started one with just the right name 
    %% but which was not linked to any previous instance of this replication-group serve.

    PortCoordPid = start_port_coordinator(ThisSitePortCoordName, ThisSiteEnvDir, PortCoordConfig),

    ?INFO_LOGGER("local port_coordinator: ~w (registered as ~w)~n", [PortCoordPid, ThisSitePortCoordName]),
    
    %% Spawn a helper process to start trying to connect to peer sites that are 'unknown' (and spawn them if they are not running).
    ThisSiteReplicationGroupPid = self(),
    FindOrSpawnPeersSupportProc = 
        #support_proc{name     = find_or_spawn_peers, 
                      pid      = proc_lib:spawn_link(
                                   fun() -> find_or_spawn_peers_process(
                                              ReplicationGroupName, 
                                              ThisSiteId,
                                              ThisSiteReplicationGroupPid, 
                                              [SiteId || {SiteId, _, _, _} <- OtherSites],  % SiteIds to find/spawn
                                              get_value_for_key(find_other_sites_period_milliseconds, ConfigProplist),
                                              get_value_for_key(rpc_call_timeout_milliseconds, ConfigProplist),
                                              case get_value_for_key(spawn_other_sites, ConfigProplist) of
                                                  enabled ->                       
                                                      %% info required when spawning a site
                                                      {[ThisSite | OtherSites],             % AllSites
                                                       ConfigProplist};         
                                                  disabled ->
                                                      disabled
                                              end)
                                   end),
                      stop_fun = fun(#support_proc{pid = Pid}) -> 
                                         exit(Pid, kill)  % we don't need a clean exit (and we want it to be fast)
                                 end},

    State = #state{current_master     = unknown,
                   this_site_status   = #site_status{status            = {replica, not_in_sync, unknown}, % looking for master
                                                     last_updated_time = now()},                          % other members are intentionally left as 'undefined'                   
                   other_sites_status = [{OtherSiteId, #site_status{status                = unknown,
                                                                    last_updated_time     = unknown,      % we don't use the default of 'undefined' as we want to detect bugs of omission (failing to set a field)
                                                                    replication_group_pid = unknown,
                                                                    monitor_ref           = unknown}}
                                         || {OtherSiteId, _, _, _} <- OtherSites],
                   other_sites        = OtherSites,
                   notify_clients     = [],
                   perm_failed_times  = [],
                   support_procs      = [FindOrSpawnPeersSupportProc]},

    %% Save immutable state
    put(config_proplist,            ConfigProplist),
    put(replication_group_name,     ReplicationGroupName),
    put(this_site_port_coord_name,  ThisSitePortCoordName),
    put(this_site_id,               ThisSiteId),
    put(this_site_name,             ThisSiteName),
    put(this_site_port_coord_pid,   PortCoordPid),

    ?seed_rng(),

    {ok, ?NEW_STATE(State)}.


display_site(Name, {SiteId, _SiteEnvDir, BDBHAHostPort, SitePortCoordConfig}) -> 
    ?INFO_LOGGER_NO_TIMESTAMP("init: ~w ~w ~258p ~258p~n", [Name, SiteId, BDBHAHostPort, SitePortCoordConfig]).
%%     ?INFO_LOGGER_NO_TIMESTAMP("init: ~w site_id           = ~200p~n", [Name, SiteId]),
%%     ?INFO_LOGGER_NO_TIMESTAMP("init: ~w bdb_ha_host_port  = ~200p~n", [Name, BDBHAHostPort]),
%%     ?INFO_LOGGER_NO_TIMESTAMP("init: ~w port_coord_config = ~200p~n", [Name, SitePortCoordConfig]).


%% See comments in init()
start_port_coordinator(ThisSitePortCoordName, ThisSiteEnvDir, PortCoordConfig) ->

    case ?BDB_PC:start_link(ThisSitePortCoordName, ThisSiteEnvDir, PortCoordConfig) of
        {ok, Pid} -> 
            Pid;
        {error, {already_started, ExistingPid}} ->
            ?WARNING_LOGGER("found existing port_coordinator: ~w (registered as ~w); telling it to exit...~n", [ExistingPid, ThisSitePortCoordName]),
            exit(ExistingPid, replication_group_server_restarting),
            repeat_until_true(fun() ->
                                      case whereis(ThisSitePortCoordName) of
                                          undefined   -> 
                                              ?INFO_LOGGER("existing port_coordinator ~w (registered as ~w) has exited~n", [ExistingPid, ThisSitePortCoordName]),
                                              true;
                                          ExistingPid ->     % any other pid/value and we crash, as something very fishy is going on
                                              ?INFO_LOGGER("existing port_coordinator ~w (registered as ~w) has NOT yet exited~n", [ExistingPid, ThisSitePortCoordName]),
                                              timer:sleep(200),
                                              false      
                                      end
                              end),

            %% We expect it to work now -- i.e. give it one chance, we don't loop
            {ok, Pid} = ?BDB_PC:start_link(ThisSitePortCoordName, ThisSiteEnvDir, PortCoordConfig),
            Pid
    end.


%%--------------------------------------------------------------------
%% Function: handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------

handle_call({get_master_location_hint, FindMasterTimeout}, 
            From, 
            #state{current_master     = CurrentMaster,
                   other_sites_status = OtherSitesStatus,
                   notify_clients     = NotifyClients} = State) ->

    case CurrentMaster of 

        {this_site, available} ->
            ?INFO_LOGGER("get_master_location_hint: returning this_site~n"),
            {reply, 
             {ok, #master_location_hint{site_locator = {get(replication_group_name), get(this_site_name), node()},
                                        pid          = self()}},
             State};

        {{MasterSiteName, MasterSiteNode} = MasterSiteId, available} ->

            ?INFO_LOGGER("get_master_location_hint: returning ~w~n", [MasterSiteId]),
            {value, {_, #site_status{replication_group_pid = MasterReplicationGroupPid}}} 
                = lists:keysearch(MasterSiteId, 1, OtherSitesStatus),

            {reply, 
             {ok, #master_location_hint{site_locator = {get(replication_group_name), MasterSiteName, MasterSiteNode},
                                        pid          = MasterReplicationGroupPid}},
            State};

        _ ->
            ?INFO_LOGGER("get_master_location_hint: waiting to discover master~n"),

            case FindMasterTimeout of
                0 ->
                    %% caller was just polling -- reply immediately
                    {reply, {error, {bdb_replication_group_error, no_master}}, State};

                infinity ->
                    %% Add this client to the list of clients to notify if we do discover a master.
                    ?INFO_LOGGER("get_master_location_hint: will notify client ~w when master is discovered (infinite timeout)~n", [From]),
                    {noreply, ?NEW_STATE(State#state{notify_clients = [{From, infinity} | NotifyClients]})};

                _ ->
                    %% Add this client to the list of clients to notify if we do discover a master,
                    %% *and* also send a message to ourself on a timer to tell the client that there is no_master 
                    %% (if one has not been discovered by then).
                    %% (Note: we use erlang:start_timer not timer:send_after as apparently the latter is very inefficient).
                    %% start_timer sends a message: {timeout, TimerRef, Msg}.
                    %% We store the TimerRef in notify_clients so we can cancel the timer if we do discover a master.

                    TimerRef = erlang:start_timer(FindMasterTimeout, self(), {notify_client_timer_expired, From}), 
                    ?INFO_LOGGER("get_master_location_hint: will notify client ~w (~w ms timeout, TimerRef ~w)~n", [From, FindMasterTimeout, TimerRef]),

                    {noreply, ?NEW_STATE(State#state{notify_clients = [{From, TimerRef} | NotifyClients]})}
            end
    end;

    
handle_call({do_txn_on_master, TxnFun, NumDeadlockRetries, Timeout} = Command, 
            From, 
            #state{current_master     = CurrentMaster,
                   other_sites_status = OtherSitesStatus} = State) ->

    case CurrentMaster of 

        {this_site, available} ->
            %%?INFO_LOGGER("do_txn_on_master on local site~n"),
            async_do_txn_on_master_local_site(Command, From),
            {noreply, State};

        {{_, _} = MasterSiteId, available} ->
            %%?INFO_LOGGER("do_txn_on_master: we are not master - forwarding to master site ~w~n", [MasterSiteId]),

            %% Forward command to remote site that is master
            %% If the destination server dies or hangs then the gen_server:call will timeout.

            %% Note: MasterSiteId is {SiteName, Node}, so we have to translate it to find 
            %% the replication_group process.
            %% There are two translations:
            %%   1.   {get_proc_name(get(replication_group_name), SiteName), Node}
            %%   2.   Look for MasterSiteId in State.other_sites_status and use the replication_group_pid element.

            {value, {_, #site_status{replication_group_pid = MasterReplicationGroupPid}}} 
                = lists:keysearch(MasterSiteId, 1, OtherSitesStatus),

            %% Note that we change CommandName to do_txn_on_master_on_local_site, to avoid any danger of forwarding more 
            %% than once.  This means that if we are being called by do_txn_on_master and the caller site 
            %% has an out-of-date notion of who the master is, and the target site has the correct information,
            %% we will _still_ execute the txn on the first (target) site -- it won't be forwarded again.
            %% There is always some (small) risk of executing a txn on the 'wrong' site,
            %% so disallowing indefinite forwarding doesn't lose much, and gains predictability, and 
            %% avoids risk of forwarding in infinite cycles if sites happen to have pathologically bad 
            %% (e.g. circular) status information.
            %% If a txn fun REALLY needs to know if it is running on the master then it should 
            %% always begin a write operation (e.g. take a RMW lock) -- that will only ever succeed on the 
            %% master.
            %%
            %% Note: if the destination site (which we believe to be master) is actually down, 
            %% then the client will receive a 'site_is_down' exception, and NOT a no_master
            %% exception.  i.e. We don't have code to convert "site_is_down from the site that 
            %% we believe to be master" to no_master. 
            %% This may change in future, but currently, the reason for *not* doing the conversion 
            %% is that no_master means that the client can should be able to call get_master_location_hint 
            %% (on this site) to find the master, and we would immediately return the same master.
            %% So we should not return no_master without changing Site.current_master to 'unknown', 
            %% and that means overriding the BDB-HA state-tracking FSM, which is non-trivial 
            %% (e.g. a 'DOWN' message should be on the way from the crashed/unreachable master).
            %% However, this is clearly less convenient for the client, so we might revisit this in future.

            forward_command_to_remote_site(MasterSiteId, 
                                           MasterReplicationGroupPid, 
                                           {do_txn_on_master_on_local_site, TxnFun, NumDeadlockRetries, Timeout},
                                           From),
            {noreply, State};

        _ ->
            {reply, {error, {bdb_replication_group_error, no_master}}, State}
    end;


%% This is only used internally, when do_txn_on_master forwards to another site
handle_call({do_txn_on_master_on_local_site, _TxnFun, _NumDeadlockRetries, _Timeout} = Command, 
            From, 
            #state{current_master = CurrentMaster} = State) ->

    case CurrentMaster of 

        {this_site, available} ->
            %%?INFO_LOGGER("do_txn_on_master_on_local_site~n"),
            async_do_txn_on_master_local_site(Command, From),
            {noreply, State};

        {{_, _} = MasterSiteId, available} ->
            ?WARNING_LOGGER("do_txn_on_master_on_local_site: we are not master! We believe ~w is master~n", [MasterSiteId]),
            {reply, {error, {bdb_replication_group_error, {do_txn_on_master_on_local_site, get(this_site_id), MasterSiteId}}}, State};

        _ ->
            {reply, {error, {bdb_replication_group_error, no_master}}, State}
    end;


handle_call({do_txn_on_site, _TxnFun, _NumDeadlockRetries, _Timeout} = Command, 
            From, 
            State) ->

    async_do_txn_on_local_site(Command, From),
    {noreply, State};


%% We prefer sites in the following order
%%   - local site if it is in-sync (even if it happens to be the master)
%%   - any other in-sync replica (i.e. excluding the master)
%%   - the master
%%   - any site that we can find that is running (prefering the local site of course)
%%
%% TODO: accept a parameter that determines whether we will consider the last option
%% (i.e. a not-in-sync site).  This is critical for consistency.

handle_call({do_txn_on_any_site, _TxnFun, _NumDeadlockRetries, _Timeout} = Command, 
            From, 
            #state{current_master = unknown} = State) ->

    %% There is not even a master, so from our perspective there are currently NO in-sync
    %% sites, even if other sites have told us that they have found the master 
    %% and sync'd with it.
    %%
    %% If the client really want an in-sync site, we should return an error here.
    %% TODO: that should be a command parameter.   
    %%
    %% For the moment we choose to say that possibly stale data is better than failure,
    %% so we execute on the local site.
    %%
    %% [sidebar: this site might actually be {replica, in_sync, unknown}
    %% which means that it has received NEWMASTER and STARTUPDONE but not 
    %% received the status-update from the new master yet. If we ever do 
    %% have a parameter to disallow reads from not_in_sync sites, then 
    %% that case should be tested for, and the read allowed.
    %% Also, we _might_ want to prefer an other site that claims to be 
    %% in_sync with some master, even if this site has not actually heard 
    %% from that other master yet).

    ?WARNING_LOGGER("do_txn_on_any_site: using local site as the least-worst option (**there are no in-sync sites and no master**)~n"),
    async_do_txn_on_local_site(Command, From),
    {noreply, State};

handle_call({do_txn_on_any_site, _TxnFun, _NumDeadlockRetries, _Timeout} = Command, 
            From, 
            %% As do_txn_on_any_site is use for read-only transactions, 
            %% we DON'T require the master to be able to do durable writes 
            #state{current_master     = {CurrentMasterSiteId, _},
                   this_site_status   = #site_status{status = ThisSiteStatusValue},
                   other_sites_status = OtherSitesStatus} = State) ->

    %% We do know of a master, but we may or may not be in sync with it.

    case ThisSiteStatusValue of 

        {replica, in_sync, _} ->
            %%?INFO_LOGGER("do_txn_on_any_site: using local site as it is an in-sync replica~n"),
            async_do_txn_on_local_site(Command, From),
            {noreply, State};

        {master, _} ->
            %%?INFO_LOGGER("do_txn_on_any_site: using local site as it is master~n"),
            %% TODO: in future perhaps have an command-parameter to suppress this 
            %% option; i.e. force use of other in-sync replicas first, and then master 
            %% as a last resort (which is what the clause below this one does).
            %% That would move read-request load off the master, possibly improving 
            %% write latency at the expense of an extra hop for some reads.
            %% Whether this is a win or not depends on the application's access patterns.
            async_do_txn_on_local_site(Command, From),
            {noreply, State};

        _ -> 
            %% Local site is not in sync
            %% Are there any replicas that are in-sync with OUR view of the current master?
            %% TODO: should we just look for ANY in-sync replica, regardless of the master,
            %% as if there are more than one (including ours) our view is arbitrary).
            %% Or if there are more than one (including ours), should we refuse to run 
            %% the txn until things stabilize?
            %% Using OUR current_master _is_ better than arbitrary I think, as it is updated  
            %% by recent events (e.g. directly by BDB-HA NEWMASTER events). 
            %% i.e. it is the best knowledge we have.  Any information 
            %% in other_sites_status is a little more likely to be out of date.
            %% So we stick with this.

            case [SiteId || {SiteId, #site_status{status = {replica, in_sync, {MasterSiteId, _}}}} <- OtherSitesStatus, 
                            MasterSiteId == CurrentMasterSiteId] 
            of
                [] ->
                    %% There are not any in-sync replicas, so use the master (the only in-sync site).
                    %% (We know that the master cannot be the local site, as we checked earlier.)
                    ?WARNING_LOGGER("do_txn_on_any_site: using master as the least-worst option (**there are no in-sync replicas**)~n"),

                    {value, {_, #site_status{replication_group_pid = MasterReplicationGroupPid}}} 
                        = lists:keysearch(CurrentMasterSiteId, 1, OtherSitesStatus),

                    forward_command_to_remote_site(CurrentMasterSiteId, MasterReplicationGroupPid, Command, From);

                InSyncReplicas ->
                    %% We have a choice of in-sync sites.
                    %% Pick one at random (the RNG was seeded in init()).
                    SiteId = lists:nth( random:uniform(length(InSyncReplicas)), InSyncReplicas),

                    ?WARNING_LOGGER("do_txn_on_any_site: using another replica (~w) (**this replica is not in_sync**)~n", [SiteId]),

                    {value, {_, #site_status{replication_group_pid = ReplicationGroupPid}}} 
                        = lists:keysearch(SiteId, 1, OtherSitesStatus),

                    forward_command_to_remote_site(SiteId, ReplicationGroupPid, Command, From)
            end,
            {noreply, State}
    end;


handle_call(remove_self, 
            _, 
            #state{support_procs      = SupportProcs,
                   other_sites_status = OtherSitesStatus} = State) ->

    ?INFO_LOGGER("remove_self starting~n"),

    %% TODO: 
    %% 1. run a txn on the master to change group membership
    %% 2. if this site happens to be master then perform graceful failover
    %%    - suspend writes on this site only
    %%    - call an immediate election in which this site will not participate (e.g. set this site's priority to 0 and downgrade to replica)

    %% We are shutting down this site so we CAN block the server while we do the following:

    %% Stop trying to spawn other sites
    {value, FindOrSpawnPeersSupportProc} = lists:keysearch(find_or_spawn_peers, #support_proc.name, SupportProcs),
    StateOut = ?NEW_CONSISTENT_STATE(stop_support_procs_UNSAFE([FindOrSpawnPeersSupportProc], State)),

    %% Tell all other sites not to reconnect to / respawn this site
    %%
    %% Note: even when BDB (repmgr) allows us to break connections to sites (and prevent retries) 
    %% we don't need to call that API from this site as we are about to call terminate() and 
    %% shutdown the port_coordinator.
    %%
    %% Do this in parallel as the  group might be large
    ?BDB_U:pmap(fun({OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid}}) 
                   when is_pid(OtherSiteReplicationGroupPid) -> 

                        case (catch gen_server:call(OtherSiteReplicationGroupPid, {forget_other_site, get(this_site_id)})) of
                            ok ->
                                ?INFO_LOGGER("remove_self: successfully told other site ~w to forget this site~n", [OtherSiteId]);
                            Other ->
                                ?WARNING_LOGGER("remove_self: failed to tell other site ~w to forget this site (~w)~n", [OtherSiteId, Other])
                        end;

                   ({OtherSiteId, _}) ->
                        ?INFO_LOGGER("remove_self: not connected to ~w so don't need to tell it to forget about this site~n", [OtherSiteId])
                end,
                OtherSitesStatus),

    %% Terminate
    {stop, normal, ok, StateOut};


%% This is an internal command used by remove_self
handle_call({forget_other_site, OtherSiteId} = Msg, 
            {FromPid, _}, 
            #state{other_sites        = OtherSites,
                   other_sites_status = OtherSitesStatus} = State) ->

    ?INFO_LOGGER("forget_other_site ~w~n", [OtherSiteId]),

    %% Stop trying to connect to/spawn the other site
    %% Note race condition (we might be connecting right now). So subscribe needs to handle case where OtherSiteId has been forgotten.
    send_msg_to_find_or_spawn_peers_process(Msg, State),

    %% If we have a connection to OtherSite then we need to de-monitor it.
    case lists:keysearch(OtherSiteId, 1, OtherSitesStatus) of

        {value, {OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                           monitor_ref           = MonitorRef} = OtherSiteStatus}} when is_reference(MonitorRef) ->

            ?INFO_LOGGER("handle_call(forget_other_site): de-monitoring OtherSiteId ~w (Status = ~w)~n", [OtherSiteId, OtherSiteStatus]),
            OtherSiteReplicationGroupPid = FromPid,    % assert: should only be told to forget a site by that site
            ok = erlang:demonitor(process, MonitorRef);

        {value, {OtherSiteId, OtherSiteStatus}} ->
            ?INFO_LOGGER("handle_call(forget_other_site): no need to de-monitor OtherSiteId ~w (Status = ~w)~n", [OtherSiteId, OtherSiteStatus])

        %% We should never get 'false' (never be asked to forget the same site twice)
    end,

    %% TODO: when BDB (repmgr) allows us to break connections to sites (and prevent retries) we need to call that API here

    {reply, 
     ok, 
     ?NEW_STATE(State#state{other_sites_status = lists:keydelete(OtherSiteId, 1, OtherSitesStatus),
                               other_sites        = [OtherSite || {SiteId, _, _, _} = OtherSite <- OtherSites, SiteId /= OtherSiteId]})};


handle_call(get_group_status, 
            _, 
            #state{this_site_status   = #site_status{status = ThisSiteStatusValue},
                   other_sites_status = OtherSitesStatus} = State) ->

    {reply, 
     {ok, [{get(this_site_id), ThisSiteStatusValue}
           | [{SiteId, StatusValue} || {SiteId, #site_status{status = StatusValue}} <- OtherSitesStatus]]},
     State}.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------

%% Another site wants to subscribe to our status updates.
%% This may be the first time we have learned of the other site (if so, we need to start monitoring it).
handle_cast({subscribe, OtherSiteId, OtherSiteReplicationGroupPid}, 
            #state{other_sites_status = OtherSitesStatus} = State) ->

    %% We may or may not already know about the caller site.
    %% And independently of that, we may or may not have 'connected' to it (asked it for it's status).
    %%
    %% If we first learn of OtherSite via this request to subscribe, then:
    %%
    %% - We record its pid/monitor ref (i.e. before find_or_spawn_peers_process finds them, 
    %%   so handle_cast(found_other_site) must be prepared for that).
    %%
    %% - We leave its status alone (as 'unknown')
    %%   TODO: the caller could include it's own status with the subscribe message, to allow us to fill 
    %%   it in here.  However, I'd like to have a single point that updates the status -- or we'd have 
    %%   to run the whole state machine here as well.
    %%
    %% -  We don't bother sending a 'found_other_site' message to ourselves, as find_or_spawn_peers_process
    %%    should do that very shortly and we don't want to do it multiple times.
    %%
    %% So the main thing we want to do here is to start monitoring the caller if we are not already 
    %% doing so.
    
    StateOut = 
        case lists:keysearch(OtherSiteId, 1, OtherSitesStatus) of

            {value, {OtherSiteId, #site_status{replication_group_pid = unknown} = OtherSiteStatus}} ->

                ?INFO_LOGGER("subscribe: Valid subscribe from entirely unknown site ~w (pid ~w)~n", [OtherSiteId, OtherSiteReplicationGroupPid]),

                %% If this doesn't get there then the other site will eventually ask us again.
                send_site_status(get(this_site_id), State, OtherSiteReplicationGroupPid),

                NewOtherSiteStatus = 
                    OtherSiteStatus#site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                                monitor_ref           = erlang:monitor(process,
                                                                                       OtherSiteReplicationGroupPid)},
                ?NEW_STATE(State#state{other_sites_status 
                                          = lists:keyreplace(OtherSiteId, 
                                                             1, 
                                                             OtherSitesStatus,
                                                             {OtherSiteId, NewOtherSiteStatus})});

            {value, {OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid} = OtherSiteStatus}} ->
                ?INFO_LOGGER("subscribe: Valid subscribe from known site ~w (OtherSiteStatus ~w)~n", [OtherSiteId, OtherSiteStatus]),
                %% We may have learned about the other site via find_or_spawn_peers.
                %% So we don't need to monitor it (we are already), but we do need to allow it to subscribe.
                %% If this doesn't get there then the other site will eventually ask us again.
                send_site_status(get(this_site_id), State, OtherSiteReplicationGroupPid),
                State;

            {value, {OtherSiteId, OtherSiteStatus}} ->
                ?INFO_LOGGER("subscribe: Discarding INCONSISTENT attempt to subscribe from site ~w for a different pid (~w) than is already registered (OtherSiteStatus ~w)~n", [OtherSiteId, OtherSiteReplicationGroupPid, OtherSiteStatus]),
                %% A different Pid is claiming to be OtherSiteId.
                %% This can legitimately happen; another site could die, be respawned and attempt to subscribe to us 
                %% before the DOWN notice for the previous one reaches us (to eradicate the old pid).
                %% So we just ignore this -- the caller will retry indefinitely and the DOWN messate is guaranteed to reach 
                %% us eventually.  See handle_info('DOWN') for more discussion.
                State;

            false ->
                %% We don't recognize OtherSiteId at all.
                %% This can happen through mis-configuration of the system; an old site was part of the group, but has been removed.
                %% The remaining members of the group are restarted to pick up the changed config, but if the old member is 
                %% not shutdown (or is later started again by accident) it will contact the others.
                ?INFO_LOGGER("subscribe: Discarding attempt to subscribe from UNRECOGNISED site ~w~n", [OtherSiteId]),
                State
        end,

    {noreply, StateOut};


%% This handles status updates from other sites.
%%
%% As find_or_spawn_peers_process subscribes to the other site on our behalf, 
%% this may be the first time we learn of the other site (if so, we need to start monitoring it).
%%
%% In certain cases we might get events out of order.  We use their last_updated timestamps (generated on 
%% the sending site -- *cannot* be compared to our own time) to order the messages and discard ones from the past.
%%
handle_cast({other_site_status, OtherSiteId, OtherSiteReplicationGroupPid, NewOtherSiteState, NewOtherSiteUpdatedAtTime}, 
            #state{current_master     = CurrentMaster,
                   other_sites_status = OtherSitesStatus} = State) ->

    StateOut = 
        case lists:keysearch(OtherSiteId, 1, OtherSitesStatus) of

            {value, {_, #site_status{replication_group_pid = unknown} = OldOtherSiteStatus}} ->

                ?INFO_LOGGER("other_site_status: initial from ~w now ~w; previously ~w (current_master ~w)~n", [OtherSiteId, NewOtherSiteState, OldOtherSiteStatus, CurrentMaster]),

                %% We need to know if the other site dies or becomes partitioned
                OtherSiteMonitorRef = erlang:monitor(process, OtherSiteReplicationGroupPid),

                %% Tell find_or_spawn_peers_process that it can stop trying OtherSiteId.
                %% (If we ever receive a DOWN message for OtherSiteMonitorRef then 
                %% we tell find_or_spawn_peers_process to start trying to find/spawn OtherSiteId again)
                send_msg_to_find_or_spawn_peers_process({stop_trying_site, OtherSiteId}, State),

                %% Now push the state change through the official state machine.
                %% We need to pre-insert the pid and monitor ref, but we let handle_other_site_status_update update the status field.

                ConnectedOtherSiteStatus = OldOtherSiteStatus#site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                                                          monitor_ref           = OtherSiteMonitorRef},

                handle_other_site_status_update(
                  OtherSiteId, 
                  ConnectedOtherSiteStatus, 
                  NewOtherSiteState, 
                  NewOtherSiteUpdatedAtTime, 
                  ?NEW_STATE(State#state{other_sites_status = lists:keyreplace(
                                                                OtherSiteId, 
                                                                1, 
                                                                OtherSitesStatus, 
                                                                {OtherSiteId, ConnectedOtherSiteStatus})}));

            {value, {_, #site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                               last_updated_time     = OldOtherSiteUpdatedAtTime} = OldOtherSiteStatus}} 
            when OldOtherSiteUpdatedAtTime < NewOtherSiteUpdatedAtTime ->

                ?INFO_LOGGER("other_site_status: from ~w now ~w; previously ~w (current_master ~w)~n", [OtherSiteId, NewOtherSiteState, OldOtherSiteStatus, CurrentMaster]),

                %% Tell find_or_spawn_peers_process that it can stop trying OtherSiteId.
                %% (This is necessary here as well as the previous clause as we might have set replication_group_pid 
                %% when the site subscribed to us).
                send_msg_to_find_or_spawn_peers_process({stop_trying_site, OtherSiteId}, State),

                handle_other_site_status_update(OtherSiteId, 
                                                OldOtherSiteStatus, 
                                                NewOtherSiteState, 
                                                NewOtherSiteUpdatedAtTime, 
                                                State);

            {value, {_, #site_status{replication_group_pid = OtherSiteReplicationGroupPid} = OldOtherSiteStatus}} ->
                ?INFO_LOGGER("other_site_status: Ignoring from ~w (~w) as it is same age or older (~w) than the one we already have (~w)~n", [OtherSiteId, NewOtherSiteState, NewOtherSiteUpdatedAtTime, OldOtherSiteStatus]),
                State;

            {value, {_, OldOtherSiteStatus}} ->
                ?INFO_LOGGER("other_site_status: Discarding INCONSISTENT from ~w (~w) as new pid (~w) does not match the pid we have recorded (~w)~n", [OtherSiteId, NewOtherSiteState, OtherSiteReplicationGroupPid, OldOtherSiteStatus]),
                %% A different Pid is already claiming to be OtherSiteId.
                %% This could legitimately happen; another site could die, be respawned and find_or_spawn_peers_process could find it 
                %% before the DOWN notice for the previous one reaches us (to eradicate the old pid).
                %% So we just ignore this -- the caller will retry indefinitely and the DOWN message is guaranteed to reach 
                %% us eventually.  See handle_info('DOWN') for more discussion.
                State;

            false ->
                %% We don't recognize OtherSiteId at all.
                %% This shouldn't happen as we reject attempts from unrecognized sites to subscribe to us, and 
                %% we should only get status updates from sites that we have subscribed to (via find_or_spawn_peers_process).
                exit({status_update_from_unrecognized_site, OtherSiteId, NewOtherSiteState})
        end,

    {noreply, StateOut};


handle_cast({stop_entire_group, FromPid}, 
            #state{other_sites_status = OtherSitesStatus,
                   support_procs      = SupportProcs} = State) ->

    ?INFO_LOGGER("stop_entire_group~n"),

    %% Coordinate a clean **temporary** shutdown of all running sites.
    %% Important: in future when we have group membership recorded in a replicated BDB, 
    %% this API will *NOT* alter that membership database.

    %% TODO:
    %% The main purpose of this API is to:
    %% - suspend writes on all sites (not just the current master).  do_txn_on_master should throw an appropriate error.
    %% - wait for all sites to sync up  (with a timeout)
    %% - then shutdown all sites (as below)

    %% Stop trying to spawn other sites
    {value, FindOrSpawnPeersSupportProc} = lists:keysearch(find_or_spawn_peers, #support_proc.name, SupportProcs),
    StateOut = ?NEW_CONSISTENT_STATE(stop_support_procs_UNSAFE([FindOrSpawnPeersSupportProc], State)),
    
    %% Tell all other sites to shutdown
    %%
    %% We DON'T simply call remove_site() on all of the sites that we know about as
    %%   - at some point remove_site will remove the site from the group membership database.
    %%   - the site that we happen to be running on might not have a connection to some site.
    %%     i.e. We need stop_entire_group to be transitive.
    %%
    %% Currently this is implemented by sending stop_entire_group to all sites that we know about 
    %% (and they in turn will send it to sites that they know about).
    %% To minimise the messages, if we received 'stop_entire_group' from a process that we recognise as a site 
    %% then we don't send the command to the site 
    %%
    %% Note: even when BDB (repmgr) allows us to break connections to sites (and prevent retries) 
    %% we don't need to call that API from this site as we are about to call terminate() and 
    %% shutdown the port_coordinator.
    %%
    %% Do this in parallel as the group might be large and we want to shutdown quickly.
    lists:foreach(fun({OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid}}) 
                     when is_pid(OtherSiteReplicationGroupPid), OtherSiteReplicationGroupPid /= FromPid -> 

                          ?INFO_LOGGER("stop_entire_group: forwarding to ~w~n", [OtherSiteId]),
                          gen_server:cast(OtherSiteReplicationGroupPid, {stop_entire_group, self()});

                     ({OtherSiteId, _}) ->
                          ?INFO_LOGGER("stop_entire_group: not connected to ~w (or command was received from that site) so don't need to tell it to stop_entire_group~n", [OtherSiteId])
                  end,
                  OtherSitesStatus),

    %% Terminate
    {stop, normal, StateOut}.


%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------

handle_info({local_sleepycat_event, Event}, State) ->
    %% This is a message from berkeley_db_replication_group_gen_event_handler 
    %% from the port_coordinator on THIS site.
    {noreply, handle_this_site_sleepycat_event(Event, State)};


handle_info(achieved_durability, #state{current_master   = {this_site, waiting_for_durability}, 
                                        this_site_status = ThisSiteStatus} = State) ->

    %% This is a message from our own probe_to_sync_replicas_process indicating that we (the master)
    %% can write durable transactions now, so we should advertize itself for use by do_txn_on_master.

    ?INFO_LOGGER("recevied achieved_durability when current_master = {this_site, waiting_for_durability}, so advertizing as master now~n"),

    State1_UNSAFE = 
        ?NEW_PARTIAL_STATE(State#state{current_master   = {this_site, available},
                                       this_site_status = ThisSiteStatus#site_status{status            = {master, available},
                                                                                     last_updated_time = now()}}),

    %% Now we tell all other sites that this site is master.
    publish_this_site_status_UNCHANGED(State1_UNSAFE),

    %% Notify any clients that are blocking in get_master_location_hint() that we discovered the master.
    %% This returns State with an empty notify_clients list.
    State2_UNSAFE = notify_clients_of_discovered_master_UNSAFE(
                      #master_location_hint{site_locator = {get(replication_group_name), get(this_site_name), node()},
                                            pid          = self()}, 
                      State1_UNSAFE),

    %% We _might_ want to shutdown the probe_to_sync_replicas_process 
    %% now, or we might need to leave it running (if there are still some connected 
    %% replicas that are not in-sync).
    State3_UNSAFE = manage_probe_to_sync_replicas_process_UNSAFE(State2_UNSAFE),

    %% The master is responsible for coordinating removal of txn logs on all sites in the group,
    %% so start a process for that.
    State4_UNSAFE = manage_remove_txn_logs_on_all_sites_process_UNSAFE(State3_UNSAFE),

    %% We also want to monitor all replicas to make sure that they don't fall too 
    %% far behind the master.  BDB-HA does not provide an event if this happens 
    %% so we have to poll REP_STAT.st_next_lsn and caclulate LSN differences ourselves.
    StateOut = ?NEW_CONSISTENT_STATE(manage_alarm_on_lagging_replicas_process_UNSAFE(State4_UNSAFE)),

    {noreply, StateOut};

handle_info(achieved_durability, #state{current_master = CurrentMaster} = State) ->
    ?INFO_LOGGER("received achieved_durability when current_master = ~w (ignoring)~n", [CurrentMaster]),
    {noreply, State};


handle_info({timeout, TimerRef, {notify_client_timer_expired, From}}, 
            #state{notify_clients = NotifyClients} = State) ->

    %% A client was only willing to wait until now to discover a master

    %% Of course, we might actually have discovered a master (and told the client about it) before this timeout happened.
    %% So we only send 'no_master' to the client if 'From' is still in the notify_clients list.
    %% (We don't have to worry about the same client process asking for notification again -- after we've previously 
    %% replied to it but before this timeout arrives -- as 'From' contains a Ref and so is unique, not just a pid.)

    {noreply,
     case ?BDB_U:lists_find_and_remove_first({From, TimerRef}, NotifyClients) of

         {found, ListAfterRemoval} ->

             ?INFO_LOGGER("notify_client_timer_expired: sending no_master to ~w~n", [From]),
             gen_server:reply(From, {error, {bdb_replication_group_error, no_master}}),    

             ?NEW_STATE(State#state{notify_clients = ListAfterRemoval});

         not_found ->
             ?INFO_LOGGER("notify_client_timer_expired: ~w is not in notify_list~n", [From]),
             State
     end};


handle_info({'DOWN', MonitorRef, process, DownPid, DownReason} = Msg, 
            #state{current_master     = CurrentMaster,
                   other_sites_status = OtherSitesStatus} = State) ->

    %% A site we were monitoring has disappeared.  
    ?INFO_LOGGER("Process ~w is DOWN : ~w~n", [DownPid, DownReason]),

    StateOut = 
        %% Find out which site
        case [{OtherSiteId, OtherSiteStatus}
              || {OtherSiteId, #site_status{replication_group_pid = Pid,
                                            monitor_ref           = Ref} = OtherSiteStatus} 
                     <- OtherSitesStatus, Pid == DownPid, Ref == MonitorRef] 
        of
            [{OtherSiteId, OtherSiteStatus}] ->

                ?INFO_LOGGER("Site ~w is DOWN (had status ~w) : ~w~n", [OtherSiteId, OtherSiteStatus, DownReason]),

                %% Tell find_or_spawn_peers_process to start attempting to find/spawn it again
                send_msg_to_find_or_spawn_peers_process({find_or_spawn_site, OtherSiteId}, State),

                case CurrentMaster of

                    {OtherSiteId, _} ->
                        ?WARNING_LOGGER("We lost the master site ~w~n", [OtherSiteId]),
                        
                        %% Just because we lost connection with the master we cannot assume that everyone else has.
                        %%
                        %% If Info is 'noconnection' (note no '_') then we have a network partition which might only be affecting this 
                        %% one node-to-node connection.
                        %%
                        %% If Info is 'noproc' then the process died between find_or_spawn finding or spawning it, and 
                        %% subscribe calling erlang:monitor() on it.
                        %%
                        %% If 'Info' is anything else then it is the ExitReason of the process.
                        %%
                        %% In the latter two cases we DO know that all sites have lost the master (at some point).
                        %% However, we DON'T act on it here as there is a race condition.  Those sites might already
                        %% have found another master, synced with it, and told us.  So we cannot legitimately 
                        %% change their status to not_in_sync.  We simply have to track their state changes as they happen -- 
                        %% we cannot make any guesses.
                        %%
                        %% However, as we _know_ that this_site_status is changing (to not_in_sync) here, we do send 
                        %% out that status update.  And all other affected sites will do the same of course.  
                        %% There is a race condition of course; the present 'DOWN' message may have arrived *after* we found and synced to 
                        %% a new master.  But we prevent that by:
                        %%
                        %% - handle_cast(found_new_site) explicitly refuses to subscribe to a new site if there is already a pid 
                        %%   registered for that site. The pid might be the same (i.e. find_or_spawn_peers just retried unecessarily 
                        %%   because of the race condition with telling it to stop; that's fine), or the pid might be a different pid
                        %%   which means that find_or_spawn_peers found or spawned the site before we got the DOWN message for the old
                        %%   site.  In that case we just ignore the found_new_site message; find_or_spawn_peers will retry indefinitely 
                        %%   and the DOWN message will eventually arrive.
                        %%
                        %% - subscribe explicitly refuses a subscribe attempt if a different pid is still registered 
                        %%   for that site.  (The caller will retry indefinitely and we will eventually get a DOWN message for the old pid).
                        %%
                        %% This is another place where we would benefit from a proper group view/epoch management system 
                        %% - to determine which messages were sent before the membership changed (master died) and which 
                        %% were sent after.
                        
                        NewOtherSiteStatus = #site_status{status                = unknown,
                                                          last_updated_time     = unknown,
                                                          replication_group_pid = unknown,
                                                          monitor_ref           = unknown},

                        StateOut1 =
                            ?NEW_STATE(State#state{current_master     = unknown,
                                                   this_site_status   = #site_status{status            = {replica, not_in_sync, unknown},
                                                                                     last_updated_time = now()},
                                                   other_sites_status = lists:keyreplace(OtherSiteId, 
                                                                                         1, 
                                                                                         OtherSitesStatus, 
                                                                                         {OtherSiteId, NewOtherSiteStatus})}),
                        %% Tell all other sites our new status (and return new state)
                        publish_this_site_status_UNCHANGED(StateOut1);

                    StillMaster ->
                        ?INFO_LOGGER("We didn't lose the master site ~w~n", [StillMaster]),

                        NewOtherSiteStatus = #site_status{status                = unknown,
                                                          last_updated_time     = unknown,
                                                          replication_group_pid = unknown,
                                                          monitor_ref           = unknown},

                        %% (note: the status of this site does not change, so there is no event to publish to other sites.)

                        State1_UNSAFE = ?NEW_PARTIAL_STATE(State#state{other_sites_status 
                                                                       = lists:keyreplace(OtherSiteId, 
                                                                                          1, 
                                                                                          OtherSitesStatus, 
                                                                                          {OtherSiteId, NewOtherSiteStatus})}),

                        %% We might be the master and we have lost the only replica that was not in-sync,
                        %% so we might need to stop the probe_to_sync_replicas_process.
                        %% Note: if we are the master then we might also have lost the ability to do durabile 
                        %% writes.  However, we only test for that ability explicitly when we first become master.
                        %% After that we rely on the perm_failed_events_limit mechanism.
                        ?NEW_CONSISTENT_STATE(manage_probe_to_sync_replicas_process_UNSAFE(State1_UNSAFE))
                end;

            _ ->
                exit({unexpected_DOWN_msg, Msg, OtherSitesStatus})
        end,

    {noreply, StateOut};


handle_info({'EXIT', FromPid, Reason}, 
            #state{support_procs = SupportProcs} = State) ->

    case get(this_site_port_coord_pid) of 

        FromPid ->
            %% Our site's port_coordinator shut down unexpectedly.
            %% We could restart it, but rather than re-invent supervisor behaviour here 
            %% we choose to shutdown and let some higher-level mechanism restart 
            %% the entire site.
            ?ERROR_LOGGER("handle_info: Received EXIT signal from this site's port_coordiantor ~w (reason ~p)~n", [FromPid, Reason]),
            %%exit({this_site_port_coordinator_exited, {FromPid, Reason}});
            {stop, {this_site_port_coordinator_exited, {FromPid, Reason}}, State};

        _ ->
            %% The only other processes we should be linked to are our support processes
            case lists:keysearch(FromPid, #support_proc.pid, SupportProcs) of 

                {value, #support_proc{name = Name} = SP} ->
                    %% If a support process dies we assume that the error must be serious, so we die too.
                    ?ERROR_LOGGER("handle_info: Received EXIT signal from support process ~w (~w) (reason ~p)~n", [Name, SP, Reason]),
                    %% Minor point: we return 'stop' instead of doing an exit() as 
                    %% gen_server will catch our exit and call terminate() anyway,
                    %% and this way we get to adjust the list of running support procs
                    %% so that terminate() doesn't fail trying to kill a process that is already dead.
                    %% We have to use NEW_PARTIAL_STATE here as removing this support proc will very likely 
                    %% violate invariants. But that's OK -- we just want to shutdown as gracefully 
                    %% as possible.  terminate() must always be written to work with whatever it finds
                    %% (i.e. it cannot assume that support procs are still running).
                    {stop, 
                     {support_process_exited, {SP, Reason}}, 
                     ?NEW_PARTIAL_STATE(State#state{support_procs = SupportProcs -- [SP]})};

                false ->
                    %% We don't recognize the source of this exit signal.
                    %% This should not happen.
                    %% If it does, it may be a bug in this module; if we intentionally 
                    %% stop a support_proc and remove it's entry from State#state.support_procs
                    %% *without* unlinking it first, and if that support-process exits with reason 
                    %% other than 'normal', we will get an EXIT message for it here 
                    %% (because we will nolonger recognize it as a support-proc).
                    exit({unknown_process_exited, {FromPid, Reason}})
            end
    end;


handle_info(Unrecognized, _State) ->
    exit({unrecognized_message, Unrecognized}).


%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------

%% Note: this function accepts a State that does not necessarily 
%% satisfy assert_state_invariants.  (e.g. an expected support-proc 
%% might have exited and been removed from support_procs).
%% Hence the name PartialState.
terminate(Reason, #state{support_procs = SupportProcs} = PartialState) ->

    ?INFO_LOGGER("terminate: (reason: ~p)~n", [Reason]),

    %% Shut down support processes.
    %% (NB we must waait for the support processes to exit, otherwise they are killed when 
    %% terminate returns (and their parent exits).

    %% As always in terminate(), the support processes might already have died
    %% and we might already have eaten the EXIT message.
    %% Fortunately erlang:monitor does what we want -- we will get a DOWN message
    %% even if the process is not running when we try to monitor it.

    ?NEW_PARTIAL_STATE(stop_support_procs_UNSAFE(SupportProcs, PartialState)), % ignore return value

    PortCoordPid = get(this_site_port_coord_pid),
    PortCoordMonitorRef = erlang:monitor(process, PortCoordPid),
    berkeley_db_port_coordinator:stop(PortCoordPid),
    receive {'DOWN', PortCoordMonitorRef, process, _, _} -> ok end,

    ?INFO_LOGGER("terminate: success~n"),
    ok.


%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, StateOut}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%%
%% Helper for handle_cast(other_site_status)
%%

handle_other_site_status_update(OtherSiteId, 
                                OldOtherSiteStatus, 
                                {replica, _, NewStatusMaster} = NewStatus, 
                                NewUpdatedAtTime, 
                                #state{current_master     = CurrentMaster,
                                       other_sites_status = OtherSitesStatus} = State) ->

    NewOtherSiteStatus = OldOtherSiteStatus#site_status{status            = NewStatus,
                                                        last_updated_time = NewUpdatedAtTime},

    %% Update OtherSite's status.
    %% (A message from a replica *never* affects the receiver's notion of current_master)
    State1_UNSAFE = 
        ?NEW_PARTIAL_STATE(State#state{other_sites_status = lists:keyreplace(OtherSiteId, 
                                                                             1, 
                                                                             OtherSitesStatus, 
                                                                             {OtherSiteId, NewOtherSiteStatus})}),

    %% If the other site claims that it is treating this site as its master in some fashion,
    %% then we can assert that we really are master.  
    %% Obviously the replica's message may be delayed, but there are still constrains, as 
    %% live downgrade is not currently supported (i.e. the steps towards becomimg master are a ratchet).

    ThisSiteId = get(this_site_id),
    case {NewStatusMaster, CurrentMaster} of

        {{ThisSiteId, available}, {this_site, available}} -> 
            %% replica thinks we are an advertizing master, and we agree
            ok;  
        {{ThisSiteId, available}, _} -> 
            %% replica thinks we are an advertizing master, and we disagree
            exit({replica_thinks_we_are_master, CurrentMaster});

        {{ThisSiteId, waiting_for_durability}, {this_site, available}} -> 
            %% replica thinks we are becoming master (transitioning_to or waiting_for_durability)
            %% and we agree (replica is a little out of date, but still valid)
            ok;
        {{ThisSiteId, waiting_for_durability}, {this_site, waiting_for_durability}} -> 
            %% replica thinks we are becoming master and we agree 
            ok;  

        {{ThisSiteId, transitioning_to}, _} -> 
            %% replica says it got NEWMASTER naming us, and even if this is the first we've heard 
            %% about us becoming master, we cannot disagree (other sites often get NEWMASTER 
            %% before the new master gets MASTER)
            ok;  

        {{ThisSiteId, _}, _} -> 
            %% replica thinks we are becoming master (transitioning_to or waiting_for_durability)
            %% and we disagree
            exit({other_site_thinks_we_are_master, OtherSiteId, NewStatusMaster, CurrentMaster});

        _ ->
            %% other site either doesn't know who is master or thinks that some other site is master 
            %% (we don't have to agree with it)
            ok   
    end,            

    %% In BDB v4.5 the STARTUPDONE (replica-is-in-sync) event is only sent on receipt of new 
    %% write transactions by a replica.  Until that is fixed, we need support process to 
    %% stimulate STARTUPDONE on the replicas iff we are master and there are any not_in_sync replicas.
    ?NEW_CONSISTENT_STATE(manage_probe_to_sync_replicas_process_UNSAFE(State1_UNSAFE));

handle_other_site_status_update(OtherSiteId, 
                                OldOtherSiteStatus, 
                                {master, waiting_for_durability} = NewStatus, 
                                NewUpdatedAtTime, 
                                #state{current_master     = CurrentMaster} = State) ->

    %% Another site is telling us that it has received DB_EVENT_REP_MASTER, 
    %% but has NOT yet confirmed that it can accept durable write transactions.
    %%
    %% So we _cannot_ yet advertize the other size as master (for use with do_txn_on_master).
    %%
    %% Note that we _might_ previously have seen DB_EVENT_REP_NEWMASTER (for this new master).
    %% And (if so) we _might_ also have got STARTUPDONE too - in that case we need to 
    %% preserve in_sync'nes when we adjust this_site_status
    %% However, if this site believes it is in-sync with a *different* master then 
    %% we need to reset this_site_status to not_in_sync, pending the arrival of the STARTUPDONE event.

    %% Note that if more than one site claims to be master at roughly the same time, 
    %% we currently do 'last message wins'.  We have no way to detect order-inversion of messages
    %% from different sites (only for two messages the same site).
    %% Of course NEWMASTER events give an authoratative decision, but presumably there 
    %% will be more than one of those (one for each new master), and one or more of them 
    %% might not have arrived yet (the status-updates might arrive before or after NEWMASTER).
    %% So we would need to introduce a time-bound window in which to make a decision 
    %% (and that still might be wrong of course).
    %%
    %% The ideal solution would be for BDB to expose it's internal 'egen' generation-id 
    %% as part of NEWMASTER and MASTER event payloads.  Then we would have both a total-ordering 
    %% and a way to pair local events with status-updates from other sites.  
    %% (egen is a durable generation id, so this would fix the problem even if sites crash and 
    %% restart rapidly.)
    %%
    %% In the absence of that, we rely on the fact that masters don't change (more than once) 
    %% in rapid succession.

    case CurrentMaster of 

        {this_site, _} ->
            %% we know that we are master (as live downgrades are not supported)
            exit({illegal_other_site_status_update, OtherSiteId, NewStatus, CurrentMaster});

        {OtherSiteId, waiting_for_durability} ->
            %% duplicate event
            exit({illegal_other_site_status_update, OtherSiteId, NewStatus, CurrentMaster});

       {OtherSiteId, available} ->
            %% OtherSite is already advertizing as master 
            exit({illegal_other_site_status_update, OtherSiteId, NewStatus, CurrentMaster});
 
        _ -> 
            %%    'unknown' 
            %% or (OtherSiteId, transitioning_to}, 
            %% or {SomeOtherMaster, _}
            %% (the latter can happen if the old master dies (or is partitioned away) and a new master is elected 
            %% before we receive the DOWN message from the old master.)
            handle_other_site_status_update_master(OtherSiteId, OldOtherSiteStatus, NewStatus, NewUpdatedAtTime, State)
    end;

handle_other_site_status_update({OtherSiteName, OtherSiteNode} = OtherSiteId, 
                                #site_status{replication_group_pid = OtherSiteReplicationGroupPid} = OldOtherSiteStatus, 
                                {master, available} = NewStatus, 
                                NewUpdatedAtTime, 
                                #state{current_master = CurrentMaster} = State) ->

    %% Another site is telling us that it has received DB_EVENT_REP_MASTER, 
    %% and can now accept durable write transactions.

    case CurrentMaster of 

        {this_site, _} ->
            %% we know that we are master (as live downgrades are not supported)
            exit({illegal_other_site_status_update, OtherSiteId, NewStatus, CurrentMaster});

        {OtherSiteId, available} ->
            %% duplicate event
            exit({illegal_other_site_status_update, OtherSiteId, NewStatus, CurrentMaster});

        _ -> 
            %%    'unknown' 
            %% or {OtherSiteId, transitioning_to}       [this _is_ legal; we might get NEWMASTER for OtherSiteId before we subscribe, so we could miss seeing OtherSiteId go through waiting_for_durability.
            %% or (OtherSiteId, waiting_for_durability}, 
            %% or {SomeOtherMaster, _}
            %% (the latter can happen if the old master dies (or is partitioned away) and a new master is elected 
            %% before we receive the DOWN message from the old master.)

            %% We need to tell any clients that are blocking in get_master_location_hint() 
            %% that we have found a new master.
            State1_UNSAFE = 
                        notify_clients_of_discovered_master_UNSAFE(
                          #master_location_hint{site_locator = {get(replication_group_name), OtherSiteName, OtherSiteNode},
                                                pid          = OtherSiteReplicationGroupPid},
                          State),

            handle_other_site_status_update_master(OtherSiteId, OldOtherSiteStatus, NewStatus, NewUpdatedAtTime, State1_UNSAFE)
    end.


handle_other_site_status_update_master(OtherSiteId, 
                                       OldOtherSiteStatus,
                                       {master, NewMasterSubState} = NewStatus, 
                                       NewUpdatedAtTime, 
                                       #state{this_site_status   = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus,
                                              other_sites_status = OtherSitesStatus} = State1_UNSAFE) ->

    %% TODO should we check (as we do above) for:
    %%   - current_master = {this_site, _}
    %%   - duplicate events from OtherSiteId

    %% Change other site's status
    NewOtherSiteStatus = OldOtherSiteStatus#site_status{status            = NewStatus,
                                                        last_updated_time = NewUpdatedAtTime},

    NewMasterAndSubState = {OtherSiteId, NewMasterSubState},

    %% Change this site's status, preserving the current in_sync/not_in_sync (STARTUPDONE)
    NewThisSiteStatus = 
        case ThisSiteStatusValue of 
            {replica, ThisSiteSyncStatus, _} -> 
                ThisSiteStatus#site_status{status            = {replica, ThisSiteSyncStatus, NewMasterAndSubState},
                                           last_updated_time = now()}
        end,

    StateOut = 
        ?NEW_CONSISTENT_STATE(State1_UNSAFE#state{current_master     = NewMasterAndSubState,
                                                  this_site_status   = NewThisSiteStatus,
                                                  other_sites_status = lists:keyreplace(
                                                                         OtherSiteId, 
                                                                         1, 
                                                                         OtherSitesStatus,
                                                                         {OtherSiteId, NewOtherSiteStatus})}),
    publish_this_site_status_UNCHANGED(StateOut).


%%
%% Helper for handle_info(sleepycat_event)
%%

handle_this_site_sleepycat_event({_PortCoordName, bdb_DB_EVENT_REP_PERM_FAILED}, 
                                 #state{current_master    = CurrentMaster,
                                        perm_failed_times = PermFailedTimes} = State) ->

    %% skip the assert for this event; it really has no chance of firing and we might get a lot of these
    %%PortCoordName = get(this_site_port_coord_name),  % assert

    %% This site is master, and we failed to get enough acks for some transaction, so that transaction may not be durable under replication.
    %% Important: by the time we get this event it is **too late** to prevent potential inconsistency.
    %% The txn_commit/abort may have already returned to the caller (without error), 
    %% all write locks for the txn will have been released, and the potentially-not-durable data 
    %% may have been read or modified by any number of other transactions.  
    %% If this site fails then it is possible that a new master will be elected 
    %% which does not have the transaction in question.  That new master will then diverge from this ex-master. 
    %% When this site restarts and syncs with a master, the committed transaction will be rolled back on this site
    %% during sync-up.
    %% To protect against exposing data that may be rolled back we have 2 options
    %%
    %% 1. Handle PERM_FAILED in the event_notify callback (in the driver layer, before txn_commit/abort returns)
    %%    By 'handle' we could either
    %%    a. ard-abort the OS process, thereby avoiding returning from txn_commit and preventing reads to the data on this site.
    %%       A new master will be elected, which may or may not have the txn.
    %%       To avoid exposing the data, reads must not be allowed on the restarted site until it has found a master 
    %%       and become in_sync.  
    %%       NOTE: if reads are allowed on ANY replicas we may still expose data which is later rolled back.
    %%       Therefore, for perfect consistency, all reads must be done on the master, and only on data for 
    %%       which the master did NOT get PERM_FAILED.
    %%
    %%    b. We could use thread-local storage to pass back a flag from the event-notify callback to our txn_commit/abort 
    %%       wrappers (the C code) to say that this txn is not durable -- and the wrapper could then throw.
    %%       However, write locks will still be dropped and the data exposed.  This does not avoid the problem 
    %%       at all, merely tells the client that it is happening (and which data is potentially involved).
    %%
    %% Note that PERM_FAILED may well mean that the master has been partitioned into a minority and the remaining replicas 
    %% elect another master, which may be visible to service clients.   Therefore the current master really should 
    %% downgrade and call an election.
    %%
    %% 2.  Wait for Sleepycat/Oracle to implement master-leases in v4.6, which are designed to solve this out of the box 
    %%     but with much less ovhead; a master that fails to write durable txns is quarantined until it discovers 
    %%     the new master and syncs with it.  The main difference is that the OS process runnint the site does not need to be 
    %%     hard aborted.
    %%
    %% We choose to wait for master-leases, particularly as I have seen sporadic PERM_FAILED occasionally (e.g.
    %% replicated checkpoints can be sufficiently intrusive to block the required number of acks for a while 
    %% although recent patches to v4.5.20 have greatly improved that case).
    %%
    %% Util we have master-leases we do want to guard against *gross* failure of replication -- e.g. replicas dying over time, or the 
    %% master being partitioned in a minority.  We don't want to continue writing non-durable txns for minutes or hours.
    %% 
    %% So we currently implement a MaxCount in MaxTime tolerance window, similar to that implemented OTP supervisors for 
    %% crashing processes.
    %% We receive more than MaxCount PERM_FAILED events in any contiguous period MaxTimeMilliseonds then we tell 
    %% the master to shutdown (gracefully -- there's no pressing need to hard-kill it), and allow its presumed 
    %% supervisor to restart it.  Such restarts are unlikely to happen in rapid succession to the same site 
    %% as another site will presumably be elected master.

    case get(perm_failed_events_limits) of

        disabled -> 
            State;

        {WaitingForDurability, Available} ->

            {MaxCount, WindowMs} = case CurrentMaster of
                                       {this_site, available} -> 
                                           Available;
                                       _ -> 
                                           %% anything else is taken as the same as {this_site, waiting_for_durability} 
                                           %% - e.g. BDB-HA might send PERM_FAILED events before we get (or process) MASTER
                                           %% so current_master might be unknown or {OtherSiteId, _}
                                           WaitingForDurability
                                   end,

            Now = erlang:now(),
            NewPermFailedTimes = add_perm_failed_event([Now|PermFailedTimes], Now, WindowMs),

            case length(NewPermFailedTimes) of
                Count when Count =< MaxCount ->
                    State#state{perm_failed_times = NewPermFailedTimes};
                _ ->
                    ?ERROR_LOGGER("Exiting as received more than ~w PERM_FAILED events in ~w milliseconds: ~w~n", [MaxCount, WindowMs, NewPermFailedTimes]),
                    exit({perm_failed_events_rate_too_high, NewPermFailedTimes})
            end
    end;

handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_CLIENT}, 
                                 #state{current_master = CurrentMaster} = State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    ?INFO_LOGGER("sleepycat event: Local site is a replica, and might not be in-sync yet~n"),

    StateOut = 
        case CurrentMaster of 
            {this_site, _} ->
                exit(attempt_to_downgrade_from_master); % live downgrades are not supported
            _ ->
                %% We have no reason to believe that the master has changed.
                ?NEW_STATE(State#state{this_site_status = #site_status{status            = {replica, not_in_sync, CurrentMaster},
                                                                       last_updated_time = now()}})
        end,
    publish_this_site_status_UNCHANGED(StateOut);

handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_MASTER},
                                 #state{current_master     = CurrentMaster} = State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    ?INFO_LOGGER("sleepycat event: This site is now the master (OLD master was ~w)~n", [CurrentMaster]),

    case CurrentMaster of 

        {this_site, _} ->
            ?ERROR_LOGGER("sleepycat event: This site got DB_EVENT_REP_MASTER but is already master~n"),
            exit({bad_MASTER_event, State});

        _PreviousMaster -> % 'unknown' or {ExMasterSiteId, _}

            %% We have been elected as new master.
            %%
            %% We may still think we are connected to an existing master (seems unlikely, but does happen in tests).
            %% This does NOT necessarily mean more than one master -- e.g. we might be elected as the new master 
            %% before we get the DOWN message for the ex-master (this does happen in tests).
            %%
            %% So if PreviousMaster is not 'unknown', other_sites_status will still contain
            %%    - the ex-master site, still with status 'master'
            %%    - possibly some other sites claiming to be not_in_sync or in_sync with the ex-master site
            %%
            %% However, we only change 
            %%     this_site_status       to {master,    waiting_for_durability} 
            %% and current current_master to {this_site, waiting_for_durability}, 
            %%
            %% We DON"T change the ex-masters' status, or any other site that claims to be using the ex-master as its master.
            %% We are attempting to track the internal state of BDB-HA, by responding to events that it generates 
            %% on each site.  We rely on Erlang's messaging guarantees to ensure that we receive all status updates 
            %% (in order from each site, but without any other ordering), or 'DOWN' messages.
            %% There is no such thing as a consistent picture of the global state. What is important is that we converge 
            %% on the same state that BDB-HA has internally.
            %% (If BDB-HA's state is screwed up internally, we do need to expose/reflect that at this level, so that 
            %% we can understand the problem).  If we start making local assumptions about what 'should' have 
            %% happened to other sites, then our local picture may drift away from BDB-HA's internal state,
            %% and we will have true confusion.

            State1_UNSAFE =
                ?NEW_PARTIAL_STATE(State#state{current_master     = {this_site, waiting_for_durability},
                                               this_site_status   = #site_status{status            = {master, waiting_for_durability},
                                                                                 last_updated_time = now()}}),

            %% Before we can advertize ourselves as available for do_txn_on_master we need to start writing 
            %% dummy transactions to see if we can do durable writes.
            StateOut = ?NEW_CONSISTENT_STATE(manage_probe_to_sync_replicas_process_UNSAFE(State1_UNSAFE)),

            %% Tell all other sites that we are {master, waiting_for_durability}
            publish_this_site_status_UNCHANGED(StateOut)
    end;

handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_NEWMASTER, MasterHostPort}, 
                                 #state{current_master = CurrentMaster} = State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert

    %% We use our own other_site_status_update messages to discover which site is master (and that 
    %% it is ready to write durable transactions).  
    %% So we don't really need BDB-HA's NEWMASTER events.
    %% We only use them as a safety/diagnostic device -- to confirm that our other_site_status_update messages 
    %% are correct (i.e. that the the SiteId of the master is the smae).
    %% We DON'T change current_master to NewMasterSiteId, as transitioning to that state means that 
    %% we are advertising that site as being able to accept durable write txns (i.e. we would have to call 
    %% notify_clients_of_discovered_master), and that is not true until receive other_site_status_update 
    %% with 'master' status (not {master, waiting_for_durability}).

    PeerHostNamesPortsToSiteIds = get(peer_hostnameports_to_site_ids),
    NewMasterSiteId = case lists:keysearch(MasterHostPort, 1, PeerHostNamesPortsToSiteIds) of 
                          {value, {_, SiteId}} -> SiteId;
                          _                    -> exit({unrecognized_site_with_NEWMASTER_event, MasterHostPort, PeerHostNamesPortsToSiteIds})
                      end,

    ?INFO_LOGGER("sleepycat event: got {DB_EVENT_REP_NEWMASTER, ~w}. current master is ~w~n", [NewMasterSiteId, CurrentMaster]),

    case CurrentMaster of 

        {this_site, _} ->
            %% This would appear to mean that another master exists
            exit(this_site_got_NEWMASTER_while_it_is_master);

        {NewMasterSiteId, transitioning_to} ->
            %% current_master can only have this value if we've already seen this same event recently, which should not happen
            exit({this_site_got_double_NEWMASTER_event, NewMasterSiteId});

        {NewMasterSiteId, waiting_for_durability} ->
            %% This can happen if we receive {other_site_status_update, {master, waiting_for_durability}} from the new master 
            %% before BDB-HA delivered this local NEWMASTER event.
            %% Nothing to do.
            State;

        {NewMasterSiteId, available} ->
            %% This can happen if we receive {other_site_status_update, {master, available}} from the new master 
            %% before BDB-HA delivered this local NEWMASTER event.
            %% Nothing to do.
            State;

        _ -> % 'unknown' or {ExMasterSiteId, _}

            %% Note: as we got this event (and NewMasterSiteId) from BDB-HA and not our own code,
            %% other_sites_status for OtherSiteId might be anything at this point 
            %% -- we might not even be connected to that site yet.
            %% So we can't adjust other_site_status for that site to {master, transitioning_to}
            %% even if we wanted to.   (And if we did there would be the problem that we 
            %% cannot change the last_updated_time, as we don't know the other site's clock.)

            NewMasterStatus = {NewMasterSiteId, transitioning_to},
            StateOut = 
                ?NEW_STATE(State#state{current_master     = NewMasterStatus,
                                       this_site_status   = #site_status{status            = {replica, not_in_sync, NewMasterStatus},
                                                                         last_updated_time = now()}}),
            publish_this_site_status_UNCHANGED(StateOut)
    end;

handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_STARTUPDONE}, 
                                 #state{current_master   = CurrentMaster, % might be 'unknown'; see comments below
                                        this_site_status = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus} = State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    ?INFO_LOGGER("sleepycat event: This site (state was ~w) but is is now in-sync with the master (master = ~w)~n", [ThisSiteStatus, CurrentMaster]),

    %% See comments in NEWMASTER clause, above
    %% We must change this-site's status from not_in_sync to in_sync (even if master is 'unknown').
    %% If this site's status is already in_sync then we exit, as it appears that STARTUPDONE 
    %% has been delivered before (the next) NEWMASTER. (That is fixed in BDB v4.6)

    case ThisSiteStatusValue of 

        {replica, not_in_sync, Master} ->

            %% This is the only legal set of states in which we should get STARTUPDONE
            %% -- any other state will exit with badmatch
            StateOut =
                ?NEW_STATE(State#state{this_site_status = #site_status{status            = {replica, in_sync, Master},
                                                                       last_updated_time = now()}}),
            publish_this_site_status_UNCHANGED(StateOut)
    end;

handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_WRITE_FAILED}, _State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert

    ?ERROR_LOGGER("sleepycat event: This site's BDB failed to write to disk~n"),
    exit(bdb_DB_EVENT_WRITE_FAILED).


%% Helpers for handle_sleepycat_event

add_perm_failed_event([Timestamp | T], Now, WindowMs) ->
    case within_time_window(Now, Timestamp, WindowMs) of
	true ->
	    [Timestamp | add_perm_failed_event(T, Now, WindowMs)];
	_ ->
	    []
    end;
add_perm_failed_event([], _, _) ->
    [].

within_time_window(Timestamp, Now, WindowMs) ->
    case ?BDB_U:now_diff_milliseconds(Timestamp, Now) of
	T when T > WindowMs -> false;
	_                   -> true
    end.


%% This is called for its side-effects.
%% It returns State unchanged, purely as a convenience to the caller
%% so it has a name that reflects that semantic.
publish_this_site_status_UNCHANGED(#state{this_site_status   = #site_status{status = ThisSiteStatusValue},
                                          other_sites_status = OtherSitesStatus} = State) ->

    ThisSiteId = get(this_site_id),
    lists:foreach(fun({OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid}}) 
                     when is_pid(OtherSiteReplicationGroupPid) ->

                          ?INFO_LOGGER("publish_this_site_status: ~w sending to ~w (~w)~n", [ThisSiteStatusValue, OtherSiteId, OtherSiteReplicationGroupPid]),
                          send_site_status(ThisSiteId, State, OtherSiteReplicationGroupPid);

                     ({OtherSiteId, OtherSiteStatus}) ->
                          ?INFO_LOGGER("publish_this_site_status: ~w NOT sending to ~w as replication_group_pid is unknown ~w~n", [ThisSiteStatusValue, OtherSiteId, OtherSiteStatus])
                  end,
                  OtherSitesStatus),
    State.

send_site_status(ThisSiteId, 
                 #state{this_site_status = #site_status{status            = ThisSiteStatusValue,
                                                        last_updated_time = ThisSiteStatusLastUpdatedTime}},
                 DestinationReplicationGroupPid) ->

    gen_server:cast(DestinationReplicationGroupPid, 
                    {other_site_status, ThisSiteId, self(), ThisSiteStatusValue, ThisSiteStatusLastUpdatedTime}).


notify_clients_of_discovered_master_UNSAFE(_MasterLocationHint, #state{notify_clients = []} = State) ->
    %% minor optimization; no clients to notify, so no need to modify state
    State;

notify_clients_of_discovered_master_UNSAFE(MasterLocationHint, #state{notify_clients = NotifyClients} = State) ->
    notify_clients_of_discovered_master2_UNSAFE(NotifyClients, MasterLocationHint, State).


notify_clients_of_discovered_master2_UNSAFE([], _MasterLocationHint, State) ->
    %% Finished notifying clients
    ?INFO_LOGGER("notify_clients_of_discovered_master: done all clients~n"),
    ?NEW_PARTIAL_STATE(State#state{notify_clients = []});

notify_clients_of_discovered_master2_UNSAFE([{From, TimerRef} | T], MasterLocationHint, State) ->
    ?INFO_LOGGER("notify_clients_of_discovered_master: cancelling timer ~w and sending notification to ~w~n", [TimerRef, From]),
    cancel_timer(TimerRef),
    gen_server:reply(From, {ok, MasterLocationHint}),
    notify_clients_of_discovered_master2_UNSAFE(T, MasterLocationHint, State).

cancel_timer(infinity) -> ok;
cancel_timer(TimerRef) -> erlang:cancel_timer(TimerRef).



%%
%% Helpers for handle_call(do_txn_on_...)
%%

forward_command_to_remote_site({SiteName, SiteNode}, 
                               PortCoordPid, 
                               {_CommandName, _TxnFun, _NumDeadlockRetries, Timeout} = Command, 
                               From) ->

    %% TODO: possibly significant optimization (avoid 1 network hop for reply):
    %%
    %% If we passed 'From' to the destination site then the destination could reply directly to 
    %% the original client (rather than indirectly via the intermediate process we spawn here)
    %% as From contains the original client Pid.  
    %% i.e. We ouldn't need the call to gen_server:reply() here.  
    %% In fact we wouldn't even need to spawn a process here -- that's only necessary to avoid avoid 
    %% blocking this replication_group process while we send the command and wait for the response,
    %% and in the new scheme we would not be waiting for the response. 
    %% Of course, gen_server:call would need to become gen_server:cast() in that scheme.
    %%
    %% The above assumes that gen_server replies can be sent from a different process to that to which 
    %% the request was sent.  I have not tested that.
    %%
    %% One possibly significant difference: the original client (still in it's call to gen_server:call)
    %% will be monitoring the original destination process.  So if the process to which we are forwarding 
    %% crashes, the client won't get a reply at all, and will incur it's full gen_server:call timeout.  
    %% In the current scheme, the process we spawn here does another gen_server:call, and so probably 
    %% finds out much sooner (and tell the original client) if there is a problem with the destination.
    %% I don't currently know whether this is sufficiently important to warrant the overhead of an 
    %% extra network hop for the reply.
    %%
    %% Important: Note that an indirect hop (use of forward_command_to_remote_site) should be 
    %% quite rare (for do_txn_on_master at least) as the reply includes a new MasterLocationHint
    %% that tells the client where to go directly next time.  Current do_txn_on_any_site does 
    %% not have any such mechanism though (e.g. if the current site is not in sync, it will 
    %% a always silently forward to another in-sync site if there is one).
    
    %% I am undecided whether to call plain spawn() or proc_lib:spawn() here -- proc_lib is certainly 
    %% slower, and do we really need the error reports for this worker (which is catch-all anyway,
    %% unless it is given malformed arguments).
    %%
    %% It's a _little_ bit clearer that we should call spawn() rather than spawn_link() here.
    %% If we did the latter then we would have to keep the pid (associated with From presumably)
    %% and handle any EXIT messages etc.  That's a _lot_ of churning state management 
    %% for very little real benefit.
    %% If the spawned process dies then the client will eventually get a timeout, and the original 
    %% replication_group server (to which the client sent the request) won't care anyway.
    %% But that should never happen, as we carefully catch all possible errors from attempting to 
    %% forward the command.
    %% If the replication_server dies then the spawned process is not killed
    %% -- but unless Timeout is infinity is is guaranteed to return at some point.

    ReplicationGroupName = get(replication_group_name),
    spawn(
      fun() -> 
              %%?INFO_LOGGER("forward_command_to_remote_site (for ~w): this worker is forwarding to site ~w, ~w~n", [_CommandName, SiteId, PortCoordPid]),

              %% We use the same timeout (assume we have used zero time so far).
              %% We need to emulate THROW_TXN_REMOTE_SITE_ERRORS 
              %% to do the same conversion of noproc etc to 'site_is_down' 
              %% (with the *correct* SiteLocator, i.e. that of the site that 
              %% we are *forwarding* too, *not* that of the site doing the forwarding).
              %% But we can't (easily) use THROW_TXN_REMOTE_SITE_ERRORS itself, 
              %% as it removes the ok/error tag, and throws; we want to keep the 
              %% tag and return the term, so that the rethrow happens on the client.

              Reply =
                  try gen_server:call(PortCoordPid, Command, Timeout) of

                      {ok, _} = Result -> Result;

                      {error, {bdb_helper_error, {acquire_bdb_port_handle, {{exit, {noproc,{gen_server,call,_}}}, _}}} = Reason} ->
                          {error, {bdb_replication_group_error, {site_is_down, {ReplicationGroupName, SiteName, SiteNode}, Reason}}};

                      {error, {bdb_helper_error, {release_bdb_port_handle, {{exit, {bdb_receive_reply_timeout, _}}}, _}} = Reason} ->
                          {error, {bdb_replication_group_error, {site_is_down, {ReplicationGroupName, SiteName, SiteNode}, Reason}}};

                      {error, {bdb_helper_error, {release_bdb_port_handle, {port_is_unreachable, _}}} = Reason} ->
                          {error, {bdb_replication_group_error, {site_is_down, {ReplicationGroupName, SiteName, SiteNode}, Reason}}};

                      {error, _} = Result -> Result
                  catch
                      exit:Reason ->
                          {error, {bdb_replication_group_error, {site_is_down, {ReplicationGroupName, SiteName, SiteNode}, Reason}}}
                  end,

              gen_server:reply(From, Reply)

              %%?INFO_LOGGER("forward_command_to_remote_site: worker ~w finished~n", [self()])
      end).


async_do_txn_on_master_local_site({_CommandName, TxnFun, NumDeadlockRetries, _Timeout}, 
                                  From) ->

    %% We don't need to send TxnFun to the port_coordinator, we can execute it ourselves and avoid the extra hop.
    %% The earlier comments about proc_lib vs plain, and spawn vs spawn_link apply here too.
    PortCoordPid = get(this_site_port_coord_pid),

    %% TODO: to avoid generating (small) garbage on every txn, only create a new MasterLocationHint 
    %% if the client didn't specify one, or if the new one would be different 
    %% (most of the time neither of those will be true).
    ReplicationGroupName = get(replication_group_name),
    SiteName             = get(this_site_name),
    MasterLocationHint   = #master_location_hint{site_locator = {ReplicationGroupName, SiteName, node()},
                                                 pid          = self()},

    %% See comments in forward_command_to_remote_site about use of spawn vs proc_lib:spawn or spawn_link.
    spawn(
      fun() -> 
              %%?INFO_LOGGER("async_do_txn_on_master_local_site: worker starting~n"),

              %% Pass out-of-band parameters to TxnFun.
              %% See comments in berkeley_db_replication_group_helpers.erl
              %% for txn_fun_gtm_*.
              %% A TxnFun that is used to call a weakly-replicated application needs to 
              %% know the name of the group and site that it is running on, in order to 
              %% construct the process name of the application.
              %% We currently don't pass this as a real parameter to the TxnFun, as 
              %% most TxnFun's don't need to know that and we don't want a different interface 
              %% (and version of do_txn_no_throw) for replication-aware code.
              %% It is safe to write to this process' dictionary as we just spawned it.
              %% If we ever adopt the (possible) optimization of executing the TxnFun 
              %% in the process that originally called do_txn_on_master/any_site then 
              %% this would be unacceptable.
              put(txn_fun_replication_context, {ReplicationGroupName, SiteName}),

              gen_server:reply(
                From, 
                try 
                    ?BDB_H:do_txn_no_throw(PortCoordPid, TxnFun, NumDeadlockRetries) 
                of 
                    {ok, Result}            -> {ok, {Result, MasterLocationHint}};
                    {error, _} = ErrorReply -> ErrorReply
                catch
                    %% do_txn_no_throw tries hard to be completely 'no throw'.
                    %% But just incase there is a bug somewhere, we catch-all here
                    %% otherwise the spawned worker will silently exit and the client 
                    %% will be left waiting for a full gen_server timeout 
                    %% and it won't be clear why.
                    Class:Reason -> 
                        {error, {bdb_helper_error, {internal_error, {Class, Reason, erlang:get_stacktrace()}}}}
                end)
              %%?INFO_LOGGER("async_do_txn_on_master_local_site: worker finished~n")
      end).


async_do_txn_on_local_site({_CommandName, TxnFun, NumDeadlockRetries, _Timeout}, 
                           From) ->
    %% Execute TxnFun on local site 
    %% See comments in do_txn_on_master -- the only difference here is that we don't return a new MasterLocationHint.

    ReplicationGroupName = get(replication_group_name),
    SiteName             = get(this_site_name),
    PortCoord            = get(this_site_port_coord_pid),

    %% See comments in forward_command_to_remote_site about use of spawn vs proc_lib:spawn or spawn_link.
    spawn(
      fun() -> 
              %%?INFO_LOGGER("async_do_txn_on_local_site (~w): worker starting~n", [_CommandName]),

              %% See comments for this same line in async_do_txn_on_master_local_site.
              put(txn_fun_replication_context, {ReplicationGroupName, SiteName}),

              gen_server:reply(
                From, 
                try 
                    ?BDB_H:do_txn_no_throw(PortCoord, TxnFun, NumDeadlockRetries)
                catch
                    %% see comments in async_do_txn_on_master_local_site
                    Class:Reason -> 
                        {error, {bdb_helper_error, {internal_error, {Class, Reason, erlang:get_stacktrace()}}}}
                end)
              %%?INFO_LOGGER("async_do_txn_on_local_site: worker finished~n")
      end).


%%
%% Support processes
%%

%% Returns modified State
stop_support_procs_UNSAFE(StopSupportProcs, #state{support_procs = SupportProcs} = State) ->
    lists:foreach(fun(#support_proc{name = Name, pid = Pid, stop_fun = StopFun} = StopSupportProc) ->
                          ?INFO_LOGGER("stopping support proc ~w (pid ~w)~n", [Name, Pid]),

                          MonitorRef = erlang:monitor(process, Pid),                                                  
                          unlink(Pid), % Important: we nolonger want the exit signal (handle_info will terminate if it receives it)

                          StopFun(StopSupportProc),
                          receive
                              {'DOWN', MonitorRef, process, _, Reason} -> 
                                  ?WARNING_LOGGER("support proc ~w exited with: ~w~n", [Name, Reason])
                          end
                  end,
                  StopSupportProcs),
    ?NEW_PARTIAL_STATE(State#state{support_procs = SupportProcs -- StopSupportProcs}).


%%
%% find_or_spawn_peers_process
%%

find_or_spawn_peers_process(ReplicationGroupName, 
                            ThisSiteId, 
                            ThisSiteReplicationGroupPid, 
                            OtherSitesIds,
                            FindOtherSitesPeriodMs,
                            RpcCallTimeoutMs,
                            SpawnOtherSitesInfo) ->

    ?INFO_LOGGER("find_or_spawn_peers_process starting~n"),

    put(find_other_sites_period_milliseconds, FindOtherSitesPeriodMs),
    put(rpc_call_timeout_milliseconds, RpcCallTimeoutMs),

    find_or_spawn_peers_loop(ReplicationGroupName, 
                             ThisSiteId,
                             ThisSiteReplicationGroupPid, 
                             SpawnOtherSitesInfo,
                             [{SiteId, 0} || SiteId <- OtherSitesIds],  % ThisPassSiteIds : [{SiteId, NumAttempts}, ...]
                             [],                                        % NextPassSiteIds
                             0).                                        % ReceiveTimeout -- start first pass immediately

%% The desired behaviour is:
%%
%% Every N milliseconds attempt to find or spawn all of the FindSiteIds
%% Don't pause between each one (or large groups will take a long time to spawn).
%% But do check (between each one) for requests (messages) to start or stop trying other sites,
%% and handle those requests immediately.
%% Eventually we will want all nodes to be out of phase (or use different retry periods),
%% so that the attempts to spawn other sites are well distributed in time (to minimize recovery latency).

find_or_spawn_peers_loop(ReplicationGroupName, 
                         ThisSiteId,
                         ThisSiteReplicationGroupPid, 
                         SpawnOtherSitesInfo,
                         [], 
                         [],
                         ReceiveTimeout) when ReceiveTimeout /= infinity ->

    %% no work to do, so wait forever (for a message)
    %%?INFO_LOGGER("find_or_spawn_peers: transitioned to having no work to do: setting timeout to infinity~n"),
    find_or_spawn_peers_loop(ReplicationGroupName,
                             ThisSiteId,
                             ThisSiteReplicationGroupPid, 
                             SpawnOtherSitesInfo,
                             [],
                             [],
                             infinity);


find_or_spawn_peers_loop(ReplicationGroupName, 
                         ThisSiteId,
                         ThisSiteReplicationGroupPid, 
                         SpawnOtherSitesInfo,
                         [], 
                         NextPassSiteIds,
                         _) when NextPassSiteIds /= [] ->

    %% All ThisPassSiteIds are exhausted -- start a new pass
    %%?INFO_LOGGER("find_or_spawn_peers: starting new pass after ~w ms with ~w~n", [get(find_other_sites_period_milliseconds), NextPassSiteIds]),
    find_or_spawn_peers_loop(ReplicationGroupName,
                             ThisSiteId,
                             ThisSiteReplicationGroupPid, 
                             SpawnOtherSitesInfo,
                             NextPassSiteIds,
                             [],
                             get(find_other_sites_period_milliseconds));  %% wait for a while before starting


find_or_spawn_peers_loop(ReplicationGroupName, 
                         ThisSiteId,
                         ThisSiteReplicationGroupPid, 
                         SpawnOtherSitesInfo,
                         ThisPassSiteIds,
                         NextPassSiteIds,
                         ReceiveTimeout) ->


    %% Invariant: ReceiveTimeout == infinity iff there is some work to do
    if (ReceiveTimeout == infinity) and (ThisPassSiteIds == []) and (NextPassSiteIds == [])  -> ok;
       (ReceiveTimeout /= infinity) and ((ThisPassSiteIds /= []) or (NextPassSiteIds /= [])) -> ok;
       true -> erlang:error({receive_timeout_invalid, ReceiveTimeout, ThisPassSiteIds, NextPassSiteIds})
    end,

    %% As our replication_group process sends us a 'stop_trying_site' message every time 
    %% it receives a site update, a constant stream of those messages would starve the 
    %% retry mechanism if we re-used the same receive timeout each time.
    %% so we adjust the the timeout each time.
    StartReceiveTime = now(),

    %%?INFO_LOGGER("find_or_spawn_peers: ThisPassSiteIds = ~w, NextPassSiteIds = ~w, ReceiveTimeout = ~w~n", [ThisPassSiteIds, NextPassSiteIds, ReceiveTimeout]),
    receive

        {find_or_spawn_site, OtherSiteId} -> 
            ?INFO_LOGGER("find_or_spawn_peers: received find_or_spawn_site ~w~n", [OtherSiteId]),
            %% This generally means that we were monitoring a site and we got a DOWN message.
            %% Note that we start monitoring sites when they subscribe to us, not just 
            %% when we subscribe to them.  Therefore it is possible that we get this 
            %% message for a site that we are already looking for, so we need to prevent
            %% duplicate entries.
            %% We immediately try the site (if it crashed we want to respawn it 
            %% as quickly as possible), and then continue with this pass
            %% (placing the new site at the end of NextPassSiteIds).
            %% We also reset NumAttempts for that site.

            find_or_spawn_site(ReplicationGroupName, 
                               ThisSiteId,
                               ThisSiteReplicationGroupPid, 
                               SpawnOtherSitesInfo,
                               OtherSiteId),

            find_or_spawn_peers_loop(ReplicationGroupName,
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     SpawnOtherSitesInfo, 
                                     lists:keydelete(OtherSiteId, 1, ThisPassSiteIds),
                                     lists:keydelete(OtherSiteId, 1, NextPassSiteIds) ++ [{OtherSiteId, 0}],
                                     case ReceiveTimeout of 
                                         infinity -> get(find_other_sites_period_milliseconds);                  % start retries
                                         0        -> 0;                                                % in the middle of a pass (continue)
                                         _        -> reduced_timeout(ReceiveTimeout, StartReceiveTime) % waiting for next pass (adjust time)
                                     end);

        {stop_trying_site, OtherSiteId} -> 
            ?INFO_LOGGER("find_or_spawn_peers: received stop_trying_site ~w~n", [OtherSiteId]),
            find_or_spawn_peers_loop(ReplicationGroupName,
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     SpawnOtherSitesInfo,
                                     lists:keydelete(OtherSiteId, 1, ThisPassSiteIds),
                                     lists:keydelete(OtherSiteId, 1, NextPassSiteIds),
                                     case ReceiveTimeout of 
                                         infinity -> infinity;
                                         0        -> 0;                                                  % in the middle of a pass (continue)
                                         _        -> reduced_timeout(ReceiveTimeout, StartReceiveTime)   % waiting for next pass (adjust time)
                                     end);

        {forget_other_site, OtherSiteId} -> 
            ?INFO_LOGGER("find_or_spawn_peers: received forget_other_site ~w~n", [OtherSiteId]),
            find_or_spawn_peers_loop(ReplicationGroupName,
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     case SpawnOtherSitesInfo of
                                         {AllSites, ConfigProplist} ->
                                             {[OtherSite || {SiteId, _, _, _} = OtherSite <- AllSites, SiteId /= OtherSiteId], 
                                              ConfigProplist};
                                         _ -> 
                                             SpawnOtherSitesInfo
                                     end,
                                     lists:keydelete(OtherSiteId, 1, ThisPassSiteIds),
                                     lists:keydelete(OtherSiteId, 1, NextPassSiteIds),
                                     case ReceiveTimeout of 
                                         infinity -> infinity;
                                         0        -> 0;                                                 % in the middle of a pass (continue)
                                         _        -> reduced_timeout(ReceiveTimeout, StartReceiveTime)  % waiting for next pass (adjust time)
                                     end);

        stop -> 
            ?INFO_LOGGER("find_or_spawn_peers exiting gracefully~n");

        Other -> 
            %% We shouldn't get any delayed replies from rpc:call (in find_or_spawn_site) 
            %% because we use rpc:call/5, the variant that spawns a middleman process 
            %% to swallow any late messages
            exit({unexpected_msg, Other})

    after
        ReceiveTimeout ->
            
            %%?INFO_LOGGER("find_or_spawn_peers: working on this pass. ThisPassSiteIds = ~w, NextPassSiteIds = ~w, ReceiveTimeout = ~w~n", [ThisPassSiteIds, NextPassSiteIds, ReceiveTimeout]),

            %% As timeout was not infinity there must be some work to do.
            [ {SiteToFindId,  NumAttempts} | TailThisPassSiteIds ] = ThisPassSiteIds,

            %% Try the next site.
            %%
            %% Even if successful, this we do NOT drop SiteId from the work queues as 
            %% 'success' isn't achieved until we have received the first status update 
            %% from the found/spawned site (at which point our replication_group will 
            %% send us a 'stop_trying_site' message)
            %% Until then anything might go wrong, so we keep retrying.
            %% Note: it would seem that we could stop retrying as soon as replication_group is 
            %% monitoring the found site's pid.  If replication_group gets a DOWN 
            %% message it will send us a 'find_or_spawn_site' message.
            %% However, we also begin monitoring a site if it subscribes to us
            %% -- but that doesn't mean that we are subscribed to that site yet.

            find_or_spawn_site(ReplicationGroupName, 
                               ThisSiteId,
                               ThisSiteReplicationGroupPid, 
                               SpawnOtherSitesInfo,
                               SiteToFindId),

            %% Move this SiteToFind to the end of the NextPass queue.
            %%
            %% We always set ReceiveTimeout to 0 in order to process the next site in TailThisPassSiteIds 
            %% immediately.  
            %% If TailThisPassSiteIds is empty then an earlier function clause will 
            %% move NextPassSiteIds to ThisPassSiteIds and 
            %% set ReceiveTimeout to the full between-passses timeout.

            find_or_spawn_peers_loop(ReplicationGroupName, 
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     SpawnOtherSitesInfo,
                                     TailThisPassSiteIds,
                                     NextPassSiteIds ++ [{SiteToFindId, NumAttempts + 1}],
                                     0)

    end.


reduced_timeout(ReceiveTimeout, StartReceiveTime) ->
    case ReceiveTimeout - ?BDB_U:now_diff_milliseconds(now(), StartReceiveTime) of
        T when T >= 0 -> T;
        _             -> 0
    end.
             

find_or_spawn_site(ReplicationGroupName, 
                   ThisSiteId,
                   ThisSiteReplicationGroupPid, 
                   SpawnOtherSitesInfo,
                   {_, SiteToFindNode} = SiteToFindId) ->

    %%?INFO_LOGGER("find_or_spawn_site trying ~w~n", [SiteToFindId]),

    %% We use rpc:call/5 (the variant with the timeout) as we don't want to 
    %% hang indefinitely if the other node doesn't reply.
    %% Also this variant promises to not pollute our message queue with 
    %% delayed replies (which I did see happen with rpc:call/4)
    case rpc:call(SiteToFindNode, 
                  ?MODULE, 
                  remote_find_or_spawn_site, 
                  [ReplicationGroupName, 
                   SiteToFindId, 
                   SpawnOtherSitesInfo],
                  get(rpc_call_timeout_milliseconds)) 
    of
        {ok, SiteToFindReplicationGroupPid} ->
            %%?INFO_LOGGER("find_or_spawn_peers found ~w for ~w~n", [SiteToFindReplicationGroupPid, SiteToFindId]),

            %% Point the two sites at each other (mutually subscribe)
            gen_server:cast(SiteToFindReplicationGroupPid, {subscribe, ThisSiteId,   ThisSiteReplicationGroupPid}),
            gen_server:cast(ThisSiteReplicationGroupPid,   {subscribe, SiteToFindId, SiteToFindReplicationGroupPid});

        Other ->
            ?INFO_LOGGER("find_or_spawn_site failed on ~w with ~w~n", [SiteToFindId, Other])
    end.

%% This helper must be exported for rpc to find it
remote_find_or_spawn_site(ReplicationGroupName, 
                          {SiteToFindName, SiteToFindNode} = SiteToFindId, 
                          SpawnOtherSitesInfo) ->

    SiteToFindNode = node(), % assert that we are spawning on the correct node

    case whereis( get_proc_name(ReplicationGroupName, SiteToFindName) ) of

        Pid when is_pid(Pid) -> 
            ?INFO_LOGGER("find_or_spawn_site found ~w for ~w~n", [Pid, SiteToFindId]),
            {ok, Pid};

        undefined -> 
            case SpawnOtherSitesInfo of

                {AllSites, ConfigProplist} ->

                    %% We should attempt to respawn the other site
                    [SiteToFind] = [Site || {SiteId, _, _, _} = Site <- AllSites, SiteId == SiteToFindId],

                    try ?MODULE:start(ReplicationGroupName, 
                                      SiteToFind, 
                                      AllSites -- [SiteToFind],  % AllSites from the perspective of SiteToFind
                                      ConfigProplist) of 

                        Pid -> 
                            ?INFO_LOGGER("find_or_spawn_site successfully spawned ~w for ~w~n", [Pid, SiteToFindId]),
                            {ok, Pid}

                    catch
                        {error, {already_started, Pid}} -> 
                            %% More than one site will try to spawn a missing site at the same time, so some of them may 
                            %% find that it does exists, even though the 'whereis' check above failed.
                            ?INFO_LOGGER("find_or_spawn_site spawned found ~w while attempting to spawn ~w~n", [Pid, SiteToFindId]),
                            {ok, Pid};

                        Error -> Error
                    end;

                disabled ->
                    %% We are not allowed to spawn the other site
                    undefined
            end
    end.


send_msg_to_find_or_spawn_peers_process(Msg, #state{support_procs = SupportProcs}) ->   
    case lists:keysearch(find_or_spawn_peers, #support_proc.name, SupportProcs) of 
        {value, #support_proc{pid = Pid}} -> Pid ! Msg
    end.



%%
%% Master-only support procs
%%

%% We need to coordinate remove of transaction logs across the entire group.
%% See the top-of-file comment for this.
%% Implementation choice: we leverage the master as a single decision-point/coordinator.
%% The master starts a remove_txn_logs_on_all_sites_process when it is elected, 
%% which runs txn-funs on all other sites to call log_archive.
%%
%% To avoid spreading the logic to handle this across many state-transition handlers
%% we just have all of the state-transition handlers (that could possibly affect this) 
%% call this single 'manage' function (passing their exit State).
%% This function which works out what to do (as that depends entirely on the value of 
%% State, not the delta between two States), does it, and returns a new exit State.
%% (Also see assert_remove_txn_logs_on_all_site_invariants)

manage_remove_txn_logs_on_all_sites_process_UNSAFE(#state{current_master = {this_site, available},
                                                          support_procs  = SupportProcs} = State) ->
    %% start the process if it is not already running.
    case lists:keysearch(remove_txn_logs_on_all_sites, #support_proc.name, SupportProcs) of
        false -> start_remove_txn_logs_on_all_sites_process_UNSAFE(State);
        _     -> State
    end;

manage_remove_txn_logs_on_all_sites_process_UNSAFE(#state{support_procs = SupportProcs} = State) ->
    %% We are not master so the process should not be running
    case lists:keysearch(remove_txn_logs_on_all_sites, #support_proc.name, SupportProcs) of
        false       -> State;
        {value, SP} -> stop_support_procs_UNSAFE([SP], State)
    end.

%% internal helpers
start_remove_txn_logs_on_all_sites_process_UNSAFE(#state{support_procs = SupportProcs} = State) ->

    ReplicationGroupName = get(replication_group_name),
    ThisSiteName         = get(this_site_name),

    Config = get(config_proplist),

    %% We handle 'disabled' by starting the process (so manage_remove_txn_logs_on_all_sites_process_UNSAFE 
    %% and assert_remove_txn_logs_on_all_site_invariants are satisfied) 
    %% but using a period of infinity so we never actually do any work 
    PeriodMs = case get_value_for_key(log_removal_period_seconds, Config) of
                   disabled -> infinity;
                   Seconds  -> Seconds * 1000
               end,
    DeleteFilePeriodMilliseconds      = get_value_for_key(log_removal_delete_file_period_milliseconds, Config),
    MaxNumTrailingLogFilesIfAnyErrors = get_value_for_key(log_removal_max_num_trailing_log_files, Config),

    SupportProc = 
        #support_proc{name     = remove_txn_logs_on_all_sites, 
                      pid      = proc_lib:spawn_link(
                                   fun() ->
                                           remove_txn_logs_on_all_sites_process(
                                             ReplicationGroupName,
                                             ThisSiteName,
                                             PeriodMs,
                                             DeleteFilePeriodMilliseconds,
                                             MaxNumTrailingLogFilesIfAnyErrors)
                                   end),
                      stop_fun = fun(#support_proc{pid = Pid}) -> 
                                         exit(Pid, kill)  % we don't need a clean exit (and we want it to be fast)
                                 end},

    ?NEW_PARTIAL_STATE(State#state{support_procs = [SupportProc | SupportProcs]}).


remove_txn_logs_on_all_sites_process(ReplicationGroupName, ThisSiteName, PeriodMs, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors) ->
    ?INFO_LOGGER("remove_txn_logs_on_all_sites: ReplicationGroupName = ~w, ThisSiteName = ~w, PeriodMs = ~w, DeleteFilePeriodMilliseconds = ~w, MaxNumTrailingLogFilesIfAnyErrors = ~w~n", [ReplicationGroupName, ThisSiteName, PeriodMs, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors]),
    remove_txn_logs_on_all_sites_loop(ReplicationGroupName, ThisSiteName, PeriodMs, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors).

remove_txn_logs_on_all_sites_loop(ReplicationGroupName, ThisSiteName, PeriodMs, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors) ->
    receive
        Msg -> ?INFO_LOGGER("remove_txn_logs_on_all_sites_process: received uexpected msg ~p; exiting~n", [Msg]),
               exit({unexpected_msg, Msg})
    after
        PeriodMs ->
            do_remove_txn_logs_on_all_sites(ReplicationGroupName, ThisSiteName, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors),
            remove_txn_logs_on_all_sites_loop(ReplicationGroupName, ThisSiteName, PeriodMs, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors)
    end.
    

do_remove_txn_logs_on_all_sites(ReplicationGroupName, ThisSiteName, DeleteFilePeriodMilliseconds, MaxNumTrailingLogFilesIfAnyErrors) ->

    %%?INFO_LOGGER("do_remove_txn_logs_on_all_sites: starting~n"),

    %% Ask all sites (inc. this site, the master) for their reverse-sorted 
    %% list of removable txn log filenames.
    %% If any sites are 'unknown' or return errors, record that fact.

    SiteLocatorsResponses = 
        ?BDB_U:pmap(
           fun({{SiteName, SiteNode}, unknown}) ->
                   %% Probably no point in asking this site
                   SiteLocator = {ReplicationGroupName, SiteName, SiteNode},
                   ?INFO_LOGGER("remove_txn_logs_on_all_sites: ~w is unknown~n", [SiteLocator]),
                   {SiteLocator, {error, site_unknown}};

              %% The site is connected, so we should ask it about it's logs,
              %% - even if it is a replica that is still looking for a master, or in the process of synching.
              %% (Remember, the goal is to make sure that we don't remove any logs from any other sites.)
              ({{SiteName, SiteNode}, _}) ->
                   SiteLocator = {ReplicationGroupName, SiteName, SiteNode},
                   Result = 
                       try 
                           do_txn_on_site(SiteLocator,
                                          {?MODULE, txn_fun_rev_list_removable_txn_logs, []}, 
                                          0)
                       catch
                           Reason -> {error, Reason}
                       end,                                        
                   %%?INFO_LOGGER("remove_txn_logs_on_all_sites: ~w returned ~p~n", [SiteLocator, Result]),
                   {SiteLocator, Result}
           end,
           %% call our own available API on this site, to get group status
           get_group_status({ReplicationGroupName, ThisSiteName, node()})),

    %% Find the most recent logfile that can safely be removed on ALL of the sites
    %% from which we got a response.
    %% This is the 'lowest' (i.e. oldest) file from the heads of all of the 
    %% per-site file lists (as those lists are in reverse order, with most 
    %% recent file first).
    %%
    %% We will use the identified file as a filter on the per-site lists returned by log_archive.
    %% Any file more recent that the LatestSafeRemovable file will be kept, and 
    %% only the older (removable) files will actually be removed.
    %%
    %% However, if even one site failed to return a list of removeable files then 
    %% we will allow up to 'log_removal_max_num_trailing_log_files' of the most 
    %% recent files to accumulate on each site.  This provides a window of log 
    %% history in which the down site can restart and do incremental sync-up.
    %% (If it restarts after falling past the oldest available log file then it will 
    %% need to do a full bootstrap of the environment before it can join 
    %% the replication group).  When all sites become in_sync again,
    %% the extra trailing logfiles will automatically be deleted on the 
    %% next pass of this function.
    %%
    %% Note that sites with zero removable files don't require any special treatment
    %% -- they simply fail to affect LatestSafeRemovable, and filtering their 
    %% list of removable files based on the the final LatestSafeRemovable obviously 
    %% has no effect.

    case 
        lists:foldl(
          fun({_SiteId, {removable_logs, [ThisRemovable]}}, {LatestSafeRemovable, Errors}) when ThisRemovable < LatestSafeRemovable ->
                  %%?INFO_LOGGER("remove_txn_logs_on_all_sites: site ~w with one removable log overrides LatestSafeRemovable (~p) with ThisRemovable (~p)~n", [SiteId, LatestSafeRemovable, ThisRemovable]),
                  {ThisRemovable, Errors};

             ({_SiteId, {removable_logs, [ThisRemovable|_] = _RemovableLogs}}, {LatestSafeRemovable, Errors}) when ThisRemovable < LatestSafeRemovable ->
                  %%?INFO_LOGGER("remove_txn_logs_on_all_sites: site ~w with ~w removable log overrides LatestSafeRemovable (~p) with ThisRemovable (~p)~n", [SiteId, length(RemovableLogs), LatestSafeRemovable, ThisRemovable]),
                  {ThisRemovable, Errors};

             ({_SiteId, {no_removable_logs, OldestNonRemovable}}, {LatestSafeRemovable, Errors} = Acc) ->                  
                  %% this site doesn't have ANY removable logs, but it might still be a 'lagging' site
                  %% -- i.e. it's oldest in-use log might well be the oldest in-use log of all sites,
                  %% so we want all sites to keep their logs back to that point.
                  %% So we need to calculate what LatestSafeRemovable _would_ have been if this 
                  %% site did in fact have some even older logs that were removable.
                  case log_filename_one_less_than(OldestNonRemovable) of 
                      WouldBeRemovable when WouldBeRemovable < LatestSafeRemovable ->
                          %%?INFO_LOGGER("remove_txn_logs_on_all_sites: lagging site ~w with no removable logs overrides LatestSafeRemovable (~p) with WouldBeRemovable = ~p~n", [SiteId, LatestSafeRemovable, WouldBeRemovable]),
                          {WouldBeRemovable, Errors};

                      _WouldBeRemovable ->
                          %%?INFO_LOGGER("remove_txn_logs_on_all_sites: site ~w with no removable logs, but LatestSafeRemovable (~p) is earlier or same as than this site's WouldBeRemovable (~p)~n", [SiteId, LatestSafeRemovable, WouldBeRemovable]),
                          Acc
                  end;

             ({_, {error, _}}, {LatestSafeRemovable, _}) ->
                  {LatestSafeRemovable, error}; % at least 1 site failed to return a list

             (_, LatestSafeRemovable_AnyErrors) ->
                  LatestSafeRemovable_AnyErrors
          end,
          %% initial {LatestSafeRemovable, AnyErrors}
          %% all log file names sort < "sentinel" (cannot be an atom, as atom() < list())
          {"sentinel", ok},
          SiteLocatorsResponses)
    of
        {"sentinel", _} ->
            %% No sites have any removable log files, so we are done.
            %%?INFO_LOGGER("remove_txn_logs_on_all_sites: no sites have any removable logs.\n"),
            ok;
            
        {LatestSafeRemovable, AnyErrors} ->
            %% Normal case: tell all sites to remove logs with 
            %%   id <= LatestRemovableLogId
            ?BDB_U:pmap(
               fun({SiteLocator, {removable_logs, RemovableTxnLogs}}) ->

                       %% We only want to delete logs older than (or same as) LatestSafeRemovable
                       SafeToRemoveLogs  = lists:filter(fun(Log) -> Log =< LatestSafeRemovable end,
                                                        RemovableTxnLogs),

                       %% If any site had an error, we can keep the first 
                       %% N of the remaining logs.
                       LogsToRemove = 
                           case AnyErrors of 
                               ok    -> 
                                   SafeToRemoveLogs;
                               error -> 
                                   ?INFO_LOGGER("remove_txn_logs_on_all_sites: some sites were unreachable, so leaving ~w trailing removable logs~n", [MaxNumTrailingLogFilesIfAnyErrors]),
                                   case length(SafeToRemoveLogs) - (MaxNumTrailingLogFilesIfAnyErrors + 1) of 
                                       N when N > 0 -> lists:nthtail(N, SafeToRemoveLogs);
                                       _            -> []
                                   end
                           end,

                       %%?INFO_LOGGER("remove_txn_logs_on_all_sites: telling site ~w to remove logs ~p~n", [SiteLocator, LogsToRemove]),

                       %% We always do this even if if LogsToRemove is [], as we want to log 
                       %% on the destination site that there were no removable logs (to help 
                       %% diagnose any problems)
                       do_txn_on_site(SiteLocator,
                                      {?MODULE, txn_fun_remove_specified_txn_logs, {LogsToRemove, DeleteFilePeriodMilliseconds}}, 
                                      0,
                                      %% allow enough time (even if there are no logs to remove allow some time)
                                      5000 + (length(LogsToRemove) * DeleteFilePeriodMilliseconds));

                  ({_SiteLocator, {no_removable_logs, _} = _NoRemovableLogs}) ->
                       %%?INFO_LOGGER("remove_txn_logs_on_all_sites: skipping site ~w as it returned with ~p\n", [SiteLocator, NoRemovableLogs])
                       ok;

                  ({SiteLocator, {error, Error}}) ->
                       ?INFO_LOGGER("remove_txn_logs_on_all_sites: skipping site ~w as it returned error ~500p\n", [SiteLocator, Error])
               end,
               SiteLocatorsResponses)
    end.

log_filename_one_less_than(LogFilename) ->
    %% BDB txn log filenames are "log.NNNNNNNNNN" -- i.e. 10-digit monotonically increasing numbers starting from 1.
    %% The number is padded with zeroes, which is critical -- it makes lexographic sort-order equivalent to numerical sort-order.
    lists:flatten(io_lib:format("log.~10..0w", [log_filename_to_number(LogFilename) - 1])).

log_filename_to_number(LogFilename) ->
    list_to_integer(string:substr(LogFilename, string:str(LogFilename, ".") + 1)).

%% Returns:
%%   If the site has one or more removable logs:
%%
%%      {removable_logs, ReverseSortedLogFilenamesList}
%% 
%%      LogFilename -> "log.0000000001".
%%      (Note that due to 0-padding, lexographic sort order is the same 
%%      as numerical sort order.)
%%       
%%      The returned list is sorted in REVERSE (decreasing) sorted order
%%      (so most recent logfile is first in list).
%%
%%   If site has NO removable logs:
%%
%%      {no_removable_logs, OldestNonRemovableLog}
%%
%%      This site might be lagging behind the other sites, so we 
%%      need to ensure that all other sites retain their local logs 
%%      back to at least the earliest in-use log from this site.
txn_fun_rev_list_removable_txn_logs(BdbPort, _) ->

    %% We only want to allow one of these processes at a time, so 
    %% register it, and crash (badarg) if another process currently 
    %% exists with the same name,
    register_worker_process("remove_txn_logs"),

    case ?BDB:log_archive(BdbPort, []) of
        [] ->
            {no_removable_logs, lists:min(?BDB:log_archive(BdbPort, [bdb_DB_ARCH_LOG]))};  % earliest log from set of ALL logs on this site

        RemovableLogs ->
            {removable_logs, lists:sort(fun(A, B) -> A > B end, RemovableLogs)}
    end.

%% helper -- must be called from a TxnFun
%% As a convenience, returns {ReplicationGroupName, SiteName}
register_worker_process(NameStr) ->
    {ReplicationGroupName, SiteName} = ReplicationContext = get(txn_fun_replication_context),
    register(list_to_atom(get_proc_name_as_list(ReplicationGroupName, SiteName) ++ "_" ++ NameStr), self()),
    ReplicationContext.

txn_fun_remove_specified_txn_logs(BdbPort, {RemoveLogFilenames, DeleteFilePeriodMilliseconds}) ->

    %% Again, we only want to allow one of these to exist at any given time
    register_worker_process("remove_txn_logs"),

    ?INFO_LOGGER("Removing ~w transaction logs~n", [length(RemoveLogFilenames)]),
    BdbDir = get_value_for_key(env_dir, ?BDB_PC:get_config(BdbPort)) ++ "/",
    lists:foreach(fun(LogFilename) ->
                          FilenameToDelete = BdbDir ++ LogFilename,
                          case file:delete(FilenameToDelete) of
                              ok -> 
                                  ok;
                              Error ->
                                  ?WARNING_LOGGER("**failed** to delete ~p : ~w~n", [FilenameToDelete, Error])
                          end,
                          timer:sleep(DeleteFilePeriodMilliseconds)
                  end,
                  RemoveLogFilenames).


%% We need to alarm if any replicas start falling significantly behind the master
%% This process periodically polls DB_REP_STAT.st_next_lsn on all sites, 
%% and logs a warning if the LSN difference is too great.
manage_alarm_on_lagging_replicas_process_UNSAFE(#state{current_master = {this_site, available},
                                                       support_procs  = SupportProcs} = State) ->
    %% start the process if it is not already running.
    case lists:keysearch(alarm_on_lagging_replicas, #support_proc.name, SupportProcs) of
        false -> start_alarm_on_lagging_replicas_process_UNSAFE(State);
        _     -> State
    end;

manage_alarm_on_lagging_replicas_process_UNSAFE(#state{support_procs = SupportProcs} = State) ->
    %% We are not master so the process should not be running
    case lists:keysearch(alarm_on_lagging_replicas, #support_proc.name, SupportProcs) of
        false       -> State;
        {value, SP} -> stop_support_procs_UNSAFE([SP], State)
    end.

%% internal helpers
start_alarm_on_lagging_replicas_process_UNSAFE(#state{support_procs = SupportProcs} = State) ->

    ReplicationGroupName = get(replication_group_name),
    ThisSiteName         = get(this_site_name),

    Config = get(config_proplist),

    %% We handle 'disabled' by starting the process (so manage_alarm_on_lagging_replicas_process_UNSAFE 
    %% and assert_remove_txn_logs_on_all_site_invariants are satisfied) 
    %% but using a period of infinity so we never actually do any work 
    PeriodMs = case get_value_for_key(check_for_lagging_replicas_period_seconds, Config) of
                   disabled -> infinity;
                   Seconds  -> Seconds * 1000
               end,

    LogFileSize                  = get_value_for_key(log_file_size_bytes, ?BDB_PC:get_config(get(this_site_port_coord_pid))),
    LaggingReplicaThresholdBytes = get_value_for_key(lagging_replica_threshold_bytes, Config),
    LaggingReplicaRestartAfterN  = get_value_for_key(lagging_replica_restart_after_N_consecutive, Config),

    SupportProc = 
        #support_proc{name     = alarm_on_lagging_replicas, 
                      pid      = proc_lib:spawn_link(
                                   fun() ->
                                           alarm_on_lagging_replicas_process(
                                             ReplicationGroupName,
                                             ThisSiteName,
                                             PeriodMs,
                                             LogFileSize,
                                             LaggingReplicaThresholdBytes,
                                             LaggingReplicaRestartAfterN)
                                   end),
                      stop_fun = fun(#support_proc{pid = Pid}) -> 
                                         exit(Pid, kill)  % we don't need a clean exit (and we want it to be fast)
                                 end},

    ?NEW_PARTIAL_STATE(State#state{support_procs = [SupportProc | SupportProcs]}).


alarm_on_lagging_replicas_process(ReplicationGroupName, ThisSiteName, PeriodMs, LogFileSize, 
                                  LaggingReplicaThresholdBytes, LaggingReplicaRestartAfterN) ->

    ?INFO_LOGGER("alarm_on_lagging_replicas: ReplicationGroupName = ~w, ThisSiteName = ~w, PeriodMs = ~w, LogFileSize = ~w, LaggingReplicaThresholdBytes = ~w, LaggingReplicaRestartAfterN = ~w~n", 
                 [ReplicationGroupName, ThisSiteName, PeriodMs, LogFileSize, LaggingReplicaThresholdBytes, LaggingReplicaRestartAfterN]),

    alarm_on_lagging_replicas_loop([], 
                                   ReplicationGroupName, {ThisSiteName, node()}, PeriodMs, LogFileSize, 
                                   LaggingReplicaThresholdBytes, LaggingReplicaRestartAfterN).

alarm_on_lagging_replicas_loop(PersistentlyLaggingReplicas, 
                               ReplicationGroupName, ThisMasterSiteId, PeriodMs, LogFileSize, 
                               LaggingReplicaThresholdBytes, LaggingReplicaRestartAfterN) ->
    receive
        Msg -> ?INFO_LOGGER("alarm_on_lagging_replicas_process: received uexpected msg ~p; exiting~n", [Msg]),
               exit({unexpected_msg, Msg})
    after
        PeriodMs ->

            LaggingReport = check_for_currently_lagging_replicas(ReplicationGroupName, ThisMasterSiteId, LogFileSize, LaggingReplicaThresholdBytes),

            %% Consider restarting any in_sync replicas (of *this* master) that are persistently lagging.
            %% We only do this for in_sync replicas as we want to allow replicas 
            %% a chance to catch up after restarting them (or joining).

            %% LaggingReport                = [{SiteLocator, SiteStatus, lagging|not_lagging|failed_to_report}, ...]
            %% PersistentlyLaggingReplicas  = [{SiteLocator, Count}, ...]
            
            NewPersistentlyLaggingReplicas 
                = lists:foldl(
                    fun({SiteLocator, {replica, in_sync, {MasterSiteId, available}}, lagging}, Acc) when MasterSiteId =:= ThisMasterSiteId ->

                            case lists:keysearch(SiteLocator, 1, PersistentlyLaggingReplicas) of 

                                {value, {SiteLocator, Count}}  ->
                                    case Count + 1 of 
                                        NewCount when NewCount >= LaggingReplicaRestartAfterN ->
                                            stop_site(SiteLocator, ThisMasterSiteId),
                                            Acc;     % replica is nolonger considered to be lagging until after it is in_sync again
                                        NewCount ->
                                            [{SiteLocator, NewCount} | Acc]
                                    end;

                                false ->
                                    %% Site was not lagging before, but is lagging now
                                    %% (we might want to restart it immediately)
                                    if LaggingReplicaRestartAfterN == 1 ->
                                            stop_site(SiteLocator, ThisMasterSiteId),
                                            Acc;
                                       true ->
                                            [{SiteLocator, 1} | Acc]
                                    end
                            end;

                       (_, Acc) -> 
                            %% we don't consider restarting sites that have not received STARTUPDONE for this master
                            %% (also, this master must be 'available', not waiting_for_durability)
                            Acc
                    end,
                    [],
                    LaggingReport),

            %%?INFO_LOGGER("NewPersistentlyLaggingReplicas = ~w~n", [NewPersistentlyLaggingReplicas]),

            alarm_on_lagging_replicas_loop(NewPersistentlyLaggingReplicas, 
                                           ReplicationGroupName, ThisMasterSiteId, PeriodMs, LogFileSize, 
                                           LaggingReplicaThresholdBytes, LaggingReplicaRestartAfterN)
    end.

stop_site(StopSiteLocator, ThisMasterSiteId) ->    
    ?ERROR_LOGGER("**Shutting down (restarting) site ~w because it is lagging too much**~n", [StopSiteLocator]),
    try 
        do_txn_on_site(StopSiteLocator,
                       {?MODULE, txn_fun_stop_site, ThisMasterSiteId}, 
                       0)
    catch
        {bdb_replication_group_error, {site_is_down, StopSiteLocator, _Reason}} ->
            %% the site's terminate function might kill our do_txn worker before it returns
            %% -- that's fine
            ok
    end.

txn_fun_stop_site(#bdb_port_handle{port_coordinator_pid = PortCoordPid}, ThisMasterSiteId) ->
    ?ERROR_LOGGER("**site is being shutdown (restarted) by master ~w because it is lagging too much**~n", [ThisMasterSiteId]),
    ?BDB_PC:stop(PortCoordPid).


%% Returns LaggingReport = [{SiteLocator, SiteStatus, lagging|not_lagging|failed_to_report}, ...]
check_for_currently_lagging_replicas(ReplicationGroupName, {MasterSiteName, MasterSiteNode}, LogFileSize, LaggingReplicaThresholdBytes) ->

    MasterSiteLocator = {ReplicationGroupName, MasterSiteName, MasterSiteNode},
    MasterLSN = do_txn_on_site(MasterSiteLocator,
                               {?MODULE, txn_fun_rep_stat_next_lsn, []}, 
                               0),

    %% we use pmap (rather than foreach/spawn) so that we don't return until all responses are in.
    %% We log the results as they arrive, rather than after all the results are in (when pmap returns),
    %% so that we get fairlh accurate timestamps with the LSNs (i.e. they are not distorted 
    %% if one response is delayed).
    ?BDB_U:pmap(
       fun({{SiteName, SiteNode}, unknown = SiteStatus}) ->
               %% Probably no point in asking this site
               SiteLocator = {ReplicationGroupName, SiteName, SiteNode},
               ?INFO_LOGGER("alarm_on_lagging_replicas: ~w is unknown~n", [SiteLocator]),
               {SiteLocator, SiteStatus, failed_to_report};

          %% The site is connected, so we should ask it for it's next_LSN
          %% - even if it is a replica that is still looking for a master, or in the process of synching.
          %% (Remember, the goal is to make sure that we don't remove any logs from any other sites.)
          ({{SiteName, SiteNode}, SiteStatus}) ->
               SiteLocator = {ReplicationGroupName, SiteName, SiteNode},
               try 
                   do_txn_on_site(SiteLocator,
                                  {?MODULE, txn_fun_compare_rep_stat_next_lsn_to_master, 
                                   {MasterLSN, LogFileSize, LaggingReplicaThresholdBytes}}, 
                                  0)
               of
                   {ReplicaLSN, DiffLSNs} ->
                       if DiffLSNs >= LaggingReplicaThresholdBytes ->
                               ?WARNING_LOGGER("alarm_on_lagging_replicas: **LAGGING REPLICA** master LSN = ~w, replica LSN = ~w, difference = ~w bytes, site = ~w,~w status = ~w~n", [MasterLSN, ReplicaLSN, DiffLSNs, SiteName, SiteNode, SiteStatus]),
                               {SiteLocator, SiteStatus, lagging};
                          true ->
                               %%?INFO_LOGGER("alarm_on_lagging_replicas: master LSN = ~w, replica LSN = ~w, difference = ~w bytes, site = ~w,~w~n", [MasterLSN, ReplicaLSN, DiffLSNs, SiteName, SiteNode])
                               {SiteLocator, SiteStatus, not_lagging}
                       end
               catch
                   Class:Reason ->
                       ?INFO_LOGGER("alarm_on_lagging_replicas: ~w failed to report LSN: ~w:~p~n", [SiteLocator, Class, Reason]),
                       {SiteLocator, SiteStatus, failed_to_report}
               end                                        
       end,
       %% call our own public API on this site, to get group status
       get_group_status(MasterSiteLocator)).

txn_fun_rep_stat_next_lsn(BdbPort, _) ->
    ?BDB:rep_stat_next_lsn(BdbPort, []).

%% We do the comparison on the replica, so we can log on the replica if it is behind.
%% Then we return the replica's LSN, so the master can log a summary.
txn_fun_compare_rep_stat_next_lsn_to_master(BdbPort, {MasterLSN, LogFileSize, LaggingReplicaThresholdBytes}) ->

    %% We only want one of these processes at a time on each site
    register_worker_process("alarm_on_lagging_replicas"),

    ThisSiteLSN = ?BDB:rep_stat_next_lsn(BdbPort, []),
    DiffLSNs    = diff_LSNs_bytes(MasterLSN, ThisSiteLSN, LogFileSize),
    if DiffLSNs >= LaggingReplicaThresholdBytes ->
            ?WARNING_LOGGER("**LAGGING REPLICA** master LSN = ~w, this-site LSN = ~w, difference = ~w bytes~n", [MasterLSN, ThisSiteLSN, DiffLSNs]);
       true ->
            %%?INFO_LOGGER("master LSN = ~w, this-site LSN = ~w, difference = ~w bytes~n", [MasterLSN, ThisSiteLSN, DiffLSNs])
            ok
    end,
    {ThisSiteLSN, DiffLSNs}.


%% This computes LSN_A - LSN_B in bytes, assuming that *all* log files are size LogFileSize
%% (it is worth avoiding bignum arithmetic by subtracting before multiply)
diff_LSNs_bytes({A_File, A_Offset}, {B_File, B_Offset}, LogFileSize) ->
    ((A_File - B_File) * LogFileSize) + A_Offset - B_Offset.
    


%% The probe_to_sync_replicas_process does periodic write transactions
%% to a 'reserved' database/key, purely to stimulate the STARTUPDONE
%% event on replicas if the application isn't doing any write
%% transactions of it's own.
%%
%% As we do with the remove_txn_logs_on_all_sites_process, to 
%% to avoid spreading the logic to handle this across many state-transition handlers
%% we just have all of the state-transition handlers (that could possibly affect this) 
%% call this single manage_probe_to_sync_replicas function (passing their exit State).
%% This function which works out what to do (as that depends entirely on the value of 
%% State, not the delta between two States), does it, and returns a new exit State.
%% (Also see assert_probe_to_sync_replicas_invariants)

%% If probe_to_sync_replicas should be running and it is not, then start it.
%% If it is running and it shouldn't be, then stop it.
%% Returns new State (or old State unchanged)

manage_probe_to_sync_replicas_process_UNSAFE(#state{current_master = {this_site, waiting_for_durability}} = State) ->
    %% We have received the MASTER event and are waiting for probe_to_sync_replicas 
    %% to tell us that we can write durable transactions.
    start_probe_to_sync_replicas_process_if_not_running_UNSAFE(State);

manage_probe_to_sync_replicas_process_UNSAFE(#state{current_master     = {this_site, available},
                                                    other_sites_status = OtherSitesStatus} = State) ->

    %% We are master, so probe_to_sync_replicas should only be running if any other site is 
    %% {replica, not_in_sync, ThisSiteId}.
    %% Note: we don't run probe_to_sync_replicas if any replicas are 'unknown' (i.e. not connected)
    %% even if all the others are in_sync. We only start probe_to_sync_replicas when the missing replica
    %% connects and transitions to {replica, not_in_sync, ThisSiteId}.
    %% We also don't start the probe if some other sites are {replica, not_in_sync, SomeOtherMaster}
    %% or {replica, in_sync, SomeOtherMaster}.  We wait for them to realize that we are master and 
    %% transition to {replica, not_in_sync ThisSite}.
    %% Also, this won't start the probe if there are no replicas in the group.

    ThisSiteId = get(this_site_id),

    case [SiteId || {SiteId, #site_status{status = {replica, not_in_sync, {MasterSiteId, _}}}} <- OtherSitesStatus,
                    MasterSiteId == ThisSiteId] 
    of
        [] -> 
            %% All replicas are either 'unknown' (not connected) or in_sync, so stop probe if it is running.
            stop_probe_to_sync_replicas_process_if_running_UNSAFE(State);
        _ -> 
            %% At least 1 replica is not_in_sync, so start it if it is not already running.
            start_probe_to_sync_replicas_process_if_not_running_UNSAFE(State)
    end;

manage_probe_to_sync_replicas_process_UNSAFE(State) ->  
    %% We are not master so probe_to_sync_replicas should not be running
    stop_probe_to_sync_replicas_process_if_running_UNSAFE(State).

%% internal helpers
start_probe_to_sync_replicas_process_if_not_running_UNSAFE(#state{support_procs = SupportProcs} = State) ->

    case lists:keysearch(probe_to_sync_replicas, #support_proc.name, SupportProcs) of
        false -> 
            ThisSitePid          = self(),
            ThisSitePortCoordPid = get(this_site_port_coord_pid),
            PeriodsMs            = get(probe_to_sync_replicas_periods_milliseconds),

            ?INFO_LOGGER("Spawning probe_to_sync_replicas_process~n"),
            ProbeToSyncReplicasSupportProc =
                #support_proc{name     = probe_to_sync_replicas, 
                              pid      = proc_lib:spawn_link(
                                           fun() -> 
                                                   probe_to_sync_replicas_process(ThisSitePid, ThisSitePortCoordPid, PeriodsMs) 
                                           end),
                              stop_fun = fun(#support_proc{pid = Pid}) -> 
                                                 exit(Pid, kill)  % we don't need a clean exit (and we want it to be fast)
                                         end},

            ?NEW_PARTIAL_STATE(State#state{support_procs = [ProbeToSyncReplicasSupportProc | SupportProcs]});

        {value, #support_proc{}} -> 
            State   % is already running
    end.

stop_probe_to_sync_replicas_process_if_running_UNSAFE(#state{support_procs = SupportProcs} = State) ->  
    %% We are not master so probe_to_sync_replicas should not be running
    case lists:keysearch(probe_to_sync_replicas, #support_proc.name, SupportProcs) of
        false       -> State;
        {value, SP} -> stop_support_procs_UNSAFE([SP], State)
    end.

probe_to_sync_replicas_process(ThisSitePid, PortCoordPid, {NotDurableMs, DurableMs} = PeriodsMs) ->
    ?INFO_LOGGER("probe_to_sync_replicas: starting NotDurablePeriods = ~w Durable = ~w~n", [NotDurableMs, DurableMs]),
    %% Initially assume that we are not durable, so we send a message when 
    %% we transition to durability.
    %% We always start by attempting a transaction immediately, not by sleeping.
    do_probe_to_sync_replicas_and_then_loop(ThisSitePid, PortCoordPid, PeriodsMs, not_durable).

probe_to_sync_replicas_loop(ThisSitePid, PortCoordPid, PeriodsMs, CurrPeriodMs, DurabilityStatus) ->
    receive
        Msg -> ?INFO_LOGGER("probe_to_sync_replicas_process: received unexpected msg ~p; exiting.~n", [Msg]),
               exit({unexpected_msg, Msg})
    after
        CurrPeriodMs -> 
            do_probe_to_sync_replicas_and_then_loop(ThisSitePid, PortCoordPid, PeriodsMs, DurabilityStatus)
    end.

do_probe_to_sync_replicas_and_then_loop(ThisSitePid, PortCoordPid, {_NotDurableMs, DurableMs} = PeriodsMs, DurabilityStatus) ->
    try 
        ?BDB_H:do_txn(PortCoordPid, {?MODULE, txn_fun_probe_to_sync_replicas, []}, 0)
    of
        ok ->
            case DurabilityStatus of 
                durable     -> 
                    ok;              % already durable; nothing to do
                not_durable -> 
                    ?INFO_LOGGER("probe_to_sync_replicas transitioned to durable transactions~n"),
                    ThisSitePid ! achieved_durability
            end,
            probe_to_sync_replicas_loop(ThisSitePid, PortCoordPid, PeriodsMs, DurableMs, durable)

    catch
        {txn_fun_error, {bdb_error, bdb_DB_REP_NOTPERM}} ->
            handle_probe_to_sync_replicas_NOTPERM(ThisSitePid, PortCoordPid, PeriodsMs, DurabilityStatus);

        {txn_fun_error, {bdb_port_coordinator_error, {bdb_error, bdb_DB_REP_NOTPERM}}} ->
            %% ... this exception occurs if ?BDB_PC:get_db_handle hits NOTPERM
            handle_probe_to_sync_replicas_NOTPERM(ThisSitePid, PortCoordPid, PeriodsMs, DurabilityStatus);

        %% If this flag is being used:
        %%  {replication, 
        %%    [{config_flags, [bdb_DB_REP_CONF_NOWAIT]}]}
        %% then this error might be returned:
        {txn_fun_error, {bdb_error, bdb_DB_REP_LOCKOUT}} ->
            handle_probe_to_sync_replicas_NOTPERM(ThisSitePid, PortCoordPid, PeriodsMs, DurabilityStatus);
        {txn_fun_error, {bdb_port_coordinator_error, {bdb_error, bdb_DB_REP_DB_REP_LOCKOUT}}} ->
            handle_probe_to_sync_replicas_NOTPERM(ThisSitePid, PortCoordPid, PeriodsMs, DurabilityStatus);

        Reason ->
            exit({probe_to_sync_replicas_failed, Reason})
    end.

handle_probe_to_sync_replicas_NOTPERM(ThisSitePid, PortCoordPid, {NotDurableMs, _DurableMs} = PeriodsMs, DurabilityStatus) ->
    case DurabilityStatus of 
        durable -> 
            %% Transition from durable to not_durable!
            %% We handle this with perm_failed_events_limit, so we only log here.
            ?WARNING_LOGGER("probe_to_sync_replicas transitioned to durable transactions~n");
        not_durable -> 
            ?INFO_LOGGER("probe_to_sync_replicas is still getting NOTPERM~n")
    end,
    probe_to_sync_replicas_loop(ThisSitePid, PortCoordPid, PeriodsMs, NotDurableMs, not_durable).

txn_fun_probe_to_sync_replicas(BdbPort, _) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, {"replication_group_meta_data.db", "misc"}),
    ?BDB:db_put(BdbPort, DB, void, <<"probe_to_sync_replicas">>, term_to_binary(now()), []).



%%
%% Other helpers
%%

txn_fun_start_checkpointer_if_enabled(BdbPort, _) ->
    ?BDB_PC:start_checkpointer_if_enabled(BdbPort).

txn_fun_stop_checkpointer_if_running(BdbPort, _) ->
    ?BDB_PC:stop_checkpointer_if_running(BdbPort).


txn_fun_set_master_cache_throttle_io(_BdbPort, {_EventArgs, same_as_replicas}) ->
    %% nothing to do -- port_coordinator:init already set this
    %%?INFO_LOGGER("txn_fun_set_master_cache_throttle_io : same_as_replicas~n"),
    ok;
txn_fun_set_master_cache_throttle_io(BdbPort, {_EventArgs, {SequentialWriteOps, SleepMicroseconds} = Params}) ->
    ?INFO_LOGGER("txn_fun_set_master_cache_throttle_io setting ~w~n", [Params]),
    ?BDB:set_mp_max_write(BdbPort, SequentialWriteOps, SleepMicroseconds).


%% This works for assoc-lists like config_proplist and the top-level of our BDB records.
%% We use lists:keysearch is it is a bif (faster), rather than proplists:lookup.
get_value_for_key(Key, AssocList) ->
    case lists:keysearch(Key, 1, AssocList) of
        {value, {Key, Value}} -> Value;
        false                 -> throw({missing_required_key, Key, AssocList})
    end.


repeat_until_true(Fun) ->
    case Fun() of 
        true  -> ok;
        false -> repeat_until_true(Fun)
    end.
             

%%
%% Assert invariants
%%

assert_state_invariants(#state{current_master   = unknown,
                               this_site_status = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus} = State) ->
    assert_this_site_status_invariants(ThisSiteStatus),
    {replica, _, unknown} = ThisSiteStatusValue,
    assert_all_other_sites_have_status_entry(State),
    %% We cannot assert that no other sites are master, as we might not have their latest status
    %% (e.g. we might have known about a master, then learned about a newer master (and not yet received a status update or 'DOWN' from the 
    %% first master), and then received a 'DOWN' from the second master.  This would have reset us to {replica, not_in_sync, unknown}, 
    %% but other_sites_status would still show the first master still claiming to be master
    assert_support_proc_invariants(State);

assert_state_invariants(#state{current_master     = {this_site, waiting_for_durability},
                               this_site_status   = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus} = State) ->
    assert_this_site_status_invariants(ThisSiteStatus),
    {master, waiting_for_durability} = ThisSiteStatusValue,
    assert_all_other_sites_have_status_entry(State),
    %% We cannot assert that no other sites are master, as we might not have their latest status
    assert_support_proc_invariants(State);

assert_state_invariants(#state{current_master     = {this_site, available},
                               this_site_status   = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus,
                               notify_clients     = NotifyClients} = State) ->
    assert_this_site_status_invariants(ThisSiteStatus),
    {master, available} = ThisSiteStatusValue,
    assert_all_other_sites_have_status_entry(State),
    %% We cannot assert that no other sites are master, as we might not have their latest status
    [] = NotifyClients, % we should already have told all clients
    assert_support_proc_invariants(State);

%% current master is another site  ...
assert_state_invariants(#state{current_master     = {_CurrentMasterSiteId, transitioning_to} = CurrentMaster,
                               this_site_status   = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus} = State) ->
    assert_this_site_status_invariants(ThisSiteStatus),
    {replica, _, CurrentMaster} = ThisSiteStatusValue,
    assert_all_other_sites_have_status_entry(State),
    %% As this state might result from BDB-HA sending us NEWMASTER for CurrentMasterSiteId 
    %% before we even connect to CurrentMasterSiteId ourselves, 
    %% we cannot assert anything about other_sites_status.
    %% And we cannot assert that NotifyClients is empty.
    assert_support_proc_invariants(State);

%% we _must_ have had an other_site_status_update from CurrentMasterSiteId
assert_state_invariants(#state{current_master     = {CurrentMasterSiteId, waiting_for_durability = CurrentMasterSubState} = CurrentMaster,
                               this_site_status   = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus,
                               other_sites_status = OtherSitesStatus} = State) ->
    assert_this_site_status_invariants(ThisSiteStatus),
    {replica, _, CurrentMaster} = ThisSiteStatusValue,
    assert_all_other_sites_have_status_entry(State),
    %% We cannot assert that CurrentMasterSiteId is the _only_ other site claiming to be master, 
    %% as we might not have their latest status
    %% All we can assert is that CurrentMasterSiteId has its correct own state
    {value, {_, #site_status{status = {master, CurrentMasterSubState}}}} = lists:keysearch(CurrentMasterSiteId, 1, OtherSitesStatus),
    %% We cannot assert that NotifyClients is empty.
    assert_support_proc_invariants(State);

assert_state_invariants(#state{current_master     = {CurrentMasterSiteId, available = CurrentMasterSubState} = CurrentMaster,
                               this_site_status   = #site_status{status = ThisSiteStatusValue} = ThisSiteStatus,
                               other_sites_status = OtherSitesStatus,
                               notify_clients     = NotifyClients} = State) ->
    assert_this_site_status_invariants(ThisSiteStatus),
    {replica, _, CurrentMaster} = ThisSiteStatusValue,
    assert_all_other_sites_have_status_entry(State),
    %% We _must_ have had an other_site_status_update from CurrentMasterSiteId
    %% But we still cannot assert that CurrentMasterSiteId is the _only_ other site claiming to be master, 
    %% as we might not have their latest status
    %% All we can assert is that CurrentMasterSiteId has its correct own state
    {value, {_, #site_status{status = {master, CurrentMasterSubState}}}} = lists:keysearch(CurrentMasterSiteId, 1, OtherSitesStatus),
    [] = NotifyClients,  % we should already have told all clients
    assert_support_proc_invariants(State).


assert_this_site_status_invariants(#site_status{status = unknown} = ThisSiteStatus) ->
    throw({this_site_unknown, ThisSiteStatus});
assert_this_site_status_invariants(#site_status{status                = _,
                                                last_updated_time     = Time,
                                                replication_group_pid = undefined,
                                                monitor_ref           = undefined} = ThisSiteStatus) when is_tuple(Time), size(Time) == 3 ->
    assert_site_status_invariants(ThisSiteStatus).


assert_all_other_sites_have_status_entry(#state{other_sites        = OtherSites,
                                                other_sites_status = OtherSitesStatus}) ->
    lists:foreach(fun({FindSiteId, _, _, _}) ->
                          [SiteStatus] = [SiteStatus || {SiteId, SiteStatus} <- OtherSitesStatus, SiteId == FindSiteId],
                          assert_site_status_invariants(SiteStatus)
                  end,
                  OtherSites),
    %% and the inverse...
    lists:foreach(fun({FindSiteId, SiteStatus}) ->
                          assert_site_status_invariants(SiteStatus),
                          %% some restrictions for other site's status
                          case SiteStatus of 
                              #site_status{replication_group_pid = undefined} -> throw(other_site_status_cannot_have_undefined);
                              _                                               -> ok
                          end,
                          [_ExactlyOne] = [SiteId || {SiteId, _, _, _} <- OtherSites, SiteId == FindSiteId]
                  end,
                  OtherSitesStatus).
    

assert_site_status_invariants(#site_status{status = Status} = SiteStatus) ->
    assert_valid_site_status_value(Status),
    %% if status is unknown the last_updated_time must be unknown and vice-versa
    case SiteStatus of 
        #site_status{status            = unknown,
                     last_updated_time = unknown} -> ok;
        #site_status{status            = Status,
                     last_updated_time = Time} when Status =/= unknown, Time =/= unknown -> ok
    end,
    %% if replication_group_pid is not unknown then monitor_ref must not be unknown, and vice-versa
    case SiteStatus of 
        #site_status{replication_group_pid = undefined,
                     monitor_ref           = undefined} -> ok; % must be this_site_status
        #site_status{replication_group_pid = unknown,
                     monitor_ref           = unknown} -> ok;
        #site_status{replication_group_pid = Pid,
                     monitor_ref           = Ref} when is_pid(Pid), is_reference(Ref) -> ok
    end.


%% TODO: pass in list of AllSiteIds and assert that MasterSiteId is in that list
assert_valid_site_status_value(unknown)                                                  -> ok;
assert_valid_site_status_value({master, waiting_for_durability})                         -> ok;
assert_valid_site_status_value({master, available})                                      -> ok;
assert_valid_site_status_value({replica, not_in_sync, unknown})                          -> ok;
assert_valid_site_status_value({replica, not_in_sync, {{_, _}, transitioning_to}})       -> ok;
assert_valid_site_status_value({replica, not_in_sync, {{_, _}, waiting_for_durability}}) -> ok;
assert_valid_site_status_value({replica, not_in_sync, {{_, _}, available}})              -> ok;
assert_valid_site_status_value({replica, in_sync, unknown})                              -> ok;
assert_valid_site_status_value({replica, in_sync, {{_, _}, transitioning_to}})           -> ok;
assert_valid_site_status_value({replica, in_sync, {{_, _}, waiting_for_durability}})     -> ok;
assert_valid_site_status_value({replica, in_sync, {{_, _}, available}})                  -> ok.


assert_support_proc_invariants(State) ->
    assert_find_or_spawn_peers_invariants(State),
    assert_probe_to_sync_replicas_invariants(State),
    assert_remove_txn_logs_on_all_sites_invariants(State).


assert_find_or_spawn_peers_invariants(#state{support_procs  = SupportProcs}) ->
    %% Normally this should be running, but we do stop it in stop_entire_group and remove_site
    case [SP || #support_proc{name = find_or_spawn_peers} = SP <- SupportProcs] of
        []            -> ok;
        [_ExactlyOne] -> ok
    end.


assert_probe_to_sync_replicas_invariants(#state{current_master     = {this_site, waiting_for_durability},
                                                support_procs      = SupportProcs}) ->
    %% probe_to_sync_replicas should be running
    [_ExactlyOne] = [SP || #support_proc{name = probe_to_sync_replicas} = SP <- SupportProcs];

assert_probe_to_sync_replicas_invariants(#state{current_master     = {this_site, available},
                                                other_sites_status = OtherSitesStatus,
                                                support_procs      = SupportProcs}) ->
    %% We are master, so probe_to_sync_replicas should be running IFF any replica is not_in_sync
    %% Note: we don't run probe_to_sync_replicas if any replicas are 'unknown' (i.e. not connected)
    %% even if all the others are in_sync. We only start probe_to_sync_replicas when the missing replica
    %%connects as not_in_sync.
    ThisSiteId = get(this_site_id),
    case [SiteId || {SiteId, #site_status{status = {replica, not_in_sync, {MasterSiteId, _}}}} <- OtherSitesStatus,
                    MasterSiteId == ThisSiteId] 
    of
        [] -> 
            [] = [SP || #support_proc{name = probe_to_sync_replicas} = SP <- SupportProcs];
        _ -> 
            [_ExactlyOne] = [SP || #support_proc{name = probe_to_sync_replicas} = SP <- SupportProcs]
    end;

assert_probe_to_sync_replicas_invariants(#state{support_procs = SupportProcs}) ->  
    %% process should not be running
    [] = [SP || #support_proc{name = probe_to_sync_replicas} = SP <- SupportProcs].


assert_remove_txn_logs_on_all_sites_invariants(#state{current_master = {this_site, available},
                                                      support_procs  = SupportProcs}) ->
    %% We are master, so the process should be running
    [_ExactlyOne] = [SP || #support_proc{name = remove_txn_logs_on_all_sites} = SP <- SupportProcs];

assert_remove_txn_logs_on_all_sites_invariants(#state{support_procs = SupportProcs}) ->  
    %% should not be running
    [] = [SP || #support_proc{name = remove_txn_logs_on_all_sites} = SP <- SupportProcs].


